//
//  fairness_prolog_callback_functions.c
//  
//
//  Created by Ivaylo Dobrikov on 28/09/15.
//
//

#include "fairness_prolog_callback_functions.h"

/*
 * Converts a Prolog list with transition ids to a sorted array of SP_integers
 * The transition ids must all be integers. (I hope this is reallistic)
 * The array is returned in rlist and rsize.
 */
SP_integer* extract_transids_from_prolog_list(SP_term_ref list, int *rsize) {
    SP_integer *transids;
    SP_term_ref head = SP_new_term_ref();
    SP_term_ref tail = SP_new_term_ref();
    SP_term_ref tmplist = list;
    const char *termtype;
    /* Compute the length of the list: */
    int size = 0, i, success;
    while(SP_is_list(tmplist)) {
        size += 1;
        SP_get_list(tmplist,head,tail);
        tmplist = tail;
    }
    /* Allocate the array */
    transids = (SP_integer*) ltl_malloc(sizeof(SP_integer)*size);
    
    // in case the list of transids is empty no further
    //  if (transids == NULL) {
    //    fprintf(stderr,"malloc() returned NULL for list of transition ids, halt.\n");
    //    exit(1);
    //  }
    
    /* Iterate through the list and convert each Prolog term ref to a SP_integer */
    tmplist = list;
    for( i=0; i<size; i++) {
        SP_get_list(tmplist,head,tail);
        success = SP_get_integer(head,transids + i);
        if (!success) {
            switch (SP_term_type(head)) {
                case SP_TYPE_VARIABLE: termtype="a variable"; break;
                case SP_TYPE_INTEGER: termtype="an integer"; break;
                case SP_TYPE_FLOAT: termtype="a float"; break;
                case SP_TYPE_ATOM: termtype="an atom"; break;
                case SP_TYPE_COMPOUND: termtype="a coumpound term"; break;
                default: termtype="something else (unknown type)"; break;
            }
            fprintf(stderr,"Expected an integer as transition id but was %s, halt.\n", termtype);
            exit(1);
        }
        tmplist = tail;
    }
    /* sort the list */
    qsort(transids,size,sizeof(SP_integer),sp_int_cmp);
    /* and set the return values */
    *rsize = size;
    return transids;
}

static void check_status_and_initialise_call_result(int status, int *abort, const char *msg) {
    
    switch (status) {
        case SP_SUCCESS: /* ok, call succeeded */
            break;
        case SP_FAILURE: /* ok, call failed */
            ltl_reraise_exception();
            *abort=1;
            break;
        default: /* exception */
            IFLTLDEBUG( ltlmsg(msg); );
            ltl_reraise_exception();
            *abort=1;
            break;
    }
}

// clock_time: for clocking just the checking time of the fairness algorithm
SP_term_ref get_enabled_actions(SP_pred_ref predicate, SP_term_ref scc_state_ids, ltl_formula *formula, int clock_time) {
    int abort = 0;
    
    // the result variable for getting the list of enabled actions from prolog
    SP_term_ref actions_list;
    SP_put_variable(actions_list = SP_new_term_ref());
    
    
    clock_t now;
    IFLTLTIMING ( clock_t timer_start = clock(); );
    // calling the predicate_name(state_ids,action_list) predicate on the found SCC scc_state_ids list
    int status = SP_query(predicate, scc_state_ids, actions_list);
    now = clock();
    if (clock_time) {
        IFLTLTIMING( formula->fairness_callback_time += (now - timer_start);
                    formula->fairness_callback_calls += 1; );
    }

    const char* msg = "SCC weak fairness check raised exception!\n";
    check_status_and_initialise_call_result(status,&abort,msg);
    
    if (abort) {
        fprintf(stderr, "Unexpected call back result for getting transitions ids (get_enabled_actions)!");
        exit(1);
    }
    return actions_list;
}

// clock_time: for clocking just the checking time of the fairness algorithm
SP_integer* get_matching_transitions(SP_pred_ref predicate, SP_term_ref argument, SP_term_ref scc_state_ids, int *msize, ltl_formula *formula, int clock_time) {
    
    int abort = 0;
    
    // the result variable for getting the transition IDs from prolog
    SP_term_ref list;
    SP_put_variable(list = SP_new_term_ref());
    
    // calling the get_transition_ids/3 predicate on the found SCC scc_state_ids list
    
    clock_t now;
    IFLTLTIMING ( clock_t timer_start = clock(); );
    int status = SP_query(predicate, argument, scc_state_ids, list);
    now = clock();
    if (clock_time) {
        IFLTLTIMING( formula->fairness_callback_time += (now - timer_start);
                    formula->fairness_callback_calls += 1; );
    }
    
    const char* msg = "SCC weak fairness check raised exception!\n";
    check_status_and_initialise_call_result(status,&abort,msg);
    
    if (abort) {
        fprintf(stderr, "Unexpected call back result for getting transitions ids (get_matching_transitions)!");
        exit(1);
    }
    return extract_transids_from_prolog_list(list,msize);
}
