% (c) 2016-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
:- module(ltsmin,[ltsmin_init/2,ltsmin_init/3,ltsmin_loop/1,
                  ltsmin_teardown/2, start_ltsmin/4, generate_bindings/3]).

:- use_module(library(fastrw),[ fast_buf_read/2]).
:- use_module(library(codesio), [write_to_codes/2]).
:- use_module(library(lists)).

:- use_module(extension('ltsmin/ltsmin_c_interface')).
:- use_module(extension('ltsmin/msg_interop')).


:- use_module('../../src/module_information.pl').

:- use_module(extension('plspec/plspec/plspec_core')).
:- enable_spec_check(write_matrix/3).

:- use_module(probsrc(specfile)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(b_read_write_info), [tcltk_read_write_matrix/1]).

:- use_module(probsrc(typing_tools)).
%:- enable_all_spec_checks.

:- module_info(group,experimental).
:- module_info(description,'This is the interface to C code for LTSmin integration.').
:- use_module(probsrc(b_state_model_check), [get_guard/2]).
:- use_module(probsrc(bsyntaxtree), [conjunction_to_list/2, same_texpr/2,
                                     conjunct_predicates/2, is_truth/1, create_negation/2]).
%:- use_module(probsrc(system_call), [system_call_keep_open/7]).
:- use_module(library(process), [process_wait/2]).

:- dynamic label_to_reads/2.

:- spec_pre(msg_interop:write_matrix/3, [c_pointer,
                                         one_of([ltsmin_matrix(ltsmin_state_label, ltsmin_variable), % state label / guard / LTL matrices
                                                 ltsmin_matrix(ltsmin_transition_group, ltsmin_variable), % read/write matrices
                                                 ltsmin_matrix(ltsmin_state_label, ltsmin_transition_group), % NES / NDS
                                                 ltsmin_matrix(ltsmin_transition_group, ltsmin_state_label), % guard info
                                                 ltsmin_matrix(ltsmin_state_label, ltsmin_state_label), % may be co-enabled
                                                 ltsmin_matrix(ltsmin_transition_group, ltsmin_transition_group)]), % DNA
                                         any]).

:- defspec(ltsmin_flag, one_of([atom(por),
                                atom(g),
                                atom(nocache),
                                compound(ltl_formula(any))])).
:- defspec(ltsmin_backend, one_of([atom(sequential), atom(symbolic)])).
:- defspec(atom_bool, one_of([atom(true), atom(false)])).
:- defspec(c_pointer, int).
:- defspec(ltsmin_matrix(X, Y), [tuple([X, list(Y)])]).
:- defspec(b_pred_ish, compound(bpred(any))).

ltsmin_variable(X) :-
    varname_list(Variables), member(X, Variables).
:- defspec_pred(ltsmin_variable, ltsmin_variable).

ltsmin_transition_group(X) :-
    get_transition_groups(TGs), member(X, TGs).
:- defspec_pred(ltsmin_transition_group, ltsmin_transition_group).

ltsmin_state_label(X) :- state_label(X, _, _).
:- defspec_pred(ltsmin_state_label, ltsmin_state_label).

:- spec_pre(ltsmin_loop/1, [c_pointer]).
ltsmin_loop(Zocket) :-
    h_zocket2(Zocket, Ret), % this is a C function in ltsmin_c_interface, which does the call-backs to Prolog
    Ret == 0, !, ltsmin_loop(Zocket).
ltsmin_loop(_).

interpose([], _, []).
interpose([X], _, [X]) :- !.
interpose([H|T], I, [H,I|IT]) :-
    interpose(T, I, IT).

atom_concat([X], X) :- !.
atom_concat([H1, H2|T], R) :-
    atom_concat(H1, H2, H),
    atom_concat([H|T], R).

:- spec_pre(get_ltsmin_invariant/2, [var, atom_bool]).
:- spec_invariant(get_ltsmin_invariant/2, ['return value: an atom containing the flags for invariant checking for ltsmin':atom,
                                           'input: true or false describing whether POR is disabled or enabled':atom_bool]).
get_ltsmin_invariant(LTSminInv, Por) :-
    (Por == false -> Type = invariant ; Type = por_invariant),
    findall(InvLabel, state_label(InvLabel, Type, _AST), InvariantNames),
    (InvariantNames == [] -> Invariant = ['true']; interpose(InvariantNames, ' && ', Invariant)),
    atom_concat(Invariant, Invariant2),
    atom_concat(['--invariant=!is_init || (', Invariant2, ')'], LTSminInv).


backend_specific_flags(symbolic, ['--lace-workers=1'], _).
backend_specific_flags(sequential, FlagsSeq, MoreFlags) :-
    (member(nocache, MoreFlags) -> FlagsSeq = [] ; FlagsSeq = ['-c']).

:- spec_pre(property_flags/4, [atom_bool, atom_bool, atom_bool, var]).
:- spec_invariant(property_flags/4, ['no deadlock checking':atom_bool,
                                     'no invariant checking':atom_bool,
                                     'POR active':atom_bool,
                                     'resulting flags':[atomic]]).
property_flags(true, true, _, []).
property_flags(false, true, _, ['-d']).
property_flags(true, false, POR, [InvariantFlags]) :-
    get_ltsmin_invariant(InvariantFlags, POR).


% TODO: maybe the path should be chosen in another way
get_path_to_ltl_file('/tmp/prob-ltsmin-formula.ltl').
get_path_to_trace_file('/tmp/prob-ltsmin-trace.gcf').
get_path_to_csv_trace('/tmp/prob-ltsmin-trace.csv').

:- spec_pre(ltsmin_flag/2, [ltsmin_flag, var]).
:- spec_invariant(ltsmin_flag/2, [ltsmin_flag, atom]).
ltsmin_flag(por, '--por').
ltsmin_flag(g, '-g').
ltsmin_flag(ltl_formula(_), Flag) :-
    get_path_to_ltl_file(Path),
    atom_concat('--ltl=', Path, Flag).


% MoreFlags can contain: por, nocache, g (for guard splitting), ltl_formula(String)
:- spec_pre(get_ltsmin_flags/5, [ltsmin_backend, atom_bool, atom_bool, [ltsmin_flag], var]).
:- spec_invariant(get_ltsmin_flags/5, ['symbolic_or_sequential':ltsmin_backend,
                                       % at least one of the checks has to be false
                                       'no deadlock checking':atom_bool,
                                       'no invariant checking':atom_bool,
                                       'more flags that should be converted to LTSmin flags':[ltsmin_flag],
                                       'resulting flags for LTSmin':[atom]]).
get_ltsmin_flags(Backend, NoDead, NoInv, MoreFlags, ResFlags) :-
    backend_specific_flags(Backend, BackendFlags, MoreFlags),
    (member(por, MoreFlags) -> POR = true ; POR = false),
    property_flags(NoDead, NoInv, POR, PropertyFlags),
    convlist(ltsmin_flag, MoreFlags, EvenMoreFlags),
    get_path_to_trace_file(Tracefile),
    TraceFlags = ['--trace', Tracefile],
    append([TraceFlags, BackendFlags, PropertyFlags, EvenMoreFlags], ResFlags).



:- use_module(probsrc(tools_printing),[print_error/1]).
:- use_module(probsrc(debug), [debug_println/2]).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(error_manager), [add_warning/3, add_internal_error/2]).
get_tool_name(symbolic, '/prob2lts-sym').
get_tool_name(sequential, '/prob2lts-seq').
get_tool_name(printtrace, '/ltsmin-printtrace').

:- spec_pre(start_ltsmin/4, [ltsmin_backend, tuple([atom_bool, atom_bool]), [ltsmin_flag],var]).
:- spec_invariant(start_ltsmin/4, ['which backend should be used (symbolic or sequential)':ltsmin_backend,
                                   'tuple of length 2: no deadlock checking, no invariant checking':tuple([atom_bool, atom_bool]),
                                   'more ltsmin flags':[ltsmin_flag],
                                   'resulting of LTSmin model-checking':any ]).
start_ltsmin(Backend, [NoDead, NoInv], MoreFlags,Result) :-
    (NoDead = false, NoInv = false ->
        add_warning(start_ltsmin,'Turning off deadlock checking for LTSmin command: ',Backend),
        NoDead2 = true
     ; NoDead2 = NoDead),
    Endpoint = '/tmp/ltsmin.probz',
    ltsmin:ltsmin_init(Endpoint, Zocket),
    get_ltsmin_flags(Backend, NoDead2, NoInv, MoreFlags, Flags),
    (member(ltl_formula(Formula), MoreFlags)
      -> get_path_to_ltl_file(LTLFile),
         ltsmin_generate_ltlfile(Formula, LTLFile)
     ; true),
    preferences:get_preference(path_to_ltsmin, Path),
    get_tool_name(Backend, Tool),
    atom_concat(Path, Tool, LTSminCmd),
    on_exception(error(existence_error(_,_),E),
        process:process_create(LTSminCmd, [Endpoint, '--stats', '--when'|Flags], [process(PID)]),
        (debug_println(20,E),
         add_error_fail(ltsmin,'Path to LTSmin incorrect or LTSmin command not installed (copy prob2lts-sym, prob2lts-seq, ltsmin-printtrace from https://github.com/utwente-fmt/ltsmin/releases to ProB\'s lib folder),  cannot execute: ',LTSminCmd))),
    ltsmin:ltsmin_loop(Zocket),
    process_wait(PID, ExitStatus),
    (ExitStatus == exit(0)
      -> Result = ltsmin_model_checking_ok
       ; get_tool_name(printtrace, PrintTrace),
         atom_concat(Path, PrintTrace, PrintTraceCmd),
         get_path_to_trace_file(TraceFile),
         get_path_to_csv_trace(CsvFile),
         format('writing trace to ~w~n', CsvFile),
         on_exception(error(existence_error(_,_),E),
             process:process_create(PrintTraceCmd, [TraceFile, CsvFile, '-s,,'], [process(PID2)]),
             (debug_println(20,E),
             add_error_fail(ltsmin,'Path to LTSmin incorrect or LTSmin command not installed (copy command from https://github.com/utwente-fmt/ltsmin/releases to ProB\'s lib folder), cannot execute: ',PrintTraceCmd))),
         process_wait(PID2, _ExitStatus2),
         Result = ltsmin_counter_example_found(CsvFile)
    ),
    format('~w with ~w exited with status: ~w~n',[Tool, Flags, ExitStatus]),
    ltsmin:ltsmin_teardown(Zocket, Endpoint).




:- dynamic varname_list/1.

:- use_module(probltlsrc(ltl_tools), [temporal_parser/3]).

:- dynamic(inv_state_label_matrix/1).
:- dynamic(inv_state_label_matrix_for_por/1).

write_to_ltl_file(Formula, Path) :-
    open(Path, write, Stream),
    tell(Stream),
    print(Formula),
    told.

:- public ltsmin_generate_ltlfile/2.
ltsmin_generate_ltlfile(Formula, OutputFile) :-
    %% firstly, wrap the Formula into X(Formula)
    %% this avoids that a property is already considered true
    %% because it is in the artificial initial state
    atom_concat('X(', Formula, XFormula),
    atom_concat(XFormula, ')', FinalFormula),
    format('parsing LTL formula ~a~N', [FinalFormula]),
    setup_ltl_formula(FinalFormula, ResAtom),
    format('done. writing to: ~a~N', [OutputFile]),
    write_to_ltl_file(ResAtom, OutputFile).


setup_ltl_formula(Formula, ResAtom) :-
    temporal_parser(Formula, ltl, ParsedFormula),
    ltl_safety:pp_formula_to_ltl2ba_syntax(ParsedFormula, ResAtom).


:- spec_pre(ltsmin_init/3, [atom, var, atom]).
:- spec_pre(ltsmin_init/3, [atom, var, var]).
:- spec_invariant(ltsmin_init/3, ['the path to the ZMQ endpoint':atom,
                                  'return value: a C pointer to a ZMQ zocket':c_pointer,
                                  'an LTL formula':atom]).
ltsmin_init(Endpoint,Zocket, LtlFormula) :-
    (var(LtlFormula) -> true ; setup_ltl_formula(LtlFormula, _)),
    ltsmin_init(Endpoint, Zocket).

:- spec_pre(ltsmin_init/2, [atom, var]).
:- spec_pre(ltsmin_init/2, [atom, var]).
:- spec_invariant(ltsmin_init/2, ['the path to the ZMQ endpoint':atom,
                                  'return value: a C pointer to a ZMQ zocket':c_pointer]).
ltsmin_init(Endpoint,Zocket) :-
    loadfr,
    reset_ltsmin_extension,
    split_invariant(ListOfInvariantConjuncts, InvNames, ListOfInvariantConjunctsForPOR, InvNamesForPOR),
    get_state_labels(ListOfInvariantConjuncts, InvNames, invariant, InvStateLabelMatrix),
    get_state_labels(ListOfInvariantConjunctsForPOR, InvNamesForPOR, por_invariant, InvStateLabelMatrixForPOR),
    assert(inv_state_label_matrix(InvStateLabelMatrix)),
    assert(inv_state_label_matrix_for_por(InvStateLabelMatrixForPOR)),
    ltsmin_initialize(Endpoint, Zocket).

%assert_dir :- (user:library_directory('.') -> true ;
%assert(user:library_directory('.'))).


:- public get_next_states/4. % not sure how this is called
get_next_states(LongOrPattern, Msg, State,Op) :- % State is a list of bindings
    % TO DO: set_context_state(SomeID) and look if state_error(SomeID,Id,Error) were added
    compute_successors(Op, State, Succs),
    length(Succs,NrStates),
    %    format(" ~w successor states for ~w~n  ~w~n",[NrStates,Op,Succs]),nl,
    put_number(Msg, NrStates),
    put_successors(LongOrPattern, Succs, Msg).


read_state([],[]).
read_state([Addr|T], [Arg|R]) :- read_state(T,R), fast_buf_read(Arg,Addr).

:- meta_predicate mnf1(0).
mnf1(Call) :- %format('Calling ~w~n',[Call]), flush_output,
  if( on_exception(E,Call,(format('Exception: ~w~n',[E]),flush_output,fail)),
      true, (format('Failed ~w~n~n',[Call]),flush_output,fail)).

:- use_module(probsrc(specfile),[csp_with_bz_mode/0]).
% convert a list of values into a ProB environment [bind(x,Vx),...]
generate_bound_long_state([CSP|BState],Res) :- csp_with_bz_mode,!,
    Res=csp_and_b(CSP,BoundState),
    varname_list(L),
    %print(gen_bind(BState,L)),nl,
    generate_bindings(BState, L, BoundState).
/* this needs to be fixed for short states ;
 *  it does not get the entire varname list */
generate_bound_long_state(BState,BoundState) :-
    varname_list(L),
    generate_bindings(BState, L, BoundState).

generate_bound_state_short_label(BState,Label,BoundState) :-
    label_to_reads(Label, L),
    generate_bindings(BState, L, BoundState).

:- spec_pre(generate_bindings/3, [list(any), list(atom), list(compound(bind(atom, any)))]).
:- spec_pre(generate_bindings/3, [list(any), list(atom), var]).
:- spec_pre(generate_bindings/3, [var, var, list(compound(bind(atom, any)))]).
:- spec_invariant(generate_bindings/3, ['a list of values':list(any),
                                        'a list of identifiers':list(atom),
                                        'list of identifiers and values, wrapped in functor bind/2':list(compound(bind(atom, any)))]).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
:- use_module(probsrc(bmachine),[b_is_variable/2]).
:- use_module(probsrc(btypechecker), [unify_types_strict/2]).
:- use_module(probsrc(error_manager),[add_internal_error/2]).
:- debug_println(19,'TYPE CHECKING LTSMIN BINDINGS:').
generate_bindings([], [], []) :- !.
generate_bindings([H|T], [Name|TN], [bind(Name, H)|R]) :-
    (get_type(Name,Type)
     -> kernel_objects:infer_value_type(H,HType),
        (unify_types_strict(HType,Type) -> true
          ; add_internal_error('Unexpected type for:',id(Name,HType,Type,H))
        )
      ; add_internal_error('Unknown identifier:',Name)
    ),
    generate_bindings(T, TN, R).

get_type(Name,Type) :- b_is_variable(Name,Type).
get_type(Name,Type) :- b_get_machine_constants(Csts),
  member(b(identifier(Name),Type,_),Csts).
:- else.
generate_bindings([], [], []) :- !.
generate_bindings([H|T], [Name|TN], [bind(Name, H)|R]) :-
    generate_bindings(T, TN, R).
:- endif.

:- public get_successors_short/3. % called from C code
% implements R2W
get_successors_short(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(get_successors_aux(short, Msg, OperationA, StateA)).

:- public get_successors/3. % called from C code
get_successors(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(get_successors_aux(long, Msg, OperationA, StateA)).

get_successors_aux(Type, Msg, OperationA, StateA) :-
    %trace,
    read_state(StateA,State),
    fast_buf_read(Operation,OperationA),
    %print(operation(Operation)),nl,
    %    print(successors(Operation, State)),nl,
    (Type == short -> black_magic_short_transformation(Operation, BoundState, State, LongPattern, ShortSucc, Defaults, _, _, _, _, _), ShortOrLong = LongPattern-ShortSucc-Defaults
                    ; generate_bound_long_state(State, BoundState), ShortOrLong = long),
                        %print(generate(BoundState)),nl,
    (debug:debug_mode(on)
     -> format('~n ** get_successors for ~w~n',[Operation]),
        translate:print_state(BoundState),nl
     ; true),
    get_next_states(ShortOrLong, Msg, BoundState,Operation).

:- public get_state_label/3. % called by C
get_state_label(Msg, LabelA, StateA) :- mnf1(get_state_label_aux(long, Msg, LabelA, StateA)).

:- public get_state_label_short/3. % called by C
get_state_label_short(Msg, LabelA, StateA) :- mnf1(get_state_label_aux(short, Msg, LabelA, StateA)).

get_state_label_aux(Type, Msg, LabelA, StateA) :-
    read_state(StateA, State),
    fast_buf_read(Label,LabelA), %print(get_state_label(Label)),nl,
    (Type == long -> generate_bound_long_state(State,BoundState)
                   ; generate_bound_state_short_label(State, Label, BoundState)),
    %print(get_state_label_state(Label, BoundState)),nl,
    eval_label(Msg, BoundState,Label).

% NOTE: LTSMin assumes the first transition group is the initialisation action, it is called only once
% all other events are not called in the initial state
get_transition_groups(['$init_state'|L]) :-
    (specfile:b_or_z_mode
     -> findall(OperationName, bmachine:b_top_level_operation(OperationName), BL),
        (csp_with_bz_mode -> L = ['$CSP'|BL] ; L=BL)
     ;  L=[]).  % FIXME: provide list for CSP-M

extract_name_and_type([],[], []).
extract_name_and_type([b(identifier(HT),TypeTerm,_)|T], [HT|RT], [TypeTerm|RType]) :-
    extract_name_and_type(T,RT,RType).

/* is that already defined somewhere? */
get_ltsmin_variable_names(Names, Types) :-
    bmachine:b_get_machine_variables(V),
    % bmachine:b_get_machine_all_constants(C), % this includes constants which have been moved to deferred sets; causes issue with hanoi model
    bmachine:b_get_machine_constants(C), % this does not include global set constants
    append(C,V,CV),
    extract_name_and_type(CV,Names,Types).




extract_tg_and_nth_entry([], _, R, R).
extract_tg_and_nth_entry([list(H)|T], Arg, Acc, R) :-
    nth1(1, H, TransitionGroup),
    nth1(Arg, H, X),
    extract_tg_and_nth_entry(T, Arg, [[TransitionGroup, X]|Acc], R).

% patch matrix for CSP||B
put_patched_matrix(Msg, Type,M) :-
    patch_matrix(M,PM),
    (debug:debug_mode(on) -> print(put_matrix(Type,PM)),nl ; true),
    write_matrix(Msg, PM, Type).

% add $CSP as read/write dependency to matrixes
% [ [Odd,[]], [Even,[]], [Inc,[count]], [Reset,[count]]]
patch_matrix(M,Res) :- csp_with_bz_mode, !,
   Res = [ ['$CSP',['$CSP_CHUNK']] | PM], % add virtual CSP transition for CSP only events; TO DO: also allow/treat other pure CSP events; the CSP transition can only modify the $CSP chunk
   maplist(patch_row,M,PM).
patch_matrix(M,M).

patch_row([Op,Vars], [Op,['$CSP_CHUNK'|Vars]]).


build_initial_state(_, VType, VVal) :-
    any_value_for_type(VType, VVal).
    % do not use term undefined in order to avoid guard errors in the initial state
    %VVal = term(undefined).


term_to_atom(T, A) :-
    write_to_codes(T, Codes),
    atom_codes(A, Codes).


:- use_module(probsrc(predicate_debugger), [get_operation_propositional_guards/3]).


guard_seen(G, [H|_]) :-
    same_texpr(G, H), !.
guard_seen(G, [_|LofG]) :-
    guard_seen(G, LofG).


setify_guards(L, R) :-
    setify_guards(L, [], R).
setify_guards([], Seen, Seen).
setify_guards([G|RGuards], Seen, Res) :-
    (guard_seen(G, Seen)
      -> setify_guards(RGuards, Seen, Res)
      ;  setify_guards(RGuards, [G|Seen], Res)).


guard_vector(ListOfListOfGuards, Res) :-
    append(ListOfListOfGuards, ListOfGuards),
    setify_guards(ListOfGuards, Res).


get_name([H|_], [Name|_], G, Name) :-
    same_texpr(G, H), !.
get_name([_|GV], [_|Names], G, R) :-
    get_name(GV, Names, G, R).


guards_to_indices(GuardVector, Names, ListOfGuards, ListOfIndices) :-
    maplist(get_name(GuardVector, Names), ListOfGuards, ListOfIndices).

split_guards(Por, ['$init_state'|TransitionGroups], GuardVector, Names, [[]|IndexVectors]) :-
    % if we use POR, the transition group has to be enabled iff all its guards are true;
    % thus, for POR we need to use get_guard that also provides
    % existential quantifications for parameters as additional guards
    (Por == false
    -> maplist(get_operation_propositional_guards, TransitionGroups, RGuards, _Bodies)
     ; maplist(get_guard, TransitionGroups, Guards),
       maplist(conjunction_to_list, Guards, RGuards)),
    %% TODO: normalise guards
    guard_vector(RGuards, GuardVector),
    generate_names(GuardVector, 'guard_', Names), % get list [guardN, ..., guard1]
    maplist(guards_to_indices(GuardVector, Names), RGuards, IndexVectors).


create_matrix([], [], Res, Res) :- !.
create_matrix([H1|T1], [H2|T2], Acc, Res) :-
    create_matrix(T1, T2, [[H1,H2]|Acc], Res).
create_matrix(Headings, Entries, Res) :-
    create_matrix(Headings, Entries, [], Res).


generate_names([], _, _, Acc, Acc) :- !.
generate_names([_|T], Prefix, N, Acc, Res) :-
    number_codes(N, NC),
    append(Prefix, NC, NamedC),
    atom_codes(Named, NamedC),
    N1 is N+1,
    generate_names(T, Prefix, N1, [Named|Acc], Res).
generate_names(Terms, PrefixAtom, Names) :-
    atom_codes(PrefixAtom, PrefixChars),
    generate_names(Terms, PrefixChars, 1, [], Names).

wrap_named_inv(Type, Name, Invariant, state_label(Name, Type, Invariant)).
:- dynamic state_label/3.


assert_all([]).
assert_all([H|T]) :-
    assert(H), assert_all(T).

/* need to extract this from the atomic props of the formula,
   we can also use the prolog term of a formula as its state label */
get_state_labels(ListOfConjuncts, Names, Type, StateLabelMatrix) :-
    %% format of StateLabelMatrix is
    % [[label1, [id1, id2, ...]]
    %  [label2, [id2, id4, ...]]
    %  ...]
    maplist(wrap_named_inv(Type), Names, ListOfConjuncts, Facts),
    assert_all(Facts),
    maplist(bsyntaxtree:predicate_identifiers, ListOfConjuncts, SLSVars),
    create_matrix(Names, SLSVars, StateLabelMatrix).


:- use_module(probsrc(bmachine),[get_unproven_invariants/1]).
split_invariant(ListOfInvariantConjuncts, InvNames, ListOfInvariantConjunctsForPOR, InvNamesForPOR) :-
    get_unproven_invariants(ListOfInvariantConjuncts),!,
    shorten_inv_list(ListOfInvariantConjuncts,ListOfInvariantConjunctsForPOR),
    generate_names(ListOfInvariantConjunctsForPOR, 'porinv', InvNamesForPOR),
    generate_names(ListOfInvariantConjuncts, 'inv', InvNames).

% shorten to maximal 8 invariants, as otherwise ltsmin segfaults
shorten_inv_list([I1,I2,I3,I4,I5,I6,I7,I8|T],[I1,I2,I3,I4,I5,I6,I7,I8New]) :-  T \= [], !,
    length([I8|T],Len), NrInvs is Len+7,
    format('Merging last ~w of ~w invariants (LTSMin POR supports max 8 invariants)~n',[Len,NrInvs]),
    conjunct_predicates([I8|T],I8New).
shorten_inv_list([],R) :- !, format('No invariants to check~n',[]), R=[].
shorten_inv_list(I,I).



:- use_module(probporsrc(static_analysis), [test_path/6, dependent_actions/5]).

does_not_accord_with_aux([], _, []).
does_not_accord_with_aux([H|T], TransitionGroup, [H|R]) :-
    debug_println(9,check_dependency(H,TransitionGroup)),
    get_preference(timeout_cbc_analysis,TO),
    dependent_actions(H, TransitionGroup, 0, TO, _), !, % TODO: handle interrupts?
    does_not_accord_with_aux(T, TransitionGroup, R).
does_not_accord_with_aux([_|T], TransitionGroup, R) :-
    does_not_accord_with_aux(T, TransitionGroup, R).

do_not_accord(['$init_state'|TransitionGroups], [[]|DoNotAccordMatrix]) :-
    maplist(does_not_accord_with_aux(TransitionGroups), TransitionGroups, DoNotAccordMatrix).

can_enable(TransitionGroup, Guard, NGuard) :-
    ltsmin_test_path_possible(NGuard, [TransitionGroup], Guard).

ltsmin_test_path_possible(Conjunction,Path,Guard) :-
   get_preference(use_cbc_analysis,true),!,
   get_preference(timeout_cbc_analysis,TO),
   test_path(Conjunction, Path, Guard, 0, TO, Result),
    (Result = ok ; Result == timeout ; Result == unknown).
ltsmin_test_path_possible(_,_,_). % assume possible


calculate_enabling_set_aux([], _Guard, _NGuard, []) :- !.
calculate_enabling_set_aux([Transition|Rest], Guard, NGuard, [Transition|ResList]) :-
    can_enable(Transition, Guard, NGuard), !,
    calculate_enabling_set_aux(Rest, Guard, NGuard, ResList).
calculate_enabling_set_aux([_|Rest], Guard, NGuard, ResList) :-
    calculate_enabling_set_aux(Rest, Guard, NGuard, ResList).

calculate_enabling_set(TransitionGroupList, Guard, Res) :-
    create_negation(Guard, NGuard),
    calculate_enabling_set_aux(TransitionGroupList, Guard, NGuard, Res).

necessary_enabling_set(GuardVector, ['$init_state'|TransitionGroups], R) :-
    maplist(calculate_enabling_set(TransitionGroups), GuardVector, R).

necessary_disabling_set(GuardVector, TransitionGroups, R) :-
    maplist(create_negation, GuardVector, NGuardVector),
    necessary_enabling_set(NGuardVector, TransitionGroups, R).


may_be_coenabled_guards(G1, _G2Name, G2) :-
    conjunct_predicates([G1, G2], Conjunction),
    is_truth(Truth),
    ltsmin_test_path_possible(Conjunction, [], Truth).
         % fail on interrupt

may_be_coenabled2([], [], []).
may_be_coenabled2([_], [_], []).
may_be_coenabled2([H|T], [Name|NameRest], [[Name,R]|RT]) :-
    include(may_be_coenabled_guards(H), NameRest, T, R),
    may_be_coenabled2(T, NameRest, RT).
may_be_coenabled(GuardVector, Names, MayBeCoEnabledMatrix) :-
    % calculates an upper triangle matrix without the elements on the diagonal
    may_be_coenabled2(GuardVector, Names, MayBeCoEnabledMatrix).

:- dynamic(guard_labels/1).

get_label_group_guards(Msg, State) :-
    guard_labels(Labels),
    maplist(eval_label(Msg, State), Labels).


:- public get_state_label_group/3. % not sure how this is called
get_state_label_group(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(get_state_label_group_aux(Msg, OperationA, StateA)).
get_state_label_group_aux(Msg, TypeA, StateA) :-
    read_state(StateA,State),
    %print(generate(State)),nl,
    generate_bound_long_state(State,BoundState),
    fast_buf_read(Type,TypeA),
    (debug:debug_mode(on)
     -> format('~n ** get_state_label_group for ~w~n',[Type]),
        translate:print_state(BoundState),nl
     ; true),
    (Type == '1' -> get_label_group_guards(Msg, BoundState)
                 ;  print(unhandled_group_type(Type)),nl,fail).

get_ltl_labels(ASTs, Names) :-
    findall(ap_representation(Hash, AST), ltl_safety:ap_representation(Hash, AST), APReprs),
    maplist(arg(2), APReprs, ASTs),
    maplist(arg(1), APReprs, Names).


:- use_module(probsrc(error_manager), [add_error_fail/3]).

op_to_reads_guard(Op, ReadSet) :-
    get_opname(Op, OpName),
    op_guards(OpName, GuardList),
    maplist(guard_reads, GuardList, ReadSets),
    append(ReadSets, ReadSet).

oplist_to_reads_guards(OpList, ReadSet) :-
    maplist(op_to_reads_guard, OpList, ReadSets),
    append(ReadSets, ReadSet).
% identifiers might be contained multiple times due to append
get_ltl_readsets(bpred(X), ReadSet) :-
    bsyntaxtree:predicate_identifiers(X, ReadSet).
get_ltl_readsets(enabled(Op), ReadSet) :-
    op_to_reads_guard(Op, ReadSet). % get readset of guards of op
get_ltl_readsets(det(OpList), ReadSet) :-
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(dlk(OpList), ReadSet) :-
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(ctrl(OpList), ReadSet) :- !,
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(deadlock, ReadSet) :-
    get_transition_groups(OpList),
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(AP,_) :- !,
    add_error_fail(ltsmin, 'operation is not supported for LTSmin', AP).

:- dynamic(guard_reads/2).

assert_guard_readsets([]).
assert_guard_readsets([[GuardName, ReadSet]|R]) :-
    assert(guard_reads(GuardName, ReadSet)),
    assert_guard_readsets(R).

:- dynamic(op_guards/2).

assert_op_guards([], []).
assert_op_guards([Op|T], [GuardList|T2]) :-
    assert(op_guards(Op, GuardList)),
    assert_op_guards(T, T2).

:- spec_pre(ltl_to_state_label_matrix/3, [[b_pred_ish], [atom], var]).
:- spec_invariant(ltl_to_state_label_matrix/3, [[b_pred_ish], [atom], ltsmin_matrix(ltsmin_state_label, ltsmin_variable)]).
:- spec_post(ltl_to_state_label_matrix/3, [[b_pred_ish], [atom], var], [[b_pred_ish], [atom], ltsmin_matrix(ltsmin_state_label, ltsmin_variable)]).
ltl_to_state_label_matrix(ListOfAPs, LtlApNames, LtlStateLabelMatrix) :-
    %% format of StateLabelMatrix is
    % [[label1, [id1, id2, ...]]
    %  [label2, [id2, id4, ...]]
    %  ...]
    maplist(wrap_named_inv(ltl_ap), LtlApNames, ListOfAPs, Facts),
    assert_all(Facts),
    maplist(get_ltl_readsets, ListOfAPs, SLSVars),
    create_matrix(LtlApNames, SLSVars, LtlStateLabelMatrix).


calc_and_append_por_info(false, _Msg, _, _, _) :- !.
calc_and_append_por_info(_Por, Msg, GuardVector, GuardNames, TransitionGroups) :-
    may_be_coenabled(GuardVector, GuardNames, MayBeCoEnabledMatrix),
    write_matrix(Msg, MayBeCoEnabledMatrix, may_be_coenabled_matrix),

    necessary_enabling_set(GuardVector, TransitionGroups, NES),
    create_matrix(GuardNames, NES, NesMatrix),
    write_matrix(Msg, NesMatrix, necessary_enabling_set),

    necessary_disabling_set(GuardVector, TransitionGroups, NDS),
    create_matrix(GuardNames, NDS, NdsMatrix),
    write_matrix(Msg, NdsMatrix, necessary_disabling_set),

    do_not_accord(TransitionGroups, DNA),
    create_matrix(TransitionGroups, DNA, DnaMatrix),
    write_matrix(Msg, DnaMatrix, do_not_accord_matrix).


sort_by_and_distinct(SortByList, ToBeSortedMayContainDups, SortedUniqs) :-
    findall(X, (member(X, SortByList), member(X, ToBeSortedMayContainDups)), SortedWithDups),
    distinct(SortedWithDups, SortedUniqs).



:- dynamic black_magic_short_transformation/11.
% arg 1: Name of Operation
% arg 2: a pattern in order to create a bound state
% arg 3: a list of variables that unify the second argument to a bound state
% arg 4: a long state pattern that can be unified in order to get a short state in arg 5
% arg 5: the corresponding short state
% arg 6: default values in case variables are not written
% arg 7-11: see arg 2-6, but for next_action


% the last argument should be unified with the LTSmin short state values;
% this way, we acquire a bound state in the second to last argument.
prepare_binding([], [], [], []).
prepare_binding([H|AccessedTail], [H|ReadTail], [bind(H, V)|BindingTail], [V|ValTail]) :- !,
    prepare_binding(AccessedTail, ReadTail, BindingTail, ValTail).
prepare_binding([H|AccessedTail], Reads, [bind(H, term(undefined))|BindingTail], Vals) :-
    prepare_binding(AccessedTail, Reads, BindingTail, Vals).

% TODO: maybe refactor this (pk, 07.03.2018)
back_to_short([], [], [], [], [], _InitialState).
back_to_short([bind(H, _)|T], [H|WriteTail], [bind(H, V)|LongStateTail], [bind(H, V)|ShortTail], [bind(H, Default)|DefaultBindingsTail], InitialState) :- !,
    % variable is written to
    memberchk(bind(H, Default), InitialState),
    back_to_short(T, WriteTail, LongStateTail, ShortTail, DefaultBindingsTail, InitialState).
back_to_short([bind(H, _)|T], Writes, [bind(H, _)|LongStateTail], Short, DefaultBindings, InitialState) :-
    back_to_short(T, Writes, LongStateTail, Short, DefaultBindings, InitialState).

assert_short_state_transformations([], InitialState) :-
    varname_list(L),
    prepare_binding(L, [], PreparedBinding, PreparedVarListForUnification),
    back_to_short(PreparedBinding, L, LongPattern, PreparedShort, InitialState, InitialState),
    assert(black_magic_short_transformation('$init_state', PreparedBinding, PreparedVarListForUnification, LongPattern, PreparedShort, InitialState,
                                                           PreparedBinding, PreparedVarListForUnification, LongPattern, PreparedShort, InitialState)).
assert_short_state_transformations([list([Op, ReadsGuard, ReadsAction, MustWrite, MayWrite|_])|T], InitialState) :-
    varname_list(L),
    sort_by_and_distinct(L, ReadsAction, SortedReadsAction),
    append([ReadsGuard, ReadsAction, MustWrite, MayWrite], AllAccessedVarsWithDups),
    append([ReadsAction, MustWrite, MayWrite], AllAccessedVarsForActionWithDups),
    append([ReadsGuard, ReadsAction], AllReadsWithDups),
    append([MustWrite, MayWrite], AllWritesWithDups),
    sort_by_and_distinct(L, AllAccessedVarsWithDups, AllAccessedVars),
    sort_by_and_distinct(L, AllAccessedVarsForActionWithDups, AllAccessedVarsForAction),
    sort_by_and_distinct(L, AllReadsWithDups, AllReads),
    sort_by_and_distinct(L, AllWritesWithDups, AllWrites),
    prepare_binding(AllAccessedVars, AllReads, PreparedBinding, PreparedVarListForUnification),
    prepare_binding(AllAccessedVarsForAction, SortedReadsAction, PreparedBindingForNextAction, PreparedVarListForUnificationForNextActionFactoryProviderSingleton),
    back_to_short(PreparedBinding, AllWrites, LongPattern, PreparedShort, Defaults, InitialState),
    back_to_short(PreparedBindingForNextAction, AllWrites, LongPatternForAction, PreparedShortForNextAction, DefaultsForAction, InitialState),
    assert(black_magic_short_transformation(Op, PreparedBinding, PreparedVarListForUnification, LongPattern, PreparedShort, Defaults,
                                          PreparedBindingForNextAction, PreparedVarListForUnificationForNextActionFactoryProviderSingleton, LongPatternForAction, PreparedShortForNextAction, DefaultsForAction)),
    assert_short_state_transformations(T, InitialState).


assert_label_to_accessed_vars([]).
assert_label_to_accessed_vars([[Label,ListOfVars]|T]) :-
    varname_list(L),
    findall(AccessedOrderedVars, (member(AccessedOrderedVars, L),
                                  member(AccessedOrderedVars, ListOfVars)),
                                  AccessedVarSet), % poor man's sort-by
    assert(label_to_reads(Label, AccessedVarSet)),
    assert_label_to_accessed_vars(T).






distinct(X, Y) :-
    distinct(X, [], Y).
distinct([], _, []).
distinct([H|T], Seen, Res) :-
    member(H, Seen), !,
    distinct(T, Seen, Res).
distinct([H|T], Seen, [H|Res]) :-
    distinct(T, [H|Seen], Res).



put_read_write_matrices(Msg, MR) :-
            %% read write matrices
            tcltk_read_write_matrix(M),
            (debug:debug_mode(on) -> print(M),nl,flush_output ; true),
            M = list([_Headings|MR]),
            (debug:debug_mode(on) -> print(MR),nl ; true),
            extract_tg_and_nth_entry(MR, 5, [],  MayWriteMatrix),
            extract_tg_and_nth_entry(MR, 4, [],  MustWriteMatrix),
            extract_tg_and_nth_entry(MR, 3, [],  ReadsActionMatrix),
            extract_tg_and_nth_entry(MR, 2, [],  ReadsGuardMatrix),
            put_may_write_matrix(Msg, MayWriteMatrix),
            put_patched_matrix(Msg, must_write_matrix,MustWriteMatrix),
            put_patched_matrix(Msg, reads_action_matrix,ReadsActionMatrix),
            put_patched_matrix(Msg, reads_guard_matrix,ReadsGuardMatrix).

assert_short_state_info(MR, InitialState) :-
    varname_list(L),
    generate_bindings(InitialState, L, BoundInitialState),
    assert_short_state_transformations(MR, BoundInitialState).

reset_ltsmin_extension :-
            reset_c_counters,
            retractall(black_magic_short_transformation(_,_,_,_,_,_,_,_,_,_,_)),
            retractall(label_to_reads(_,_)),
            retractall(inv_state_label_matrix(_)),
            retractall(inv_state_label_matrix_for_por(_)),
            retractall(varname_list(_)),
            retractall(op_guards(_, _)),
            retractall(state_label(_, _, _)),
            retractall(guard_labels(_)),
            retractall(guard_reads(_, _)).

:- public get_init/2.
get_init(Msg, PorInt) :-
            (PorInt == 0 -> Por = false ; Por = true),
            print(pl_init_for_lts_min),nl,flush_output,
            get_ltsmin_variable_names(VS, VTypes),
            assert(varname_list(VS)),
            print(vars(VS)),nl,
            maplist(build_initial_state, VS, VTypes, BInitialState),
            %maplist(any_value_for_type,VTypes,InitialState),
            maplist(term_to_atom, VTypes, VTypeAtoms),
            get_transition_groups(TransitionGroups),
            (csp_with_bz_mode
              -> InitialState = [stop(no_loc_info_available) | BInitialState],  % DUMMY Values
                 Vars = ['$CSP_CHUNK'|VS], Types = ['csp_process'|VTypeAtoms]
               ; InitialState = BInitialState,
                 Vars = VS, Types = VTypeAtoms
            ),
            debug_println(9,initial(InitialState)),
            write_chunklist(Msg, InitialState, initial_state),
            write_chunklist(Msg, TransitionGroups, transition_group_list),
            write_chunklist(Msg, Vars, variables_name_list), % we use $CSP_CHUNK as name for the CSP chunk
            write_chunklist(Msg, Types, variables_type_list),


            put_read_write_matrices(Msg, MR),
            assert_short_state_info(MR, InitialState),

            (Por == false -> inv_state_label_matrix(InvStateLabelMatrix)
                           ; inv_state_label_matrix_for_por(InvStateLabelMatrix)),

            %print(slm(InvStateLabelMatrix)),nl,
            %% split guards to state labels
            split_guards(Por, TransitionGroups, GuardVector, GuardNames, IndexVectors),
            get_state_labels(GuardVector, GuardNames, guard, GuardStateLabelMatrix),

            % memoize mapping OpName -> GuardNames
            assert_op_guards(TransitionGroups, IndexVectors),
            % memoize mapping GuardName -> ReadSet(Guard)
            assert_guard_readsets(GuardStateLabelMatrix),

            reverse(GuardNames, GuardNamesInOrder),
            assert(guard_labels(GuardNamesInOrder)),

            get_ltl_labels(ListOfLabelASTs, LtlApNames),
            ltl_to_state_label_matrix(ListOfLabelASTs, LtlApNames, LtlStateLabelMatrix),

            % accessed variable names in order (for short states)
            assert_label_to_accessed_vars(InvStateLabelMatrix),
            assert_label_to_accessed_vars(GuardStateLabelMatrix),
            assert_label_to_accessed_vars(LtlStateLabelMatrix),

            %% write all state labels (invariant conjuncts and guard conjuncts)
            write_matrix(Msg, InvStateLabelMatrix, state_label_matrix),
            write_matrix(Msg, GuardStateLabelMatrix, guard_label_matrix),
            write_matrix(Msg, LtlStateLabelMatrix, ltl_label_matrix),

            %% create guard info matrix, i.e. which transition depends on which guards
            create_matrix(TransitionGroups, IndexVectors, GuardInfoMatrix),
            write_matrix(Msg, GuardInfoMatrix, guard_info_matrix),
            calc_and_append_por_info(Por, Msg, GuardVector, GuardNames, TransitionGroups).



compute_successors('$init_state',_,Result) :- !,
    findall(S,init_state(S),Result),
    (debug:debug_mode(on) -> print(res_init_state(Result)), nl, flush_output ; true).
compute_successors(OpName,L,Result) :-
    (debug:debug_mode(on) -> nl,print(compute_successors(OpName)),nl,flush_output ; true),
    findall(S,specfile:specfile_trans(L,OpName,_,S,_,_),Result).
    % this in turn will call b_execute_top_level_operation_update for normal operations

:- public next_action/3. % called from C
next_action(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(next_action_aux(long, Msg, OperationA, StateA)).
next_action_short(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(next_action_aux(short, Msg, OperationA, StateA)).
next_action_aux(Type, Msg, OperationA, StateA) :-
    read_state(StateA,State),
    fast_buf_read(Operation,OperationA),
    %print(generate(Type, State, Operation)),nl,
    (Type == short -> black_magic_short_transformation(Operation, _, _, _, _, _, BoundState, State, LongPattern, ShortSucc, Defaults), LongOrPattern = LongPattern-ShortSucc-Defaults
                    ; generate_bound_long_state(State, BoundState), LongOrPattern = long),
    (debug:debug_mode(on)
     -> format('~n ** get_successors for ~w~n',[Operation]),
        translate:print_state(BoundState),nl
     ; true),
    get_next_action(LongOrPattern, Msg, BoundState,Operation).

:- use_module(probsrc(preferences),[get_preference/2,temporary_set_preference/3,reset_temporary_preference/2]).
get_next_action(Type, Msg, State,Op) :- % State is a list of bindings
    temporary_set_preference(ltsmin_do_not_evaluate_top_level_guards,true,CHNG),
    get_next_states(Type, Msg, State,Op),
    reset_temporary_preference(ltsmin_do_not_evaluate_top_level_guards,CHNG).

init_state(Res) :- specfile:specfile_trans(root,_,_,S,_,_),
                   init_state2(S,Res). % , nl,print(init(Res)),nl.

init_state2(csp_and_b_root,Res) :- !,
    specfile:specfile_trans(csp_and_b_root,_,_,S,_,_),
    init_state2(S,Res).
init_state2(concrete_constants(L),FullRes) :- !,
    specfile:specfile_trans(concrete_constants(L),_,_,Res,_,_),
    expand_const_and_vars_to_full_store(Res,FullRes).
init_state2(R,R).

eval_label(Msg, State,invariant) :-
    bmachine:b_get_invariant_from_machine(Invariant),!, % Should we use: get_unproven_invariants
    (b_interpreter:b_test_boolean_expression_for_ground_state(Invariant,[],State) ->
        put_number(Msg, 1); put_number(Msg, 0)).


eval_label(Msg, State,Label) :-
    (debug:debug_mode(on) -> print(eval(State,Label)),nl,flush_output ; true),
    state_label(Label, Type, AST),
    eval_label2(Type, Msg, State, AST).


%:- use_module(probltlsrc(ltl_prepositions), [check_ap/2]).
eval_label2(ltl_ap, Msg, State, AST) :-
    (check_ap_bound_state(AST,State) ->
         put_number(Msg, 1); put_number(Msg, 0)).
eval_label2(_, Msg, State, AST) :-
    (b_interpreter:b_test_boolean_expression_for_ground_state(AST,[],State) ->
         put_number(Msg, 1) ; put_number(Msg, 0)).
%  foreign predicate  put_number(Msg, X) :: 0 = false, 1=true


get_opname(bop(Name, _, _, _, _), Name).
get_opname(btrans(event(Name)), Name).
get_opname(btrans(event(Name, _, _)), Name).

op_enabled(Op, State) :-
    get_opname(Op, OpName),
    get_guard(OpName, Guard),
    b_interpreter:b_test_boolean_expression_for_ground_state(Guard, [], State).

% check an atomic proposition in a given node (i.e. state)
check_ap_bound_state(bpred(Property),Node) :-
    b_interpreter:b_test_boolean_expression_for_ground_state(Property, [], Node).
check_ap_bound_state(enabled(Op),Node) :- !,
    op_enabled(Op, Node).
check_ap_bound_state(det(OpList),Node) :- !,
    max_enabled(OpList, Node, 1, 0, R),
    R < 2.
check_ap_bound_state(dlk(OpList),Node) :- !,
    max_enabled(OpList, Node, 0, 0, 0).
check_ap_bound_state(ctrl(OpList),Node) :- !,
    max_enabled(OpList, Node, 1, 0, 1).
check_ap_bound_state(deadlock,Node) :- !,
    get_transition_groups(OpList),
    check_ap_bound_state(dlk(OpList), Node).

%max_enabled(OpList, Max, EnabledSoFar, ResEnabled)
max_enabled([], _State, _Max, R, R).
max_enabled([Op|T], State, Max, EnabledSoFar, R) :-
    (op_enabled(Op, State)
     -> Enabled is EnabledSoFar + 1,
        Max >= Enabled,
        max_enabled(T, State, Max, Enabled, R)
     ;  max_enabled(T, State, Max, EnabledSoFar, R)).


