#include <iostream>
#include <string>
#include <map>
#include <cstdlib>
#include <time.h>
#include <regex>
#include <cassert>

#include <sicstus/sicstus.h>
#include "regexp_glue.h"

#include "utf8_v2_3_4/source/utf8.h"

using namespace std;


string parseCode(regex_constants::error_type etype) {
    switch (etype) {
    case regex_constants::error_collate:
        return "error_collate: invalid collating element request";
    case regex_constants::error_ctype:
        return "error_ctype: invalid character class";
    case regex_constants::error_escape:
        return "error_escape: invalid escape character or trailing escape";
    case regex_constants::error_backref:
        return "error_backref: invalid back reference";
    case regex_constants::error_brack:
        return "error_brack: mismatched bracket([ or ])";
    case regex_constants::error_paren:
        return "error_paren: mismatched parentheses(( or ))";
    case regex_constants::error_brace:
        return "error_brace: mismatched brace({ or })";
    case regex_constants::error_badbrace:
        return "error_badbrace: invalid range inside a { }";
    case regex_constants::error_range:
        return "erro_range: invalid character range(e.g., [z-a])";
    case regex_constants::error_space:
        return "error_space: insufficient memory to handle this regular expression";
    case regex_constants::error_badrepeat:
        return "error_badrepeat: a repetition character (*, ?, +, or {) was not preceded by a valid regular expression";
    case regex_constants::error_complexity:
        return "error_complexity: the requested match is too complex";
    case regex_constants::error_stack:
        return "error_stack: insufficient memory to evaluate a match";
    default:
        return "";
    }
}


SP_integer cpp_regex_match(char const *str, char const *pat, SP_integer printErrors) {
 //	std::cout<<"enter regexp_match fonction"<<"\n"<<std::flush;
 std::locale old;
 #ifdef _WIN32
	std::setlocale(LC_ALL, "english");
 #else
	std::locale::global(std::locale("en_US.UTF-8"));
 #endif
 try {
      	  // convert pattern to utf16
      std::string srcpat(pat); 
      std::wstring utf16line;
      utf8::utf8to16(srcpat.begin(),srcpat.end(),back_inserter(utf16line));
	  std::wregex e (utf16line, std::regex_constants::collate);
      // convert source string to utf16
      std::wstring utf16src;
      std::string s(str); 
      utf8::utf8to16(s.begin(),s.end(),back_inserter(utf16src));
	  if (std::regex_match (utf16src, e )) { //std::cout << "string literal matched: " << s << "\n"<<std::flush;
	     std::locale::global(old);
		 return 1;
	  } else { //std::cout << "string literal NOT matched: " << s << "\n"<<std::flush;
	     std::locale::global(old);
		 return 0;
	  }
 } catch (regex_error &e) {
	    std::locale::global(old);
        if (printErrors>0) {
             cout << "### Regexp Match Error: " << e.what() << "\n### Error Code: " << parseCode(e.code()) << endl;
        }
        return 2;
    }
}


void raise_regex_error(regex_error &e) {
        cout << "### Regexp Error: " << e.what() << "\n### Error Code: " << parseCode(e.code()) << endl;
		SP_term_ref eterm = SP_new_term_ref();
		SP_term_ref arg = SP_new_term_ref();
		int ret = SP_put_atom(arg,SP_atom_from_string(e.what()));
		assert(ret);
		ret = SP_cons_functor(eterm, SP_atom_from_string("regex_exception"),1, arg);
		assert(ret);
        SP_raise_exception(eterm);
}

void assert_and_raise (int ret, std::string msg, std::string arg) {
  if(!ret) {
        cout << "### Assertion failed: " << msg << arg << endl;
		SP_term_ref eterm = SP_new_term_ref();
		ret = SP_cons_functor(eterm, SP_atom_from_string("regex_assertion_failed"),0);
		assert(ret); 
        SP_raise_exception(eterm);
  }
}

SP_atom cpp_regex_replace(char const *str,char const *pat,char const *replacementStr) {
 std::locale old;
  #ifdef _WIN32
	std::setlocale(LC_ALL, "english");
 #else
	std::locale::global(std::locale("en_US.UTF-8"));
 #endif
 try{
      // convert pattern to utf16
      std::string srcpat(pat); 
      std::wstring utf16line;
      utf8::utf8to16(srcpat.begin(),srcpat.end(),back_inserter(utf16line));
	  std::wregex e (utf16line, std::regex_constants::collate); // | std::regex_constants::optimize
      // convert source string to utf16
      std::wstring utf16src;
      std::string s(str); 
      utf8::utf8to16(s.begin(),s.end(),back_inserter(utf16src));
      // convert source string to utf16
      std::wstring utf16repl;
      std::string rep(replacementStr); 
      utf8::utf8to16(rep.begin(),rep.end(),back_inserter(utf16repl));
      
	 std::wstring res (std::regex_replace (utf16src, e, utf16repl ));
	 //std::cout << "pattern replaced: " << res << "\n";
	 std::string resultstr;
	 utf8::utf16to8(res.begin(), res.end(), back_inserter(resultstr));
	 SP_atom result = SP_atom_from_string(resultstr.c_str());
	 assert_and_raise(result,"SP_atom_from_string (cpp_regex_replace): ",resultstr);
	 std::locale::global(old);
	 return result;
 } catch (regex_error &e) {
	    std::locale::global(old);
        raise_regex_error(e);
        return SP_atom_from_string("exception"); // to make the compiler happy
 }
}


// only find first match string
SP_atom cpp_regex_search_first_str(char const *str,char const *pat) {
 std::locale old;
  #ifdef _WIN32
	std::setlocale(LC_ALL, "english");
 #else
	std::locale::global(std::locale("en_US.UTF-8"));
 #endif
 try {
      // convert pattern to utf16
      std::string srcpat(pat); 
      std::wstring utf16line;
      utf8::utf8to16(srcpat.begin(),srcpat.end(),back_inserter(utf16line));
	  std::wregex e (utf16line, std::regex_constants::collate);
      // convert source string to utf16
      std::wstring utf16src;
      std::string s(str); 
      utf8::utf8to16(s.begin(),s.end(),back_inserter(utf16src));

	  // std::cout << "Target sequence: " << s << " pat: " << srcpat << std::endl;
	  std::wsmatch m;

	  if (std::regex_search (utf16src,m,e)) {
	      std::wstring wmatch = utf16src.substr(m.position(),m.length());
	      std::string matchstr;
	      utf8::utf16to8(wmatch.begin(), wmatch.end(), back_inserter(matchstr));
	      //std:: cout << "Match Found: " << matchstr << std::endl;
	      SP_atom result = SP_atom_from_string(matchstr.c_str());
		  assert_and_raise(result,"SP_atom_from_string (cpp_regex_search_first_str): ",matchstr);
	      std::locale::global(old);
		  return result;
	  } else {
	      std::locale::global(old);
	      return SP_atom_from_string("");
	  }
 } catch (regex_error &e) {
	    std::locale::global(old);
        raise_regex_error(e);
        return 2;
 }
}

// returns string found as well as position and length of match
SP_term_ref cpp_regex_search_first(char const *str, SP_integer offset, char const *pat) {
 std::locale old;
  #ifdef _WIN32
	std::setlocale(LC_ALL, "english");
 #else
	std::locale::global(std::locale("en_US.UTF-8"));
 #endif
 try {
      // convert pattern to utf16
      std::string srcpat(pat); 
      std::wstring utf16line;
      utf8::utf8to16(srcpat.begin(),srcpat.end(),back_inserter(utf16line));
	  std::wregex e (utf16line, std::regex_constants::collate);
      // convert source string to utf16
      std::wstring utf16src;
      std::string s(str); 
      utf8::utf8to16(s.begin(),s.end(),back_inserter(utf16src));

     auto sbegin = utf16src.begin();
     assert(offset>=0);
     auto maxdist(std::distance(sbegin, utf16src.end()));
     std::advance(sbegin, offset > maxdist ? maxdist : offset); // advance offset but not beyond end of array
     std::wsregex_iterator rit ( sbegin, utf16src.end(), e );
     std::wsregex_iterator rend;
  
	  SP_term_ref resterm = SP_new_term_ref();
	  if (rit!=rend) {
	     std::wsmatch m = *rit;
	  //if (std::regex_search (s,m,e)) {
	      // the match is now returned as first element of the list
	      //std::string matchstr (s.substr(m.position(),m.length()));
	      //std:: cout << "Match Found: " << matchstr << std::endl;
		  //SP_term_ref match = SP_new_term_ref();
		  //ret = SP_put_string(match,matchstr.c_str());
		  SP_term_ref pos = SP_new_term_ref();
		  SP_term_ref len = SP_new_term_ref();
		  int ret = SP_put_integer(pos,m.position()+offset);
		  assert(ret);
		  ret = SP_put_integer(len,m.length());
		  assert(ret);
	      //std:: cout << "Match Found: " << m.position() << " with length " << m.length() << " with submatches " << m.size() << std::endl;
	      SP_term_ref innermatches = SP_new_term_ref();
		  auto iterator = m.end();
		   while(iterator != m.begin()) {
		      --iterator;
	          //std::cout << "inner match: " << *iterator  << " : len="<< (*iterator).length() << std::endl;
	          SP_term_ref lstptr = SP_new_term_ref();
	          SP_term_ref head = SP_new_term_ref();
	          std::wstring wmatch (*iterator);
	          std::string innermatchstr;
	          utf8::utf16to8(wmatch.begin(), wmatch.end(), back_inserter(innermatchstr));
		      ret = SP_put_string(head,innermatchstr.c_str());
		      //assert(ret); // if it is nil we could not convert string back to Prolog format
		      assert_and_raise(ret,"SP_put_string (cpp_regex_search_first): ",innermatchstr);
	          ret = SP_cons_list(lstptr,head,innermatches);
		      assert(ret);
		      innermatches = lstptr;
	       }
	      ret = SP_cons_functor(resterm, SP_atom_from_string("match"),3,pos,len,innermatches);
		  assert(ret);
	  } else {
	      int ret = SP_put_string(resterm,"no-match");
		  assert(ret);
	  }
	  std::locale::global(old);
	  return resterm;
 } catch (regex_error &e) {
	    std::locale::global(old);
        raise_regex_error(e);
	    SP_term_ref resterm = SP_new_term_ref(); //  not reachable
	    SP_put_string(resterm,"error");
	    return resterm;
 }
}

SP_term_ref cpp_regex_search_all_str(char const *str,char const *pat) {
 std::locale old;
  #ifdef _WIN32
	std::setlocale(LC_ALL, "english");
 #else
	std::locale::global(std::locale("en_US.UTF-8"));
 #endif
 try {
      // convert pattern to utf16
      std::string srcpat(pat); 
      std::wstring utf16line;
      utf8::utf8to16(srcpat.begin(),srcpat.end(),back_inserter(utf16line));
	  std::wregex e (utf16line, std::regex_constants::collate); // | std::regex_constants::optimize
      // convert source string to utf16
      std::wstring utf16src;
      std::string s(str); 
      utf8::utf8to16(s.begin(),s.end(),back_inserter(utf16src));

	  //std::cout << "Target sequence: " << s << std::endl;
	  std::wsmatch m;

	  SP_term_ref matches = SP_new_term_ref();
	  while (std::regex_search (utf16src,m,e)) {
		//for (auto x:m) std::cout << x << " ";
	    std::wstring wmatch = utf16src.substr(m.position(),m.length());
	    std::string matchstr;
	    utf8::utf16to8(wmatch.begin(), wmatch.end(), back_inserter(matchstr));
	    //std::cout << "match found: " << matchstr << std::endl;
	    SP_term_ref lstptr = SP_new_term_ref();
	    SP_term_ref head = SP_new_term_ref();
		int ret = SP_put_string(head,matchstr.c_str());
		assert_and_raise(ret,"SP_put_string (cpp_regex_search_all_str): ",matchstr);
	    ret = SP_cons_list(lstptr,head,matches);
	    assert(ret);
		utf16src = m.suffix().str();
		matches = lstptr;
		if (m.length()==0) {
		    cout << "INFINITELY MANY MATCHES !!" << endl;
		    return matches;
		}
	  }
	  std::locale::global(old);
	  return matches;
 } catch (regex_error &e) {
	    std::locale::global(old);
        raise_regex_error(e);
	    SP_term_ref resterm = SP_new_term_ref(); //  not reachable
	    SP_put_string(resterm,"error");
	    return resterm;
 }
}
