/*
* Exposes Glucose API to SICStus Prolog
*
* Based on SWI Prolog Minisat Integration by
* Michael Codish, Vitaly Lagoon, Peter J. Stuckey
*/

:- module(satsolver, [init_satsolver/0, sat/1, sat/2, bt_sat/1, multi_sat/3, b_sat/2]).

:- use_module('../../src/module_information.pl').
:- module_info(group,experimental).
:- module_info(description,'This is the interface between ProB and the Glucose SAT solver').

:- use_module(b_to_cnf).

:- use_module(library(lists), [maplist/2,maplist/3]).

foreign_resource('satsolver', [new_solver,solve,delete_solver,
                               add_clause,assign_model,get_model]).

foreign(new_solver, c, new_solver([-integer])).
foreign(solve, c, solve(+integer)).
foreign(delete_solver, c, delete_solver(+integer)).
foreign(add_clause, c, add_clause(+integer,+term)).
foreign(assign_model, c, assign_model(+integer,+term)).
foreign(get_model, c, get_model(+integer,+term)).

:- dynamic is_initialised/0.

init_satsolver :- is_initialised,!.
init_satsolver :-
    catch(load_foreign_resource(library(satsolver)),_,(print('*** LOADING satsolver library failed'),nl,fail)),
    assert(is_initialised).

b_sat(BFormula,State) :-
  init_satsolver,
  b_to_cnf(BFormula,State,CNF),!,
  bt_sat(CNF).

sat(CNF):-
    sat(CNF,Solved),!,
    Solved=1.

sat([],1):-!.
sat([[]],0):-!.
sat(F,Solved):-
    new_solver(SolverID),
    setup_solver(SolverID),
    (addCnf2Solver(SolverID,F,FVars)
    -> (solve(SolverID)
        -> assign_model(SolverID,[1|FVars]), Solved = 1
        ;  Solved = 0)
    ; Solved=0),
    delete_solver(SolverID),!.

bt_sat(F) :-
    new_solver(SolverID),
    bt_sat_aux(SolverID,F).

bt_sat_aux(SolverID,F) :-
    setup_solver(SolverID),
    addCnf2Solver(SolverID,F,FVars),
    solve(SolverID),
    get_model(SolverID,[_|Model]),
    bt_sat(SolverID,FVars,Model).
bt_sat_aux(SolverID,_) :-
    delete_solver(SolverID), fail.

bt_sat(SolverID,FVars,_Model) :-
    assign_model(SolverID,[1|FVars]).
bt_sat(SolverID,FVars,Model) :-
    add_negated_model(SolverID,Model),
    solve(SolverID),
    get_model(SolverID,[_|NewModel]),
    bt_sat(SolverID,FVars,NewModel).

multi_sat([],_,1):-!.
multi_sat([[]],_,0):-!.
multi_sat(F,MaxSols,SolCount):-
    new_solver(SolverID),
    setup_solver(SolverID),
    (addCnf2Solver(SolverID,F,FVars)
    -> satMultiModels(SolverID,MaxSols,Models),
       delete_solver(SolverID),!,
       length(Models,SolCount),
       (SolCount == 0 ; assignMultiSols(Models,FVars))
    ;  delete_solver(SolverID),!,SolCount=0),!.

satMultiModels(SolverID,MaxSols,Models):-
    MaxSols > 0,!,
    (solve(SolverID)
    -> get_model(SolverID,[_|Model]),
       Models=[Model|MoreModels],
       MaxSols1 is MaxSols - 1,
       ((MaxSols1 > 0, add_negated_model(SolverID,Model))
       -> satMultiModels(SolverID,MaxSols1,MoreModels)
       ; MoreModels=[])
    ; Models=[]).
satMultiModels(_,_,[]):-!.

add_negated_model(SolverID,Model):-
    maplist(neg,Model,NoAsgn),
    add_clause(SolverID,NoAsgn).

neg(V,VN) :- VN is -V.

assignMultiSols(Models,FVars):-!,
    length(FVars,VarLen),
    length(SoFar,VarLen),
    maplist(=([]),SoFar),!,
    assignMultiSols(Models,SoFar,FVars).

assignMultiSols([],SoFar,SoFar):-!.
assignMultiSols([M|Models],SoFar,Vars):-!,
    addModel2Vars(M,SoFar,NVars),!,
    assignMultiSols(Models,NVars,Vars).

addModel2Vars([],[],[]).
addModel2Vars([M|Ms],[V|Vs],[[NV|V]|NVs]):-
    (M>0 -> NV=1 ; NV= -1),
    addModel2Vars(Ms,Vs,NVs).

addCnf2Solver(SolverID,Cnf,FVars):-
    term_variables(Cnf,FVars),!,
    \+ \+ (bind2index(FVars,3,FN),
           add_cnf_clauses(Cnf,SolverID),
           neg(FN,FNeg),
           add_cnf_clauses([[FN,FNeg]],SolverID)).

add_cnf_clauses([Cl|Cls],SolverID):-!,
    maplist(to_minisat_cl,Cl,MiniSatCl),
    add_clause(SolverID,MiniSatCl),
    add_cnf_clauses(Cls,SolverID).
add_cnf_clauses([],_):-!.

setup_solver(SolverID) :-
    add_clause(SolverID,[1]), % true
    add_clause(SolverID,[-2]).

to_minisat_cl(pred_true,1) :- !.
to_minisat_cl(pred_false,2) :- !.
to_minisat_cl(X,X).

bind2index([N|Ns],N,FN) :- N1 is N+1, bind2index(Ns,N1,FN).
bind2index([],N,FN):-!, FN is N - 1.
