#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/hashtrie.h"
#include "../common/msg_headers.h"
#include "../common/assert_helper.h"

#include "mstats.h"

#include "mpull.h"

#include "../common/verbose.h"
#include "../common/sequence_numbers.h"
#include "../common/magic_constants.h"


extern size_t published_message_count;
size_t published_hashes_count = 0;
size_t published_hashes_size = 0;
static int seq_num_PUB_CMDS = SEQ_NOT_SET;
static int seq_num_PUB_HASHES = SEQ_NOT_SET;

void shutdown_everything(zsock_t * const pub) {
    seq_num_PUB_CMDS = next_seq(seq_num_PUB_CMDS);
    int rc = zsock_send(pub, "ii", CTRL_TERM, seq_num_PUB_CMDS);
    assert(rc == 0);
    published_message_count++;

    // TODO: maybe wait for proxies to acknowledge
}

void teardown_worker_queuesizes(struct worker_queuesizes * print_queues) {
    int i;
    for (i = 1; i <= print_queues->len; i++) {
        zmsg_destroy(&(print_queues->queue_msgs[i]));
    }
    free(print_queues->queue_msgs);
    print_queues->len = 0;
    print_queues->update_count = 0;
}

static void dump_worker_queues(struct worker_queuesizes * const print_queues,
                        const int fd_logfile, const hashtrie trie,
                        mStats * const stats) {
    size_t checked_states = trie_get_checked_states_count(trie);
    size_t known_states = trie_count_elements(trie);
    dprintf(fd_logfile, "{:checked %lu, :known %lu, :sent %lu :hashes-sent %lu :hashes-size %lu :queues {", checked_states, known_states, published_message_count, published_hashes_count, published_hashes_size);

    int i;
    for (i = 1; i <= print_queues->len; i++) {
        dprintf(fd_logfile, ":proxy-%d {", i);
        zmsg_t *msg = print_queues->queue_msgs[i];

        if (msg != NULL) {
            char *recvd_msgs = zmsg_popstr(msg);
            dprintf(fd_logfile, ":recvd-sub %s, ",recvd_msgs);
            free(recvd_msgs);
            char *recvd_hashmsgs = zmsg_popstr(msg);
            dprintf(fd_logfile, ":recvd-subhash %s, ",recvd_hashmsgs);
            free(recvd_hashmsgs);
            char *recvd_pull= zmsg_popstr(msg);
            dprintf(fd_logfile, ":recvd-pull %s, ",recvd_pull);
            free(recvd_pull);
            char *recvd_iterations= zmsg_popstr(msg);
            dprintf(fd_logfile, ":iterations %s, ",recvd_iterations);
            free(recvd_iterations);
            char *sent_push = zmsg_popstr(msg);
            dprintf(fd_logfile, ":sent-push %s, ",sent_push);
            free(sent_push);

            dprintf(fd_logfile, ":qsizes [");

            char *qs;
            while ((qs = zmsg_popstr(msg)) != NULL) {
                dprintf(fd_logfile, "%s ", qs);
                free(qs);
            }
            dprintf(fd_logfile, "]");
        }
        zmsg_destroy(&msg);
        print_queues->queue_msgs[i] = NULL;;
        dprintf(fd_logfile, "} ");
    }
    dprintf(fd_logfile, "}}\n");
    //fsync(fd_logfile);
    print_queues->update_count = 0;

    print_mstats(stats, fd_logfile);
}

void put_wqueues_msg(struct worker_queuesizes * const print_queues,
                     zmsg_t * const msg) {
    char *proxy_id = zmsg_popstr(msg);
    int id = atoi(proxy_id);
    free(proxy_id);

    if (print_queues->len < id) {
        size_t old_len = print_queues->len; 
        print_queues->len = id;
        print_queues->queue_msgs = realloc(print_queues->queue_msgs, (id + 1) * sizeof(zmsg_t*));

        int i;
        for (i = old_len + 1; i <= id; i++) {
            print_queues->queue_msgs[i] = NULL;
        }
    }
    if (print_queues->queue_msgs[id]) {
        zmsg_destroy(&(print_queues->queue_msgs[id]));
    }
    print_queues->queue_msgs[id] = msg;
    print_queues->update_count++;
}

static int enough_states_checked(const size_t so_far, const int max_states) {
    // not enough states, if we checked none (so_far == 0)
    // not enough states, if we have to check all (max_states == 0)
    // not enough states, if so_far < max_states
    return so_far && max_states && (so_far >= max_states);
}

static int generate_successors = 1;

void send_hashes(zsock_t *pub_hashes, zlist_t *list, const int fd_logfile) {
    size_t size = zlist_size(list);
    size_t bufsize = size * HASH_INFO;
    char buf[bufsize];
    size_t i;
    for (i = 0; i < size; i++) {
        char *hash = (char *) zlist_pop(list);
        memcpy(buf + HASH_INFO * i, hash, HASH_INFO);
        free(hash);
        if (VERBOSE_MASTER_STATES) {
            char buffi[READABLE_HASH_LEN];
            print_readable_digest((unsigned char*) hash, buffi);
            dprintf(fd_logfile, "{:type :send-hash, :hash :%s}\n", buffi);
        }
    }
    seq_num_PUB_HASHES = next_seq(seq_num_PUB_HASHES);
    int rc = zsock_send(pub_hashes, "iib", CTRL_ADD_HASHES_COMPACT, seq_num_PUB_HASHES, buf, bufsize);
    published_hashes_count++;
    assert(rc == 0);
}

int master_h_pull(zsock_t * const pull,
                  zsock_t * const pub_cmds,
                  zsock_t * const pub_hashes,
                  const int strict,
                  mStats * const stats,
                  struct id_map const * const ips,
                  const int fd_logfile,
                  struct worker_queuesizes * const print_queues,
                  const int max_states,
                  hashtrie trie,
                  FILE *msg_log,
                  zlist_t *hash_accumulator) {
    int header;
    zmsg_t *msg;
    // FIXME: if every message includes the ID, we can already receive it here
    int rc = zsock_recv(pull, "im", &header, &msg);
    assert(rc == 0);


    switch(header) {
        case P_IS_HASH_MSG: {
            size_t known = trie_count_elements(trie);
            if (generate_successors && known > max_states && max_states > 0) {
                dprintf(fd_logfile, "stop succing\n");
                // stop generating successors
                generate_successors = 0;
                seq_num_PUB_CMDS = next_seq(seq_num_PUB_CMDS);
                rc = zsock_send(pub_cmds, "ii", CTRL_STOP_SUCC, seq_num_PUB_CMDS);
                assert(rc == 0);
                break;
            }
            size_t size = zmsg_size(msg);
            published_hashes_size += size;
            for (; size ; size--) {
                char *hash = zmsg_popstr(msg);
                if (VERBOSE_MASTER_STATES) {
                    char buf[READABLE_HASH_LEN];
                    print_readable_digest((unsigned char*) hash, buf);
                    dprintf(fd_logfile, "{:type :recv-hash, :hash :%s}\n", buf);
                }
                zlist_append(hash_accumulator, hash);
            }
            //seq_num_PUB_HASHES = next_seq(seq_num_PUB_HASHES);
            //rc = zsock_send(pub_hashes, "iim", CTRL_ADD_HASHES, seq_num_PUB_HASHES, msg);
            //published_hashes_count++;
            //assert(rc == 0);
            break;
        }
        case P_RESULT_MSG: {
            char *res_str;
            while ((res_str = zmsg_popstr(msg)) != NULL) {
                dprintf(fd_logfile, "{:type :result :data %s}", res_str);
                //fsync(fd_logfile);
                free(res_str);
            }

            if (strict) {
                zmsg_destroy(&msg);
                return -1;
            }
            break;
        }
        case P_TRANSFER_DONE:
        case PSTATS_HAS_FULL:
        case PSTATS_HAS_EMPTY:
        case PSTATS_COMPLETELY_EMPTY:
        case PSTATS_HAS_PMS: {
            //fprintf(msg_log, "HEADER: %d\n", header);
            //zmsg_print(msg);
            zmsg_t *dup = zmsg_dup(msg);
            update_queue_digests(stats, header, dup, fd_logfile);
            seq_num_PUB_CMDS = m_do_balance(stats, seq_num_PUB_CMDS, pub_cmds, (char const * const *) ips->id_map, fd_logfile);
            break;
        }
        case PSTATS_QUEUES: {
            zmsg_t *dup = zmsg_dup(msg);
            put_wqueues_msg(print_queues, dup);
            // hopefully good enough rule of thumb
            if (print_queues->update_count >= 4*print_queues->len) {
                dump_worker_queues(print_queues, fd_logfile, trie, stats);
            }
            break;
        }
        default: 
            assert_esf(0, "Protocol error: unknown header %d", header);
    }
    zmsg_destroy(&msg);

    size_t checked_states = trie_get_checked_states_count(trie);
    if ((stats_can_terminate(stats) && checked_states > 0)
            || enough_states_checked(checked_states, max_states)) {
        dprintf(fd_logfile, "can terminate; checked: %ld, max states: %d\n", checked_states, max_states);
        return -1;
    }
    return 0;
}

zsock_t *master_setup_pull(const int port) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "tcp://*:%d", port);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_pull(endpoint);
    free(endpoint);
    return zocket;
}

zsock_t *master_setup_pub(const int port) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "tcp://*:%d", port);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_pub(endpoint);
    free(endpoint);
    return zocket;
}

int mpull_reader(zloop_t *loop, zsock_t *reader, void *args) {
    struct mpull_arg *arg = (struct mpull_arg*) args;

    return master_h_pull(reader, arg->pub_cmds, arg->pub_hashes, arg->strict, arg->stats, arg->ips, arg->fd_logfile, arg->worker_queues, arg->max_states, arg->trie, arg->msg_log, arg->hash_accumulator);
}
