#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdio.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/magic_constants.h"
#include "../common/msg_headers.h"
#include "../common/hashtrie.h"
#include "pstats.h"
#include "p_pull_push.h"
#include "p_pub_sub.h"
#include "../common/verbose.h"


#include "../common/get_timestamp.h"
#include "../common/assert_helper.h"
#include "../common/helpers.h"
#include "../common/sequence_numbers.h"



extern size_t sub_message_count;
extern size_t subhashes_message_count;
static int last_seq_num_SUB_master = SEQ_NOT_SET;
static int last_seq_num_SUB_stream = SEQ_NOT_SET;
static int seq_num_PUB_GLOBAL = SEQ_NOT_SET;
static int seq_num_PUB_LOCAL  = SEQ_NOT_SET;

void put_hashes_compact(zmsg_t *msg, const hashtrie trie, int fd_logfile) {
    assert(zmsg_size(msg) == 1);
    zframe_t *frame = zmsg_pop(msg);
    size_t size = zframe_size(frame);
    char *ptr = (char *) zframe_data(frame); // still owned by frame

    size_t i;
    for (i = 0; i < size; i += HASH_INFO) {
        trie_put(trie, ptr + i, ptr + i + HASH_LEN);
        if (VERBOSE_PROXY_STATES) {
            char buffi[READABLE_HASH_LEN];
            print_readable_digest((unsigned char*) ptr + i, buffi);
            dprintf(fd_logfile, "{:type :recv-hash, :hash :%s}\n", buffi);
        }
    }

    zframe_destroy(&frame);
}

void put_hashes(zmsg_t *msg, const hashtrie trie, int fd_logfile) {
    zframe_t *frame;
    while ((frame = zmsg_pop(msg)) != NULL) {
        char *new_digest = zframe_strdup(frame);
        if (zframe_size(frame) != HASH_INFO) {
            // print failing message frame
            int i;
            for (i = 0; i < zframe_size(frame); i++) {
                printf("%c", new_digest[i]);
            }
            puts("");
            for (i = 0; i < zframe_size(frame); i++) {
                printf("%d ", new_digest[i]);
            }
            puts("");
            assert(zframe_size(frame) == HASH_INFO);
        }
        if (VERBOSE_PROXY_STATES) {
            char buf[READABLE_HASH_LEN];
            print_readable_digest((unsigned char*) new_digest, buf);
            dprintf(fd_logfile, "{:type :recv-hash, :hash :%s}\n", buf);
        }
        trie_put(trie, new_digest, new_digest + HASH_LEN);
        free(new_digest);
        zframe_destroy(&frame);
    }
}

void balance(zsock_t * const pub,
            pStats * const wqs,
            const size_t thresh,
            const int header,
            const int index_has,
            const char *endpoint,
            int fd_logfile,
            uint64_t timestamp) {
    assert(index_has != -1);
    ddprintf(VERBOSE_TRANSFER, fd_logfile, "{:type :transfer, :from %d, :to \"%s\", :timestamp %ld}\n", index_has, endpoint, timestamp);
    seq_num_PUB_LOCAL = next_seq(seq_num_PUB_LOCAL);
    int rc = zsock_send(pub, "iii8s", header, seq_num_PUB_LOCAL, index_has, timestamp, endpoint);
    assert(rc == 0);
}

void fake_transfer(char const * const endpoint) {
    zsock_t *zock = zsock_new_req(endpoint);
    assert(zock);
    
    int rc = zsock_send(zock, "i", GLOBAL_TRANSFER);
    assert(rc == 0);

    int peer_ack_header;
    rc = zsock_recv(zock, "im", &peer_ack_header, NULL);
    assert(rc == 0);
    assert(peer_ack_header == TRANSFER_ACK);


    rc = zsock_disconnect(zock, "%s", endpoint);
    assert(rc == 0);
    zsock_destroy(&zock);
}

int handle_sub(zsock_t * const sub,
               const int is_stream_sub,
               zsock_t * const pub_local,
               zsock_t * const pub_global,
               zsock_t * const push,
               pStats * const stats,
               const int id,
               const size_t thresh,
               const int port_start,
               const int fd_logfile,
               const hashtrie trie,
               const int proxynumber) {
    int header;
    zmsg_t *msg;
    int seq_num;
    int rc = zsock_recv(sub, "iim", &header, &seq_num, &msg);
    if (is_stream_sub) {
        assert(verify_seq(seq_num, last_seq_num_SUB_stream));
        last_seq_num_SUB_stream = seq_num;
    } else {
        assert(verify_seq(seq_num, last_seq_num_SUB_master));
        last_seq_num_SUB_master = seq_num;
    }
    assert(rc == 0);
    sub_message_count++;

    switch (header) {
        case CTRL_ADD_HASHES_COMPACT: {
            subhashes_message_count++;
            zmsg_t *dup = zmsg_dup(msg); // forward hashes
            
            seq_num_PUB_GLOBAL = next_seq(seq_num_PUB_GLOBAL);
            rc = zsock_send(pub_global, "iim", CTRL_ADD_HASHES_COMPACT, seq_num_PUB_GLOBAL, dup);
            assert(rc == 0);
            zmsg_destroy(&dup);

            put_hashes_compact(msg, trie, fd_logfile);
            break;
        }
        case CTRL_ADD_HASHES: {
            subhashes_message_count++;
            zmsg_t *dup = zmsg_dup(msg); // forward hashes

            seq_num_PUB_GLOBAL = next_seq(seq_num_PUB_GLOBAL);
            rc = zsock_send(pub_global, "iim", CTRL_ADD_HASHES, seq_num_PUB_GLOBAL, dup);
            assert(rc == 0);
            zmsg_destroy(&dup);

            put_hashes(msg, trie, fd_logfile);
            break;
        }
        case CTRL_TERM:
            seq_num_PUB_LOCAL = next_seq(seq_num_PUB_LOCAL);
            rc = zsock_send(pub_local, "ii", P_TERM, seq_num_PUB_LOCAL);
            assert(rc == 0);
            zmsg_destroy(&msg);
            return -1;
        case CTRL_TRANSFER_TO: {
            char *target_id_s = zmsg_popstr(msg);
            int target_id = atoi(target_id_s);

            char *timestamp_s = zmsg_popstr(msg);
            size_t timestamp;
            rc = sscanf(timestamp_s, "%lu", &timestamp);
            assert(rc == 1);
            ddprintf(VERBOSE_TRANSFER, fd_logfile, "#transfer: target_id \"%s\" or as an int %d. My ID is %d. Timestamp is \"%s\" or %ld as an size_t.\n",
                                                   target_id_s, target_id, id, timestamp_s, timestamp);
            free(timestamp_s);
            
            free(target_id_s);

            if (id == target_id)  {
                int amount = zmsg_size(msg);
                int i;
                for (i = 0; i < amount; i++) {
                    char *endpoint = zmsg_popstr(msg);
                    int index_has = get_max_not_transfer_queue(stats);
                    if (index_has != -1) {
                        // a worker can balanced
                        balance(pub_local, stats, thresh, G_TRANSFER_TO, index_has, endpoint, fd_logfile, timestamp);
                    } else {
                        index_has = get_max_queue(stats);
                        if (index_has != -1) {
                            balance(pub_local, stats, thresh, G_TRANSFER_TO, index_has, endpoint, fd_logfile, timestamp);
                        } else {
                            p_send_transfer_done(push, id, stats, port_start, proxynumber, fd_logfile, timestamp);
                            fake_transfer(endpoint);
                        }
                       

                    }
                    free(endpoint);
             
                }
            }
            break;
        }
        case CTRL_TRANSFER_TERM:
            // TODO: initiate more transfers, count them
            assert_es(0, "Command CTRL_TRANSFER_TERM not implemented.");
            break;
        case CTRL_STOP_SUCC:
            dprintf(fd_logfile, "{:type :info, :recvd-msg :stop-succing}\n");
            seq_num_PUB_LOCAL = next_seq(seq_num_PUB_LOCAL);
            rc = zsock_send(pub_local, "ii", P_STOP_SUCC, seq_num_PUB_LOCAL);
            assert(rc == 0);

            seq_num_PUB_GLOBAL = next_seq(seq_num_PUB_GLOBAL);
            rc = zsock_send(pub_global, "ii", P_STOP_SUCC, seq_num_PUB_GLOBAL);
            assert(rc == 0);
            break;
        default: 
            printf("wtf is this message type: %d\n", header);
            assert(0);
    }
    zmsg_destroy(&msg);
    return 0;
}

int local_balance(zsock_t * const pub,
                  pStats * const wqs,
                  const size_t thresh,
                  const int port_start,
                  const int fd_logfile,
                  const int proxynumber) {
    int nmbr_transfers = proxy_can_balance(wqs, thresh);
    size_t timestamp = get_sizettime();
    
    int i;
    for (i = 0; i < nmbr_transfers; i++) {
        int index_need = get_empty_queue(wqs);
        int index_has = get_max_not_transfer_queue(wqs);
        assert(index_has != -1);
        mark_transfer_begin(wqs, index_need, thresh, fd_logfile);
        mark_transfer_begin(wqs, index_has, thresh, fd_logfile);
        ddprintf(VERBOSE_TRANSFER, fd_logfile, "{:type :transfer-flag, :worker %d}\n", index_need);
        ddprintf(VERBOSE_TRANSFER, fd_logfile, "{:type :transfer-flag, :worker %d}\n", index_has);
        
        char *host = "localhost";
        int port = wsink_port(port_start, proxynumber, index_need);
        // TODO: this should be calculated by the same / similar code as the port in wsink.c and proxy.c
        char *endpoint;
        int rc = asprintf(&endpoint, "tcp://%s:%d", host, port);
        assert(rc != -1);

        balance(pub, wqs, thresh, L_TRANSFER_TO, index_has, endpoint, fd_logfile, timestamp);
        free(endpoint);
    }
    return nmbr_transfers;
}


int proxy_sub_reader(zloop_t *loop, zsock_t *reader, void *args) {
    struct p_sub_args *arg = (struct p_sub_args*) args;

    return handle_sub(reader, arg->is_stream_sub, arg->pub_local, arg->pub_global, arg->push,
                      arg->stats, arg->proxy_id, arg->thresh, arg->port_start,
                      arg->fd_logfile, arg->trie, arg->proxynumber);

}

zsock_t *proxy_setup_pub(const int proxy_id) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "ipc:///tmp/zmq-proxy-%d-pubsub", proxy_id);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_pub(endpoint);
    free(endpoint);
    return zocket;
}

zsock_t *proxy_setup_pub_global(const int port_number) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "tcp://*:%d", port_number);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_pub(endpoint);
    free(endpoint);
    assert(zocket);
    return zocket;
}

zsock_t *proxy_setup_sub(char const * const endpoint) {
    zsock_t *zocket = zsock_new_sub(endpoint, "");
    return zocket;
}

zsock_t *proxy_setup_sub2(char const * const ip, const int port) {
    char *endpoint = NULL;
    int rc = asprintf(&endpoint, "tcp://%s:%d", ip, port);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_sub(endpoint, "");
    free(endpoint);
    assert(zocket);
    return zocket;
}

void proxy_add_subscription(zsock_t *sub, char const * const master_ip, int port) {
    int rc = zsock_connect(sub, "tcp://%s:%d", master_ip, port);
    assert(rc == 0);
}
