% (c) 2017-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
:- module(alloy2b, [load_alloy_ast_prolog_file/1,
                    load_alloy_model/2,
                    load_alloy_model/3,
                    translate_alloy_model/2]).

:- use_module(probsrc(debug)).
:- use_module(library(lists)).
:- use_module(library(plunit)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(translate)).
:- use_module(library(file_systems)).
:- use_module(probsrc(error_manager)).
:- use_module(library(sets), [subtract/3]).
:- use_module(probsrc(specfile), [set_animation_minor_mode/1]).
:- use_module(library(ordsets), [ord_intersection/3,ord_union/3,ord_subtract/3]).

:- dynamic enumerated_set/1,singleton_set/1,total_function/2,ordered_signature/1,command_counter/1,extending_signatures/1,is_sub_signature_of/3,is_seq/1,current_command/1,maxseq/1,uses_seqs/0.
:- volatile enumerated_set/1,singleton_set/1,total_function/2,ordered_signature/1,command_counter/1,extending_signatures/1,is_sub_signature_of/3,is_seq/1,current_command/1,maxseq/1,uses_seqs/0.

:- meta_predicate map_translate(3, -, -, -).

command_counter(0).

% An automated translation from Alloy to classical B.
% The Alloy abstract syntax tree is translated to an untyped B AST as supported by ProB.
% Afterwards, the untyped AST can be typechecked and loaded by ProB.
% The used positions are from the Alloy parser and thus refer to the Alloy model.
% Usage:
% load_alloy_model/2: input is an Alloy model represented as a Prolog term as provided by the Kotlin translation; the model is translated to an untyped B AST and loaded by ProB
% translate_alloy_model/2: input is an Alloy model represented as a Prolog term as provided by the Kotlin translation, output is an untyped B AST
% Run tests:
% - launch ProB Tcl/Tk or ProB CLI
% - 'use_module(extension('alloy2b/alloy2b')).'
% - 'error_manager:reset_errors , plunit:run_tests(full_machine_translation).'.
% - run a single test: 'error_manager:reset_errors , plunit:run_tests(full_machine_translation:directctrl).'
% - 'bmachine:full_b_machine(L) , translate:print_machine(L)'.
% KNOWN ISSUES / TO DOs
% ----------------------
% - nested quantifiers: the usage of facts for singleton_set is error prone, and cannot deal with
% nested quantifiers with variable clashes, in particular when in the inner quantifier the
% identifier is not a singleton set
% - multiplicity annotations in signatures and nested multiplicity annotations for non-binary relations
% - variable capture due to use of DEFINITIONS
% - ProB does not allow arithmetic operators on non-singleton sets;
% this is a design decision to detect most probably unintended uses of arithmetic
% - ProB does use mathematical integers, bit-width is used for MAXINT and MININT, but only influences
% enumeration scopes
load_alloy_ast_prolog_file(File) :-
    \+ file_exists(File, read),
    !,
    add_error(load_alloy_ast_prolog_file, 'Alloy AST file has not been created:', [File]),
    fail.
load_alloy_ast_prolog_file(File) :-
    formatsilent('Opening Alloy AST file: ~w~n', [File]),
    open(File, read, Stream),
    safe_read_term(Stream, AlloyTerm),
    close(Stream),
    load_alloy_model(AlloyTerm, 'alloytranslation').

safe_read_term(Stream, Term) :-
    on_exception(error(E,_), read(Stream, Term),
                 (   add_internal_error('Exception while reading Prolog term from stream:', E),
                     fail
                 )).

load_alloy_model(AlloyTerm, AlloyFile) :-
    % use the first run or check command by default for models that use sequences, otherwise all commands are translated
    load_alloy_model(0, AlloyTerm, AlloyFile).

load_alloy_model(CurrentCommand, AlloyTerm, AlloyFile) :-
    integer(CurrentCommand),
    reset_errors,
    debug_format(19, 'Translating AST to B.~n', []),
    retract_state,              % in case things not properly reseted or initialised
    assert(current_command(CurrentCommand)),
    % tools_printing:nested_print_term(AlloyTerm,10),
    translate_alloy_model(AlloyTerm, UntypedBAst),
    !,
    UntypedBAst = machine(generated(none,AbstractMachine)),
    b_load_machine_from_term(AlloyFile, complete_machine('alloytranslation',[AbstractMachine],[AlloyFile])),
    set_animation_minor_mode(alloy).
load_alloy_model(CurrentCommand, _, _) :-
    \+ integer(CurrentCommand),
    add_error(load_alloy_model, 'Current command has to be an integer:', CurrentCommand),
    fail.
load_alloy_model(_, AlloyTerm, _) :-
    add_error(load_alloy_model, 'Translating Alloy AST failed:', AlloyTerm),
    fail.

process_options(Options) :-
    member(sequences:B, Options),
    B == true,
    !,
    asserta(uses_seqs).
process_options(_).

% TODO: handle all modules in the model but exclude integer, relation, ordering and sequences
% Currently we only extract and translate the main module
translate_alloy_model(alloy(RootModule,Models), UntypedBAst) :-
    member(alloy_model(RootModule,Facts,Assertions,Commands,Functions,Signatures,OSigNames,Options), Models),
    !,
    translate_alloy_model(alloy_model(RootModule,Facts,Assertions,Commands,Functions,Signatures,OSigNames,Options), UntypedBAst).

% TODO: use atelierb_pp_mode (e.g. we use let expressions which are not supported by native B)
translate_alloy_model(alloy_model(_ModuleName,facts(Facts),assertions(Assertions),commands(Commands),functions(Functions),
                                  signatures(Signatures),ordered_signatures(OSigNames),Options), UntypedBAst) :-
    % singleton sets are asserted at runtime using singleton_set/1
    % accumulate all translations, afterwards we build the untyped machine ast
    process_options(Options),
    maplist(assert_ordered_sig_name, OSigNames),
    empty_machine_acc(MAcc),
    debug_println(19, translating_signatures),
    maplist(preprocess_signature, Signatures),                       % assert some signature names as singleton
    map_translate(translate_signature, MAcc, Signatures, MAcc0),
    debug_println(19, translating_assertions),
    map_translate(translate_assertion, MAcc0, Assertions, MAcc1),
    debug_println(19, translating_commands),
    % TO DO: extract common scopes for all commands, to set deferred set sizes:
    % NOTE: extract_common_scopes can be removed in case we stick to translating a single command at a time
    extract_common_scopes(Commands, GS, ExactScopes, UpBoundScopes, BitWidth, TMaxSeq, Con),
    (   TMaxSeq = maxSeq(MaxSeq)
    ;   MaxSeq = -1
                  ),
    debug_format(19, 'Common scopes to all commands : global=~w, exact=~w, bw=~w, ms=~w, conflicts=~w~n', [GS,ExactScopes,BitWidth,MaxSeq,Con]),
    (   number(GS),
        GS > 0 ->
        preferences:temporary_set_preference(globalsets_fdrange, GS) % TO DO: improve
    ;   true
    ),
    maplist(generate_exact_scope_predicate, ExactScopes, ScopePreds1),
    maplist(generate_leq_scope_predicate, UpBoundScopes, ScopePreds2),
    append(ScopePreds1, ScopePreds2, ScopePreds),
    extend_machine_with_conjuncts(properties, ScopePreds, MAcc1, MAcc2),
    (   BitWidth = bitwidth(BW),
        number(BW),
        BW > 0,
        MaxInt is integer(2**(BW-1))-1,
        MinInt is integer(-(2**(BW-1))),
        preferences:temporary_set_preference(maxint, MaxInt),
        preferences:temporary_set_preference(minint, MinInt)
    ->
        true
    ;   true
    ),
    (   MaxSeq \= -1
                   ->
        translate_seq_scopes(MaxSeq, SeqScopesPred),
        extend_machine_acc(properties, MAcc2, SeqScopesPred, MAcc3)
    ;   MAcc3 = MAcc2
    ),
    map_translate(translate_command, MAcc3, Commands, MAcc4),
    debug_println(19, translating_functions),
    map_translate(translate_function, MAcc4, Functions, MAcc5),
    debug_println(19, translating_facts),
    map_translate(translate_fact, MAcc5, Facts, MAcc6),
    !,
    finalize_extending_signatures(MAcc6, MAcc7),
    debug_println(19, build_machine_ast),
    build_machine_ast(MAcc7, UntypedBAst),
    % print(UntypedBAst),nl,nl,
    % tools_printing:nested_print_term(UntypedBAst,20),
    (   debug_mode(off) ->
        true
    ;   nl,
        translate:print_raw_machine_terms(UntypedBAst),
        nl
    ),
    retract_state.

map_translate(_, MAcc, [], MAcc).
map_translate(TypePred, MAcc, [Part|T], Res) :-
    call(TypePred, MAcc, Part, NewMAcc),
    map_translate(TypePred, NewMAcc, T, Res).

% things that need to be done before translating the signatures, e.g., assert singleton set info
preprocess_signature(signature(Name,_Fields,_Facts,Options,_)) :-
    % assert signatures for singleton checks
    assert_singleton_set(Options, Name),
    assert_enumerated_set(Options, Name).

% translate signature declaration sig Name {Fields}
translate_signature(MAcc, signature(Name,Fields,Facts,Options,pos(Col,Row)), NewMAcc) :-
    extend_machine_acc(signatures, MAcc, [Name], MAcc1),
    translate_signature_aux(Name, Options, pos(Col,Row), MAcc1, MAcc2),
    translate_signature_fields(Fields, Name, MAcc2, MAcc3),
    translate_e_p(Name, TName),
    map_translate(translate_signature_fact(TName), MAcc3, Facts, NewMAcc).

translate_signature_aux(Name, Options, Pos, MAcc, NewMAcc) :-
    % ordered signatures are defined as distinct sets of integer when translating a command
    member(ordered, Options),
    !,
    define_ordered_signature_functions(Pos, MAcc, Name, MAcc1),
    extend_machine_acc(signatures, MAcc1, [Name], MAcc2),
    translate_e_p(Name, TID),
    extend_machine_acc(properties, MAcc2,
                       member(none,TID,pow_subset(none,integer_set(none,'INTEGER'))), MAcc3),
    define_sig_as_set_or_constant_aux(constants, MAcc3, Name, Options, Pos, NewMAcc).
translate_signature_aux(Name, Options, Pos, MAcc, NewMAcc) :-
    define_sig_as_set_or_constant(MAcc, Name, Options, Pos, NewMAcc).

define_ordered_signature_functions(POS, MAcc, Sig, NewMAcc) :-
    translate_pos(POS, BPOS),
    gen_identifier(0, "_c_", TIDX),
    TIDS = identifier(BPOS,s),
    translate_e_p(Sig, TSig),
    MemberX = member(BPOS,TIDX,TSig),
    % next_Sig(s)  == IF succ[s] <: Sig THEN succ[s] ELSE {} END
    atom_concat(next_, Sig, NextName),
    SuccessorImg = image(BPOS,successor(BPOS),TIDS),
    PredecessorImg = image(BPOS,predecessor(BPOS),TIDS),
    extend_machine_acc(definitions, MAcc,
                       expression_definition(BPOS,NextName,[TIDS],if_then_else(BPOS,subset(BPOS,SuccessorImg,TSig),SuccessorImg,empty_set(BPOS))), MAcc1),
    % nexts_Sig(s) == {x|x:Sig & min({x} \/ s) /: {x}}
    atom_concat(nexts_, Sig, NextsName),
    NextsBody = conjunct(BPOS,MemberX,not_member(BPOS,min(BPOS,union(BPOS,set_extension(BPOS,[TIDX]),TIDS)),set_extension(BPOS,[TIDX]))),
    extend_machine_acc(definitions, MAcc1,
                       expression_definition(BPOS,NextsName,[TIDS],comprehension_set(BPOS,[TIDX],NextsBody)), MAcc2),
    atom_concat(prev_, Sig, PrevName),
    % prev_Sig(s)  == IF pred[s] <: Sig THEN pred[s] ELSE {} END
    extend_machine_acc(definitions, MAcc2,
                       expression_definition(BPOS,PrevName,[TIDS],if_then_else(BPOS,subset(BPOS,PredecessorImg,TSig),PredecessorImg,empty_set(BPOS))), MAcc3),
    % prevs_Sig(s) == {x|x:Sig & max({x} \/ s) /: {x}}
    atom_concat(prevs_, Sig, PrevsName),
    PrevsBody = conjunct(BPOS,MemberX,not_member(BPOS,max(BPOS,union(BPOS,set_extension(BPOS,[TIDX]),TIDS)),set_extension(BPOS,[TIDX]))),
    extend_machine_acc(definitions, MAcc3,
                       expression_definition(BPOS,PrevsName,[TIDS],comprehension_set(BPOS,[TIDX],PrevsBody)), NewMAcc).

translate_signature_fields(Fields, SignatureName) -->
    { translate_e_p(SignatureName, TSignatureName) },
    translate_signature_fields_aux(Fields, SignatureName, TSignatureName).

translate_signature_fields_aux([], _, _) -->
    [].
translate_signature_fields_aux([Field|T], SignatureName, TSignatureName) -->
    translate_signature_field(SignatureName, TSignatureName, Field),
    translate_signature_fields_aux(T, SignatureName, TSignatureName).

translate_signature_field(SignatureName, TSignatureName, Field, MAcc, NewMAcc) :-
    translate_signature_field_aux(SignatureName, TSignatureName, Field, TField, MAcc, MAcc1),
    extend_machine_acc(properties, MAcc1, TField, NewMAcc).

translate_signature_field_aux(SignatureName, TSignatureName,
                              field(FieldName,Expr,FieldType,Options,POS), TFieldPred, MAcc, NewMAcc) :-
    FieldType = type(TypeList,Arity),
    A1 is Arity+1,              % add SignatureName to type:
    FieldID = identifier(FieldName,type([SignatureName|TypeList],A1),POS),
    translate_e_p(FieldName, TFieldName),
    extend_machine_acc(constants, MAcc, TFieldName, MAcc1),
    translate_pos(POS, BPOS),
    disjoint_field_declaration(TSignatureName, BPOS, Options, MAcc1, TFieldName, NewMAcc),
    translate_field_decl(BPOS, SignatureName, TSignatureName, Expr, FieldID, TFieldName, TFieldPred),
    !.

translate_function_field(field(FieldName,Expr,type(_Type,_Arity),_Options,POS), TField) :-
    % (Functor = function -> assert_singleton_set(FieldName) ; true) , % NO LONGER NEEDED
    translate_e_p(FieldName, TTFieldName),
    strip_singleton_set_from_this(TTFieldName, TFieldName), % necessary for nested definition calls with 'this'
    translate_pos(POS, BPOS),
    fun_or_pred_decl_special_cases(Expr, TFieldName, BPOS, TField, _IsSingleton),
    !.
% TO DO: !! what if IsSingleton is set !!
disjoint_field_declaration(TSignatureName, TPos, Options, MAcc, TFieldName, NewMAcc) :-
    member(disj, Options),
    !,
    % all a, b: S | a != b implies no a.f & b.f
    IdA = identifier(TPos,a),
    IdB = identifier(TPos,b),
    ImplLhs = conjunct(TPos,conjunct(TPos,member(TPos,IdA,TSignatureName),member(TPos,IdB,TSignatureName)),
                       not_equal(TPos,IdA,IdB)),
    ImplRhs = equal(TPos,intersection(TPos,image(TPos,TFieldName,set_extension(none,[IdA])),
                                      image(TPos,TFieldName,set_extension(none,[IdB]))),empty_set(TPos)),
    Disjoint = forall(TPos,[IdA,IdB],implication(TPos,ImplLhs,ImplRhs)),
    extend_machine_acc(properties, MAcc, Disjoint, NewMAcc).
disjoint_field_declaration(_, _, _, MAcc, _, MAcc).


is_disjoint_quantifier(Fields) :-
    Fields = [F|_],
    is_disjoint_quantifier_field(F). % assumption: all or no field is disjoint
is_disjoint_quantifier_field(field(_FieldName,_Expr,type(_Type,_Arity),Options,_POS)) :-
    member(disj, Options).

translate_quantifier_field(field(FieldName,Expr,type(_Type,_Arity),_Options,POS), TField, FieldName) :-
    translate_pos(POS, BPOS),
    fun_or_pred_decl_special_cases(Expr, TFieldName, BPOS, TField, IsSingleton),
    !,
    (   IsSingleton = singleton ->
        assert_singleton_set(FieldName)
    ;   true
    ),
    translate_e_p(FieldName, TFieldName).


translate_fact(MAcc, fact(Expr,_Pos), NewMAcc) :-
    translate_e_p(Expr, TExpr),
    extend_machine_acc(properties, MAcc, TExpr, NewMAcc).
translate_assertion(MAcc, fact(Expr,_Pos), NewMAcc) :-
    translate_e_p(Expr, TExpr),
    extend_machine_acc(assertions, MAcc, TExpr, NewMAcc).

% Signature facts: Identifiers may refer to the signature and are joined with 'this'.
% We then need a universal quantification using 'this'.
translate_signature_fact(TSignatureName, MAcc, Expr, NewMAcc) :-
    alloy_expr_contains_join(Expr),
    translate_e_p(Expr, TExpr), % we quantify in such a way that this is a singleton_set_id; this is always considered a singleton set id
    ThisID = identifier(none,'this'),
    TImplication = implication(none,subset(none,set_extension(none,[ThisID]),TSignatureName),TExpr),
    FORALL = forall(none,[ThisID],TImplication),
    extend_machine_acc(properties, MAcc, FORALL, NewMAcc).
translate_signature_fact(_TSignatureName, MAcc, Expr, NewMAcc) :-
    translate_e_p(Expr, TExpr),
    extend_machine_acc(properties, MAcc, TExpr, NewMAcc).

% TO DO: write generic AST traversal predicate
alloy_expr_contains_join(AndOr) :-
    AndOr =.. [Functor,ListOfAsts,_Pos],
    !,
    member(Functor, [and,or]),
    member(Ast, ListOfAsts),
    alloy_expr_contains_join(Ast),
    !.
alloy_expr_contains_join(identifier('this',_,_)).
% alloy_expr_contains_join_with_this(join(Arg1,Arg2,_Type,_Pos)) :-
% Arg1 = identifier('this',_,_) ; Arg2 = identifier('this',_,_).
alloy_expr_contains_join(Expr) :-
    Expr =.. [_Functor,Arg1,Arg2,_Type,_Pos],
    !,
    (   alloy_expr_contains_join(Arg1) ->
        true
    ;   alloy_expr_contains_join(Arg2)
    ).
alloy_expr_contains_join(Expr) :-
    Expr =.. [_Functor,Arg,_Type,_Pos],
    alloy_expr_contains_join(Arg).
alloy_expr_contains_join(Expr) :-
    Expr =.. [Functor,_Params,Fields,Body,_Type,_POS],
    member(Functor, [all,no,some,one,lone,comprehension]),
    (   alloy_expr_contains_join(Body) ->
        true
    ;   member(F, Fields),
        alloy_expr_contains_join(F) ->
        true
    ).

% translate fun NAME Decls { Body }  or pred Name Decls { Body }
translate_function(MAcc, FunctionOrPredicate, NewMAcc) :-
    FunctionOrPredicate =.. [Functor,Name,Params,Decls,Body,POS],
    alloy_to_b_operator(Functor, BFunctor),
    maplist(translate_e_p, Params, TParams),
    % identifier like 'this' are singleton sets, parameters have to be identifiers
    maplist(strip_singleton_set, TParams, TTParams),
    maplist(translate_function_field, Decls, TDecls),
    translate_e_p(Body, TBody),
    translate_pos(POS, BPOS),
    (   Functor == function
    ->
        (   TDecls == []
        ->
            TResult = TBody     % no need to create set comprehension
        ;                       % create {temp | TDecls & temp : TBody }
            append(TDecls, [member(none,identifier(none,'_fun_res_'),TBody)], BehaviorList),
            join_untyped_ast_nodes(conjunct, BehaviorList, Behavior),
            TResult = comprehension_set(BPOS,[identifier(none,'_fun_res_')],Behavior)
        ),
        UAst =.. [BFunctor,BPOS,Name,TTParams,TResult]
    ;   append(TDecls, [TBody], BehaviorList),
        join_untyped_ast_nodes(conjunct, BehaviorList, Behavior),
        UAst =.. [BFunctor,BPOS,Name,TTParams,Behavior]
    ),
    extend_machine_acc(definitions, MAcc, UAst, NewMAcc).

% extract scopes common to all commands
extract_common_scopes([Cmd|T], GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts) :-
    get_scopes(Cmd, GlobalScopes1, ExactScopes1, UpBound1, BitWidth1, MaxSeq1),
    !,
    extract_common_aux(T, GlobalScopes1, ExactScopes1, UpBound1, BitWidth1, MaxSeq1, [],
                       GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts).
extract_common_scopes(_, -1, [], [], none, none, []).
extract_common_aux([], GS, ES, UpS, BW, MS, C, GS, ES, UpS, BW, MS, C).
extract_common_aux([Cmd|T], GlobalScope0, ExactScopes0, UpBoundScopes0, BitWidth0, MaxSeq0, Conflicts0,
                   GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts) :-
    get_scopes(Cmd, GlobalScope1, ExactScopes1, UpBoundScopes1, BitWidth1, MaxSeq1),
    (   GlobalScope0 = GlobalScope1 ->
        GlobalScope2 = GlobalScope0
    ;   GlobalScope2 = none
    ),
    (   BitWidth0 = BitWidth1 ->
        BitWidth2 = BitWidth1
    ;   BitWidth2 = none
    ),
    (   MaxSeq0 = MaxSeq1 ->
        MaxSeq2 = MaxSeq1
    ;   MaxSeq2 = none
    ),
    join_scopes(ExactScopes0, ExactScopes1, ExactScopes2, Conflicts1),
    ord_union(Conflicts0, Conflicts1, Conflicts2),
    join_scopes(UpBoundScopes0, UpBoundScopes1, UpBoundScopes2, Conflicts3),
    ord_union(Conflicts2, Conflicts3, Conflicts4),
    !,
    extract_common_aux(T, GlobalScope2, ExactScopes2, UpBoundScopes2, BitWidth2, MaxSeq2, Conflicts4,
                       GlobalScopes, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, Conflicts).

join_scopes(S1, S2, Inter, Conflicts) :-
    sort(S1, SS1),
    sort(S2, SS2),
    ord_intersection(SS1, SS2, Inter),
    ord_union(SS1, SS2, Union),
    ord_subtract(Union, Inter, Conflicts).

get_scopes(Command, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, MaxSeq) :-
    Command =.. [_Functor,_Body,global_scope(GlobalScope),
                 exact_scopes(ExactScopes),upper_bound_scopes(UpBoundScopes),BitWidth,MaxSeq|_],
    !,
    debug_format(19, 'Command scopes: global=~w, exact=~w, upbound=~w,bitwidth=~w~n,maxseq=~w~n',
                 [GlobalScope,ExactScopes,UpBoundScopes,BitWidth,MaxSeq]).
get_scopes(Command, -1, [], [], none, none) :-
    add_error(alloy2b, 'Cannot process Alloy Command:', Command).

translate_command(MAcc, Command, NewMAcc) :-
    Command =.. [Functor,Body,GlobalScope,ExactScopes,UpBoundScopes,BitWidth,MaxSeq,index(CommandIndex),POS],
    % either a model uses no sequences and we translate all commands or only the chosen one
    (   \+ uses_seqs
           ;   (   current_command(CurrentIndex),
                   CurrentIndex == CommandIndex
               )
    ),
    !,
    translate_command_aux(MAcc, Functor, Body, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, POS, NewMAcc).
translate_command(MAcc, Command, MAcc) :-
    % only translate the current command if using sequences (either selected by the user or the first one by default)
    Command =.. [_,_,_,_,_,_,_,index(CommandIndex),_],
    current_command(CurrentIndex),
    CurrentIndex =\= CommandIndex,
    !.
translate_command(MAcc, Command, MAcc) :-
    add_error('Unknown Alloy command (be sure Alloy2B library is up-to-date):', Command).

translate_command_aux(MAcc, Functor, Body, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, POS, NewMAcc) :-
    debug_println(19, translating_command(Functor,GlobalScope,ExactScopes)),
    (   Functor = check
    ;   Functor = run
    ),
    translate_e_p(Body, TBody),
    % we need all signature names to define the global scope
    get_signature_names_from_machine_acc(MAcc, SignatureNames),
    debug_println(19, translate_scopes(SignatureNames)),
    translate_scopes(SignatureNames, GlobalScope, ExactScopes, UpBoundScopes, BitWidth, MaxSeq, TScopesPred),
    translate_pos(POS, BPOS),
    Precondition = conjunct(BPOS,TScopesPred,TBody),
    get_command_counter_atom(CommandCounter),
    atom_concat(Functor, CommandCounter, OperationName),
    Operation = operation(BPOS,identifier(BPOS,OperationName),[],[],precondition(BPOS,Precondition,skip(none))),
    extend_machine_acc(operations, MAcc, Operation, NewMAcc),
    MaxSeq = maxseq(MaxSeqSize),
    assert(maxseq(MaxSeqSize)).

% global scope, exact scopes and bitwidth
% We do not need to set the bitwidth since we have real integers in B.
% Note: If only one command is defined in the Alloy model we could set min and max int of ProB in the definitions.
translate_scopes(SignatureNames, global_scope(GlobalScope), exact_scopes(ExactScopes), upper_bound_scopes(UpBoundScopes), _BitWidth, maxseq(MaxSeq),
                 conjunct(none,SeqScopePred,conjunct(none,TScopes,TGlobalScopes))) :-
    % translate exact and upper bound scopes first
    translate_exact_and_upper_bound_scopes(0, ExactScopes, UpBoundScopes, EUSignatureNames, TScopes, OrdSigCounter1),
    % if present, define the global scope for the remaining signatures
    translate_global_scope(OrdSigCounter1, GlobalScope, SignatureNames, EUSignatureNames, TGlobalScopes, _),
    translate_seq_scopes(MaxSeq, SeqScopePred).

translate_seq_scopes(MaxSeq, ScopePred) :-
    findall((Seq,MaxSeq), is_seq(Seq), Seqs),
    maplist(generate_leq_scope_predicate, Seqs, ScopePreds),
    join_predicates(conjunct, ScopePreds, ScopePred).

% no exact scopes defined
translate_exact_and_upper_bound_scopes(OrdSigCounter, [], [], ExactSignatureNames, TGlobalScopes, OrdSigCounter) :-
    !,
    ExactSignatureNames = [],
    TGlobalScopes = truth(none).
translate_exact_and_upper_bound_scopes(OrdSigCounter, ExactScopes, UpBoundScopes, SignatureNames, TScopes, NewOrdSigCounter) :-
    translate_exact_and_upper_bound_scopes_aux(OrdSigCounter, UpBoundScopes, [],
                                               UpperBoundSignatureNames, [], TempTUpScopes, OrdSigCounter1),
    translate_exact_and_upper_bound_scopes_aux(OrdSigCounter1, ExactScopes, [],
                                               ExactSignatureNames, [], TempTExactScopes, NewOrdSigCounter),
    append(TempTUpScopes, TempTExactScopes, TempTScopes),
    append(ExactSignatureNames, UpperBoundSignatureNames, SignatureNames),
    join_predicates(conjunct, TempTScopes, TScopes).

translate_exact_and_upper_bound_scopes_aux(OrdSigCounter, [], NamesAcc, NamesAcc, TScopeAcc, TScopeAcc, OrdSigCounter).
translate_exact_and_upper_bound_scopes_aux(OrdSigCounter, [(SignatureName,Scope)|T],
                                           NameAcc, ExactSignatureNames, TScopeAcc, TempTExactScopes, NewOrdSigCounter) :-
    is_ordered_signature(SignatureName),
    !,
    define_ordered_signature_as_integer_set(OrdSigCounter, SignatureName, Scope,
                                            TScope, TNewOrdSigCounter),
    translate_exact_and_upper_bound_scopes_aux(TNewOrdSigCounter, T, [SignatureName|NameAcc],
                                               ExactSignatureNames, [TScope|TScopeAcc], TempTExactScopes, NewOrdSigCounter).
translate_exact_and_upper_bound_scopes_aux(OrdSigCounter, [(SignatureName,Scope)|T],
                                           NameAcc, ExactSignatureNames, TScopeAcc, TempTExactScopes, NewOrdSigCounter) :-
    generate_exact_scope_predicate((SignatureName,Scope), TScope),
    translate_exact_and_upper_bound_scopes_aux(OrdSigCounter, T, [SignatureName|NameAcc],
                                               ExactSignatureNames, [TScope|TScopeAcc], TempTExactScopes, NewOrdSigCounter).

generate_exact_scope_predicate((SignatureName,Scope), TScopePred) :-
    translate_e_p(SignatureName, TSignatureName),
    % debug_format(19,'Exact scope for SignatureName=~w with scope ~w~n',[SignatureName,Scope]),
    TScopePred = equal(none,card(none,TSignatureName),integer(none,Scope)).
generate_leq_scope_predicate((SignatureName,Scope), TScopePred) :-
    translate_e_p(SignatureName, TSignatureName),
    TScopePred = less_equal(none,card(none,TSignatureName),integer(none,Scope)).

define_ordered_signatures_as_integer_set(OrdSigCounter, [], _, [], OrdSigCounter).
define_ordered_signatures_as_integer_set(OrdSigCounter, [TSignatureName|T], Scope, [TScope|TScopes], NewOrdSigCounter) :-
    define_ordered_signature_as_integer_set(OrdSigCounter, TSignatureName, Scope, TScope, TNewOrdSigCounter),
    define_ordered_signatures_as_integer_set(TNewOrdSigCounter, T, Scope, TScopes, NewOrdSigCounter).

define_ordered_signature_as_integer_set(OrdSigCounter, TSignatureName, Scope, TScope, NewOrdSigCounter) :-
    NewOrdSigCounter is OrdSigCounter+Scope,
    NewOrdSigCounter1 is NewOrdSigCounter-1,
    TScope = equal(none,identifier(none,TSignatureName),interval(none,integer(none,OrdSigCounter),integer(none,NewOrdSigCounter1))).

translate_global_scope(OrdSigCounter, -1, _, _, TGlobalScopes, OrdSigCounter) :-
    !,
    TGlobalScopes = truth(none).                          % no global scope defined
translate_global_scope(OrdSigCounter, GlobalScope, SignatureNames, ExactSignatureNames, TGlobalScopes, NewOrdSigCounter) :-
    split_ordered_unordered_signatures(SignatureNames, ExactSignatureNames, OrderedSignatureNames, RestSignatureNames),
    translate_global_scope_aux(RestSignatureNames, GlobalScope, [], TempTGlobalScopes),
    join_predicates(conjunct, TempTGlobalScopes, TTGlobalScopes),
    define_ordered_signatures_as_integer_set(OrdSigCounter, OrderedSignatureNames, GlobalScope, TOrderedScopes, NewOrdSigCounter),
    join_predicates(conjunct, [TTGlobalScopes|TOrderedScopes], TGlobalScopes).

translate_global_scope_aux([], _, Acc, Acc).
translate_global_scope_aux([SignatureName|T], GlobalScope, Acc, TGlobalScopes) :-
    translate_e_p(SignatureName, TSignatureName),
    TScope = less_equal(none,card(none,TSignatureName),integer(none,GlobalScope)),
    translate_global_scope_aux(T, GlobalScope, [TScope|Acc], TGlobalScopes).
translate_global_scope_aux([_|T], GlobalScope, Acc, TGlobalScopes) :-
    translate_global_scope_aux(T, GlobalScope, Acc, TGlobalScopes).

split_ordered_unordered_signatures(SNames, ExactSNames, OSNames, Rest) :-
    split_ordered_unordered_signatures(SNames, [], [], TOSNames, TRest),
    % remove exact signatures which are defined separately
    subtract(TOSNames, ExactSNames, OSNames),
    subtract(TRest, ExactSNames, Rest).

split_ordered_unordered_signatures([], OAcc, RAcc, OAcc, RAcc).
split_ordered_unordered_signatures([SName|T], OAcc, RAcc, OSNames, Rest) :-
    is_ordered_signature(SName),
    \+ member(SName, OAcc),
    !,
    split_ordered_unordered_signatures(T, [SName|OAcc], RAcc, OSNames, Rest).
split_ordered_unordered_signatures([SName|T], OAcc, RAcc, OSNames, Rest) :-
    \+ is_ordered_signature(SName),
    \+ member(SName, RAcc),
    !,
    split_ordered_unordered_signatures(T, OAcc, [SName|RAcc], OSNames, Rest).
split_ordered_unordered_signatures([_|T], OAcc, RAcc, OSNames, Rest) :-
    split_ordered_unordered_signatures(T, OAcc, RAcc, OSNames, Rest).

% signature in (subset)
define_sig_as_set_or_constant(MAcc, Name, Options, POS, NewMAcc) :-
    memberchk(subset(Parents), Options),
    !,
    % maplist(assert_extending_signature(Name),Parents) , % subset signatures do not need to be distinct
    translate_pos(POS, BPOS),
    define_sig_as_set_or_constant_aux(constants, MAcc, Name, Options, BPOS, MAcc1),
    % TODO: consider several parents -> we need the universe type
    Parents = [Parent|_],
    translate_e_p(Name, TName),
    translate_e_p(Parent, TParent),
    TNode = subset(BPOS,TName,TParent),
    extend_machine_acc(properties, MAcc1, TNode, NewMAcc).
% signature extends
define_sig_as_set_or_constant(MAcc, Name, Options, POS, NewMAcc) :-
    member(subsig(Parent), Options),
    !,
    (   member(one, Options) ->
        One = one
    ;   One = set
    ),
    assert_extending_signature(Name, Parent, One),
    translate_pos(POS, BPOS),
    define_sig_as_set_or_constant_aux(constants, MAcc, Name, Options, BPOS, MAcc1),
    translate_e_p(Name, TName),
    translate_e_p(Parent, TParent),
    TNode = subset(BPOS,TName,TParent),
    extend_machine_acc(properties, MAcc1, TNode, NewMAcc).
% default signature
define_sig_as_set_or_constant(MAcc, Name, Options, POS, NewMAcc) :-
    translate_pos(POS, BPOS),
    define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc).

define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc) :-
    member(one, Options),
    !,
    atom_concat('_', Name, EnumElement),
    % TODO: couldn't figure out untyped ast of enumerated set, define enumerated set S = {_S} instead of constant etc.
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), MAcc1),
    extend_machine_acc(constants, MAcc1, identifier(BPOS,EnumElement), MAcc2),
    extend_machine_acc(properties, MAcc2, equal(BPOS,identifier(BPOS,Name),set_extension(BPOS,[identifier(BPOS,EnumElement)])), NewMAcc).
define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc) :-
    member(some, Options),
    !,
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), MAcc1),
    extend_machine_acc(properties, MAcc1, greater_equal(BPOS,card(BPOS,identifier(BPOS,Name)),integer(BPOS,1)), NewMAcc).
define_sig_as_set_or_constant_aux(sets, MAcc, Name, Options, BPOS, NewMAcc) :-
    member(lone, Options),
    !,
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), MAcc1),
    extend_machine_acc(properties, MAcc1, less_equal(BPOS,card(BPOS,identifier(BPOS,Name)),integer(BPOS,1)), NewMAcc).
define_sig_as_set_or_constant_aux(sets, MAcc, Name, _Options, BPOS, NewMAcc) :-
    extend_machine_acc(sets, MAcc, deferred_set(BPOS,Name), NewMAcc).
define_sig_as_set_or_constant_aux(constants, MAcc, Name, _Options, BPOS, NewMAcc) :-
    extend_machine_acc(constants, MAcc, identifier(BPOS,Name), NewMAcc).

% --------
% Semantic translation rules: translate expression or predicate
% --------
translate_e_p(A, TA) :-         %% functor(A,F,N),format('~n TRANSLATE: ~w/~w :  ~w ~n',[F,N,A]),
    translate_quantifier_e(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_cst_e_p(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_unary_e_p(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_binary_e_p(A, TA),
    !.
translate_e_p(A, TA) :-
    translate_if_then_else(A, TA).
translate_e_p(A, _) :-
    add_error(load_alloy_model, 'Translation failed for:', A),
    fail.

% Translation in the context where we expect an integer
translate_e_p_integer(integer(A,POS), integer(BPOS,A)) :-
    !,
    translate_pos(POS, BPOS).
translate_e_p_integer(identifier(ID,_Type,POS), identifier(BPOS,ID)) :- % translate without the set_extension
    is_singleton_set_id(ID),
    !,
    translate_pos(POS, BPOS).
translate_e_p_integer(A, TA) :-
    translate_e_p(A, TSetA),
    (   TSetA = set_extension(_,[Single])
    ->
        TA = Single
    ;   get_position(A, BPOS),
        add_message(translate_e_p_integer, 'Notice: using MU operator, could not statically extract integer singleton set for: ', TSetA, BPOS),
        % construct_choose(TSetA,TA)
        TA = mu(BPOS,TSetA)     % use the Zed MU operator
    ).

% construct_choose(TSetA,external_function_call(none,'CHOOSE',[TSetA],string(none,'a member of X'),TypeParams,Decl)) :-
% % format of raw AST of CHOOSE copied from btype2
% TypeParams = rewrite_protected([identifier(none,T)]),
% Decl = rewrite_protected(total_function(none,pow_subset(none,identifier(none,T)),identifier(none,T))).
translate_if_then_else(if_then_else(ConditionPred,TruthExpr,FalsityExpr,_Type,POS), TIfThenElse) :-
    translate_e_p(ConditionPred, TConditionPred),
    translate_e_p(TruthExpr, TTruthExpr),
    translate_e_p(FalsityExpr, TFalsityExpr),
    translate_pos(POS, BPOS),
    get_type_and_arity_from_alloy_term(TruthExpr, Type, _),
    (   (   Type = [untyped]
        ;   is_primitive_type(Type, _)
        )
    ->
        TIfThenElse = conjunct(BPOS,
                               implication(BPOS,TConditionPred,TTruthExpr),
                               implication(BPOS,negation(BPOS,TConditionPred),TFalsityExpr))
    ;   TIfThenElse = if_then_else(BPOS,TConditionPred,TTruthExpr,TFalsityExpr)
    ).

translate_quantifier_e(Quantifier, TQuantifier) :-
    Quantifier =.. [Functor,Params,Fields,Body,_Type,POS],
    member(Functor, [all,no,some,one,lone,comprehension]),
    maplist(translate_e_p, Params, TParams),
    maplist(strip_singleton_set, TParams, TTParams), % more of a hack; we should ensure singleton_set is properly set
    translate_pos(POS, BPOS),
    maplist(translate_quantifier_field, Fields, TFieldsList, SingletonSetNames),
    (   is_disjoint_quantifier(Fields)
    ->
        maplist(translate_e_p, SingletonSetNames, TSets),
        AllDiff = equal(BPOS,card(BPOS,general_union(BPOS,set_extension(BPOS,TSets))),integer(BPOS,Len)),
        length(Fields, Len),
        % print(disjoint(Len)),nl,
        join_untyped_ast_nodes(conjunct, [AllDiff|TFieldsList], TFields)
    ;   join_untyped_ast_nodes(conjunct, TFieldsList, TFields)
    ),
    translate_e_p(Body, TBody),
    translate_quantifier_e_aux(BPOS, Functor, TTParams, TFields, TBody, TQuantifier),
    maplist(retract_singleton_set(BPOS), Params),
    maplist(retract_singleton_set(BPOS), SingletonSetNames).

translate_quantifier_e_aux(Pos, all, TParams, TFields, TBody, forall(Pos,TParams,implication(none,TFields,TBody))).
translate_quantifier_e_aux(Pos, no, TParams, TFields, TBody, negation(Pos,exists(none,TParams,conjunct(none,TFields,TBody)))).
translate_quantifier_e_aux(Pos, some, TParams, TFields, TBody, exists(Pos,TParams,conjunct(none,TFields,TBody))).
translate_quantifier_e_aux(Pos, one, TParams, TFields, TBody, equal(Pos,card(none,comprehension_set(none,TParams,conjunct(none,TFields,TBody))),integer(none,1))).
translate_quantifier_e_aux(Pos, lone, TParams, TFields, TBody, less_equal(Pos,card(none,comprehension_set(none,TParams,conjunct(none,TFields,TBody))),integer(none,1))).
translate_quantifier_e_aux(Pos, comprehension, TParams, TFields, TBody, comprehension_set(Pos,TParams,conjunct(none,TFields,TBody))).

% Translate Constants
translate_cst_e_p(iden(POS), event_b_identity(BPOS)) :- % has no Span Info !
    translate_pos(POS, BPOS).
translate_cst_e_p(identifier(Cst,_,POS), BConstruct) :-
    alloy_constant_to_b(Cst, BPOS, BConstruct),
    !,
    translate_pos(POS, BPOS).
translate_cst_e_p(this, _) :-
    print('this not yet translated'),
    nl,
    fail.


alloy_constant_to_b(none, BPOS, empty_set(BPOS)).
alloy_constant_to_b('boolean''False', BPOS, boolean_false(BPOS)).
alloy_constant_to_b('boolean''True', BPOS, boolean_true(BPOS)).


translate_unary_e_p(string(StringCodes,POS), set_extension(BPOS,[string(BPOS,String)])) :-
    atom_codes(String, StringCodes),
    !,
    translate_pos(POS, BPOS).
translate_unary_e_p(seq(ID), sequence_extension(none,TID)) :-
    translate_unary_e_p(ID, TID),
    !.
translate_unary_e_p(Int, integer(none,Int)) :-
    integer(Int),
    !.
translate_unary_e_p(identifier(ID,_Type,POS), set_extension(BPOS,[identifier(BPOS,NID)])) :-
    % singleton enumerated set like S = {_S}
    enumerated_set(ID),
    is_singleton_set_id(ID),
    !,
    atom_concat('_', ID, NID),
    translate_pos(POS, BPOS).
translate_unary_e_p(identifier(ID,_Type,POS), set_extension(BPOS,[identifier(BPOS,ID)])) :-
    is_singleton_set_id(ID),
    !,
    translate_pos(POS, BPOS).
translate_unary_e_p(ID, set_extension(none,[identifier(none,NID)])) :-
    enumerated_set(ID),
    is_singleton_set_id(ID),
    !,
    atom_concat('_', ID, NID).
translate_unary_e_p(ID, set_extension(none,[identifier(none,ID)])) :-
    is_singleton_set_id(ID),
    !.                          % DOES THIS EVER TRIGGER ?
translate_unary_e_p(integer(A,POS), set_extension(BPOS,[integer(BPOS,A)])) :-
    !,
    translate_pos(POS, BPOS).
translate_unary_e_p(identifier('Int',_Type,POS), integer_set(BPOS,'INTEGER')) :-
    !,
    translate_pos(POS, BPOS).
translate_unary_e_p(identifier('String',_,_), string_set(none)) :-
    !.
translate_unary_e_p('Int', integer_set(none,'INTEGER')) :-
    !.
translate_unary_e_p(identifier(ID,_Type,POS), identifier(BPOS,ID)) :-
    !,
    translate_pos(POS, BPOS).
translate_unary_e_p(boolean(true,POS), truth(BPOS)) :-
    !,                          % was boolean_true, but Alloy has no booleans
    translate_pos(POS, BPOS).
translate_unary_e_p(boolean(false,POS), falsity(BPOS)) :-
    !,                          % was boolean_false
    translate_pos(POS, BPOS).
translate_unary_e_p(Type, integer_set(none,'INTEGER')) :-
    % lhs of sequence is integer, Alloy Types are like [seqInt,A] for set(couple(INTEGER,A))
    atom(Type),
    atom_concat(seq, _, Type),
    !.
translate_unary_e_p(ID, identifier(none,ID)) :-
    atom(ID),
    !.
translate_unary_e_p(UnaryP, TUnaryP) :-
    % and/or defines a list of ast nodes
    UnaryP =.. [Op,ArgList|_],
    memberchk(Op, [and,or]),    % translated to conjunct/disjunct
    is_list(ArgList),
    !,
    maplist(translate_e_p, ArgList, TArgList),
    join_predicates(Op, TArgList, TUnaryP).
translate_unary_e_p(UnaryP, TUnaryP) :-
    UnaryP =.. [Op,Arg,_Type,POS],
    member(Op, [no,one,some,lone]),
    !,
    translate_e_p(Arg, TArg),
    translate_pos(POS, BPOS),
    translate_quantified_e(BPOS, Op, TArg, TUnaryP).
translate_unary_e_p(card(Arg,_Type,POS), set_extension(BPOS,[card(BPOS,TArg)])) :-
    !,                          % wrap card into set extension
    translate_e_p(Arg, TArg),
    translate_pos(POS, BPOS).
translate_unary_e_p(UnaryP, TUnaryP) :-
    UnaryP =.. [Op,Arg,_Type,POS],
    translate_e_p(Arg, TArg),
    alloy_to_b_unary_operator(Op, BOp),
    translate_pos(POS, BPOS),
    TUnaryP =.. [BOp,BPOS,TArg].


% FUNCTIONS defined by func in util:
% alloy_utility_function(Name,Arity,BOperator,ArgType,ReturnType).
alloy_utility_function(Name, Arity, BOperator, integer, integer) :-
    alloy_to_b_integer_to_integer_func(Name, Arity, BOperator).
alloy_utility_function(Name, Arity, BOperator, set, integer) :-
    alloy_to_b_integer_func(Name, Arity, BOperator).
alloy_utility_function(Name, Arity, BOperator, integer, set) :-
    alloy_to_b_integer_operator_func(Name, Arity, BOperator).
alloy_utility_function(Name, Arity, BOperator, set, set) :-
    alloy_to_b_operator_func(Name, Arity, BOperator).

% functions with integer arguments and result
alloy_to_b_integer_to_integer_func('integer''minus', 2, minus).
alloy_to_b_integer_to_integer_func('integer''plus', 2, add).
alloy_to_b_integer_to_integer_func('integer''div', 2, div).
alloy_to_b_integer_to_integer_func('integer''mul', 2, multiplication).
alloy_to_b_integer_to_integer_func('integer''rem', 2, modulo).
% alloy_to_b_integer_to_integer_func('integer''sum',sigma). % TO DO: translate sum
alloy_to_b_integer_to_integer_func('integer''add', 2, add).
alloy_to_b_integer_to_integer_func('integer''sub', 2, minus).
% alloy_to_b_integer_to_integer_func('integer''smaller',max).  % TO DO: if Arg1 < Arg2 then Arg1 else Arg2 end
% Unary:
alloy_to_b_integer_to_integer_func('integer''negate', 1, unary_minus).
alloy_to_b_integer_to_integer_func('integer''next', 1, successor). % not sure this is called; seems to call arity 0 and use join
alloy_to_b_integer_to_integer_func('integer''prev', 1, predecessor).


% functions with integer result but set arguments
alloy_to_b_integer_func('integer''min', 1, min).
alloy_to_b_integer_func('integer''max', 1, max).
alloy_to_b_integer_func('integer''min', 0, min_int). % TO DO: investigate bit-width related translation
alloy_to_b_integer_func('integer''max', 0, max_int). % maybe we should have a preference for Alloy2B about INT/INTEGER
% Note: for run Arithmetic for 6 Int the Alloy Evaluator produces {-32} for min and {+31} for max
% predicate functions with integer arguments
alloy_to_b_operator_func('integer''next', 0, successor).
alloy_to_b_operator_func('integer''prev', 0, predecessor).

% predicate functions with integer arguments
alloy_to_b_integer_operator_func('integer''gt', 2, greater).
alloy_to_b_integer_operator_func('integer''lt', 2, less).
alloy_to_b_integer_operator_func('integer''gte', 2, greater_equal).
alloy_to_b_integer_operator_func('integer''lte', 2, less_equal).
alloy_to_b_integer_operator_func('integer''eq', 2, equal).

% dom[r] = dom(r)
alloy_to_b_relational_utility_function('relation''dom', [Param], _Type, BPOS, TCall) :-
    !,
    translate_e_p(Param, TParam),
    TCall = dom(BPOS,TParam).
% ran[r] = ran(r)
alloy_to_b_relational_utility_function('relation''ran', [Param], _Type, BPOS, TCall) :-
    !,
    translate_e_p(Param, TParam),
    TCall = range(BPOS,TParam).
% total[r,d] = !x.(x:d => r[{x}] /= {}})
alloy_to_b_relational_utility_function('relation''total', [Relation,Domain], _Type, BPOS, TCall) :-
    !,
    translate_e_p(Relation, TRelation),
    translate_e_p(Domain, TDomain),
    gen_identifier(0, "_c_", ID0),
    LHS = member(BPOS,ID0,TDomain),
    RHS = not_equal(BPOS,image(BPOS,TRelation,set_extension(BPOS,[ID0])),empty_set(BPOS)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% functional[r,d] = !x.(x:d => card(r[{x}]) <= 1)
alloy_to_b_relational_utility_function('relation''functional', [Relation,Domain], _Type, BPOS, TCall) :-
    !,
    translate_e_p(Relation, TRelation),
    translate_e_p(Domain, TDomain),
    gen_identifier(0, "_c_", ID0),
    LHS = member(BPOS,ID0,TDomain),
    RHS = less_equal(BPOS,card(BPOS,image(BPOS,TRelation,set_extension(BPOS,[ID0]))),integer(BPOS,1)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% function[r,d] = !x.(x:d => card(r[{x}]) = 1)
alloy_to_b_relational_utility_function('relation''function', [Relation,Domain], _Type, BPOS, TCall) :-
    !,
    translate_e_p(Relation, TRelation),
    translate_e_p(Domain, TDomain),
    gen_identifier(0, "_c_", ID0),
    LHS = member(BPOS,ID0,TDomain),
    RHS = equal(BPOS,card(BPOS,image(BPOS,TRelation,set_extension(BPOS,[ID0]))),integer(BPOS,1)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
% injective[r,d] = !(x).(x:d => card(r~[{x}]) <= 1)
alloy_to_b_relational_utility_function('relation''injective', [Relation,Domain], _Type, BPOS, TCall) :-
    !,
    translate_e_p(Relation, TRelation),
    translate_e_p(Domain, TDomain),
    gen_identifier(0, "_c_", ID0),
    LHS = member(BPOS,ID0,TDomain),
    RHS = less_equal(BPOS,card(BPOS,image(BPOS,reverse(BPOS,TRelation),set_extension(BPOS,[ID0]))),integer(BPOS,1)),
    TCall = forall(BPOS,[ID0],implication(BPOS,LHS,RHS)).
alloy_to_b_relational_utility_function(FunName, _, _, BPOS, truth(BPOS)) :-
    atom_concat('relation', _, FunName),
    !,
    add_error(alloy_to_b_relational_utility_function, 'Function from util/relation currently not supported', [], BPOS).
% TODO: implement rest as described in the article
% remove_prime(Old,New) :- atom_codes(Old,OldC) , delete(OldC,39,NewC) , atom_codes(New,NewC).
get_clean_ordering_fun_name(OrderingFunctionName, CleanOrdFunName) :-
    member(CleanOrdFunName, [first,last,min,max,next,nexts,prev,prevs,larger,smaller,lt,lte,gt,gte]),
    atom_concat(_, CleanOrdFunName, OrderingFunctionName).

get_ordering_definition_name(TempFunctionName, SignatureName, FunctionName) :-
    member(TempFunctionName, [next,prev,nexts,prevs]),
    atom_concat(TempFunctionName, '_', TTempFunctionName),
    atom_concat(TTempFunctionName, SignatureName, FunctionName).

translate_binary_e_p(in(Arg1,Arg2,_Type,POS), Translation) :-
    !,
    translate_pos(POS, BPOS),
    translate_e_p(Arg1, TArg1),
    translate_binary_in(TArg1, Arg2, BPOS, Translation).
translate_binary_e_p(CartTerm, TCartesian) :-
    is_cartesian_product_term(CartTerm, Arg1, Arg2, POS),
    !,
    translate_pos(POS, BPOS),
    translate_cartesian(BPOS, Arg1, Arg2, TCartesian).
% S <: Rel --> {x1,..,xk| x1:S & (x1,...,xk):Rel}
translate_binary_e_p(dom_restr(Arg1,Arg2,_Type,POS), TDom) :-
    \+ is_binary_relation(Arg2), % for binary relation translation to B domain_restriction works
    !,
    get_type_and_arity_from_alloy_term(Arg2, _, Arity),
    gen_ids_and_couple(Arity, IDS, Couple),
    translate_pos(POS, BPOS),
    TDom = comprehension_set(BPOS,IDS,conjunct(BPOS,Mem1,Mem2)),
    IDS = [ID1|_],
    Mem1 = member(TPos,ID1,TArg1),
    Mem2 = member(TPos,Couple,TArg2),
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2).
translate_binary_e_p(join(Arg1,Arg2,_Type,POS), TJoin) :-
    !,
    translate_pos(POS, BPOS),
    translate_join(BPOS, Arg1, Arg2, TJoin).
translate_binary_e_p(let(VarName,Expr,Sub,Type,POS), TLet) :-
    !,
    translate_e_p(VarName, TVarName),
    translate_e_p(Expr, TExpr),
    (   Type == type(['PrimitiveBoolean'],0) ->
        NewFunctor = let_predicate
    ;   NewFunctor = let_expression
    ),
    translate_e_p(Sub, TSub),
    translate_pos(POS, BPOS),
    TLet =.. [NewFunctor,BPOS,[TVarName],equal(BPOS,TVarName,TExpr),TSub].
translate_binary_e_p(Call, TCall) :-
    Call =.. [Functor,Name,Params,Type,POS],
    (   Functor = pred_call
    ;   Functor = fun_call
    ),
    !,
    translate_pos(POS, BPOS),
    translate_function_call(Name, Params, Type, BPOS, TCall).
translate_binary_e_p(Binary, TBinary) :-
    Binary =.. [Op,Arg1,Arg2,_Type,POS],
    alloy_to_b_integer_operator(Op, BOp),
    !,
    translate_e_p_integer(Arg1, TArg1),
    translate_e_p_integer(Arg2, TArg2),
    translate_pos(POS, BPOS),
    TBinary =.. [BOp,BPOS,TArg1,TArg2].
translate_binary_e_p(Binary, TBinary) :-
    Binary =.. [Op,Arg1,Arg2,_Type,POS],
    alloy_to_b_binary_operator(Op, BOp),
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2),
    translate_pos(POS, BPOS),
    TBinary =.. [BOp,BPOS,TArg1,TArg2].

% translate TArg1 in Arg2  (multplicity are not allowed for not in)
translate_binary_in(TArg1, Arg2, BPOS, Translation) :-
    compound(Arg2),
    functor(Arg2, Functor, 4),              % binary operator,
    alloy_to_b_special_multiplicity_type(Functor, _, _, _, _),
    !,
    Translation = member(BPOS,TArg1,TArg2), % use element of rather than subset for special type opererators
    translate_special_multiplicity_binary_e_p(Arg2, TArg2).
translate_binary_in(TArg1, Arg2, BPOS, subset(BPOS,TArg1,TArg2)) :-
    translate_e_p(Arg2, TArg2).

/*
   not yet finished:
   % replace Arg1 M -> N Arg2  to pure cartesian product and generate predicates
   translate_multiplicity_to_cartesian_and_pred(cartesian(A,B,Type,Pos),LHS,Left,Right,Res) :- !,
   Res = cartesian(CA,CB,Type,Pos),
   get_type_and_arity_from_alloy_term(CA,_Type,ArityA), L2 is Left+ArityA,
   get_type_and_arity_from_alloy_term(CB,_Type,ArityB), R1 is Right-ArityB,
   translate_multiplicity_to_cartesian_and_pred(A,LHS,Left,R1,CA),
   translate_multiplicity_to_cartesian_and_pred(B,LHS,L2,Right,CB).
   translate_multiplicity_to_cartesian_and_pred(Binary,LHS,Left,Right,Res) :-
   Binary =.. [Op,Arg1,Arg2,Type,POS] ,
   alloy_to_b_special_multiplicity_type(Op,BOp,LHSMult,RHSMult,ok),
   !,
   % TO DO: add extra predicates
   create_dom(Left,LHS,Pos,Proj1), % project away leftmost tuples
   create_ran(Right,Proj1,Pos,Proj2), % project away rightmost tuples

   translate_multiplicity_to_cartesian_and_pred(cartesian(A,B,Type,Pos),LHS,Left,Right,Res).
   translate_multiplicity_to_cartesian_and_pred(A,_,_,_,A).

   % create_dom(N,Expr,Pos,Res): create dom(...(dom(Expr))) to project away the leftmost N Arguments of Expr
   create_dom(0,E,_,Res) :- !, Res=E.
   create_dom(N,E,Pos,domain(Pos,D1)) :- N>0, N1 is N-1,
   create_dom(N1,E,Pos,D1).

   create_ran(0,E,_,Res) :- !, Res=E.
 */
% translate expressions of the form Arg1 MULT -> MULT Arg2  with MULT : {some,lone, one, ''}
% The result is to be used in member(Pos,LHS,TBinary) not for subset checks !
translate_special_multiplicity_binary_e_p(Binary, TBinary) :-
    Binary =.. [Op,Arg1,Arg2,_Type,POS],
    alloy_to_b_special_multiplicity_type(Op, BOp, _, _, ok),
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2),
    translate_pos(POS, BPOS),
    TBinary =.. [BOp,BPOS,TArg1,TArg2].

% TODO: update sequence implementation as described in the article
% Note: Sequences in B are indexed from 1 to n, in Alloy from 0 to n-1.
% Documentation available at http://alloy.lcs.mit.edu/alloy/documentation/quickguide/seq.html
% Sequence function calls without a parameter but the sequence itself.
% butlast: seq[{size(seq)-1}]
sequence_function_to_b('seq\'butlast', BPOS, [Seq], image(BPOS,Seq,set_extension(BPOS,[minus(BPOS,size(BPOS,Seq),integer(BPOS,1))]))).
% isEmpty: seq = <>
sequence_function_to_b('seq\'isEmpty', BPOS, [Seq], equal(BPOS,Seq,empty_sequence(BPOS))).
% hasDups: size(seq) /= card(ran(seq))
sequence_function_to_b('seq\'hasDups', BPOS, [Seq], not_equal(BPOS,size(BPOS,Seq),card(BPOS,range(BPOS,Seq)))).
% inds: {x|x:1..size(seq)}
sequence_function_to_b('seq\'inds', BPOS, [Seq], comprehension_set(BPOS,[identifier(BPOS,x)],member(BPOS,identifier(BPOS,x),interval(BPOS,integer(BPOS,1),size(BPOS,Seq))))).
% lastIdx: LET x BE x = size(seq) IN (IF x > 0 THEN {x} ELSE {} END) END
% TO DO: use let like let_expression(BPOS,[identifier(BPOS,x)],equal(BPOS,identifier(BPOS,x),size(BPOS,Seq)),if_then_else(BPOS,greater(BPOS,identifier(BPOS,x),integer(BPOS,0)),set_extension(BPOS,[identifier(BPOS,x)]),empty_set(BPOS)))
sequence_function_to_b('seq\'lastIdx', BPOS, [Seq], let_expression(BPOS,[identifier(BPOS,x)],equal(BPOS,identifier(BPOS,x),size(BPOS,Seq)),if_then_else(BPOS,greater(BPOS,identifier(BPOS,x),integer(BPOS,0)),set_extension(BPOS,[identifier(BPOS,x)]),empty_set(BPOS)))).
% afterLastIdx: LET x BE x = size(seq) IN (IF x < MAXSEQ THEN {x} ELSE {} END) END
% TO DO: use let like let_expression(BPOS,[identifier(BPOS,x)],equal(BPOS,identifier(BPOS,x),size(BPOS,Seq)),if_then_else(BPOS,less(BPOS,identifier(BPOS,x),integer(BPOS,MaxSeq)),set_extension(BPOS,[identifier(BPOS,x)]),empty_set(BPOS)))
sequence_function_to_b('seq\'afterLastIdx', BPOS, [Seq], let_expression(BPOS,[identifier(BPOS,x)],equal(BPOS,identifier(BPOS,x),size(BPOS,Seq)),if_then_else(BPOS,less(BPOS,identifier(BPOS,x),integer(BPOS,MaxSeq)),set_extension(BPOS,[identifier(BPOS,x)]),empty_set(BPOS)))) :-
    maxseq(MaxSeq).
% Sequence function calls with one parameter and the sequence itself.
% idxOf[elm]: IF size(seq) > 0 THEN {min(seq~[{elm}])} ELSE {} END
sequence_function_to_b('seq\'idxOf', BPOS, [Seq,P1], if_then_else(BPOS,not_equal(BPOS,image(BPOS,reverse(BPOS,Seq),P1),empty_set(BPOS)),set_extension(BPOS,[min(BPOS,image(BPOS,reverse(BPOS,Seq),P1))]),empty_set(BPOS))).
% lastIdxOf[elm]: IF size(seq) > 0 THEN {max(seq~[{elm}])} ELSE {} END
sequence_function_to_b('seq\'lastIdxOf', BPOS, [Seq,P1], if_then_else(BPOS,not_equal(BPOS,image(BPOS,reverse(BPOS,Seq),P1),empty_set(BPOS)),set_extension(BPOS,[max(BPOS,image(BPOS,reverse(BPOS,Seq),P1))]),empty_set(BPOS))).
% indsOf[elm]: seq~[{elm}]
sequence_function_to_b('seq\'indsOf', BPOS, [Seq,P1], image(BPOS,reverse(BPOS,Seq),set_extension(BPOS,[P1]))).
% append[seq2]: LET x BE x = s^s2 IN (IF size(x) < MAXSEQ THEN x ELSE x /|\ MAXSEQ END) END
% TO DO: use let like let_expression(BPOS,[identifier(BPOS,x)],equal(BPOS,identifier(BPOS,x),concat(Seq,P1)),if_then_else(BPOS,less_equal(BPOS,size(BPOS,identifier(BPOS,x)),integer(BPOS,MaxSeq)),identifier(BPOS,x),restrict_front(BPOS,identifier(BPOS,x),integer(BPOS,MaxSeq))))
sequence_function_to_b('seq\'append', BPOS, [Seq,P1], let_expression(BPOS,[identifier(BPOS,x)],equal(BPOS,identifier(BPOS,x),concat(Seq,P1)),if_then_else(BPOS,less_equal(BPOS,size(BPOS,identifier(BPOS,x)),integer(BPOS,MaxSeq)),identifier(BPOS,x),restrict_front(BPOS,identifier(BPOS,x),integer(BPOS,MaxSeq))))) :-
    maxseq(MaxSeq).
% add[elm]: IF size(seq) < MAXSEQ THEN seq <- elm ELSE seq END
sequence_function_to_b('seq\'add', BPOS, [Seq,P1], if_then_else(BPOS,less(BPOS,size(BPOS,Seq),integer(BPOS,MaxSeq)),insert_tail(BPOS,Seq,P1),Seq)) :-
    maxseq(MaxSeq).
% delete[i]: IF (i >= 0 & i < size(seq)) THEN (seq /|\ i) ^ (seq \|/ i+1) ELSE seq END
sequence_function_to_b('seq\'delete', BPOS, [Seq,P1], if_then_else(BPOS,conjunct(BPOS,greater_equal(BPOS,P1,integer(BPOS,0)),less(BPOS,P1,size(BPOS,Seq))),concat(BPOS,restrict_front(BPOS,Seq,P1),restrict_tail(BPOS,Seq,add(BPOS,P1,integer(BPOS,1)))),Seq)).
% Sequence function calls with two parameters and the sequence itself.
% setAt[i,elm]: IF (i >= 0 & i < size(seq)) THEN (seq /|\ i) ^ [elm] ^ (seq \|/ i+1) ELSE seq END
sequence_function_to_b('seq\'setAt', BPOS, [Seq,P1,P2], if_then_else(BPOS,conjunct(BPOS,greater_equal(BPOS,P1,integer(BPOS,0)),less(BPOS,P1,size(BPOS,Seq))),concat(concat(BPOS,restrict_front(BPOS,Seq,P1),sequence_extension(BPOS,[P2])),restrict_tail(BPOS,Seq,add(BPOS,P1,integer(BPOS,1)))),Seq)).
% insert[i,elm]: LET x BE x = (IF size(seq) = MAXSEQ THEN seq /|\ (MAXSEQ - 1) ELSE seq END) IN (x /|\ i) ^ [elm] ^ (x \|/ i) END
sequence_function_to_b('seq\'insert', BPOS, [Seq,P1,P2], let_expression(BPOS,[identifier(BPOS,x)],equal(BPOS,identifier(BPOS,x),if_then_else(BPOS,equal(BPOS,size(BPOS,Seq),integer(BPOS,MaxSeq)),restrict_front(BPOS,Seq,minus(BPOS,integer(BPOS,MaxSeq),integer(BPOS,1))),Seq)),concat(concat(BPOS,restrict_front(BPOS,Seq,P1),sequence_extension(BPOS,[P2])),restrict_tail(BPOS,Seq,add(BPOS,P1,integer(BPOS,1)))))) :-
    maxseq(MaxSeq).
% subseq[from,to]: IF 0 <= from & from <= to & to < size(seq) THEN (s \|/ from) /|\ ((to-from)+1) ELSE seq end
sequence_function_to_b('seq\'subseq', BPOS, [Seq,P1,P2], if_then_else(BPOS,conjunct(BPOS,conjunct(BPOS,less_equal(BPOS,integer(BPOS,0),P1),less_equal(BPOS,P1,P2)),less(BPOS,P2,size(BPOS,Seq))),restrict_front(BPOS,restrict_tail(BPOS,Seq,P1),add(BPOS,minus(BPOS,P2,P1),integer(BPOS,1))))).

sequence_function_to_b_functor(Name, BFunctor) :-
    atom_concat('seq\'', Functor, Name),
    sequence_function_to_b_functor_aux(Functor, BFunctor).

sequence_function_to_b_functor_aux(first, first).
sequence_function_to_b_functor_aux(last, last).
sequence_function_to_b_functor_aux(rest, tail).
sequence_function_to_b_functor_aux(elems, range).

translate_seq_function_param(join(identifier(this,type([Type],1),IPos),Arg2,JType,JPos), TParam) :-
    !,
    % remove join with this for signature fields
    translate_e_p(join(identifier(Type,type([Type],1),IPos),Arg2,JType,JPos), TParam).
translate_seq_function_param(Param, TParam) :-
    translate_e_p(Param, TParam).


% sequence function calls: first, last, rest, elems
translate_function_call(Name, [Param], _Type, BPOS, TCall) :-
    sequence_function_to_b_functor(Name, BFunctor),
    !,
    translate_seq_function_param(Param, TParam),
    TCall =.. [BFunctor,BPOS,TParam].
translate_function_call(Function, Params, _Type, BPOS, TCall) :-
    maplist(translate_seq_function_param, Params, TParams),
    sequence_function_to_b(Function, BPOS, TParams, TCall),
    !.
translate_function_call(Name, Params, Type, BPOS, Result) :-
    alloy_to_b_relational_utility_function(Name, Params, Type, BPOS, Result).
translate_function_call(Name, Params, _Type, BPOS, Result) :-
    length(Params, Arity),
    alloy_utility_function(Name, Arity, BOp, ArgType, ReturnType),
    !,                                       % functions like plus, minus, mul, ... with integer arguments and result
    (   ArgType = integer ->
        maplist(translate_e_p_integer, Params, TParams)
    ;   maplist(translate_e_p, Params, TParams)
    ),
    (   ReturnType = integer ->
        Result = set_extension(BPOS,[TCall]) % we add set_extension wrapper
    ;   Result = TCall
    ),
    TCall =.. [BOp,BPOS|TParams].
% ordering function calls
translate_function_call(FunName, Params, type([SignatureName|_],_), BPOS, TCall) :-
    translate_e_p(SignatureName, TSignatureName),
    get_clean_ordering_fun_name(FunName, CleanOrdFunName),
    translate_ordering_function(CleanOrdFunName, Params, SignatureName, TSignatureName, BPOS, TCall).
translate_function_call(Name, Params, _, BPOS, TCall) :-
    % predicate and function call
    maplist(translate_e_p, Params, TParams),
    maplist(strip_singleton_set_from_this, TParams, TTParams), % identifier 'this' is added for fields which is already wrapped in a set extension
    TCall = definition(BPOS,Name,TTParams).

% first: IF S_ord /= {} THEN {min(S_ord)} ELSE {}
translate_ordering_function(first, [], _, TSignatureName, BPOS, if_then_else(BPOS,not_equal(BPOS,TSignatureName,empty_set(BPOS)),set_extension(BPOS,[min(BPOS,TSignatureName)]),empty_set(BPOS))) :-
    !.
% last: IF S_ord /= {} THEN {max(S_ord)} ELSE {}
translate_ordering_function(last, [], _, TSignatureName, BPOS, if_then_else(BPOS,not_equal(BPOS,TSignatureName,empty_set(BPOS)),set_extension(BPOS,[max(BPOS,TSignatureName)]),empty_set(BPOS))) :-
    !.
% min[p]: IF p /= {} THEN {min(p)} ELSE {}
% max[p]: IF p /= {} THEN {min(p)} ELSE {}
translate_ordering_function(FunName, [Param], _, _, BPOS, TCall) :-
    (   FunName == min
    ;   FunName == max
    ),
    !,
    translate_e_p(Param, TParam),
    InnerCall =.. [FunName,BPOS,TParam],
    TCall = if_then_else(BPOS,not_equal(BPOS,TParam,empty_set(BPOS)),set_extension(BPOS,[InnerCall]),empty_set(BPOS)).
translate_ordering_function(FunName, [Param], SignatureName, _, BPOS, TCall) :-
    (   FunName == next
    ;   FunName == nexts
    ;   FunName == prev
    ;   FunName == prevs
    ),
    !,
    get_ordering_definition_name(FunName, SignatureName, DefName),
    translate_e_p(Param, TParam),
    TCall = definition(BPOS,DefName,[TParam]).
% larger[p1,p1]: IF p1 \/ p2 /= {} THEN {max(p1 \/ p2)} ELSE {} END
% smaller[p1,p2]: IF p1 \/ p2 /= {} THEN {min(p1 \/ p2)} ELSE {} END
translate_ordering_function(FunName, [P1,P2], _, _, BPOS, TCall) :-
    (   FunName == larger
    ;   FunName == smaller
    ),
    get_partial_b_functor_for_ordering(FunName, BFunctor),
    translate_e_p(P1, TP1),
    translate_e_p(P2, TP2),
    ParamUnion = union(BPOS,TP1,TP2),
    TInnerCall =.. [BFunctor,BPOS,ParamUnion],
    InnerCall = set_extension(BPOS,[TInnerCall]),
    TCall = if_then_else(BPOS,not_equal(BPOS,ParamUnion,empty_set(BPOS)),InnerCall,empty_set(BPOS)).
% lt[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & p1 /= p2 & {min(p1 \/ p2)} = p1))
% gt[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & p1 /= p2 & {max(p1 \/ p2)} = p1))
translate_ordering_function(FunName, [P1,P2], _, _, BPOS, TCall) :-
    (   FunName == lt
    ;   FunName == gt
    ),
    get_partial_b_functor_for_ordering(FunName, BFunctor),
    translate_e_p(P1, TP1),
    translate_e_p(P2, TP2),
    TInnerCall =.. [BFunctor,BPOS,union(BPOS,TP1,TP2)],
    InnerCall = set_extension(BPOS,[TInnerCall]),
    LHS = equal(BPOS,TP1,empty_set(BPOS)),
    RHS = conjunct(BPOS,conjunct(BPOS,conjunct(BPOS,not_equal(BPOS,TP1,empty_set(BPOS)),not_equal(BPOS,TP2,empty_set(BPOS))),not_equal(BPOS,TP1,TP2)),equal(BPOS,InnerCall,TP1)),
    TCall = disjunct(BPOS,LHS,RHS).
% lte[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & {min(p1 \/ p2)} = p1)
% gte[p1,p2]: (p1 = {}) or (p1 /= {} & p2 /= {} & {max(p1 \/ p2)} = p1)
translate_ordering_function(FunName, [P1,P2], _, _, BPOS, TCall) :-
    (   FunName == lte
    ;   FunName == gte
    ),
    get_partial_b_functor_for_ordering(FunName, BFunctor),
    translate_e_p(P1, TP1),
    translate_e_p(P2, TP2),
    TInnerCall =.. [BFunctor,BPOS,union(BPOS,TP1,TP2)],
    InnerCall = set_extension(BPOS,[TInnerCall]),
    LHS = equal(BPOS,TP1,empty_set(BPOS)),
    RHS = conjunct(BPOS,conjunct(BPOS,not_equal(BPOS,TP1,empty_set(BPOS)),not_equal(BPOS,TP2,empty_set(BPOS))),equal(BPOS,InnerCall,TP1)),
    TCall = disjunct(BPOS,LHS,RHS).
translate_ordering_function(FunName, _, _, _, BPOS, empty_set(BPOS)) :-
    add_error(translate_ordering_function, 'Cannot translate ordering function', [FunName], BPOS).

get_partial_b_functor_for_ordering(larger, max).
get_partial_b_functor_for_ordering(smaller, min).
get_partial_b_functor_for_ordering(lt, min).
get_partial_b_functor_for_ordering(lte, min).
get_partial_b_functor_for_ordering(gt, max).
get_partial_b_functor_for_ordering(gte, max).

strip_singleton_set(set_extension(_,[identifier(Pos,Name)]), identifier(Pos,Name)) :-
    !.
strip_singleton_set(A, A).

strip_singleton_set_from_this(set_extension(_,[identifier(Pos,this)]), identifier(Pos,this)) :-
    !.
strip_singleton_set_from_this(A, A).

translate_cartesian(TPos, Arg1, Arg2, Res) :-
    translate_e_p(Arg1, TArg1),
    translate_cartesian2(TPos, TArg1, Arg2, Res).

translate_cartesian2(TPos, TArg1, Arg2, Res) :-
    translate_e_p(Arg2, TArg2),
    translate_cartesian3(TPos, TArg1, Arg2, TArg2, Res).
% P -> Q  --> P*Q
translate_cartesian3(TPos, TArg1, Arg2, TArg2, cartesian_product(TPos,TArg1,TArg2)) :-
    is_unary_relation(Arg2),
    !.
% P -> Q  --> {c0, c1,...,ck | c0:P & (c1,...,ck):Q}
translate_cartesian3(TPos, TArg1, Arg2, TArg2, comprehension_set(TPos,[ID0|IDs2],conjunct(TPos,Mem1,Mem2))) :-
    get_type_and_arity_from_alloy_term(Arg2, _Type, Arity),
    gen_ids_and_couple(Arity, IDs2, Couple2),
    gen_identifier(0, "_c_", ID0),
    Mem1 = member(TPos,ID0,TArg1),
    Mem2 = member(TPos,Couple2,TArg2).

gen_ids_and_couple(Arity, [ID1|IDListRest], NestedCouple) :-
    Arity > 0,
    gen_identifier(1, "_c_", ID1),
    gen_ids_and_couple(2, Arity, IDListRest, ID1, NestedCouple).

gen_ids_and_couple(Nr, Arity, L, Acc, Res) :-
    Nr > Arity,
    !,
    L = [],
    Res = Acc.
gen_ids_and_couple(Nr, Arity, [IDN|T], Acc, Res) :-
    gen_identifier(Nr, "_c_", IDN),
    N1 is Nr+1,
    gen_ids_and_couple(N1, Arity, T, couple(none,Acc,IDN), Res).

gen_identifier(Nr, Prefix, identifier(none,ID)) :-
    number_codes(Nr, NrC),
    append(Prefix, NrC, AC),
    atom_codes(ID, AC).

% ordering function calls like s.next.next are converted to fun_call
translate_join(_, Arg1, Arg2, TBinaryJoin) :-
    Arg2 = fun_call(Name,[],Type,A2Pos),
    \+ atom_concat('integer', _, Name),
    translate_e_p(fun_call(Name,[Arg1],Type,A2Pos), TBinaryJoin).
% Translation of the dot join operator has several special cases depending on the arity of the arguments.
translate_join(Pos, Arg1, Arg2, TBinaryJoin) :-
    translate_e_p(Arg1, TArg1),
    translate_e_p(Arg2, TArg2),
    % print(translate_join(Arg1,Arg2)),nl,
    translate_join_aux(Pos, Arg1, Arg2, TArg1, TArg2, TBinaryJoin).

% Note: N-ary relation are encoded in B as follows:
% { (e1 |-> (e2 |-> (e3 |-> ...))) ,... }, for example:
% {(tracy|->(spanish|->french)),(carol|->(spanish|->french)), ...}
% univ.P  -->  ran(P)  % this does NOT work for ternary,... relations
translate_join_aux(TPos, Arg1, Arg2, _TArg1, TArg2, range(TPos,TArg2)) :-
    is_univ(Arg1),
    is_binary_relation(Arg2),
    !.
% P.univ  -->  dom(P)  % this works for ternary,... relations
translate_join_aux(TPos, _Arg1, Arg2, TArg1, _TArg2, domain(TPos,TArg1)) :-
    is_univ(Arg2),
    !.
% {el}.R --> {R(el)}
translate_join_aux(Pos, Arg1, Arg2, TArg1, TArg2, set_extension(Pos,[function(Pos,TArg2,[TTArg1])])) :-
    get_type_and_arity_from_alloy_term(Arg1, Type1, Arity),
    Arity = 1,                  % is_unary_relation(Arg1),
    is_binary_relation(Arg2),
    is_total_function(TArg2, Type1),
    can_remove_singleton_set(TArg1, TTArg1),
    !.
% unary.R --> R[unary]
translate_join_aux(Pos, Arg1, Arg2, TArg1, TArg2, image(Pos,TArg2,TArg1)) :- % Unary.Arg2 --> Arg2[Unary]
    is_unary_relation(Arg1),
    is_binary_relation(Arg2),
    !.
% binary.unary
% Next RULE is broken, can lead to Type errors and test is wrong:
% translate_join_aux(Pos,Arg1,Arg2,TArg1,TArg2,function(Pos,reverse(Pos,TArg1),[TTArg2])) :-
% % function call if lhs is a total function and rhs a unary relation
% is_binary_relation(Arg1),
% is_unary_relation(Arg2),
% is_total_function(TArg1) , ! ,  % TEST SHOULD BE SURJECTIVE and INJECTIVE !!
% remove_singleton_set(TArg2,TTArg2).
% Arg1.Unary --> Arg1~[Unary]
translate_join_aux(Pos, _Arg1, Arg2, TArg1, TArg2, image(Pos,reverse(Pos,TArg1),TArg2)) :-
    is_unary_relation(Arg2),
    !.
% binary.binary:  Arg1.Arg2 --> (Arg1 ; Arg2)
translate_join_aux(Pos, Arg1, Arg2, TArg1, TArg2, composition(Pos,TArg1,TArg2)) :-
    is_binary_relation(Arg1),
    is_binary_relation(Arg2),
    !.
translate_join_aux(Pos, Arg1, Arg2, TArg1, TArg2, comprehension_set(Pos,IDS,Body)) :-
    Body = conjunct(Pos,TypingPred,exists(Pos,[JoinID],conjunct(Pos,JoinTypingPred,conjunct(Pos,Mem1,Mem2)))),
    get_type_and_arity_from_alloy_term(Arg2, [JoinIDType|RTypes], Arity), % first type refers to outter variable
    gen_ids_and_couple(Arity, IDs2, Couple2),
    IDs2 = [JoinID|Rest2],
    get_typing_pred_for_ids(Pos, RTypes, Rest2, TypingPred),
    get_typing_pred_for_ids(Pos, [JoinIDType], [JoinID], JoinTypingPred),
    (   is_univ(Arg1) ->
        Mem1 = truth(TPos),
        IDS = Rest2
    ;   is_unary_relation(Arg1) ->
        strip_singleton_set_from_this(TArg1, TTArg1),
        Mem1 = member(TPos,JoinID,TTArg1),
        IDS = Rest2
    ;   gen_identifier(0, "_c_", ID0),
        IDS = [ID0|Rest2],
        Mem1 = member(TPos,couple(TPos,ID0,JoinID),TArg1)
    ),
    Mem2 = member(TPos,Couple2,TArg2).
translate_join_aux(Pos, Arg1, Arg2, _TArg1, _TArg2, empty_set(Pos)) :-
    format('~nJoin not supported this way:~nLeft: ~w~nRight: ~w~n', [Arg1,Arg2]),
    add_error(load_alloy_model, 'Cannot translate this type of join yet', [Arg1,Arg2], Pos).

% This predicate relies on the order of elements in 2. and 3. argument for typing.
% TO DO: improve
get_typing_pred_for_ids(Pos, [], [], truth(Pos)).
get_typing_pred_for_ids(Pos, [Type|TT], [ID|IT], TypingPred) :-
    translate_e_p(Type, TType),
    get_typing_pred_for_ids_acc(Pos, TT, IT, member(Pos,ID,TType), TypingPred).

get_typing_pred_for_ids_acc(_, [], [], Acc, Acc).
get_typing_pred_for_ids_acc(Pos, [Type|TT], [ID|IT], Acc, TypingPred) :-
    translate_e_p(Type, TType),
    get_typing_pred_for_ids_acc(Pos, TT, IT, conjunct(Pos,Acc,member(Pos,ID,TType)), TypingPred).

% translate Alloy position to B position
translate_pos(pos(Col,Row), Res) :-
    !,
    Res = pos(0,1,Row,Col,Row,Col).
translate_pos(_, none).

get_position(Term, BPos) :-
    functor(Term, _, Arity),
    Arity > 0,
    arg(Arity, Term, Pos),
    !,
    translate_pos(Pos, BPos).
get_position(_, none).

translate_quantified_e(Pos, no, TArg, equal(Pos,TArg,empty_set(none))).
translate_quantified_e(Pos, one, TArg, equal(Pos,card(none,TArg),integer(none,1))).
translate_quantified_e(Pos, some, TArg, greater(Pos,card(none,TArg),integer(none,0))).
translate_quantified_e(Pos, lone, TArg, less_equal(Pos,card(none,TArg),integer(none,1))).

% Join a list of types using cartesian_product like ((A*B)*C) for types [A,B,C].
create_cartesian_type([ID], TID) :-
    !,
    translate_unary_e_p(ID, TID).
create_cartesian_type([ID1,ID2], cartesian_product(none,TID1,TID2)) :-
    !,
    translate_unary_e_p(ID1, TID1),
    translate_unary_e_p(ID2, TID2).
create_cartesian_type([ID1,ID2|T], cartesian_product(none,cartesian_product(none,TID1,TID2),NT)) :-
    translate_unary_e_p(ID1, TID1),
    translate_unary_e_p(ID2, TID2),
    create_cartesian_type(T, NT).

% Field declarations have several special cases depending on the keywords set, one, some or lone.
% We are inside Signature TSignatureName  sig NAME {  field : DeclTerm }
translate_field_decl(BPOS, SignatureName, TSignatureName, DeclTerm, FieldID, TFieldID, TFieldPred) :-
    % DeclTerm =.. [_,SetID|_],
    replace_this_in_signature_field(DeclTerm, NewDeclTerm),
    debug_format(19, 'Translating Field ~w inside signature ~w,~n   Term=~w,~n~n', [FieldID,SignatureName,NewDeclTerm]),
    (   translate_special_multiplicity_binary_e_p(NewDeclTerm, TFUNC)
    ->
        % we could do: translate_cartesian2(BPOS,TSignatureName,NewDeclTerm,Cart), TFieldPred = subset(BPOS,TFieldID,Cart), and add predicate
        % TO DO: better treatment, also detect nested special multiplicity annotations inside NewDeclTerm
        % generate forall x:SIG => x.field : TFUNC
        ID = identifier(BPOS,'_x_'),
        AID = identifier('_x_',type([SignatureName],1),none),
        (   singleton_set('_x_') ->
            true
        ;   assert_singleton_set('_x_')
        ),
        translate_e_p(AID, TAID), % will add set-extension wrapper
        LHS = member(BPOS,ID,TSignatureName),
        RHS = member(BPOS,TJoin,TFUNC),
        TFieldPred = conjunct(BPOS,FieldTyping,forall(BPOS,[ID],implication(none,LHS,RHS))),
        translate_join_aux(BPOS, AID, FieldID, TAID, TFieldID, TJoin),
        get_type_and_arity_from_alloy_term(FieldID, TypeList, _),
        create_cartesian_type(TypeList, CartesianType),
        FieldTyping = member(Pos,TFieldID,pow_subset(Pos,CartesianType))
    ;   translate_sig_field_rhs(FieldID, NewDeclTerm, BPOS, TSignatureName, TFieldID, FieldTypeExpr)
    ->
        add_constraint_for_someof_field(NewDeclTerm, member(BPOS,TFieldID,FieldTypeExpr), TFieldPred)
    ;   add_error(alloy2b, 'Field declaration not implemented:', NewDeclTerm, BPOS),
        TFieldPred = truth(none)
    ).

add_constraint_for_someof_field(someof(_,_,_), TempPred, conjunct(BPOS,not_equal(BPOS,TFieldID,empty_set(BPOS)),TempPred)) :-
    !,
    TempPred = member(BPOS,TFieldID,_).
add_constraint_for_someof_field(_, Pred, Pred).

% sig SIG { field : set Set}  -->  field :  SIG <-> Set
translate_sig_field_rhs(_, setof(Set,_,_), _, SIG, _, relations(none,SIG,TSet)) :-
    translate_e_p(Set, TSet).
% sig SIG { field : one Set}  -->  field :  SIG --> Set
translate_sig_field_rhs(_, oneof(Set,_,_), _, SIG, TFieldID, total_function(none,SIG,TSet)) :-
    translate_e_p(Set, TSet),
    assert_total_function(TFieldID, SIG).
% sig SIG { field : lone Set}  -->  field :  SIG +-> Set
translate_sig_field_rhs(_, loneof(Set,_,_), _, SIG, _, partial_function(none,SIG,TSet)) :-
    translate_e_p(Set, TSet).
% sig SIG { field : some Set}  -->  field :  SIG +-> Set
translate_sig_field_rhs(_, someof(Set,_,_), _, SIG, _, relations(none,SIG,TSet)) :-
    translate_e_p(Set, TSet).
% sig SIG { field : seq Set}  -->  field :  (SIG * Integer) +-> Set
translate_sig_field_rhs(FieldID, 'isseq->lone'(_,Set,_,_), _, SIG, _, partial_function(none,cartesian_product(none,SIG,integer_set(none,'INTEGER')),TSet)) :-
    assert(is_seq(FieldID)),
    translate_e_p(Set, TSet).
% sig SIG { field : FieldTerm}  -->  field :  POW(SIG * FieldTerm)
translate_sig_field_rhs(_, FieldTerm, Pos, SIG, _, pow_subset(none,Cart)) :-
    % TO DO: check whether lone, one, ... can appear inside FieldTerm
    translate_cartesian2(Pos, SIG, FieldTerm, Cart).

replace_this_in_signature_field(identifier('this',type([SignatureName],1),Pos), Res) :-
    !,
    Res = identifier(SignatureName,type([SignatureName],1),Pos).
replace_this_in_signature_field(DeclTerm, NewDeclTerm) :-
    DeclTerm =.. [Functor,Arg1,Type,Pos],
    !,
    replace_this_in_signature_field(Arg1, NArg1),
    NewDeclTerm =.. [Functor,NArg1,Type,Pos].
replace_this_in_signature_field(DeclTerm, NewDeclTerm) :-
    DeclTerm =.. [Functor,Arg1,Arg2,Type,Pos],
    !,
    replace_this_in_signature_field(Arg1, NArg1),
    replace_this_in_signature_field(Arg2, NArg2),
    NewDeclTerm =.. [Functor,NArg1,NArg2,Type,Pos].
replace_this_in_signature_field(DeclTerm, DeclTerm).

% Translate TFieldID: Declaration
fun_or_pred_decl_special_cases(setof(Expr,_,_), TFieldID, Pos, subset(Pos,TFieldID,TExpr), set) :-
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(oneof(Expr,_,_), TFieldID, Pos, subset(Pos,TFieldID,TExpr), singleton) :- % Default
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(loneof(Expr,_,_), TFieldID, Pos, conjunct(Pos,subset(Pos,TFieldID,TExpr),ExtraPred), set) :-
    % x : lone S  ->  x <: S & card(x) <= 1
    ExtraPred = less_equal(Pos,card(Pos,TFieldID),integer(Pos,1)),
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(someof(Expr,_,_), TFieldID, Pos, conjunct(Pos,subset(Pos,TFieldID,TExpr),ExtraPred), set) :-
    % x : some S  ->  x <: S & x \= {}
    ExtraPred = not_equal(Pos,TFieldID,empty_set(Pos)),
    translate_e_p(Expr, TExpr).
fun_or_pred_decl_special_cases(Expr, TFieldID, Pos, Translation, IsSingleton) :-
    % treat identifier, cartesian, union, ...  use translate of TFieldID in Expr
    translate_binary_in(TFieldID, Expr, Pos, Translation),
    get_type_and_arity_from_alloy_term(Expr, _Type, Arity),
    !,
    (   Arity = 1 ->
        IsSingleton = singleton
    ;   IsSingleton = set
    ).
fun_or_pred_decl_special_cases(Term, _, Pos, falsity(Pos), set) :-
    add_error(alloy2b, 'Field declaration for function or predicate not implemented:', Term, Pos).

alloy_to_b_operator(Op, BOp) :-
    alloy_to_b_unary_operator(Op, BOp),
    !.
alloy_to_b_operator(Op, BOp) :-
    alloy_to_b_binary_operator(Op, BOp),
    !.
alloy_to_b_operator(Op, Op) :-
    format('~n~nTRANSLATING ~w to B without modification!~n', [Op]).

alloy_to_b_unary_operator(not, negation).
alloy_to_b_unary_operator(closure1, closure).
alloy_to_b_unary_operator(closure, reflexive_closure).
alloy_to_b_unary_operator(inverse, reverse). % works for binary relation only; indeed Alloy says: ~ can be used only with a binary relation.


alloy_to_b_binary_operator(in, subset).
alloy_to_b_binary_operator(not_in, not_subset).
alloy_to_b_binary_operator(plus, union).
alloy_to_b_binary_operator(or, disjunct). % actually treated as unary operator
alloy_to_b_binary_operator(and, conjunct). % actually treated as unary operator
alloy_to_b_binary_operator(minus, set_subtraction).
% alloy_to_b_binary_operator(cartesian,cartesian_product). % only works for  X->UNARY !
alloy_to_b_binary_operator(function, expression_definition).
alloy_to_b_binary_operator(predicate, predicate_definition).
alloy_to_b_binary_operator(not_equal, not_equal).
alloy_to_b_binary_operator(dom_restr, domain_restriction). % not ok for ternary, set must be unary
alloy_to_b_binary_operator(ran_restr, range_restriction). % also ok for ternary, ... relations
alloy_to_b_binary_operator(equal, equal).
alloy_to_b_binary_operator(conjunct, conjunct).
alloy_to_b_binary_operator(intersection, intersection).
alloy_to_b_binary_operator(implication, implication).
alloy_to_b_binary_operator(iff, equivalence).
alloy_to_b_binary_operator(override, overwrite). % ++

is_cartesian_product_term(Term, Arg1, Arg2, POS) :-
    functor(Term, Functor, 4),
    is_cartesian_product(Functor),
    arg(1, Term, Arg1),
    arg(2, Term, Arg2),
    arg(4, Term, POS).
is_cartesian_product(cartesian).
is_cartesian_product(Special) :- % translate as Cartesian product for typing predicates ONLY ! (generates unsound translations on its own, needs to be augmented with predicates which check the multiplicity !)
    alloy_to_b_special_multiplicity_type(Special, _, _, _, POS),
    translate_pos(POS, BPOS),
    add_warning(alloy2b, 'Treating operator as Cartesian product (cannot translate nested multiplicity in signature yet):', Special, BPOS).

alloy_to_b_special_multiplicity_type(total_function, total_function, set, one, ok). % TO DO: requires element rather than subset, fix, see birthday.als; can only be used in context of in, !in
alloy_to_b_special_multiplicity_type(partial_function, partial_function, set, lone, ok).
alloy_to_b_special_multiplicity_type(partial_injection, partial_injection, set, lone, ok).
alloy_to_b_special_multiplicity_type(total_bijection, total_bijection, one, one, ok).
alloy_to_b_special_multiplicity_type(total_injection, total_injection, lone, one, ok).
alloy_to_b_special_multiplicity_type(partial_surjection, partial_surjection, some, lone, ok).
alloy_to_b_special_multiplicity_type(total_surjection, total_surjection, some, one, ok).
% relations:
alloy_to_b_special_multiplicity_type(total_relation, total_relation, set, some, ok).
alloy_to_b_special_multiplicity_type(total_surjection_relation, total_surjection_relation, some, some, ok).
alloy_to_b_special_multiplicity_type(surjection_relation, surjection_relation, some, set, ok).
alloy_to_b_special_multiplicity_type(injection_relation, injection_relation, lone, set, not_supported). % "lone->" NOT SUPPORTED BY PROB
alloy_to_b_special_multiplicity_type(injection_surjection_relation, injection_surjection_relation, lone, some, not_supported). % "lone->some" NOT SUPPORTED BY PROB
alloy_to_b_special_multiplicity_type(total_bijection_relation, total_bijection_relation, one, some, not_supported). % "one->some" NOT SUPPORTED BY PROB
alloy_to_b_special_multiplicity_type(total_bijection_relation, bijection_relation, one, set, not_supported). % "one->" NOT SUPPORTED BY PROB


% predicates with integer arguments
alloy_to_b_integer_operator(less, less).
alloy_to_b_integer_operator(greater, greater).
alloy_to_b_integer_operator(less_equal, less_equal).
alloy_to_b_integer_operator(greater_equal, greater_equal).


%%%
% Accumulate the translated machine parts and signature names during the translation and build the machine AST afterwards.
% We may need the signature names later on if translating a global scope.
build_machine_ast(b_machine(ListOfMachineParts,_SignatureNames), machine(generated(none,AbstractMachine))) :-
    % filter empty machine parts
    findall(MachinePart, (   member(MachinePart, ListOfMachineParts),
                             MachinePart =.. [_,_,L],
                             L \= []
                         ), NonEmptyMachineSections),
    % properties need to be conjoined
    (   select(properties(none,L), NonEmptyMachineSections, RestListOfUsedMachineParts) ->
        true
    ;   L = [],
        RestListOfUsedMachineParts = NonEmptyMachineSections
    ),
    join_predicates(conjunct, L, FlatL),
    AbstractMachine = abstract_machine(none,machine(none),
                                       machine_header(none,alloytranslation,[]),
                                       [properties(none,FlatL)|RestListOfUsedMachineParts]),
    !.

empty_machine_acc(b_machine([sets(none,[]),constants(none,[]),
                             definitions(none,[]),properties(none,[]),
                             assertions(none,[]),operations(none,[])],[])).

% extend a machine section with a give list of conjuncts
extend_machine_with_conjuncts(_Section, [], Acc, Res) :-
    !,
    Res = Acc.
extend_machine_with_conjuncts(Section, Conjuncts, MAcc1, MAcc2) :-
    join_untyped_ast_nodes(conjunct, Conjuncts, TPred),
    extend_machine_acc(Section, MAcc1, TPred, MAcc2).

extend_machine_acc(signatures, b_machine(MachineParts,SignatureNames), New,
                   b_machine(MachineParts,NewSignatureNames)) :-
    !,
    append(SignatureNames, New, TSignatureNames),
    remove_dups(TSignatureNames, NewSignatureNames).
extend_machine_acc(Functor, b_machine(MachineParts,SignatureNames), New,
                   b_machine([NewMachinePart|RestMachineParts],SignatureNames)) :-
    MachinePart =.. [Functor,none,List],
    select(MachinePart, MachineParts, RestMachineParts), % print(add_new(Functor,New)),nl,nl,
    append(List, [New], NewList),                        % keeps order, but is inefficient!
    NewMachinePart =.. [Functor,none,NewList],
    !.
extend_machine_acc(Functor, _, _, _) :-
    add_internal_error('Failed: ', extend_machine_acc(Functor,_,_,_)),
    fail.

get_signature_names_from_machine_acc(b_machine(_MachineParts,SignatureNames), SignatureNames).

get_command_counter_atom(CommandCounterAtom) :-
    command_counter(CommandCounter),
    number_codes(CommandCounter, CommandCounterCodes),
    atom_codes(CommandCounterAtom, CommandCounterCodes),
    retractall(command_counter(_)),
    NewCommandCounter is CommandCounter+1,
    asserta(command_counter(NewCommandCounter)).

finalize_extending_signatures(MAcc, NewMAcc) :-
    extending_signatures(List),
    !,
    finalize_extending_signatures_aux(MAcc, List, NewMAcc).
finalize_extending_signatures(MAcc, MAcc).

finalize_extending_signatures_aux(MAcc, [], MAcc).
finalize_extending_signatures_aux(MAcc, [(Parent,SubSigs)|T], NewMAcc) :-
    length(SubSigs, AmountOfSubSigs),
    AmountOfSubSigs > 1,
    !,
    maplist(translate_sub_sig, SubSigs, TSubSigs),
    translate_e_p(Parent, TParent),
    extend_machine_acc(properties, MAcc, partition(none,TParent,TSubSigs), MAcc2),
    % equivalent to
    % parent = sub1 \/ sub2 \/ ... \/ subN   &  sub1 /\ sub2 = {} &  ...
    findall(OneSub, member(OneSub/one, SubSigs), OneSubSigs),
    length(OneSubSigs, AmountOneSubSigs),
    (   AmountOneSubSigs = AmountOfSubSigs % we basically have an enumerated set with only singleton elements
    ->
        extend_machine_acc(properties, MAcc2, equal(none,card(none,TParent),integer(none,AmountOneSubSigs)), MAcc3)
    ;   extend_machine_acc(properties, MAcc2, greater_equal(none,card(none,TParent),integer(none,AmountOneSubSigs)), MAcc3)
    ),
    finalize_extending_signatures_aux(MAcc3, T, NewMAcc).
finalize_extending_signatures_aux(MAcc, [_|T], NewMAcc) :-
    finalize_extending_signatures_aux(MAcc, T, NewMAcc).
%%%
translate_sub_sig(Name/_One, B) :-
    translate_e_p(Name, B).

assert_extending_signature(Name, Parent, One) :-
    (   is_sub_signature_of(Name, Parent, One) ->
        true
    ;   debug_format(19, '~w is sub signature of ~w (~w)~n', [Name,Parent,One]),
        assert(is_sub_signature_of(Name, Parent, One))
    ),
    extending_signatures(List),
    select((Parent,SubSigs), List, Rest),
    !,
    (   \+ memberchk(Name/One, SubSigs)
           ->
        retractall(extending_signatures(_)),
        asserta(extending_signatures([(Parent,[Name/One|SubSigs])|Rest]))
    ;   true
    ).
assert_extending_signature(Name, Parent, One) :-
    retract(extending_signatures(List)),
    !,
    asserta(extending_signatures([(Parent,[Name/One])|List])).
assert_extending_signature(Name, Parent, One) :-
    asserta(extending_signatures([(Parent,[Name/One])])).

assert_enumerated_set(Options, Name) :-
    % singleton top-level signatures are defined as an enumerated set like S = {_S}
    member(one, Options),
    % but extending singleton signatures are defined as constants and singleton subsets of their parent type
    \+ member(subset(_), Options),
    \+ member(subsig(_), Options),
    !,
    asserta(enumerated_set(Name)).
assert_enumerated_set(_, _).

assert_singleton_set(Options, Name) :-
    member(one, Options),
    !,
    assert_singleton_set(Name).
assert_singleton_set(_, _).

assert_singleton_set(Name) :-   % debug_format(19,'Singleton set : ~w~n',[Name]),nl,
    % \+ singleton_set(Name),
    asserta(singleton_set(Name)),
    !.
% assert_singleton_set(_).
retract_singleton_set(BPOS, Name) :-
    (   retract(singleton_set(Name)) ->
        true
    ;
        add_message(retract_singleton_set, 'WARNING: no singleton_set fact for: ', Name, BPOS) % can be ok for non-singleton ids
    ).
% retractall(singleton_set(Name)).
retract_state :-
    retractall(uses_seqs),
    retractall(enumerated_set(_)),
    retractall(current_command(_)),
    retractall(command_counter(_)),
    asserta(command_counter(0)),
    retractall(singleton_set(_)),
    retractall(ordered_signature(_)),
    retractall(total_function(_, _)),
    retractall(extending_signatures(_)),
    retractall(is_seq(_)),
    retractall(maxseq(_)),
    retractall(is_sub_signature_of(_, _, _)),
    !.

% an identifier which in B refers not to a set but an element of a set, representing a singleton set in Alloy
is_singleton_set_id(this).      % this is always singleton (TO DO: check this ;-)
is_singleton_set_id(IDName) :-
    singleton_set(IDName).

assert_total_function(identifier(_,Name), SIG) :-
    !,
    assert_total_function(Name, SIG).
assert_total_function(Name, identifier(_,SIG)) :-
    !,
    assert_total_function(Name, SIG).
assert_total_function(Name, SIG) :-
    debug_format(19, 'Total function ~w over ~w~n', [Name,SIG]),
    asserta(total_function(Name, SIG)).

is_total_function(predecessor(_), _). % translation of prev
is_total_function(successor(_), _). % translation of next
is_total_function(function(_,ID,_), ArgType) :-
    !,
    is_total_function(ID, ArgType).
is_total_function(identifier(_,Name), ArgType) :-
    !,
    is_total_function(Name, ArgType).
is_total_function(Name, [ArgType]) :-
    total_function(Name, FunType), % TO DO: check if FunctionType is Superset of ArgType
    (   strip_singleton_set(FunType, SFunType),
        SFunType = identifier(_,ArgType)
    ->
        true
    ;   is_sub_signature_of(ArgType, FunType, _)
    ).

assert_ordered_sig_name(Name) :-
    asserta(ordered_signature(Name)).

is_ordered_signature(IDName) :-
    ordered_signature(IDName).
is_ordered_signature('ordering\'').

is_univ(identifier('univ',_,_)). % universe

is_unary_relation(AlloyTerm) :-
    get_type_and_arity_from_alloy_term(AlloyTerm, _Type, Arity),
    Arity == 1.

is_binary_relation(AlloyTerm) :-
    get_type_and_arity_from_alloy_term(AlloyTerm, _Type, Arity),
    Arity == 2.

can_remove_singleton_set(set_extension(_,[Element]), Element).
% remove_singleton_set(AST,AST).
get_type_and_arity_from_alloy_term(integer(_,_), Type, Arity) :-
    !,
    Type = integer,
    Arity = 1.                  % will get translated to singleton set
get_type_and_arity_from_alloy_term(and(_,_), Type, Arity) :-
    !,
    Type = [untyped],
    Arity = 0.                  % predicate
get_type_and_arity_from_alloy_term(or(_,_), Type, Arity) :-
    !,
    Type = [untyped],
    Arity = 0.                  % predicate
get_type_and_arity_from_alloy_term(iden(_), Type, Arity) :-
    !,
    Type = univ,
    Arity = 2.
get_type_and_arity_from_alloy_term(AlloyTerm, Type, NArity) :-
    AlloyTerm =.. [_|Args],
    member(type(Type,Arity), Args),
    (   is_primitive_type(Type, NArity)
    ;
        (   length(Type, LType),
            LType == Arity,
            NArity = Arity
        )
    ),
    !.
get_type_and_arity_from_alloy_term(AlloyTerm, Type, Arity) :-
    add_warning(alloy2b, 'Could not extract type and arity from Alloy term:', AlloyTerm),
    Type = [untyped],
    Arity = 0.                  % like predicate

% TO DO: maybe there are more primitive types?
% primitive types have arity zero in Alloy for some reason
is_primitive_type(['PrimitiveBoolean'], 1).

% atom_or_identifier_term(ID,ID) :- atom(ID).
% atom_or_identifier_term(identifier(IDName,_,_),IDName).
join_predicates(Op, TArgList, TBinaryP) :-
    alloy_to_b_operator(Op, BOp),
    (   BOp = conjunct
    ;   BOp = disjunct
    ),
    join_untyped_ast_nodes(BOp, TArgList, TBinaryP).

join_untyped_ast_nodes(_, [], truth(none)).
join_untyped_ast_nodes(BOp, [H|T], Conjoined) :-
    join_untyped_ast_nodes(BOp, T, H, Conjoined).

join_untyped_ast_nodes(_, [], Acc, Acc).
join_untyped_ast_nodes(BOp, [H|T], Acc, Conjoined) :-
    NewAcc =.. [BOp,none,Acc,H],
    join_untyped_ast_nodes(BOp, T, NewAcc, Conjoined).


:- begin_tests(full_machine_translation).

test(cards, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/cards.pl').

test(filesystem, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/filesystem.pl').

test(filesystem2, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/filesystem2.pl').

test(self_grandpas, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/selfgrandpas.pl').

test(queens2, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/queens2.pl').

test(river_crossing, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/rivercrossingpuzzle.pl').

test(enum_test, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/enumtest.pl').

test(graphiso, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/graphiso.pl').

test(disj_field_test, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/disjfieldtest.pl').

test(crewAlloc, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/crewalloc.pl').

test(sudoku1, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/sudoku1.pl').

test(sequence_test, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/sequence_test.pl').

test(http, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/http.pl').

test(knights, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/knights_and_knaves.pl').

test(hanoi, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/hanoi.pl').

test(einstein, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/einstein.pl').

test(slot_data, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/slot_data_v2.pl').

test(job_puzzle, []) :-
    load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/job_puzzle.pl').

/*
   test(family_v1,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/family-v1.pl').

   test(hanoi,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/hanoi.pl').

   test(knights,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/knights_and_knaves.pl').

   test(directctrl,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/directctrl.pl').

   test(views,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/views.pl').

   test(mutex,[]) :-
   load_alloy_ast_prolog_file('../../prob_examples/public_examples/Alloy/Prolog/mutex.pl').
 */
:- end_tests(full_machine_translation).
