% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(avl_tools,[avl_height_less_than/2,
    avl_height_compare/3,
    avl_height_compare_up_to/5,
    quick_avl_approximate_size/2, avl_approximate_size_from_height/2,
    avl_fetch_pair/3,
    avl_delete_pair/4,
    avl_apply/5,
    avl_image_interval/4,
    ord_domain_list_to_avl/2,
    check_is_non_empty_avl/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides AVL-tree utilities used by the kernel.').

:- use_module(library(avl)).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(kernel_waitflags,[add_wd_error_span/4]).

% -------------------------------

test_avl_set(node(((int(2),int(3)),int(6)),true,0,node(((int(1),int(2)),int(2)),true,0,empty,empty),node(((int(3),int(4)),int(12)),true,0,empty,empty))).

:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_height_less_than(A,3) )).
:- assert_must_succeed(( avl_tools:avl_height_less_than(empty,1) )).
:- assert_must_fail(( avl_tools:test_avl_set(A), avl_tools:avl_height_less_than(A,2) )).
:- assert_must_fail(( avl_tools:avl_height_less_than(empty,0) )).

% a custom version of avl_height; advantage it will stop when reaching MaxHeight

avl_height_less_than(empty, MaxHeight) :- MaxHeight>0.
avl_height_less_than(node(_,_,B,L,R), H0) :- H0>1,
   H1 is H0-1,
   (   B >= 0 -> avl_height_less_than(R, H1)
           ;      avl_height_less_than(L, H1)
    ).


% efficient way of comparing AVL heights without having to fully traverse larger AVL

:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_height_compare(A,A,eq) )).
:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_height_compare(empty,A,lt) )).
:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_height_compare(A,empty,gt) )).

avl_height_compare(A,B,Res) :- avl_height_compare_up_to(A,B,0,0,_,Res).

:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_height_compare_up_to(empty,A,2,0,lt) )).
:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_height_compare_up_to(empty,A,3,0,eq) )).

% avl_height_compare_up_to(Avl1,Avl2,MaxDiff,ResSz,ResCmp)  -> ResSz: Height of smaller AVL, ResCmp = eq,lt,gt
avl_height_compare_up_to(Avl1,Avl2,MaxDiff,ResSz,ResCmp) :-
   avl_height_compare_up_to(Avl1,Avl2,MaxDiff,0,ResSz,ResCmp).

avl_height_compare_up_to(empty,empty,_,Sz,Sz,eq).
avl_height_compare_up_to(empty,node(_,_,B,L,R),MaxDiff,Sz,Sz,Res) :-
   (avl_height_less_than(node(_,_,B,L,R),MaxDiff) -> Res=eq ; Res=lt).
avl_height_compare_up_to(node(_,_,B,L,R),empty,MaxDiff,Sz,Sz,Res) :-
   (avl_height_less_than(node(_,_,B,L,R),MaxDiff) -> Res=eq ; Res=gt).
avl_height_compare_up_to(node(_,_,B1,L1,R1),node(_,_,B2,L2,R2),MaxDiff,AccSz,ResSz,Res) :-
   ( B1 >= 0 -> A1=R1 ; A1=L1),
   ( B2 >= 0 -> A2=R2 ; A2=L2),
   Acc1 is AccSz+1,
   avl_height_compare_up_to(A1,A2,MaxDiff,Acc1,ResSz,Res).

% -------------------------------

% compute an upper bound for the size of an AVL based on Height (log runtime):
quick_avl_approximate_size(AVL,Size) :- avl_height(AVL,Height), Size is integer(2**Height-1).
% a lower bound could be computed by =POWER(2,(HEIGHT+0.3277)/1.4405)-2  (page 460, Knuth 3)
avl_approximate_size_from_height(Height,Size) :- Size is integer(2**Height-1).


% -------------------------------


% a version of avl_fetch that looks for a pair in the AVL tree whose
%   first component is Key; it assumes that the term ordering gives the
%   first argument higher priority than the second one.
% TO DO ?: extend to records: decompose((K,KY),K,KY). decompose(rec([H|T]),H,T).
avl_fetch_pair(Key, node((K,KY),_,_,L,R),Res) :-
    my_compare(O, Key, K), %((O='=',nonvar(Res)) -> print(comp(Res,KY)),nl ; true), % see TODO below
    avl_fetch_pair_1(O, Key, L, R, KY, Res).

% order of clauses relevant so that safe_avl_member returns elements in order !
avl_fetch_pair_1(<, Key, node((K,KY),_,_,L,R), _, _, Res) :-
    my_compare(O, Key, K),  %((O='=',nonvar(Res)) -> print(comp(Res,KY)),nl ; true), % see TODO below
    avl_fetch_pair_1(O, Key, L, R, KY, Res).
avl_fetch_pair_1(=, _Key,   _L, _R, KY, KY).
avl_fetch_pair_1(>, Key, _, node((K,KY),_,_,L,R),_,Res) :-
    my_compare(O, Key, K),  %((O='=',nonvar(Res)) -> print(comp(Res,KY)),nl ; true), % see TODO below
    avl_fetch_pair_1(O, Key, L, R, KY, Res).

my_compare(O,K1,K2) :- compare(OK,K1,K2),
   (OK=('=') -> true /* leave O free variable */
                % TO DO: think about comparing KY and Res above
           ;  O=OK).

:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_delete_pair((int(3),int(4)),A,true,AA), avl:avl_size(AA,2) )).
:- assert_must_fail(( avl_tools:test_avl_set(A), avl_tools:avl_delete_pair((int(1),int(3)),A,true,_) )).
:- assert_must_succeed(( avl_tools:test_avl_set(A), avl_tools:avl_delete_pair((int(1),int(2)),A,true,AA),
                          avl_tools:avl_delete_pair((int(2),int(3)),AA,true,AAA), avl:avl_size(AAA,1) )).

% an adaptation of avl_delete which deletes a pair just knowing the first argument of the pair
% like avl_fetch_pair it assumes term ordering gives precedence to first argument

avl_delete_pair(Key, AVL0, Val, AVL) :-
    avl_delete_pair5(AVL0, Key, Val, AVL, _).

avl_delete_pair5(node(K,V,B,L,R), Key, Val, AVL, Delta) :-
    K = (KX,_KY), % crucial difference
    compare(C, Key, KX),
    avl_delete_pair(C, Key, Val, AVL, Delta, K, V, B, L, R).

avl_delete_pair(<, Key, Val, AVL, Delta, K, V, B, L, R) :-
    avl_delete_pair5(L, Key, Val, L1, D1),
        B1 is B+D1,
    avl:avl(B1, K, V, L1, R, AVL),
    avl:avl_shrinkage(AVL, D1, Delta).
avl_delete_pair(=, _, Val, AVL, Delta, _, Val, B, L, R) :-
    (   L == empty -> AVL = R, Delta = 1
    ;   R == empty -> AVL = L, Delta = 1
    ;   avl:avl_del_max(L, K, V, L1, D1),
        B1 is B+D1,
        avl:avl(B1, K, V, L1, R, AVL),
        avl:avl_shrinkage(AVL, D1, Delta)
    ).
avl_delete_pair(>, Key, Val, AVL, Delta, K, V, B, L, R) :-
    avl_delete_pair5(R, Key, Val, R1, D1),
    B1 is B-D1,
    avl:avl(B1, K, V, L, R1, AVL),
    avl:avl_shrinkage(AVL, D1, Delta).


% -------------------------------


:- load_files(library(system), [when(compile_time), imports([environ/2])]).
% avl_apply
% similar to avl_fetch_pair: but checks whether we have a function
% and whether the function is defined for the key
:- if(environ(no_wd_checking,true)).
/* faster version without WD-checking : */
:- nl,print('DISABLING WD-CHECKING FOR FUNCTION APPLICATION!'),nl,nl.
avl_apply(Key, node((K,KY),_,_,L,R),Res,_Span,_WF) :-
    compare(O, Key, K),
    avl_apply_1(O, Key, L, R, KY, Res).
avl_apply_1(<, Key, node((K,KY),_,_,L,R), _, _, Res) :-
         compare(O, Key, K),
         avl_apply_1(O, Key, L, R, KY, Res).
avl_apply_1(=, _Key,  _Left,_Right, KY, KY).
avl_apply_1(>, Key, _, node((K,KY),_,_,L,R),_,Res) :-
         compare(O, Key, K),
         avl_apply_1(O, Key, L, R, KY, Res).
:- else.
/* normal version with WD -checking : */
avl_apply(Key, node((K,KY),_,_,L,R),Res,Span,WF) :-
    compare(O, Key, K),
    avl_apply_1(O, Key, L, R, KY, Res,Span,WF).

avl_apply_1(<, Key, NODE, _, _, Res,Span,WF) :-
    (NODE=node((K,KY),_,_,L,R)
      -> compare(O, Key, K),
         avl_apply_1(O, Key, L, R, KY, Res,Span,WF)
       ; add_wd_error_span('function applied outside of domain (#6): ','@fun'(Key,[]),Span,WF)
    ).
avl_apply_1(>, Key, _, NODE,_,Res,Span,WF) :-
    (NODE=node((K,KY),_,_,L,R)
      -> compare(O, Key, K),
         avl_apply_1(O, Key, L, R, KY, Res,Span,WF)
       ; add_wd_error_span('function applied outside of domain (#7): ','@fun'(Key,[]),Span,WF)
    ).
%%avl_apply_1(=, Key,  _Left,_Right, KY, KY, _Span,_WF).
avl_apply_1(=, Key,  Left,Right, KY, Res,Span,WF) :-
   ((Left = node((Key,KY2),_,_,_,_) ;
     Right = node((Key,KY2),_,_,_,_)) % this is only a partial quick test; the next & previous elements could be deeper in the tree
     % TO DO: use optimized version of avl_fetch((Key,KY2),Left) ; avl_fetch((Key,KY2),Right) if preferences:preference(find_abort_values,true),
     ->  add_wd_error_span('function application used for relation: ','@rel'(Key,KY,KY2),Span,WF)
        % we do not  instantiate Res in this case
     ; Res=KY).

:- endif.


% -------------------------------



% similar to avl_apply but uses interval as lookup key
avl_image_interval(From,To, node((K,KY),_,_,L,R),Res) :-
    comp_interval(O, From,To, K),
    avl_image_interval_1(O, From,To, L, R, KY, Res).
avl_image_interval_1(<, From,To, NODE, _, _, Res) :-
    NODE=node((K,KY),_,_,L,R),comp_interval(O, From,To, K),
    avl_image_interval_1(O, From,To, L, R, KY, Res).
avl_image_interval_1(=, _From,_To,  _Left,_Right, KY, Res) :- Res=KY.
avl_image_interval_1(>, From,To, _, NODE,_,Res) :-
    NODE=node((K,KY),_,_,L,R),
    comp_interval(O, From,To, K),
    avl_image_interval_1(O, From,To, L, R, KY, Res).

comp_interval(O,From,To,int(Key)) :-
    (  number(From),Key<From -> O = ('>')  % could be minus_inf
     ; number(To), Key>To -> O = ('<')
     ; O = ('<') ; O = ('=') ; O = ('>')
    ).
% -------------------------------

% like ord_list_to_avl but does not require list of Dom-Range pairs, adds true automatically
:- assert_must_succeed(( avl_tools:ord_domain_list_to_avl([1,2,3],A), avl:avl_fetch(2,A))).

add_true_to_list([],[]).
add_true_to_list([H|T],[H-true|TT]) :- add_true_to_list(T,TT).

ord_domain_list_to_avl(List,Res) :-
   add_true_to_list(List,LT),
   ord_list_to_avl(LT,Res).

% -------------------------------

check_is_non_empty_avl(V) :- var(V),!, add_internal_error('Variable AVL tree:',check_is_non_empty_avl(V)).
check_is_non_empty_avl(empty) :- !, add_internal_error('Empty AVL tree:',check_is_non_empty_avl(empty)).
check_is_non_empty_avl(node(_,_,_,_,_)) :- !.
check_is_non_empty_avl(V) :- add_internal_error('Illegal AVL tree:',check_is_non_empty_avl(V)).


