% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(b_compiler,[b_compile/6, b_optimize/6]).

:- use_module(library(lists)).

:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(error_manager).
:- use_module(btypechecker,[fasttype/2]).
:- use_module(b_interpreter).
:- use_module(custom_explicit_sets).
%:- use_module(bmachine,[b_operation_reads_output_variables/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module compiles set comprehensions into closures; making them independent of the state of the B machine.').

/* compile boolean expression into a closure where the local state and the global
 state has been incorporated, but any parameter is left in the closure */

:- public test_bexpr/3, test_bexpr2/3.
test_bexpr(Expr,[bind(cc,int(1)),bind(nn,int(2))], [bind(db,[(int(1),int(2))])]) :-
    fasttype( +conjunct( +conjunct( +member( global('Name')<<identifier(nn), +identifier('Name')),
                                    +member( global('Code')<<identifier(cc), +identifier('Code'))),
                         +not_member( global('Name')<<identifier(nn),
                                      +domain(set(couple(global('Name'),global('Code')))<<identifier(db)))), Expr).

test_bexpr2(Expr,[],[]) :-
    fasttype( +implication( +equal( seq(any)<<identifier(ss), +empty_sequence),
                            +equal( set(any)<<identifier(res), +empty_set)), Expr).


% difference with b_compile: the states LS, S will not be thrown away by the interpreter
% thus: we do not have to compile all accesses, in particular lazy lookups !
%b_optimize(TExpr,Parameters,_LS,_S,NTExpr,_WF) :- TExpr=b(_,_,Infos), member(already_optimized,Infos),
%   !, NTExpr=TExpr.
b_optimize(TExpr,Parameters,LS,S,NTExpr,WF) :- %print('OPTIMIZE: '),translate:print_bexpr(TExpr),nl, trace,
   append(LS,[bind('$do_not_force_lazy_lookups',pred_true)],NewLS),
   b_compile(TExpr,Parameters,NewLS,S,NTExpr,WF).
   %bsyntaxtree:add_texpr_info_if_new(NTExpr,already_optimized,ResTExpr).

b_compile(TExpr,Parameters,NewLS,S,NTExpr) :-
  b_compile(TExpr,Parameters,NewLS,S,NTExpr,no_wf_available). % for unit tests

:- assert_must_succeed(( b_compiler:test_bexpr(E,L,S),
                         b_compiler:b_compile(E,[nn],L,S,_R)  )).
:- assert_must_succeed(( b_compiler:test_bexpr2(E,L,S),
                         b_compiler:b_compile(E,[ss,res],L,S,_R)  )).

%:- assert_pre(b_compiler:b_compile_boolean_expression(E,Parameters,LS,S,_),
%                  (type_check(E,boolean_expression),type_check(Parameters,list(variable_id)),
%                   type_check(LS,store),type_check(S,store))).
%:- assert_post(b_compiler:b_compile_boolean_expression(_,_,_,_,E), type_check(E,boolean_expression)).

/* probably not worthwhile:
b_compile(TExpr,Parameters,LS,S,NTExpr) :- get_texpr_info(TExpr,Info),
   (memberchk(compiled(Parameters),Info) -> NTExpr = TExpr  ,print(already_compiled(Parameters)),nl, translate:print_bexpr(TExpr),nl
     ; b_compile0(TExpr,Parameters,LS,S,b(E,T,I)), translate:print(compiling(Parameters)),nl, translate:print_bexpr(TExpr),nl,
       NTExpr = b(E,T,[compiled(Parameters)|I])).
*/

b_compile(TExpr,Parameters,LS,S,NTExpr,WF) :- check_is_id_list(Parameters,Parameters1),
   %  print('compile : '),translate:print_bexpr(TExpr),nl,  statistics(runtime,[Start,_]),%%
   if(b_compile1(TExpr,Parameters1,LS,S,NTExpr0,eval,_FullyKnown,WF),
      copy_pos_infos(TExpr,NTExpr0,NTExpr), % ensure position info is not deleted if full expression compiled away, see PROB-412
     (add_internal_error('Compilation failed: ',b_compile(TExpr,Parameters,LS,S,NTExpr,WF)),
      %b_compile1(TExpr,Parameters1,LS,S,NTExpr0,eval,_FullyKnown,WF),
       NTExpr = TExpr)
     ).
     %, bsyntaxtree:check_ast(NTExpr).
     %, check_infos(TExpr,NTExpr).
   %, print('compiled: '), translate:print_bexpr(NTExpr), print(' '),statistics(runtime,[Stop,_]), T is Stop-Start, print(T), print(' ms  '),nl, bsyntaxtree:check_ast(NTExpr). % , print(NTExpr),nl. check_ast

/*
check_infos(Old,New) :- bsyntaxtree:get_texpr_info(Old,IO),
    bsyntaxtree:get_texpr_info(New,IN),
    member(X,IO), \+ member(X,IN),
    bsyntaxtree:important_info(X),
    print(check_infos(Old,New)),nl,
    print('Not copied: '), print(X),fail.
check_infos(_,_). */

:- use_module(bsyntaxtree,[always_well_defined/1]).
:- use_module(translate,[translate_bexpression_with_limit/2]).
%:- use_module(hit_profiler,[add_profile_hit/1]).
b_compile1(TExpr,Parameters,LS,S,ResultTExpr,Eval,FullyKnown1,WF) :-
    TExpr = b(Expr,Type,Infos),
    NTExpr = b(NewUntypedExpr,Type,NewInfos),
    %remove_bt(TExpr,Expr,NewUntypedExpr,NTExpr), % remove top-level Type Information
    %hit_profiler:add_profile_hit(Expr),
    b_compile1_infos(Expr,Infos,Parameters,LS,S,NewUntypedExpr,NewInfos,Eval,FullyKnown,WF),
    % print('  compiled : '),translate:print_bexpr_or_subst(NTExpr),nl,
    % print(res(FullyKnown,Eval,NTExpr)),nl,
    (FullyKnown=true, evaluate_this(Eval,NewUntypedExpr,Type,NewInfos)
     ->  %  print('eval : '),translate:print_bexpr(NTExpr),nl, %
        (nonvar(NewUntypedExpr),
          b_interpreter:b_compute_expression_nowf(NTExpr,LS,S,ResultValue) % if( would be more prudent !?
          -> %print(result(ResultValue)),nl,
             %remove_bt(TExpr,Expr,value(ResultValue),ResultTExpr)
             get_texpr_type(TExpr,TT),ResultTExpr = b(value(ResultValue),TT,[]),
             FullyKnown1 = FullyKnown
              %, print(compiled_value(ResultTExpr)),nl
          ;  translate_bexpression_with_limit(NTExpr,TTS),
             add_internal_error('Undefined Value: ',b_compile1(TTS,Parameters,LS,S,_,_,_,WF)),
             %trace, b_interpreter:b_compute_expression_nowf(NTExpr,LS,S,ResultValue),
             ResultTExpr=NTExpr, FullyKnown1=false
        )
       ;    % ((FullyKnown=true, NTExpr\=b(value(_),_,_)) -> format('not eval: ~w (~w)~n',[NTExpr,Eval]) ; true),
            ResultTExpr=NTExpr,
            (ResultTExpr = b(value(_),_,_) -> FullyKnown1=FullyKnown ; FullyKnown1=false)
     ).


:- use_module(bsyntaxtree,[create_exists_or_let_predicate/3]).
:- use_module(tools_lists,[delete_first/3]).
b_compile1_infos(exists(ExistsPara,Pred),OldInfos,Parameters,LS,S,NExpr,NewInfos,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    get_texpr_ids(ExistsPara,AtomicIds),
    append(Parameters,AtomicIds,NParameters),
    b_compile1(Pred,NParameters,LS,S,NPred,Eval,_FullyKnown1,WF),
    (is_falsity(NPred) -> NExpr = falsity, NewInfos = OldInfos
     ; is_truth(NPred) -> NExpr = truth, NewInfos = OldInfos
     ; create_exists_or_let_predicate(ExistsPara,NPred,b(NExpr,pred,NI)),
       %print('COMPILED EXISTS: '), translate:print_bexpr(b(NExpr,pred,NI)),nl,
       (delete_first(OldInfos,used_ids(_),I1),
        member(used_ids(NewUsedIds),NI) -> NewInfos = [used_ids(NewUsedIds)|I1]
        ; %nl,print(missing_used_ids(OldInfos,NI)),nl, % can happen when we construt a let_predicate above
          update_infos(NExpr,OldInfos,Parameters,NewInfos))
     ).
b_compile1_infos(identifier(Id),OldInfos,Parameters,LS,S,NExpr,NewInfos,Eval,FullyKnown,WF) :- !,
    ( memberchk(Id,Parameters) ->
        NExpr = identifier(Id), FullyKnown=false, NewInfos = OldInfos
    ; Id=op(_) ->
        NExpr = identifier(Id), FullyKnown=false, NewInfos = OldInfos
    ; b_interpreter:lookup_value_in_store_and_global_sets(Id,LS,S,Value) ->
        NExpr = value(Value), (known_value(Value) -> FullyKnown=true ; FullyKnown=false),
        NewInfos = [was_identifier(Id)|OldInfos]
        % TO DO: store in info field the identifier
    ; add_internal_error('Compilation of identifier failed: ',
                          b_compile2(identifier(Id),Parameters,LS,S,NExpr,Eval,FullyKnown,WF)),
       NExpr = identifier(Id), FullyKnown=false, NewInfos = OldInfos
    ).
b_compile1_infos(Expr,OldInfos,_Parameters,_LS,_S,NewUntypedExpr,NewInfos,_Eval,FullyKnown,_WF) :-
   b_ast_cleanup:is_integer_set(Expr,_Set),!,
   NewUntypedExpr=Expr, NewInfos=OldInfos,FullyKnown=true.
b_compile1_infos(operation_call_in_expr(Operation,OpCallParas),OldInfos,Parameters,LS,S,
                 Compiled,NewInfos,Eval,FullyKnown,WF) :- !,
    def_get_texpr_id(Operation,op(OperationName)),
    FullyKnown = false,
    b_compile_l(OpCallParas,Parameters,LS,S,OpCallParaValues,Eval,_,WF),
    b_get_operation_normalized_read_write_info(OperationName,Read,Modified),
    exclude(operation_identifier,Read,ReadVars), % Read also contains operations; TO DO: check if this is ok in other places of source code of ProB; see test 1957
    (Modified=[] -> true ; add_error(b_compiler,'Calling operation that modifies state in expression:',OperationName)),
    % print(compile_call_op(OperationName,ReadVars,Modified,OpCallParaValues)),nl,
    (ReadVars=[]
     -> Compiled = operation_call_in_expr(Operation,OpCallParaValues),
        NewInfos = OldInfos
     ;  maplist(create_value_for_read_variable(LS,S),ReadVars,TRead,TValues),
        NewInfos = [generated_by_b_compiler],
        Compiled = let_expression_global(TRead,TValues,
           b(operation_call_in_expr(Operation,OpCallParaValues),any,OldInfos))
    ).
b_compile1_infos(Expr,OldInfos,Parameters,LS,S,NewUntypedExpr,NewInfos,Eval,FullyKnown,WF) :-
    b_compile2(Expr,Parameters,LS,S,NewUntypedExpr,Eval,FullyKnown,WF),
    update_infos(NewUntypedExpr,OldInfos,Parameters,NewInfos).

:- use_module(bmachine,[b_get_operation_normalized_read_write_info/3]).

operation_identifier(op(_)).
create_value_for_read_variable(LS,S,Variable,TVariable,TValue) :-
    b_interpreter:lookup_value_in_store_and_global_sets(Variable,LS,S,Value),!, % TO DO: try and find type?
    TVariable = b(identifier(Variable),any,[]),
    TValue = b(value(Value),any,[]).
create_value_for_read_variable(_LS,_S,Variable,TVariable,b(value(term(undefined)),any,[])) :-
    TVariable = b(identifier(Variable),any,[]),
    add_internal_error('Cannot find variable (read) while compiling operation:',Variable),fail.

:- use_module(library(ordsets)).
update_infos(forall(EParas,LHS,RHS),Infos,Parameters,NewInfos) :-
  % forall also has used_ids field as it may call exists in negative context
  conjunct_predicates([LHS,RHS],Cond),
  update_infos_aux(EParas,Cond,Infos,Parameters,NewInfos),!.
update_infos(exists(EParas,Cond),Infos,Parameters,NewInfos) :- % we create_exists_or_let_predicate above, but other transformations may generate an exists below:
  update_infos_aux(EParas,Cond,Infos,Parameters,NewInfos),!.
update_infos(BOP,Infos,_,NewInfos) :- bsyntaxtree:syntaxelement(BOP,         [A,B],[], [], [], _),
  wd_and_efficient(BOP), % should not generate WD errors itself
  select(contains_wd_condition,Infos,NewInfos), % there is a WD condition attached
  fast_check_wd(A), fast_check_wd(B),
  %print(removing_wd_condition(BOP)),nl,
  !.
update_infos(_,I,_,I).

update_infos_aux(EParas,Cond,Infos,Parameters,NewInfos) :-
  select_used_ids(UsedIds,Infos,I1,EParas,Cond),
  UsedIds \= [],!,
  % remove those used identifiers which will be compiled into the predicate
  sort(Parameters,SParas),
  ord_intersection(UsedIds,SParas,NewUsedIds),
  %print(update(Parameters,UsedIds,I1,new(NewUsedIds))),nl,
  NewInfos = [used_ids(NewUsedIds)|I1].

:- use_module(bsyntaxtree,[find_identifier_uses/3]).
select_used_ids(UsedIds,Infos,I1,_,_) :- select(used_ids(UsedIds),Infos,I1),!.
  % check_used_ids_info(Parameters,Condition,UsedIds,b_compiler)
select_used_ids(UsedIds,Infos,Infos,_Parameters,Condition) :-
   %add_error(bcompiler,'Expected information of used identifiers in exists operation information : ',Parameters:Infos), %% missing info can happen due to simplifcation rules below !
   find_identifier_uses(Condition, [], UsedIds). % ,print(used(UsedIds,in_exist(Parameters))),nl.
 % TO DO: avoid re-computing used-ids; we should refactor b_compile2 to return new Info field

%is_pred(b(_,pred,_)).


evaluate_this(eval,X,Type,Info) :- !, worth_it_type(Type,X,Info).
evaluate_this(_,X,_Type,_Info) :- wd_and_efficient(X).


% things which are very quick to compute
wd_and_efficient(sequence_extension(_)) :- !.
wd_and_efficient(integer_set(_)) :- !. % will be converted into value(global_set(_))
%% wd_and_efficient(value(_)) :- !. % is already computed
% the following will reduce the size of the representation; usually a good thing;
% we assume only avl_sets apprear here; inner set comprehensions are never computed into symbolic form by b_compile (see known_value below):
%wd_and_efficient(identity(_)) :- !.
wd_and_efficient(range(_)) :- !.
wd_and_efficient(domain(_)) :- !.
wd_and_efficient(domain_restriction(_,_)) :- !.
wd_and_efficient(domain_subtraction(_,_)) :- !.
wd_and_efficient(range_restriction(_,_)) :- !.
wd_and_efficient(range_subtraction(_,_)) :- !.
wd_and_efficient(intersection(_,_)) :- !.
wd_and_efficient(set_subtraction(_,_)) :- !.
wd_and_efficient(image(_,_)) :- !.
% what about union: usually the union of two avl_sets will be smaller than keeping them separate
% some other efficient operators:
wd_and_efficient(X) :- worth_it_int(X),!.
wd_and_efficient(X) :- worth_it_other(X).

fast_check_wd(b(E,_,Infos)) :-
  (nonmember(contains_wd_condition,Infos) -> true ; nonvar(E), E= value(_)).

check_wd(NTExpr) :- bsyntaxtree:always_well_defined(NTExpr). % ,print(wd(NTExpr)),nl.
  %(bsyntaxtree:always_well_defined(NTExpr) -> true ; print(not_guaranteed_wd(NTExpr)),nl).

  % (worth_it_type(Type,X,Info) -> true ;  X=value(_) -> true ; print(not_evaluating(X)),nl,fail).
worth_it_type(integer,X,Info) :- worth_it_int_wd(X),!, check_wd(b(X,integer,Info)).
worth_it_type(T,function(A,B),Info) :- !,
 (check_wd(b(function(A,B),T,Info)) -> true
   ; fail). %print('not_precompiling_function '),translate:print_bexpr(A), print(' @ '), translate:print_bexpr(B),nl,fail).
worth_it_type(integer,X,_) :- worth_it_int(X),!.
worth_it_type(T,X,Info) :- worth_it_wd(X),!, check_wd(b(X,T,Info)).
worth_it_type(set(_),X,_) :- worth_it_set(X),!.
worth_it_type(seq(_),X,_) :- worth_it_set(X),!.
worth_it_type(_,X,_) :- worth_it_other(X).

% TO DO: we should use the predicate has_wd_condition from b_ast_cleanup

% integer operations that could generate WD-errors
worth_it_int_wd(div(_,_)).
worth_it_int_wd(max(_)).
worth_it_int_wd(min(_)).
worth_it_int_wd(mod(_,_)).
worth_it_int_wd(power_of(_,_)).
worth_it_int_wd(card(_)). % has WD condition !!
worth_it_int_wd(size(_)).

% other operations that could generate WD-errors
worth_it_wd(first(_)).
worth_it_wd(last(_)).
worth_it_wd(tail(_)).
worth_it_wd(front(_)).
worth_it_wd(restrict_front(_,_)).
worth_it_wd(restrict_tail(_,_)).
worth_it_wd(rel_iterate(_,_)). %% could be expensive
worth_it_wd(general_intersection(_)).
worth_it_wd(general_concat(_)).

% to do: check other operators that could be not well-defined !

worth_it_int(unary_minus(_)).
worth_it_int(add(_,_)).
worth_it_int(minus(_,_)).
worth_it_int(multiplication(_,_)).
worth_it_int(max_int).
worth_it_int(min_int).

%worth_it_set(empty_set). % treated below
%worth_it_set(empty_sequence). % treated below
%worth_it_set(closure(_)).  % this is closure1; it is currently always kept symbolic and can be counterproductive to compile; e.g., card(closure1(%x.(x : 1 .. 200|x + 1)))
%% ?? need to be more careful here; can be expensive --> need to keep track which parts need definitely to be evaluated
worth_it_set(integer_set(_)). % will be converted into value(global_set(_))
worth_it_set(comprehension_set(A,B)) :- b_ast_cleanup:is_integer_set(comprehension_set(A,B),_).
worth_it_set(interval(_,_)).
worth_it_set(reflexive_closure(_)).
worth_it_set(reverse(_)).
worth_it_set(domain(_)). worth_it_set(range(_)).
worth_it_set(union(_,_)). worth_it_set(intersection(_,_)).  worth_it_set(set_subtraction(_,_)).
worth_it_set(image(_,_)). worth_it_set(composition(_,_)).
worth_it_set(domain_restriction(_,_)). worth_it_set(domain_subtraction(_,_)).
worth_it_set(range_restriction(_,_)). worth_it_set(range_subtraction(_,_)).
worth_it_set(direct_product(_,_)). worth_it_set(parallel_product(_,_)).
worth_it_set(iteration(_,_)).
worth_it_set(set_extension(_)).
worth_it_set(sequence_extension(_)).
worth_it_set(rev(_)).
worth_it_set(concat(_,_)).
worth_it_set(seq(_)). % will be kept symbolic anyway
worth_it_set(seq1(_)). % will be kept symbolic anyway
worth_it_set(iseq(_)). % will be kept symbolic anyway
worth_it_set(iseq1(_)). % will be kept symbolic anyway
worth_it_set(perm(_)). % will be kept symbolic anyway
worth_it_set(pow_subset(_)). % will be kept symbolic anyway
worth_it_set(pow1_subset(_)). % will be kept symbolic anyway
worth_it_set(fin_subset(_)). % will be kept symbolic anyway
worth_it_set(fin1_subset(_)). % will be kept symbolic anyway
worth_it_set(general_union(_)).
worth_it_set(identity(_)).
worth_it_set(first_of_pair(_)).  % can also produce set
worth_it_set(second_of_pair(_)). % can also produce set
worth_it_set(cartesian_product(_,_)). % will usually be kept symbolic
worth_it_set(string_set). % ensure we have values inside compiled closures, e.g., for custom_explicit_set card computations
worth_it_set(bool_set).

:- use_module(external_functions,[not_declarative/1,external_fun_has_wd_condition/1]).
worth_it_other(boolean_true).
worth_it_other(boolean_false).
worth_it_other(string(_)).
worth_it_other(first_of_pair(_)).
worth_it_other(second_of_pair(_)).
worth_it_other(couple(_,_)).
worth_it_other(record_field(_,_)).
worth_it_other(external_function_call(FunName,_)) :-
   \+ not_declarative(FunName),
   \+ external_fun_has_wd_condition(FunName). %print(compile(FunName)),nl.
worth_it_other(rec(_)).
worth_it_other(struct(_)).
%worth_it_other(value(_)). % does not have to be evaluted; already a value

/*
% To do: distinguish which parts definitely need to be evaluated
%dont_eval_subexpressions(member).  % special membership code can be used
%dont_eval_subexpressions(not_member).  % special membership code can be used
dont_eval_subexpressions(disjunct(_,_)). % it may not be necessary to eval RHS
dont_eval_subexpressions(implication(_,_)).  % it may not be necessary to eval RHS
*/

:- use_module(kernel_objects,[equal_object/3]).
:- use_module(kernel_mappings,[binary_boolean_operator/3]).
:- use_module(kernel_tools,[filter_cannot_match/4, get_template_for_filter_cannot_match/2]).
b_compile2(Exp,Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- var(Exp), !,
    add_internal_error('Variable Expression: ',b_compile2(Exp,Parameters,LS,S,NExpr,Eval,FullyKnown,WF)),
    NExpr=Exp, FullyKnown=false.
b_compile2(truth,_,_LS,_S,NExpr,_Eval,FullyKnown,_WF) :- !, NExpr=truth, FullyKnown=false. % predicates never fully known
b_compile2(falsity,_,_LS,_S,NExpr,_Eval,FullyKnown,_WF) :- !, NExpr=falsity, FullyKnown=false. % predicates never fully known
b_compile2(empty_set,_,_LS,_S,NExpr,_Eval,FullyKnown,_WF) :- !, NExpr=value([]), FullyKnown=true.
b_compile2(empty_sequence,_,_LS,_S,NExpr,_Eval,FullyKnown,_WF) :- !, NExpr=value([]), FullyKnown=true.
b_compile2(value(Val),_Parameters,_LS,_S,NExpr,_Eval,FullyKnown,_WF) :- !,
    NExpr=value(Val),
    (known_value(Val) -> % ground(Val) -> known_value ?
     FullyKnown=true ; FullyKnown=false).
b_compile2(integer(Val),_Parameters,_LS,_S,NExpr,_Eval,FullyKnown,_WF) :- !,
    NExpr=value(int(Val)), (number(Val) -> FullyKnown=true ; FullyKnown=false).
b_compile2(lazy_lookup_pred(Id),Parameters,LS,S,NExpr,_Eval,FullyKnown,_WF) :- !,
    ( memberchk(Id,Parameters) ->  %print(lazy_lookup_pred_id_as_parameter(Id,Parameters)),nl,
        NExpr = lazy_lookup_pred(Id), FullyKnown=false
    ; otherwise -> compute_lazy_lookup(Id,lazy_lookup_pred(Id),LS,S,NExpr,FullyKnown)
    ).
b_compile2(lazy_lookup_expr(Id),Parameters,LS,S,NExpr,_Eval,FullyKnown,_WF) :- !,
    ( memberchk(Id,Parameters) -> %print(lazy_lookup_expr_id_as_parameter(Id,Parameters)),nl,
        NExpr = lazy_lookup_expr(Id), FullyKnown=false
    ; otherwise -> compute_lazy_lookup(Id,lazy_lookup_expr(Id),LS,S,NExpr,FullyKnown)
    ).
% treat a few common cases, more efficiently than syntaxtransformation
b_compile2(equal(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA,WF),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB,WF),
    ( FullyKnownA=true, FullyKnownB=true,
      NExprA = b(value(VA),_,_), NExprB = b(value(VB),_,_)
      ->  %print(computing_eq(VA,VB)),nl,
          (equal_object(VA,VB,b_compile2_1) -> NExpr = truth ; NExpr = falsity)
          %, print(result(NExpr)),nl
       ;  generate_equality(NExprA,NExprB,NExpr)
    ).
b_compile2(not_equal(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA,WF),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB,WF),
    ( FullyKnownA=true, FullyKnownB=true,
      NExprA = b(value(VA),_,_), NExprB = b(value(VB),_,_)
      ->  %print(computing_neq(VA,VB)),nl,
          (kernel_objects:equal_object(VA,VB,b_compile2_2) -> NExpr = falsity ; NExpr = truth)
          %, print(result(NExpr)),nl
       ;  NExpr = not_equal(NExprA,NExprB)
    ).
b_compile2(member(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_,WF),
    (NExprB = b(value(X),BT,BI)
      -> (X==[], always_well_defined(A) -> NExpr = falsity
           ; b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF),
             % check if we can decide membership (e.g., 1:{1})
            (quick_test_membership1(NExprA,X,PRes)
              -> convert_pred_res(PRes,NExpr)
              % TO DO: maybe optimize x: {y|P(y)}  --> P(x) ; by commenting in code below
              % However, ensure test 273 succeeds and closure equality correctly detected (e.g., for BV16 = {bt|bt : BIT_VECTOR & bv_size(bt) = 16})
              % ; (get_texpr_id(NExprA,IDA),
              %    get_comprehension_set(NExprB,IDB,Pred), get_texpr_info(NExprB,Info),
              %     print(rename(IDB,IDA,Info)),nl,
              %    bsyntaxtree:rename_bt(Pred,[rename(IDB,IDA)],PredA)) -> get_texpr_expr(PredA,NExpr)
              ;  custom_explicit_sets:singleton_set(X,El) % replace x:{El} -> x=El
                -> get_texpr_type(NExprA,TA), NExpr = equal(NExprA,b(value(El),TA,[]))
               ;  get_template_for_filter_cannot_match(NExprA,VA)
                  -> filter_cannot_match(X,VA,NewX,Filtered),
                      % (Filtered=false -> true ; print(filtered(NewX,VA,X)),nl),
                     (Filtered=false -> NExpr = member(NExprA,NExprB)
                       ; (NewX==[], always_well_defined(A)) -> NExpr = falsity
                       ; custom_explicit_sets:singleton_set(NewX,El)
                          -> get_texpr_type(NExprA,TA), NExpr = equal(NExprA,b(value(El),TA,[]))
                       ; otherwise -> NExpr = member(NExprA, b(value(NewX),BT,BI))
                     )
                    % end fo filtering
               ;  NExpr = member(NExprA,NExprB)
             )
                % , print('compile: '),print(NExpr),nl
         )
      ; NExpr = member(NExprA,NExprB),
        b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF)
    ).
% TO DO: also optimize things like: (ev1 |-> 0) |-> eg : #221:{((1|->0)|->3),((2|->0)|->0),...,((233|->0)|->218),((234|->0)|->228)}

b_compile2(not_member(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_,WF),
    (NExprB = b(value(X),_,_)
      -> (X==[], always_well_defined(A) -> NExpr = truth
         ;  b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF),
            (quick_test_membership1(NExprA,X,PRes)
             -> convert_neg_pred_res(PRes,NExpr)
             ; NExpr = not_member(NExprA,NExprB))
          )
    ;  NExpr = not_member(NExprA,NExprB),
       b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF)
    ).
b_compile2(function(Fun,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    (is_extension_function_gt1(Fun) -> Eval1=false % extension function applications treated lazily
     % TO DO: first evaluate B, if known then only evaluate relevant part of extension function
     ; Eval1=Eval),
    b_compile1(Fun,Parameters,LS,S,NExprA,Eval1,FullyKnown1,WF),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown12),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnown2,WF),
    %tools_printing:print_term_summary(compiled_function(FullyKnown1,FullyKnown2,NExprA,NExprB)),nl,
    (FullyKnown2=true, preferences:preference(find_abort_values,false),
      % check if we have enough information to apply a partially specified function as a list
      % e.g. if the function is [(int(1),A),(int(2),B)] and the argument is int(2): we can apply the function without knowing A or B;  can have a dramatic importance when expanding unversal quantifiers !
     can_apply_partially_specified_function(NExprA,NExprB,ResultExpr,WF)
      -> NExpr = ResultExpr,
         (FullyKnown12=true -> FullyKnown=true
           ; ResultExpr=value(ResultValue),known_value(ResultValue) -> FullyKnown=true
           ; FullyKnown = false)
      ; % print(cannot_apply_function(NExprA,NExprB,FullyKnown2)),nl,
        NExpr = function(NExprA,NExprB), FullyKnown = FullyKnown12).

b_compile2(conjunct(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF),
    (is_falsity(NExprA) -> NExpr = falsity
     ;
    % TO DO: propagate static information from A to B, e.g. if A :: x=10  -> add x=10 to LS/Parameters
    % also: try and detect unsatisfiable predicates beforehand
      b_compile1(B,Parameters,LS,S,NExprB,Eval,_,WF),
      (is_falsity(NExprB) -> NExpr = falsity
       ; is_truth(NExprB) -> get_texpr_expr(NExprA,NExpr)
       ; is_truth(NExprA) -> get_texpr_expr(NExprB,NExpr)
       ; NExpr = conjunct(NExprA,NExprB))
    ).
b_compile2(disjunct(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF),
    (is_truth(NExprA) -> NExpr = truth
     ; b_compile1(B,Parameters,LS,S,NExprB,false,_,WF), % does not have to be executed in a successful branch: set Eval to false to avoid computing expensive expressions which may not be needed
       (is_truth(NExprB) -> NExpr = truth
       ; is_falsity(NExprA) -> get_texpr_expr(NExprB,NExpr)
       ; is_falsity(NExprB) -> get_texpr_expr(NExprA,NExpr)
       ; NExpr = disjunct(NExprA,NExprB))
    ).
b_compile2(implication(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF),
    (is_falsity(NExprA) -> NExpr = truth
     ; is_truth(NExprA) -> b_compile1(B,Parameters,LS,S,NExprB,false,_,WF), get_texpr_expr(NExprB,NExpr)
     ; NExpr = implication(NExprA,NExprB),
       b_compile1(B,Parameters,LS,S,NExprB,false,_,WF) % does not have to be executed in a succesful branch: set Eval to false to avoid computing expensive expressions which may not be needed
    ).
b_compile2(negation(A),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_,WF),
    ( get_negated_expr(NExprA,NegatedA) -> NExpr = NegatedA
      ; NExpr = negation(NExprA)
    ).

/* b_compile2(comprehension_set(CParas,Body),Parameters,LS,S,NExpr,_Eval,FullyKnown,WF) :-
     sort(Parameters,SParas),
     b_ast_cleanup:not_occurs_in_predicate(SParas,Body), %can only be applied if Body does not reference Parameters
     !,
    b_generate_inner_closure_if_necessary(Parameters,CParas,Body,LS,S,Result), % will itself call b_compile
    NExpr = value(Result),
    (ground(Result) -> FullyKnown=true ; FullyKnown=false). */
b_compile2(forall(ForallPara,A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    get_texpr_ids(ForallPara,AtomicIds),
    append(Parameters,AtomicIds,NParameters),
    b_compile1(A,NParameters,LS,S,NA,Eval,_FullyKnownA,WF),
    (is_falsity(NA) -> NExpr = truth
     ; b_compile1(B,NParameters,LS,S,NB,Eval,_FullyKnownB,WF),
       (is_truth(NB) -> NExpr = truth
         ; NExpr = forall(ForallPara,NA,NB)
           % TO DO ??: if we have an equality -> replace Body by value and remove quantifier
           %,print('COMPIlED: '),translate:print_bexpr(b(NExpr,pred,[])),nl
       )
    ).
% some special cases to avoid calling syntaxtransformation
b_compile2(record_field(A,FieldName),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA,WF),
    % First: we can try compute even if value not fully known !, we just need record field list
    (get_record_field(NExprA,FieldName,NExpr,FullyKnown) -> true
     ; NExpr = record_field(NExprA,FieldName), FullyKnown=FullyKnownA).
    %((FullyKnown=false,NExprA=b(VV,_,_),nonvar(VV),VV=value(V)) , trace, known_value(V) ; true).
b_compile2(couple(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    %NExpr = couple(NExprA,NExprB),
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown1,WF),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnown2,WF),
    (FullyKnown \== true,  % evaluation will already combine the result
     NExprA=b(value(VA),_,_),NExprB=b(value(VB),_,_)
      -> NExpr = value((VA,VB))
      ;  NExpr = couple(NExprA,NExprB)).
b_compile2(image(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    %NExpr = image(NExprA,NExprB),
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown1,WF),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnown2,WF),
    (FullyKnown \= true, NExprA=b(value(VA),BT,BI),
     NExprB=b(value(VB),_,_), nonvar(VB),
     custom_explicit_sets:singleton_set(VB,VBX)
          -> filter_cannot_match(VA,(VBX,_),NewVA,_Filtered),  %nl,print(filtered_image(VBX,VA,NewVA)),nl,
             NExpr = image(b(value(NewVA),BT,BI),NExprB)
        % TO DO: add case where VB==[] or VA==[] are the empty set
       ;  NExpr = image(NExprA,NExprB)
    ). %, print('compile image: '),translate:print_bexpr(NExpr),nl.
    % TO DO: do something like can_apply_partially_specified_function; or filter_can_match
b_compile2(domain(A),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown1,WF),
    (FullyKnown1=false,get_texpr_expr(NExprA,NA),evaluate_domain(NA,Domain) % try and compute domain even if not fully known
      ->  FullyKnown = true, NExpr = value(Domain)
      ;   FullyKnown = FullyKnown1, NExpr = domain(NExprA)).
b_compile2(range(A),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown1,WF),
    (FullyKnown1=false,evaluate_range(NExprA,Range) % try and compute range even if not fully known
      ->  FullyKnown = true, NExpr = value(Range)
      ;   FullyKnown = FullyKnown1, NExpr = range(NExprA)).
b_compile2(assign(LHS_List,RHS_List),Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- !,
    %print(compile_assign(LHS_List)),nl,
    NExpr = assign(New_LHS_List,NewRHS_List), FullyKnown = false,
    maplist(b_compile_lhs(Parameters,LS,S,Eval,WF),LHS_List,New_LHS_List),
    b_compile_l(RHS_List,Parameters,LS,S,NewRHS_List,Eval,_FullyKnown2,WF).
b_compile2(assign_single_id(ID,B),Parameters,LS,S,assign_single_id(ID,NExprB),Eval,FullyKnown,WF) :- !,
    FullyKnown = false,
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_,WF).
b_compile2(operation_call(Operation,OpCallResults,OpCallParas),Parameters,LS,S,InlinedOpCall,Eval,FullyKnown,WF) :-
    %print(opcall(Operation,OpCallResults,OpCallParas)),nl,
    % TO DO: evaluate if we should also use the let_expression_global approach used for operation_call_in_expr
    !,
    def_get_texpr_id(Operation,op(OperationName)),
    FullyKnown = false,
    b_compile_l(OpCallParas,Parameters,LS,S,OpCallParaValues,Eval,_,WF),
    TopLevel=false,
    b_get_machine_operation_for_animation(OperationName,OpResults,OpFormalParameters,Body,_OType,TopLevel),
    bsyntaxtree:replace_ids_by_exprs(Body,OpResults,OpCallResults,Body2),
    get_texpr_ids(OpFormalParameters,OpIds),
    append(OpIds,Parameters,InnerParas),
    b_compile1(Body2,InnerParas,[],S,NBody,Eval,FullyKnown,WF), % do not pass local state: this may override constants,... from the called machine
    % Note: global state S should be valid for all machines in currently loaded specification
    get_texpr_ids(OpCallResults,OpRIds),
    append(OpRIds,Parameters,LocalKnownParas),
    simplify_assignment(OpFormalParameters,OpCallParaValues,LHSFormalParas,RHSCallVals), %%
    split_formal_parameters(LHSFormalParas,RHSCallVals,LocalKnownParas,
                                FreshOpParas,FreshCallVals,
                                ClashOpParas, ClashCallVals), % only set up VAR for fresh Paras
     %   print(split(LHSFormalParas,LocalKnownParas,fresh(FreshOpParas),clash(ClashOpParas))),nl,
    (FreshOpParas=[] -> get_texpr_expr(NBody,Let1)
      ; create_equality_for_let(FreshOpParas,FreshCallVals,Equality1),
        Let1 = let(FreshOpParas,Equality1,NBody)),
    (ClashOpParas = []
      -> InlinedOpCall = Let1
      ;  %nl,print(clash(ClashOpParas)),nl,
         maplist(create_fresh_id,ClashOpParas,FreshCopy),
         create_equality_for_let(ClashOpParas,FreshCopy,Equality2), % copy Values from Fresh Ids
         create_equality_for_let(FreshCopy,ClashCallVals,Equality3), % assign Parameter Vals to Fresh Ids
         get_texpr_pos(Body,BodyPos), BodyPosInfo = [nodeid(BodyPos)],
         Let2 = let(ClashOpParas,Equality2,b(Let1,subst,[BodyPosInfo])),
         InlinedOpCall = let(FreshCopy,Equality3,b(Let2,subst,[BodyPosInfo]))
    ).
    %, print('Compiled operation call: '), translate:print_subst(b(InlinedOpCall,subst,BodyPosInfo)),nl,nl.
% Maybe this version is faster when no OpCallResults:
%b_compile2(operation_call(Operation,OpCallResults,OpCallParas),Parameters,LS,S,InlinedOpCall,Eval,FullyKnown,WF) :-
%    OpCallResults=[],
%    !,
%    def_get_texpr_id(Operation,op(OperationName)),
%    FullyKnown = false,
%%    b_compile_l(Results,Parameters,LS,S,NResults,Eval,_,WF),
%    b_compile_l(OpCallParas,Parameters,LS,S,OpCallParaValues,Eval,_,WF),
%    TopLevel=false,
%    b_get_machine_operation_for_animation(OperationName,OpResults,OpParameters,Body,_OType,TopLevel),
%    % Note: input parameters cannot be assigned to: so replace ids is ok
%    % However, we cannot replace output parameters: these can be assigned to and we can have aliasing
%    % Note: no aliasing is allowed in OpCallResults (cf Atelier-B handbook x,x <-- op not allowed)
%    bsyntaxtree:replace_ids_by_exprs(Body,OpParameters,OpCallParaValues,Body2),
%    get_texpr_ids(OpResults,OpIds),
%    append(OpIds,Parameters,InnerParas),
%    b_compile1(Body2,InnerParas,LS,S,NBody,Eval,FullyKnown,WF),
%    FinalBody = NBody.
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :-
    functor(Expr,BOP,2),
    binary_boolean_operator(BOP,_,_),!, % from kernel_mappings, slightly more efficient than syntaxtransformation
    arg(1,Expr,A),
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA,WF),
    arg(2,Expr,B),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB,WF),
    (eval_binary_bool(FullyKnownA,NExprA,FullyKnownB,NExprB,BOP,Result)
      -> NExpr = Result
       ; functor(NExpr,BOP,2), arg(1,NExpr,NExprA), arg(2,NExpr,NExprB)
    ).
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :-
    functor(Expr,BOP,2),
    kernel_mappings:binary_function(BOP,_,_),!, % slightly more efficient than syntaxtransformation
    arg(1,Expr,A),
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA,WF),
    combine_fully_known(FullyKnownA,FullyKnownB,FullyKnown),
    arg(2,Expr,B),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB,WF),
    functor(NExpr,BOP,2), arg(1,NExpr,NExprA), arg(2,NExpr,NExprB).
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :-
    functor(Expr,UOP,1),
    kernel_mappings:unary_function(UOP,_,_),!, % slightly more efficient than syntaxtransformation
    arg(1,Expr,A),
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown,WF),
    functor(NExpr,UOP,1), arg(1,NExpr,NExprA).
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown,WF) :- % GENERAL CASE
    %hit_profiler:add_profile_hit(Expr),
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr),
    get_texpr_ids(Names,Ids), append(Parameters,Ids,NParameters),
    b_compile_l(Subs,NParameters,LS,S,NSubs,Eval,FullyKnown,WF).

b_compile_l([],_,_,_,[],_Eval,true,_WF).
b_compile_l([Expr|ExprRest],P,LS,S,[NExpr|NExprRest],Eval,FullyKnown,WF) :-
    b_compile1(Expr,P,LS,S,NExpr,Eval,FullyKnown1,WF),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile_l(ExprRest,P,LS,S,NExprRest,Eval,FullyKnown2,WF).

% detecth when a parameter clashes with existing known ids
id_clash(LocalKnownParas,TID) :- def_get_texpr_id(TID,ID), member(ID,LocalKnownParas).

/* comment in if we want to compile inner comprehension sets:
:- use_module(closures,[construct_closure_if_necessary/4]).
b_generate_inner_closure_if_necessary(OuterParameters,ClosureParas,Condition,LocalState,State,Result) :-
    split_names_and_types(ClosureParas,Names,Types),
    append(OuterParameters,Names,FullNames),
    b_compile(Condition,FullNames,LocalState,State,ClosurePred),
    print('INNER COMPILE: '), translate:print_bexpr(ClosurePred),nl,
    construct_closure_if_necessary(Names,Types,ClosurePred,Result).
    */


% the LHS contains l-values which should not be looked up in the environment
% however, we need to deal with things like f(i) := RHS and compile i
b_compile_lhs(Parameters,LS,S,Eval,WF,TExpr,NewTExpr) :-
    TExpr = b(Expr,Type,Info), NewTExpr = b(NewExpr,Type,Info),
    b_compile_lhs_aux(Expr,Parameters,LS,S,NewExpr,Eval,WF).

b_compile_lhs_aux(function(A,B),Parameters,LS,S,function(NewA,NewB),Eval,WF) :- !,
    b_compile_lhs(Parameters,LS,S,Eval,WF,A,NewA), % we can have nexted function calls f(e)(g) :=
    b_compile1(B,Parameters,LS,S,NewB,Eval,_,WF).
b_compile_lhs_aux(E,_,_,_,E,_,_WF). % other LHS should be identifier -> just copy

get_record_field(b(value(Val),_,_),FieldName,value(Field),FullyKnown) :- !,
   nonvar(Val), Val=rec(Fields),
   safe_get_field(Fields,FieldName,Field),
   known_value(Field,FullyKnown).
get_record_field(b(function(FUN,ARG),_Type,_Info),FieldName,function(ReducedFUN,ARG),false) :-
   %  {1|->rec(a:1,b:2),...}(x)'b  -->  {1|->2,...}(x)
   %nl,print(get_field(FUN,ARG,FieldName)),nl,
   % TO DO: also allow nested functions calls  {1|->rec(a:1,b:2),...}(x)(y)'b or other reduction operators
   get_field_in_range(FUN,FieldName,ReducedFUN).
safe_get_field(V,_,_) :- var(V),!,fail.
safe_get_field([field(N,V)|T],FieldName,Result) :- nonvar(N),
   (N=FieldName -> Result=V ; safe_get_field(T,FieldName,Result)).

:- use_module(probsrc(custom_explicit_sets),[expand_custom_set_to_list/4, convert_to_avl/2]).
:- use_module(probsrc(avl_tools),[avl_height_less_than/2]).
get_field_in_range(b(value(OldVal),set(couple(Dom,OldRange)),Info),FieldName,
                   b(value(NewVal),set(couple(Dom,NewRange)),Info)) :-
        nonvar(OldVal), OldVal=avl_set(AVL),
        avl_height_less_than(AVL,10),
        OldRange = record(Fields),
        member(field(FieldName,NewRange),Fields),!,
        expand_custom_set_to_list(avl_set(AVL),List,_,get_field_in_range),
        maplist(safe_get_range_field(FieldName),List,NewList),
        %print(reduced_get_field_in_range(NewList)),nl,
        convert_to_avl(NewList,NewVal).

% compute the field access for the range elements:
safe_get_range_field(FieldName,(Dom,rec(Fields)),(Dom,Field)) :- safe_get_field(Fields,FieldName,Field).


eval_binary_bool(true,b(value(ValA),_,_),true,b(value(ValB),_,_),BOP,Result) :-
   eval_binary_aux(BOP,ValA,ValB,Result).
   %print(eval(BOP,ValA,ValB,Result)),nl,nl.
eval_binary_aux(less,int(A),int(B),Result) :-
   (A < B -> Result = truth ; Result = falsity).
eval_binary_aux(greater,int(A),int(B),Result) :-
   (A > B -> Result = truth ; Result = falsity).
eval_binary_aux(less_equal,int(A),int(B),Result) :-
   (A =< B -> Result = truth ; Result = falsity).
eval_binary_aux(greater_equal,int(A),int(B),Result) :-
   (A >= B -> Result = truth ; Result = falsity).
% TO DO: other operators are subset, strict_subset, ...

get_negated_expr(b(E,pred,_),Res) :- negate_expr_aux(E,Res).
negate_expr_aux(falsity,truth).
negate_expr_aux(truth,falsity).
negate_expr_aux(member(X,S),not_member(X,S)).
negate_expr_aux(not_member(X,S),member(X,S)).
negate_expr_aux(subset(X,S),not_subset(X,S)).
negate_expr_aux(not_subset(X,S),subset(X,S)).
negate_expr_aux(subset_strict(X,S),not_subset_strict(X,S)).
negate_expr_aux(not_subset_strict(X,S),subset_strict(X,S)).
negate_expr_aux(equal(X,S),not_equal(X,S)).
negate_expr_aux(not_equal(X,S),equal(X,S)).
negate_expr_aux(less(X,S),greater_equal(X,S)).
negate_expr_aux(greater_equal(X,S),less(X,S)).
negate_expr_aux(less_equal(X,S),greater(X,S)).
negate_expr_aux(greater(X,S),less_equal(X,S)).

:- use_module(store,[lookup_value_for_existing_id/4]).
compute_lazy_lookup(Id,Expr,LS,S,NExpr,FullyKnown) :-
    lookup_value_for_existing_id(Id,LS,S,(Evaluated,Value)),
    !,
    known_value(Value,FullyKnown),
   (Evaluated==pred_true -> NExpr = value(Value)
    ; FullyKnown==true -> NExpr = value(Value)
    ; memberchk('$do_not_force_lazy_lookups',LS) -> NExpr = Expr % we do not force lookup: there could be WD issues ! TO DO: examine Infos for wd_condition
     %, print(not_forcing(Id,Value)),nl
    ; if(Evaluated = pred_true, % will force evaluation ! Note: this means shared expression inside a compiled expression must be well-defined
            NExpr = value(Value),
            (add_internal_error('Compilation failed: ',compute_lazy_lookup(Id,Expr,LS,S,NExpr,FullyKnown) ),
            fail))
    ).
compute_lazy_lookup(Id,E,LS,S,LazyValue,_) :-
    add_internal_error('Compilation failed, illegal lazy_lookup_value: ',compute_lazy_lookup(Id,E,LS,S,LazyValue,_)),
    fail.

combine_fully_known(true,A,A).
combine_fully_known(false,_,false).

% is an extension function with at least two elements
is_extension_function_gt1(b(A,_,_Info)) :- is_extension_function_aux(A).
is_extension_function_aux(set_extension([_,_|_])).
is_extension_function_aux(sequence_extension([_,_|_])).

:- use_module(bsyntaxtree,[is_set_type/2]).
% check if we can apply an argument to a partially specified function (as a list)
%can_apply_partially_specified_function(Fun,X,_) :- print(fun(Fun)),nl,print(val(X)),nl,fail.
can_apply_partially_specified_function(b(Expr,TYPE,_), b(value(X),TA,_),ResultExpr,WF) :-
    can_apply_aux(Expr,TYPE,X,TA,ResultExpr,WF).

can_apply_aux(value(VAL),TYPE,X,TA,value(ResultValue),WF) :-
    is_set_type(TYPE,couple(TA,_TB)),
    lookup_result(VAL,X,ResultValue,WF).
can_apply_aux(sequence_extension(List),_,X,integer,ResultExpr,_WF) :- % important, e.g., for solving test 1551: {b|[a,b,c](2)=333} =res & a : res & b:res
    nonvar(X), X=int(Index), number(Index),
    % this will prevent computing rest of sequence extension: can hide WD issues !
    % print(apply_seq_extension(Index,List)),nl,
    nth1(Index,List,b(ResultExpr,_,_)).

:- use_module(kernel_equality,[equality_objects_wf/4]).
%lookup_result(VAR,ArgValue,Result,WF) :- var(VAR),!,kernel_mappings:kernel_call_apply_to(VAR,ArgValue,Result,unknown,unknown,WF).
% the clause above is unsound; as apply_to CAN RAISE WD ERRORS !! what if we have IF 1:dom(f) THEN f(1) ELSE f(0) END; we do not want to compute f(1) !  it can also instantiate unbound variables !
    % the idea was to avoid that we wait on the full function, before compiled closure can be evaluated; test 1552
    % important for e.g. a = {(1, {([]|->2)} ), (2, const(1, [a(1)]))}
    % or: a = {1 |-> {[] |-> 2},   2 |-> (dom({pi,ff,p|((pi : seq(INTEGER) & ff : INTEGER) & p : seq(INTEGER)) & (p |-> ff : [a(1)](1) & pi = 1 -> p)}) \/ {[] |-> 1})}
    % test 1552 is currently skipped
% TO DO: think whether there are other deterministic computations that can be done in b_compile: prj1,prj2, record-field access ?  this would ensure that closures,... can be computed earlier
lookup_result(VAR,_,_,_WF) :- var(VAR),!,fail.
lookup_result([(HFrom,HTo)|T],X,Result,WF) :-
    equality_objects_wf(HFrom,X,PredRes,WF),  %print(compare(PredRes,HFrom,X)),nl,
    nonvar(PredRes), % only proceed if we have enough information to determine the result of the comparison
    (PredRes = pred_true -> Result = HTo % we have found the value; we assume there is no other value later
      % TO DO: if find_abort_values=true: look later in the list if for all other HFrom PredRes=pred_false
      ; lookup_result(T,X,Result,WF)).

% get_comprehension_set(b(SET,_,_),ID,PRED) :- get_comprehension_set_aux(SET,ID,PRED).
% get_comprehension_set_aux(comprehension_set([TID],PRED),ID,PRED) :-
%   get_texpr_id(TID,ID).
% get_comprehension_set_aux(value(V),ID,PRED) :- nonvar(V), V = closure([ID],_,PRED).




quick_test_membership1(b(value(VA),_,_),VB,Result) :-
   quick_test_membership_aux(VB,VA,Result).
quick_test_membership_aux(VAR,_,_) :- var(VAR),!,fail.
quick_test_membership_aux([],_,pred_false).
quick_test_membership_aux(avl_set(AVL),X,Result) :-
  custom_explicit_sets:quick_test_avl_membership(AVL,X,Result). % we could also check that in case X is not ground but partially instantiated, whether there is a possible match in AVL (if not Result = pred_false)
quick_test_membership_aux(global_set(GS),X,Result) :- nonvar(X), X=int(IX), nonvar(IX),
   custom_explicit_sets:membership_global_set(GS,X,R,_WF),
   nonvar(R), Result=R.
quick_test_membership_aux(closure(P,T,B),X,Result) :- nonvar(X), X=int(IX), nonvar(IX),
   custom_explicit_sets:is_interval_closure(P,T,B,LOW,UP), number(LOW),number(UP),
   kernel_equality:in_nat_range_test(X,int(LOW),int(UP),R),
   nonvar(R), Result=R.
quick_test_membership_aux(closure(P,T,B),X,Result) :- X==[],  % check {} : POW(_) / FIN(_)  and  {} : POW1(_) / FIN1(_)
   custom_explicit_sets:is_powerset_closure(closure(P,T,B),Kind,_),
   (contains_empty_set(Kind) -> Result = pred_true ; Result=pred_false).

contains_empty_set(pow).
contains_empty_set(fin).

convert_pred_res(pred_false,falsity).
convert_pred_res(pred_true,truth).
convert_neg_pred_res(pred_true,falsity).
convert_neg_pred_res(pred_false,truth).

generate_equality(b(A,_,_),b(B,_,_),Res) :-
   ( generate_equality_aux(A,B,Res) ;
     generate_equality_aux(B,A,Res) ),
   !.
generate_equality(A,B,equal(A,B)).

generate_equality_aux(convert_bool(AA),value(PT),Res) :-
   PT==pred_true,
   get_texpr_expr(AA,TA),
   !,
   %print(simplify_bool(AA,PT)),nl,
   Res = TA.
generate_equality_aux(convert_bool(AA),value(PF),Res) :-
   PF==pred_false,!,
   %print(simplify_bool(AA,PF)),nl,
   Res = negation(AA).

known_value(X,FullyKnown) :- (known_value(X) -> FullyKnown=true ; FullyKnown=false).
known_value(X) :- nonvar(X), known_value2(X).
known_value2(global_set(GS)) :- nonvar(GS).
known_value2(freetype(GS)) :- nonvar(GS).
known_value2(closure(P,T,B)) :- known_closure(P,T,B).
   %Note: for other closures: they still may have to be computed; some of the computations in wd_and_efficient could be expensive for closures
known_value2(avl_set(_)). % already fully normalised
known_value2((X,Y)) :- known_value(X),known_value(Y).
known_value2(fd(A,B)) :- number(A),atomic(B).
known_value2(int(N)) :- number(N).
known_value2([]).
known_value2(pred_true).
known_value2(pred_false).
known_value2(string(S)) :- atomic(S).
known_value2([H|T]) :- known_value(H), known_value(T).
%known_value2(record(Fields)) :- known_fields(Fields).
known_value2(rec(Fields)) :- known_fields(Fields).


known_closure(P,T,B) :- custom_explicit_sets:is_interval_closure(P,T,B,Low,Up),  !,
   number(Low), number(Up),
   Up < Low+1000.
known_closure(P,T,B) :-
   is_cartesian_product_closure(closure(P,T,B),A1,A2),
   !,% we could call is_cartesian_product_closure_aux(P,T,B,A1,A2)
   known_value(A1), known_value(A2).

known_fields(X) :- var(X),!,fail.
known_fields([]).
known_fields([field(N,V)|T]) :- ground(N),known_value(V),known_fields(T).

%b_evaluate1(TExpr,Parameters,LS,S,ResultTExpr,FullyKnown) :-

check_is_id_list([],[]).
check_is_id_list([H|T],Res) :-
   (H=b(identifier(_),_,_)
    -> add_internal_error('Typed id list as argument: ',check_is_id_list([H|T],Res)), % use get_texpr_ids or maplist(def_get_texpr_id,P,PIDs)
       maplist(def_get_texpr_id,[H|T],Res)
    ; Res=[H|T]).

:- use_module(kernel_frozen_info,[kfi_domain/2]).

:- use_module(custom_explicit_sets,[construct_interval_closure/3]).
% try and evaluate domain of a list (avl_set already dealt with separately)
%evaluate_domain(X,_) :- print(dom(X)),nl,fail.
evaluate_domain(sequence_extension(L),Domain) :- length(L,Len),
    % this will prevent computing range of sequence extension: can hide WD issues ! TO DO: check wd info and/or preference
   construct_interval_closure(1,Len,Domain).
% TO DO: set_extension when possible ?
evaluate_domain(value(L),Domain) :-
   get_domain(L,Domain).

get_domain(V,Dom) :- var(V),!,kfi_domain(V,Dom). % try infer domain from attached co-routines
get_domain([],[]).
get_domain([V|VT],Result) :- nonvar(V), V=(D,_),
    known_value(D),get_list_domain(VT,DT),
    sort([D|DT],Result).
% we could also do this:
%get_domain(closure(Par,Typ,Body),Result) :-
%    is_lambda_value_domain_closure(Par,Typ,Body, DomainValue,Expr), check_wd(Expr),
%    Result=DomainValue.
get_list_domain(V,_) :- var(V),!,fail.
get_list_domain([],[]).
get_list_domain([V|VT],[D|DT]) :- nonvar(V), V=(D,_),
    known_value(D),get_list_domain(VT,DT).


:- use_module(kernel_frozen_info,[kfi_range/2]).

% try and evaluate range of a list (avl_set already dealt with separately)
evaluate_range(b(value(L),_T,_),Range) :-
   get_range(L,Range).
get_range(V,Dom) :- var(V),!,kfi_range(V,Dom). % try infer range from attached co-routines
get_range([],[]).
get_range([V|VT],Result) :- nonvar(V), V=(_,D),
   known_value(D),
   get_list_range(VT,DT),
   sort([D|DT],Result).

get_list_range(V,_) :- var(V),!,fail.
get_list_range([],[]).
get_list_range([V|VT],[D|DT]) :-
   nonvar(V), V=(_,D),
   known_value(D),
   get_list_range(VT,DT).

create_equality_for_let(LHS,RHS,Conj) :-
    maplist(create_eq,LHS,RHS,CL),
    conjunct_predicates(CL,Conj).
create_eq(Id,Expr,TPred) :- safe_create_texpr(equal(Id,Expr),pred,TPred).

% split formal operation call parameters into those which clash and those that do not:
split_formal_parameters([],[],_,[],[],[],[]).
split_formal_parameters([Formal1|FormalT],[Val1|VT],LocalKnownParas,Fresh,FV,Clash,CV) :-
   (id_clash(LocalKnownParas,Formal1)
    ->  Fresh = FreshT, FV=FVT, Clash = [Formal1|ClashT], CV = [Val1|CVT]
    ;   Fresh = [Formal1|FreshT], FV = [Val1|FVT], Clash = ClashT, CV = CVT
   ), split_formal_parameters(FormalT,VT,LocalKnownParas,FreshT,FVT,ClashT,CVT).

:- use_module(gensym,[gensym/2]).
create_fresh_id(b(identifier(_),T,I),b(identifier(FRESHID),T,I)) :-
   gensym('__COMPILED_ID__',FRESHID).

:- use_module(bsyntaxtree, [get_texpr_pos/2,same_id/3]).
% compute which assignments are really necessary, remove skip assignments x := x
simplify_assignment([],[],[],[]).
simplify_assignment([ID1|T1],[ID2|T2],Res1,Res2) :-
    (same_id(ID1,ID2,_) -> Res1=TR1, Res2=TR2 ; Res1=[ID1|TR1], Res2=[ID2|TR2]),
    simplify_assignment(T1,T2,TR1,TR2).
