% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(coverage_statistics,[tcltk_compute_coverage/1, tcltk_compute_coverage_and_enabling/1,
                               tcltk_compute_coverage_and_degree/1,
                               tcltk_compute_min_max/1,pretty_print_min_max_coverage_to_file/1,
                               compute_the_coverage/5,pretty_print_coverage_information_to_file/1,
                               operation_hit/2,query_node_hit/2, uncovered_operation/1]).

:- use_module(state_space).
:- use_module(tools).

:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module is responsible for computing state space coverage info.').

/* --------------------------------------------------- */
:- dynamic var_min_max/3.

tcltk_compute_min_max(_) :-
   retractall(var_min_max(_,_,_)),
   visited_expression(_ID,State), %print(treat(_ID)),nl,
   treat_state_for_min_max(State),fail.
tcltk_compute_min_max(list(Res)) :-
   findall(Info, get_info(Info,constant,true), CRes1),
   findall(Info, get_info(Info,constant,false), CRes2),
   findall(Info, get_info(Info,variable,true), VRes1),
   findall(Info, get_info(Info,variable,false), VRes2),
   combine_results(CRes1,CRes2,CRes),
   combine_results(VRes1,VRes2,VRes),
   (CRes=[] -> Res = ['VARIABLES'|VRes] ;
     append(['CONSTANTS'|CRes],['VARIABLES'|VRes],Res)).

combine_results([],[],R) :- !, R=[].
combine_results(Eq,[],R) :- R = ['/* all only one value: */'|Eq].
combine_results([],NotEq,R) :- R = ['/* all more than one value: */'|NotEq].
combine_results(Eq,NotEq,R) :-
    append(['/* with one value: */'|Eq],['/* with more than one value: */'|NotEq],R).

:- use_module(tools_files,[write_lines_to_file/2]).
pretty_print_min_max_coverage_to_file(Filename) :-
   tcltk_compute_min_max(list(Res)),
   write_lines_to_file(Filename,Res).

:- use_module(b_machine_hierarchy,[abstract_constant/2,concrete_constant/2]).

get_info(Info,Type,Equal) :- var_min_max(Var,Min,Max),
   (Min==Max -> Equal=true ; Equal=false),
   ((abstract_constant(Var,_) ; concrete_constant(Var,_))
     -> Type = constant ; Type = variable),
   translate_min_max(Var,Min,Max,Info).

translate_min_max(Var,M,M,Info) :- !, translate:translate_bvalue(M,TMin),
   ajoin([Var,'=',TMin],Info).
translate_min_max(Var,pred_false,pred_true,Info) :- !,
   ajoin([Var,':BOOL'],Info).
translate_min_max(Var,fd(X,T),fd(Y,T),Info) :- !,
   gen_fd_set(X,Y,T,Set), translate:translate_bvalue(Set,TS),
   ajoin([Var,':',TS],Info).
translate_min_max(Var,Min,Max,Info) :-
    translate:translate_bvalue(Min,TMin),
    translate:translate_bvalue(Max,TMax),
    ajoin([Var,':',TMin,'..',TMax],Info).

gen_fd_set(X,Y,_T,R) :- X>Y,!, R=[].
gen_fd_set(X,Y,T,[fd(X,T)|R]) :- X1 is X+1,
   gen_fd_set(X1,Y,T,R).

:- use_module(specfile,[property/2]).
% treat a visited state for min_max
treat_state_for_min_max(const_and_vars(_,S)) :- !, treat_state_for_min_max(S).
treat_state_for_min_max(concrete_constants(S)) :- !, treat_state_for_min_max(S).
treat_state_for_min_max(csp_and_b(_,S)) :- !, treat_state_for_min_max(S).
treat_state_for_min_max(root) :- !.
treat_state_for_min_max([]) :- !.
treat_state_for_min_max(List) :- List=[_|_],!,member(bind(VAR,VAL),List), treat_var(VAR,VAL),fail.
treat_state_for_min_max(OtherState) :- \+ b_mode,
   property(OtherState,'='(Variable,Value)),
   treat_var(Variable,Value),fail.

treat_var(Variable,Value) :- retract(var_min_max(Variable,Min,Max)),!,
    (is_smaller(Value,Min) -> NewMin = Value ; NewMin=Min),
    (is_smaller(Max,Value) -> NewMax = Value ; NewMax=Max),
    assert(var_min_max(Variable,NewMin,NewMax)).
treat_var(Variable,Value) :-
    assert(var_min_max(Variable,Value,Value)).

:- use_module(library(avl),[avl_size/2, avl_max/2]).
is_smaller(int(X),int(Y)) :- !, X < Y.
is_smaller(string(X),string(Y)) :- !, X @< Y.
is_smaller(pred_true,_) :- !,fail.
is_smaller(pred_false,_) :- !.
is_smaller(fd(X,T),fd(Y,T)) :- !, X < Y.
is_smaller((X1,X2),(Y1,Y2)) :- !, (is_smaller(X1,Y1) -> true ; X1=Y1, is_smaller(X2,Y2)).
is_smaller([],Y) :- !, Y \== [].
is_smaller(_,[]) :- !, fail.
is_smaller(avl_set(X),avl_set(Y)) :- avl_size(X,SX), avl_size(Y,SY),!,
   (SX<SY -> true
    ; SX=SY, avl_max(X,MaxX), avl_max(Y,MaxY),
      (is_smaller(MaxX,MaxY) % used to be   X@<Y
        -> true
        ; MaxX=MaxY -> X @< Y) % if same maximum; look at all values; TO DO: maybe look at avl_min first
    ).
is_smaller(X,Y) :- X @< Y.

/* --------------------------------------------------- */

pretty_print_coverage_information_to_file(FileName) :-
   tcltk_compute_coverage(list(Res)),
   write_lines_to_file(FileName,Res).

tcltk_compute_coverage(list(Res)) :-
   compute_the_coverage(Res,_,_,false,false).
tcltk_compute_coverage_and_enabling(list(Res)) :-
   compute_the_coverage(Res,_,_,true,false).
tcltk_compute_coverage_and_degree(list(Res)) :-
   compute_the_coverage(Res,_,_,false,true).

compute_the_coverage(Res,TotalNodeNr,TotalTransSum,ShowEnabledInfo,ShowDegreeInfo) :-
   retractall(operation_hit(_,_,_)),
   reset_node_hit(ShowEnabledInfo,ShowDegreeInfo),
   compute_coverage(ShowEnabledInfo,ShowDegreeInfo),
   findall(S2,
       (operation_hit(OpS,Nr),
        %translate:translate_bexpression(Op,OpS),
        string_concatenate(':',Nr,S1),
        string_concatenate(OpS,S1,S2)),Res1_),
        sort(Res1_,Res1),
   findall(S2,
       (query_node_hit(Prop,Nr),
        string_concatenate(':',Nr,S1),
        string_concatenate(Prop,S1,S2)),Res0),
   findall(OpName, uncovered_operation(OpName),Res2_),
   sort(Res2_,Res2),
   total_number_of_transitions(TotalTransSum),
   (ShowEnabledInfo=false
     -> ResDE = []
     ;
       findall(DE, def_enables(DE),EnRes1),
       findall(DE, pos_enables(DE),EnRes2),
       append(['DEFINITELY_ENABLED_AFTER'|EnRes1],['POSSIBLY_ENABLES'|EnRes2],ResDE)
   ),
   (ShowDegreeInfo=false
     -> Res00 = ResDE
     ;
       findall(DegR, (max_degree(OpDeg,MaxDeg), tools:ajoin([OpDeg,'->',MaxDeg],DegR)), DegRes),
       append(['MAXIMUM_DEGREE'|DegRes],ResDE,Res00)
   ),
   get_op_string('UNCOVERED_',UNCOVEREDSTR),
   get_op_string('TOTAL_',TOTALSTR), get_op_string('COVERED_',COVEREDSTR),
   append([UNCOVEREDSTR|Res2],Res00,Res01),
   append([TOTALSTR,TotalTransSum,COVEREDSTR|Res1],
          Res01,OpRes),
   append(['STATES'|Res0],OpRes,Res),
   query_node_hit(total,TotalNodeNr).
   %% ,pge_algo: disabled_operations_stats(ResPGE), % in case no disabling stats are given an empty list is returned
   %% append(Res_,ResPGE,Res).

:- use_module(specfile,[get_specification_description/2]).
get_op_string(PREFIX,Res) :-
   get_specification_description(operations,OP),
   string_concatenate(PREFIX,OP,Res).


def_enables(X) :- operation_hit(OpName1,_,_),
   definitely_enabled_after(OpName1,OpName2),
   string_concatenate('====>',OpName2,R1),
   string_concatenate(OpName1,R1,X).
pos_enables(X) :- operation_hit(OpName1,_,_),
   possibly_enables(OpName1,OpName2),
  % \+ definitely_enabled_after(OpName1,OpName2),
   string_concatenate('=?+=>',OpName2,R1),
   string_concatenate(OpName1,R1,X).

/* note: assumes that same operation name cannot appear with multiple arity */

total_number_of_transitions(TotalTransSum) :-
  findall(Nr,operation_hit(_,_,Nr),List),
  sum_list(List,0,TotalTransSum).

sum_list([],N,N).
sum_list([H|T],SoFar,Res) :- S1 is SoFar+H, sum_list(T,S1,Res).

operation_hit(OpTerm,Nr) :-
   operation_hit(OpTerm,_ID,Nr).

:- dynamic operation_hit/3.
operation_hit(operation(_),node__2,55).

:- use_module(specfile,[get_possible_event/1]).

uncovered_operation(OpName) :- specfile:get_possible_event(OpName),
   \+(operation_hit(OpName,_,_)).

add_hit(OpName,ID) :-
  ((OpName = partition(N,_Op,P)) -> LookupName = partition(N,_,P) ; LookupName = OpName),
  (retract(operation_hit(LookupName,NID,Nr))
   -> (N1 is Nr+1, assert(operation_hit(LookupName,NID,N1)))
   ;  (assert(operation_hit(OpName,ID,1)))
  ).

query_node_hit(Prop,Nr) :- node_hit(Prop,Nr).
query_node_hit(total,Nr) :- node_hit(open,N1), node_hit(live,N2), node_hit(deadlocked,N3),
   Nr is N1+N2+N3.
:- dynamic node_hit/2.
node_hit(property,22).
add_node_hit(OpName) :-
  (retract(node_hit(OpName,Nr))
   -> N1 is Nr+1, assert(node_hit(OpName,N1))
   ;  assert(node_hit(OpName,1))
  ).

reset_node_hit(ShowEnabledInfo,ShowDegreeInfo) :-
    retractall(node_hit(_,_)),
    assert(node_hit(deadlocked,0)),
    (specfile:b_or_z_mode -> assert(node_hit(invariant_violated,0)),
                    assert(node_hit(invariant_not_checked,0))
                  ; true),
    assert(node_hit(open,0)), assert(node_hit(live,0)),
    init_enabling_degree_info(ShowEnabledInfo,ShowDegreeInfo).

:- use_module(specfile,[get_operation_name/2]).

compute_coverage(_,_) :-
   visited_expression_id(ID),
   (not_all_transitions_added(ID)
       -> add_node_hit(open)
       ;  (transition(ID,_,_) -> add_node_hit(live)
           ; user:is_not_interesting(ID) -> add_node_hit(ignored)
           ; max_reached_or_timeout_for_node(ID) -> true  % e.g., if MAX_OPERATIONS set to 0
           ; otherwise -> add_node_hit(deadlocked))
    ),
    (max_reached_or_timeout_for_node(ID)
       -> add_node_hit(explored_but_not_all_transitions_computed)
       ; true
    ),
   (invariant_violated(ID) -> add_node_hit(invariant_violated) ; true),
   (time_out_for_invariant(ID) -> add_node_hit(invariant_time_out) ; true),
   (time_out_for_assertions(ID) -> add_node_hit(assertions_time_out) ; true),
   (invariant_not_yet_checked(ID) -> add_node_hit(invariant_not_checked) ; true),
   (state_error(ID,_,Err),Err\=invariant_violated ->
      (Err=abort_error(TYPE,_,_,_)
         -> add_node_hit(TYPE)
         ;  add_node_hit(state_error))
     ; true),
   fail.
compute_coverage(ShowEnabledInfo,ShowDegreeInfo) :-
   transition(ID,Operation,NID),
   nonvar(Operation),
   get_operation_name(Operation,OpName),
   add_hit(OpName,ID),
   (ShowEnabledInfo=true -> update_enabling_info(ID,OpName,NID) ; true),
   (ShowDegreeInfo=true -> update_degree_info(ID,OpName) ; true),
   fail.
%compute_coverage :- print_message('Covered Properties: '),
%  node_hit(OpName,Nr),
%   print(OpName), print('  nodes:'), print(Nr),nl,
%   fail.
%compute_coverage :- print_message('Covered Operations: '),
%  operation_hit(OpName,_ID,Nr),
%   print(OpName), print('  hits:'), print(Nr),nl,
%   fail.
%compute_coverage :- nl, print_message('Uncovered Operations: '),
%   uncovered_operation(OpName),
%   print(OpName), print(' '), fail.
compute_coverage(_ShowEnabledInfo,ShowDegreeInfo) :-
    (ShowDegreeInfo=true -> finalise_degree_info ; true),
    nl.




:- dynamic definitely_enabled_after/2, possibly_enables/2.  /* true if after an operation another is always enabled */
:- dynamic degree_in_id/3, max_degree/2.

:- use_module(specfile,[b_mode/0]).
:- use_module(bmachine,[b_is_operation_name/1,b_get_machine_operation/4]).
init_enabling_degree_info(ShowEnabledInfo,_ShowDegreeInfo) :-
    retractall(definitely_enabled_after(_,_)),
    retractall(possibly_enables(_,_)),
    retractall(degree_in_id(_,_,_)),
    retractall(max_degree(_,_)),
    b_mode,
    ShowEnabledInfo=true,
    (b_is_operation_name(OpName1) ; OpName1 = '$initialise_machine'),
    b_is_operation_name(OpName2),
    assert(definitely_enabled_after(OpName1,OpName2)),
    fail.
init_enabling_degree_info(_,_).

finalise_degree_info :- update_last_degree.

update_enabling_info(_ID,OpName,NewID) :- %print(upd(_ID,OpName)),nl,
    definitely_enabled_after(OpName,OpName2),
    (opname_possible(NewID,OpName2)
      -> true
      ; retract(definitely_enabled_after(OpName,OpName2))
      %(retract(definitely_enabled_after(OpName,OpName2)) -> print(retract(OpName,OpName2)))
    ),fail.
update_enabling_info(ID,OpName,NewID) :- %print(upd(_ID,OpName)),nl,
   opname_possible(NewID,OpName2),  % Note: can succeed multiple times with same OpName2 !
   \+ possibly_enables(OpName,OpName2),
   \+ opname_possible(ID,OpName2),
   assert(possibly_enables(OpName,OpName2)),
   fail.
update_enabling_info(_,_,_).

% NOTE: we suppose that the transitions are grouped by operation name for each state id; may not be true for CSP ??
update_degree_info(ID,OpName) :-
   get_cur_degree(ID,OpName,Degree), D1 is Degree+1,
   assert(degree_in_id(ID,OpName,D1)),
   fail.
update_degree_info(_,_).

update_last_degree :- (retract(degree_in_id(_,OpName,D)) -> update_degree(OpName,D) ; true).

get_cur_degree(ID,OpName,Degree) :-
   (retract(degree_in_id(ID,OpName,D)) -> Degree=D
     ; update_last_degree,
       Degree=0).
update_degree(OpName,Degree) :-
   ((max_degree(OpName,Max), Max>=Degree) -> true
    ; retractall(max_degree(OpName,_)),  format('Update degree ~w -> ~w~n',[OpName,Degree]),
     assert(max_degree(OpName,Degree))
   ).
opname_possible(ID,OpName) :- transition(ID,Op,_),get_operation_name(Op,OpName).

% -------------------------------------


