% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

/*
 * This module allows to register predicates as "debugging predicates".
 * When such a predicate is marked by the prefix operator ~~,
 * during term expansion, the term is replaced by:
 *  - if remove_debug_calls is true:
 *    one of the term's arguments, namely the one that is marked
 *    by a star (*) during registration
 *  - if remove_debug_calls is false:
 *    only the prefix operator ~~ is removed, leaving the rest
 *    of the term unchanged.
 *
 * Example:
 *
 * :- register_debugging_call log(-,*).
 * log(Info,X) :-
 *   print(Info),print(' enter: '),print(X),nl,
 *   call(X),
 *   print(Info),print(' leave: '),print(X),nl.
 *
 * :- enable_debugging_calls.
 * p1 :- ~~log(p1, member(X,[a])).
 * :- disable_debugging_calls.
 * p2 :- ~~log(p2, member(X,[a])).
 *
 * a call of p1 prints:
 *   p1 enter: member(_123,[a])
 *   p1 leave: member(a,[a])
 * while a call to p2 prints nothing
 *
 */

:- module(debugging_calls,[register_debugging_call/1,
                           enable_debugging_calls/0,
                           disable_debugging_calls/0,
                           remove_debugging_calls/4 % called by term expander
                          ]).

:- use_module(library(lists)).
:- use_module(library(codesio)).

:- use_module(module_information,[module_info/2]).

:- module_info(group,infrastructure).
:- module_info(description,'This module provides functionality to specify meta-calls that (like a must-not-fail call) that are short-circuited when compiling production code.').

:- op(1150, fx, register_debugging_call).
:- op(300, fy, ~~).

:- dynamic remove_debug_calls/0.
% remove_debug_calls.

:- dynamic is_debug_call/3.

safe_atom_codes(A,C) :-
    on_exception(error(representation_error(max_atom_length),_),atom_codes(A,C),
       (print(exception(max_atom_length)),nl,A='???')).

%avl_term(X) :- var(X),!,fail.
%avl_term(avl_set(_)).
%avl_term((A,B)):- (avl_term(A) -> true ; avl_term(B)).
%avl_term([H|_]) :- avl_term(H).

remove_debugging_calls(Layout,Term,Layout,Term) :- var(Term),!.
%remove_debugging_calls(Layout,nl,Layout,format('  line:~w~n',[Layout])) :- !.
    % comment in to see nl (newline term_expansion) info with :- prolog_flag(source_info,_,on).
remove_debugging_calls(LayoutIn,when(Cond,Body),LayoutOut,ResTerm) :- !,
   ResTerm = when(Cond,NewBody), % do not expand inside condition
   (LayoutIn = []
    -> LayoutBody = [],
       LayoutOut = []
    ;  LayoutIn = [LayoutWhen,LayoutCond,LayoutBody],
       LayoutOut = [LayoutWhen,LayoutCond,LayoutBodyOut]),
   remove_debugging_calls(LayoutBody,Body,LayoutBodyOut,NewBody).
%remove_debugging_calls(Layout,ground(Term),Layout,Term2) :- !, Term2 = (nonvar(Term),(debugging_calls:avl_term(Term) -> print(ground_check(Term)),nl, trace ; true),ground(Term)). %% comment in to observe ground checks
remove_debugging_calls(LayoutIn,Term,Layout2,Term2) :-
	functor(Term, ~~, 1),
	arg(1,Term,Term1),
	nonvar(Term1),!,functor(Term1,Functor,Arity),
	( LayoutIn = [] -> Layout1 = []
	; LayoutIn = [_,Layout1]),
	( is_debug_call(Functor,Arity,CallPos) ->
	    ( remove_debug_calls ->
		safe_arg(CallPos,Term1,TermI),
		( Layout1 = [] -> LayoutI = []
		; Layout1 = [_|ArgLayout] ->
		    nth1(CallPos,ArgLayout,LayoutI)),
		remove_debugging_calls(LayoutI,TermI,Layout2,Term2)
	    ; otherwise ->
		remove_debugging_calls(Layout1,Term1,Layout2,Term2))
	; otherwise ->
	    write_to_codes(Functor/Arity,M1),
	    append(["encountered unregistered ",M1,
		    " tagged for debugging removal"],M3),
	    safe_atom_codes(Msg1,M3),
	    print_message(error,Msg1),
	    Term2 = Term1
	).
%remove_debugging_calls(LayoutIn,on_exception(E,Call,Code),LayoutOut,Term2) :- !,
%   LayoutOut = LayoutIn,
%   print(on_exception(E,Call,Code)),nl,
%   Term2 = on_exception(E,Call,(format('~n**EXCEPTION = ~w~n~n',[E]),Code)).
%remove_debugging_calls(LayoutIn,time_out(Call,TO,Res),LayoutOut,Term2) :- !,
%   LayoutOut = LayoutIn,
%   print(time_out(Call,TO,Res)),nl,
%   Term2 = ((Call = _:C -> true ; Call=C),
%            functor(C,F,N),format('**TIMEOUT CALL ~w/~w = ~w ms~n',[F,N,TO]),
%            time_out(Call,TO,Res),
%            (Res=time_out -> format('**TIMEOUT OCCURED ~w/~w after ~w ms~n',[F,N,TO]) ; true)).
remove_debugging_calls(LayoutIn,Term1,LayoutOut,Term2) :-
	!,functor(Term1,Functor,Arity),
	functor(Term2,Functor,Arity),
	( LayoutIn = [] ->
	    LayoutOut = [],
	    remove_debugging_calls_args_nl(1,Arity,Term1,Term2)
	; LayoutIn = [Pos|Layout1] ->
	    LayoutOut = [Pos|Layout2],
	    remove_debugging_calls_args(Layout1,Term1,Layout2,Term2,1)
	; otherwise ->
	    LayoutIn = LayoutOut,
	    Term1 = Term2).

remove_debugging_calls_args([],_,[],_,_).
remove_debugging_calls_args([L1|Layout1],Term1,[L2|Layout2],Term2,N) :-
	safe_arg(N,Term1,In), safe_arg(N,Term2,Out),
	remove_debugging_calls(L1,In,L2,Out),
	N2 is N+1,
	remove_debugging_calls_args(Layout1,Term1,Layout2,Term2,N2).

safe_arg(N,Term,Arg) :- compound(Term), arg(N,Term,Arg).

remove_debugging_calls_args_nl(N,Max,_,_) :- N > Max,!.
remove_debugging_calls_args_nl(N,Max,Term1,Term2) :-
	N =< Max,
	safe_arg(N,Term1,In), safe_arg(N,Term2,Out),
	remove_debugging_calls([],In,_,Out),
	N2 is N+1,
	remove_debugging_calls_args_nl(N2,Max,Term1,Term2).

register_debugging_call((Call1,Call2)) :-
	!,register_debugging_call(Call1),
	register_debugging_call(Call2).
register_debugging_call(Call) :-
	functor(Call,Functor,Arity),
	Call =.. [_|Args],
	( nth1(Pos,Args,*) ->
	    ( nth1(NPos,Args,*), NPos \= Pos ->
		register_error_msg(Call)
	    ; otherwise ->
		retractall(is_debug_call(Functor,Arity,_)),
		assert(is_debug_call(Functor,Arity,Pos)))
	; otherwise ->
	    register_error_msg(Call)).
register_error_msg(Call) :-
	write_to_codes(Call,M),
	append("register_debug_call has wrong argument: ", M, Codes),
	safe_atom_codes(Msg,Codes),
	print_message(error,Msg).


enable_debugging_calls :-
	retractall(remove_debug_calls).
disable_debugging_calls :-
	(remove_debug_calls -> true; assert(remove_debug_calls)).

/*
 * This code uses the term_expansion/6 hook, wich
 *   is a multifile-predicate. Due to a bug in SICStus Prolog,
 *   it is not possible to mix uncompiled and compiled code
 *   for such a predicate.
 *   So we always compile the term_expansion hook
 */
:- prolog_flag(compiling,Pre,compactcode),
	compile(debugging_calls_te),
	prolog_flag(compiling,_,Pre).
