% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(fastwhen,[]).

:- use_module(library(lists)).
:- use_module(library(terms)).

% :- fastwhen( Pred, Spec, BlockedPred).
%
% creates a predicate with head Pred which will call BlockedPred as soon
% as if Spec is fulfilled. Spec is a term like the first argument in when/2.
% Pred should be a term whose arguments are all variables.
% All variables used in Spec should be arguments of Pred.
% Spec may contain the following constructs:
%  - nonvar(Var)
%  - ground(Var)
%  - SpecA ; SpecB
%  - SpecA , SpecB
%
% Please note that the functionallity depends on the term expansion
% mechanism, so one has to use strictly the form ":- fastwhen(Pred,Spec,BlockedPred)."
% Something like ":- metapredicate a(:), fastwhen(a(A),ground(A),b(A))." would
% not work. Use two lines instead:
%    ":- metapredicate a(:)."
%    ":- fastwhen(a(A),ground(A),b(A))."
%
% If Pred has the functor p, fastwhen also generates several internal predicates p__...
%
% example:
%   :- fastwhen( rule(A,B,C), nonvar(A), rule_blocked(A,B,C)  ).
%
% creates a predicate rule/3 that will call rule_blocked/3 as soon as
% A is nonvar. A call to rule/3 behaves like when( nonvar(A), rule_blocked(A,B,C) ),
% but should be faster.

:- public fastwhen_generate/4.
fastwhen_generate(Rule,Spec,Call,Rules) :-
    fastwhen_generate(Rule,Spec,Call,[],Rules).
fastwhen_generate(Rule,Spec,Call,Options,Rules) :-
    Rule =.. [Name|Args],
    cblock_without_abort(Spec,Name,Call,Args,Rules),
    ( member(verbose,Options) ->
        print_rules(Rule,Spec,Call,Rules)
    ; otherwise -> true).

print_rules(Rule,Spec,Call,Rules) :-
    print('Generated the following predicates for '),
    copy_term(fastwhen(Rule,Spec,Call),fastwhen(R,S,C)),
    numbervars(fastwhen(R,S,C),0,_),
    print(fastwhen(R,S,C)), print(':'), nl,
    print_rules2(Rules),nl.
print_rules2(Rules) :-
    member(R,Rules), portray_clause(R), fail.
print_rules2(_).

% part1: generate rule definitions
cblock_without_abort(Spec,Name,Call,Args,Rules) :-
    dnf(Spec,DNF),!,
    create_block(DNF,Name,Args,Block),
    copy_term([Args,Call],[NArgs,NCall]),
    Head =.. [Name|NArgs],
    Rules = [Block,( Head :- NCall )].
cblock_without_abort(ground(G),Name,Call,Args,[R1,Block,R2]) :-
    atom_concat(Name,'__ground2',Name2),
    create_ground1(G,Name,Name2,Call,Args,R1),
    create_ground_block(Name2,Args,Block),
    create_ground2(Name2,Call,Args,R2).
cblock_without_abort( (A;B),Name,Call,Args,[Or|Rules]) :-
    copy_term([A,B,Call,Args],[NA,NB,NCall,NArgs]),
    cblock_or(NA,NB,Name,NCall,NArgs,Rules,CallA,CallB,_AbortA,_AbortB),
    Head =.. [Name|NArgs],
    Or = ( Head :- CallA, CallB ).
cblock_without_abort( (A,B),Name,Call,Args,Rules) :-
    atom_concat(Name,'__and2',Name2),
    CallToB =.. [Name2|Args],
    cblock_without_abort(A,Name,CallToB,Args,RulesA),
    cblock_without_abort(B,Name2,Call,Args,RulesB),
    append(RulesA,RulesB,Rules).

cblock_with_abort(Spec,Name,Call,Args,Abort,Rules) :-
    dnf((nonvar(Abort);Spec),DNF),!,
    AllArgs = [Abort|Args],
    create_block(DNF,Name,AllArgs,Block),
    copy_term([AllArgs,Abort],[Args1,Abort1]),
    copy_term([AllArgs,Call],[Args2,Call2]),
    Head1 =.. [Name|Args1],
    Head2 =.. [Name|Args2],
    Rules = [Block, (Head1:-nonvar(Abort1),!), (Head2:-Call2)].
cblock_with_abort(ground(G),Name,Call,Args,Abort,[R1,Block,R2Abort,R2]) :-
    atom_concat(Name,'__ground2',Name2),
    create_ground_abort1(G,Name,Name2,Call,Args,Abort,R1),
    create_ground_abort_block(Name2,Args,Block),
    create_ground_abort2a(Name2,Args,R2Abort),
    create_ground_abort2(Name2,Call,Args,R2).
cblock_with_abort( (A;B),Name,Call,Args,Abort,[Or|Rules]) :-
    copy_term([A,B,Call,Args,Abort],[NA,NB,NCall,NArgs,NAbort]),
    cblock_or(NA,NB,Name,NCall,NArgs,Rules,CallA,CallB,AbortA,AbortB),
    Head =.. [Name,NAbort|NArgs],
    Or = ( Head :- fastwhen:or_abort(NAbort,AbortA,AbortB), CallA, CallB ).
cblock_with_abort( (A,B),Name,Call,Args,Abort,Rules) :-
    atom_concat(Name,'__and2',Name2),
    CallToB =.. [Name2,Abort|Args],
    cblock_with_abort(A,Name,CallToB,Args,Abort,RulesA),
    cblock_with_abort(B,Name2,Call,Args,Abort,RulesB),
    append(RulesA,RulesB,Rules).

create_ground1(G,Name,Name2,Call,Args,Rule) :-
    create_ground_gen_1(G,Name,Name2,Call,Args,false,_,Rule).
create_ground_abort1(G,Name,Name2,Call,Args,Abort,Rule) :-
    create_ground_gen_1(G,Name,Name2,Call,Args,true,Abort,Rule).
create_ground_gen_1(G,Name,Name2,Call,Args,WithAbort,Abort,Rule) :-
    copy_term([G,Call,Args,Abort],[NG,NCall,NArgs,NAbort]),
    ( WithAbort == false -> HeadArgs=NArgs, CallArgs=[H,T|NArgs]
    ; WithAbort == true  -> HeadArgs=[NAbort|NArgs], CallArgs=[NAbort,H,T|NArgs] ),
    Head =.. [Name|HeadArgs],
    Call2 =.. [Name2|CallArgs],
    Rule = ( Head :- term_variables(NG,Vars),
               ( Vars==[] -> NCall; Vars=[H|T] -> Call2) ).

create_ground2(Name,Call,Args,Rule) :-
    create_ground_gen_2(Name,Call,Args,false,Rule).
create_ground_abort2(Name,Call,Args,Rule) :-
    create_ground_gen_2(Name,Call,Args,true,Rule).
create_ground_gen_2(Name,Call,Args,WithAbort,Rule) :-
    copy_term([Call,Args],[NCall,NArgs]),
    ( WithAbort == false -> HeadArgs=[H,T|NArgs], CallArgs=[H2,T2|NArgs]
    ; WithAbort == true  -> HeadArgs=[Abort,H,T|NArgs], CallArgs=[Abort,H2,T2|NArgs]),
    Head =.. [Name|HeadArgs],
    RCall =.. [Name|CallArgs],
    Rule = ( Head :- term_variables([H|T],Vars),
               ( Vars==[] -> NCall; Vars=[H2|T2] -> RCall) ).

create_ground_abort2a(Name,Args,Rule) :-
    copy_term(Args,NArgs),
    Head =.. [Name,Abort,_,_|NArgs],
    Rule = ( Head :- nonvar(Abort),! ).

create_ground_block(Name,Args,Rule) :-
    length(Args,NumberArgs),
    BArity is NumberArgs+2,
    functor(Blocking,Name,BArity),
    Blocking =.. [Name,-|QMarks],
    all_same(QMarks,?),
    Rule = (:- block Blocking).

create_ground_abort_block(Name,Args,Rule) :-
    length(Args,NumberArgs),
    BArity is NumberArgs+3,
    functor(Blocking,Name,BArity),
    Blocking =.. [Name,-,-|QMarks],
    all_same(QMarks,?),
    Rule = (:- block Blocking).

all_same([],_).
all_same([H|T],H) :- all_same(T,H).

cblock_or(A,B,Name,Call,Args,Rules,CallA,CallB,AbortA,AbortB) :-
    atom_concat(Name,'__or1',NameA),
    atom_concat(Name,'__or2',NameB),
    CallInA = (AbortB=1, Call),
    CallInB = (AbortA=1, Call),
    cblock_with_abort(A,NameA,CallInA,[AbortB|Args],_,RulesA),
    cblock_with_abort(B,NameB,CallInB,[AbortA|Args],_,RulesB),
    CallA =.. [NameA,AbortA,AbortB|Args],
    CallB =.. [NameB,AbortB,AbortA|Args],
    append(RulesA,RulesB,Rules).

create_block(DNF,Name,Args,(:-block Blocking) ) :-
    length(Args,Arity),
    create_block2(DNF,Name,Args,Arity,Blocking).
create_block2([Disj|Rest],Name,Args,Arity,Blocking) :-
    functor(Block,Name,Arity),
    create_block3(Args,1,Disj,Block),
    ( Rest == [] ->
        Blocking = Block
    ; otherwise ->
        Blocking = (Block,RBlock),
        create_block2(Rest,Name,Args,Arity,RBlock)).
create_block3([],_,_,_Block).
create_block3([Arg|RestArgs],Index,Disj,Block) :-
    ( strict_member(Arg,Disj) ->
        Sym = (-)
    ; otherwise ->
        Sym = (?)),
    arg(Index,Block,Sym),
    I2 is Index+1,
    create_block3(RestArgs,I2,Disj,Block).

strict_member(A,[H|T]) :-
    (A==H -> true; strict_member(A,T)).

:- public or_abort/3.
% or abort propagation
:- block or_abort(-,-,-).
or_abort(U,A,B) :-
    ( nonvar(U) -> A=1,B=1
    ; otherwise -> true).

% create a Disjunctive Normal Form for a block
% the result is a list of disjunctions, each is a list of variables
dnf(F,[Disj]) :-
    is_disjunction(F,Disj),!.
dnf( (A;B), Konj) :-
    dnf(A,AD),dnf(B,BD),
    distribute(AD,BD,Konj).
dnf( (A,B), Konj) :-
    dnf(A,AD),append(AD,BD,Konj),dnf(B,BD).
distribute([],_,[]).
distribute([A|Rest],F2,New) :-
    distribute2(F2,A,NewFormula),
    append(NewFormula,NewRest,New),
    distribute(Rest,F2,NewRest).
distribute2([],_A,[]).
distribute2([F|Frest],A,[N|Nrest]) :-
    append(A,F,N),
    distribute2(Frest,A,Nrest).
is_disjunction(nonvar(V),[V]).
is_disjunction( (A;B), V) :-
    is_disjunction(A,AV),
    append(AV,BV,V),
    is_disjunction(B,BV).


/*
 * This code uses the term_expansion/6 hook, wich
 *   is a multifile-predicate. Due to a bug in SICStus Prolog,
 *   it is not possible to mix uncompiled and compiled code
 *   for such a predicate.
 *   So we always compile the term_expansion hook
 */
:- prolog_flag(compiling,Pre,compactcode),
    compile(fastwhen_te),
    prolog_flag(compiling,_,Pre).

:- public btest_blocked/3.
:- fastwhen( btest(A,B,C), (nonvar(A);ground(B)), btest_blocked(A,B,C),[verbose] ).
btest_blocked(A,B,C) :- print(btest(A,B,C)).
