% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(hit_profiler,[add_profile_hit/1, add_profile_hit/2,
                    add_hit/2, add_hit_if_not_covered/2, % direct calls without abstraction of terms/args
                    add_profile_hash_hit/2,
                    profile_functor/1, get_profile_statistics/2,
                    reset_profiler/0, print_hit_profile_statistics/0]).

% call with hit_profiler:profile_statistics.

print_hit_profile_statistics :- \+ profile_functor(_),!. % print nothing
print_hit_profile_statistics :- print('-- ProB hit profiler statistics --'),nl,
   profile_functor(F),
   print('Functor : '), print(F),
   findall(hits(Nr,A),hit(_,F,A,Nr),Hits),
   sort(Hits,SortedHits), sum_hits(Hits,0,TotalHits), print(' --> '), print(TotalHits),nl,
   print_hits(SortedHits),
   fail.
print_hit_profile_statistics :- print('-- end --'),nl.

get_profile_statistics(F,Hits) :-
   profile_functor(F),
   findall(hit(A,Nr),hit(_,F,A,Nr),Hits).

sum_hits([],A,A).
sum_hits([hits(Nr,_)|T],A,R) :- NA is A+Nr, sum_hits(T,NA,R).

print_hits([]).
print_hits([hits(Nr,A)|T]) :- print(' - '), print(A), print(' --> '), print(Nr),nl,
  print_hits(T).

:- dynamic hit/4.
:- dynamic profile_functor/1.
reset_profiler :- retractall(hit(_,_,_,_)), retractall(profile_functor(_)).

%:- meta_predicate add_profile_hit(0).
add_profile_hit(Call) :- add_profile_hit(Call,1).
%:- meta_predicate add_profile_hit(0,*).
add_profile_hit(Call,Level) :-
    decompose_call(Call,Functor,Args),
    abstract_args(Args,Level,AA),
    %print(hit(Call,Functor,Args,AA)),nl,
    add_hit(Functor,AA).

% add hit where term gets abstracted to hash; interesting to see if we have lots of different calls
add_profile_hash_hit(Functor,Term) :- term_hash(Term,Hash), add_hit(Functor,Hash).

decompose_call(X,F,A) :- var(X),!,F=var,A=[].
decompose_call(_Module:Call,Functor,Args) :- !, Call =.. [Functor|Args].
decompose_call(Call,Functor,Args) :- Call =.. [Functor|Args].


abstract_args([],_,[]).
abstract_args([H|T],Level,[AH|AT]) :- abstract_arg(H,Level,AH), abstract_args(T,Level,AT).

:- use_module(library(avl)).
abstract_arg(X,_,R) :- var(X),!,R=var.
abstract_arg(X,Level,R) :- number(X),!,(Level>0 -> R=X ; R=number).
abstract_arg([],_,R) :- !, R=[].
abstract_arg(X,Level,R) :- atomic(X),!,(Level>0 -> R=X ; R=atomic).
abstract_arg(b(Pred,_,_),Level,R) :- !, R=b(AP),abstract_arg(Pred,Level,AP).
abstract_arg(avl_set(A),Level,R) :- !, (Level>0 -> avl_size(A,Sz), R=avl_set(Sz) ; R=avl_set).
abstract_arg(node(A,B,C,D,E),Level,Res) :-  (Level>0 -> avl_size(node(A,B,C,D,E),Sz), R=avl_node(Sz) ; R=avl_node), !, Res=R.
abstract_arg([H|T],Level,R) :- length([H|T],L),!,
    (Level>0 -> R = list(L) ; R=list).
    %LL is L//100,R=list(LL).
abstract_arg(X,0,R) :- !, functor(X,F,N), R=F/N.
abstract_arg(X,Level,R) :- X =.. [F|Args], L1 is Level-1, abstract_args(Args,L1,AA),
   R =.. [F|AA].

:- use_module(library(terms),[term_hash/2]).
add_hit(Functor,AbsArgs) :-
   term_hash(hit(Functor,AbsArgs),Hash),
   add_hit_hash(Hash,Functor,AbsArgs).

add_hit_hash(Hash,Functor,AbsArgs) :-
   (retract(hit(Hash,Functor,AbsArgs,Old))
     -> true
      ; Old=0,
       (profile_functor(Functor) -> true ; assert(profile_functor(Functor)))
   ),
   New is Old+1,
   assert(hit(Hash,Functor,AbsArgs,New)).

% use if you do not want to count number of hits:
add_hit_if_not_covered(Functor,AbsArgs) :-
   term_hash(hit(Functor,AbsArgs),Hash),
   (hit(Hash,Functor,AbsArgs,_Old) -> true
    ; assert(hit(Hash,Functor,AbsArgs,1)),
      (profile_functor(Functor) -> true ; assert(profile_functor(Functor)))
    ).
