% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_mappings,
 [ unary_function/3,
   binary_function/3, binary_arg1_determines_value/3,
   unary_in_boolean_type/2, unary_not_in_boolean_type/2,
   unary_in_definitely_true/2,
   binary_in_boolean_type/2,
   binary_not_in_boolean_type/2,
   binary_in_definitely_true/2,
   cst_in_boolean_type/2, cst_not_in_boolean_type/2,
   binary_boolean_operator/3,
   negate_binary_boolean_operator/2,
   symbolic_closure_binary_operator/1, is_definitely_empty/3, do_not_keep_symbolic/5,
   symbolic_closure_unary_operator/1, do_not_keep_symbolic_unary/2,

   special_binary_predicate/6,

   unary_kernel_call/8,
   binary_kernel_call/9,
   unary_in_kernel_call/6,
   binary_in_kernel_call/7,
   kernel_call_predicate/4,
   kernel_call_predicate_check_element_of_wf/4, kernel_call_predicate_equal_object_optimized/3,
   kernel_call_predicate_not_equal_object/4,
   kernel_call_predicate_not_element_of_wf/3,
   kernel_call/4,
   kernel_call_apply_to/6,

   kernel_call_predicate3/7, kernel_call_predicate2/6
   ]).

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling(must_succ_kernel_call/4).
%:- enable_profiling(kernel_call_predicate/4).
%:- enable_profiling(kernel_call/4).
%:- enable_profiling(kernel_call_apply_to/5).
%:- enable_profiling(kernel_call_or/5).

% old profiler
%:- use_module(hit_profiler, [add_profile_hit/1,add_profile_hit/2]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module translates AST constructs to calls in the ProB kernel.').

:- use_module(self_check).
:- use_module(kernel_waitflags,[get_wait_flag0/2]).
:- use_module(kernel_tools,[value_variables/2]).

:- meta_predicate kernel_call_or(0,*,*,*,*).

:- use_module(debugging_calls).
:- if(debug:global_debug_flag).
:- register_debugging_call must_succ(*,-).
:- else.
:- register_debugging_call must_succ(*,-),kernel_det_call(*,-,-,-).
:- endif.




:- load_files(library(system), [when(compile_time), imports([environ/2])]).


unary_kernel_call(Module,KernelFunction,ESV1,Value,WF,Expr,Type,Span) :-
    (expects_waitflag_and_span(KernelFunction)
     ->  KernelCall =.. [KernelFunction,ESV1,Value,Span,WF]
      ;  expects_waitflag(KernelFunction) ->  KernelCall =.. [KernelFunction,ESV1,Value,WF]
      ;  KernelCall =.. [KernelFunction,ESV1,Value]
     ),
     %print(call(Type,KernelCall)),nl,
    (type_ok_for_wf0(Type) -> %print(direct_call(Type,KernelCall)),nl,
        Module:KernelCall
     ; reversible_unary_function(KernelFunction)
      -> kernel_call_or(Module:KernelCall,ESV1,Value,WF,Expr)
      ;  must_succ_kernel_call(Module:KernelCall,ESV1,WF,Expr)
    ).

% result is ok for wf0: no optimized AVL representation exists; we do not need to delay calling function
type_ok_for_wf0(integer).
type_ok_for_wf0(boolean).
type_ok_for_wf0(global(_)).
type_ok_for_wf0(string).


binary_kernel_call(_,KernelFunction,ESV1,ESV2,Value,WF,_Expr,_Type,Span) :-
    determined_binary_integer_function(KernelFunction),
    !,
    get_wait_flag0(WF,WF0),
    call_determined_integer_function(KernelFunction,ESV1,ESV2,Value,WF0,WF,Span).
binary_kernel_call(Module,KernelFunction,ESV1,ESV2,Value,WF,Expr,Type,Span) :-
    (expects_waitflag_and_span(KernelFunction)
         ->  KernelCall =.. [KernelFunction,ESV1,ESV2,Value,Span,WF]
     ;  expects_waitflag(KernelFunction)
         ->  KernelCall =.. [KernelFunction,ESV1,ESV2,Value,WF]
      ;  expects_waitflag_and_type(KernelFunction)
         -> KernelCall =.. [KernelFunction,ESV1,ESV2,Value,Type,WF]
     ;      KernelCall =.. [KernelFunction,ESV1,ESV2,Value]
    ),
    special_propagate_binary_function(KernelFunction,ESV1,ESV2,Value,WF),
    must_succ_kernel_call(Module:KernelCall,(ESV1,ESV2),WF,Expr).

binary_in_kernel_call(Module,Kernel_predicate,ElValue,SV1,SV2,WF,_Expr) :-
     %KernelCall =.. [Kernel_predicate,ElValue,SV1,SV2,WF],
     %kernel_call(Module:KernelCall,(ElValue,SV1,SV2),WF,Expr).
     get_wait_flag0(WF,WF0),
     binary_in_kernel_call2(Module,Kernel_predicate,ElValue,SV1,SV2,WF,WF0).

:- block binary_in_kernel_call2(?,?,-,?,?,?,-). % wait until either WF0 set or ElValue
binary_in_kernel_call2(Module,Kernel_predicate,ElValue,SV1,SV2,WF,WF0) :- nonvar(WF0),!,
     call(Module:Kernel_predicate,ElValue,SV1,SV2,WF).
binary_in_kernel_call2(Module,Kernel_predicate,ElValue,SV1,SV2,WF,WF0) :-
     not_symbolic_when_var(ElValue,EWait), % ensure that we do not call this in case ElValue is a closure which may need expanding
     value_variables(SV1,G1), value_variables(SV2,G2),
     when((nonvar(WF0) ; ground(EWait), ground(G1), ground(G2)),
           call(Module:Kernel_predicate,ElValue,SV1,SV2,WF)).

:- block not_symbolic_when_var(-,?).
not_symbolic_when_var(closure(_,_,_),_) :- !.
not_symbolic_when_var(X,WV) :- value_variables(X,WV).

unary_in_kernel_call(Module,Kernel_predicate,ElValue,SV1,WF,_Expr) :-
     %KernelCall =.. [Kernel_predicate,ElValue,SV1,WF],
     %kernel_call(Module:KernelCall,(ElValue,SV1),WF,Expr).
     get_wait_flag0(WF,WF0),
     unary_in_kernel_call2(Module,Kernel_predicate,ElValue,SV1,WF,WF0).

:- block unary_in_kernel_call2(?,?,-,?,?,-). % wait until either WF0 set or ElValue
unary_in_kernel_call2(Module,Kernel_predicate,ElValue,SV1,WF,WF0) :- nonvar(WF0),!,
     call(Module:Kernel_predicate,ElValue,SV1,WF).
unary_in_kernel_call2(Module,Kernel_predicate,ElValue,SV1,WF,WF0) :-
     not_symbolic_when_var(ElValue,EWait), % ensure that we do not call this in case ElValue is a closure which may need expanding; then we should at least wait until WF0 is ground
     value_variables(SV1,G1),
     when((nonvar(WF0) ; ground(EWait), ground(G1)),
           call(Module:Kernel_predicate,ElValue,SV1,WF)).


%must_succ_kernel_call(kernel_objects:equal_object_optimized(A,B),(A1,A2),WF,Expr) :- !, %%print('EQUAL'),nl,
%     Call = kernel_objects:equal_object_optimized(A,B),
%     kernel_call_or(Call,A1,A2,WF,Expr).
must_succ_kernel_call(Call,HasToBeGround,WF,_Expr) :-
     get_wait_flag0(WF,WF0),
     (nonvar(WF0)
      -> call(Call)
      ;  %hit_profiler:add_profile_hit(must_succ_kernel_call(Call,WF0),3),
         %ground_value_check(HasToBeGround,G1), % using this makes test 1562 fail; probably because co-routines remain active after WF0 has been set, leading un bound variables preventing optimisations in kernel_objects
         value_variables(HasToBeGround,VG1),
         %must_succ_kernel_call2(Call,WF0,G1)
         when((nonvar(WF0);ground(VG1)),call(Call))
     ).

%:- block must_succ_kernel_call2(?,-,-).
%must_succ_kernel_call2(Call,_,_) :- call(Call). % was ~~must_succ(~~kernel_det_call(Call,WF,WF0,Expr),Expr)


kernel_call_predicate(equal_object_optimized(A,B),kernel_objects,WF,_) :- !, %print('EQUAL'),nl,
     get_wait_flag0(WF,WF0),
     equal_obj(A,B,WF0,WF).
  % equality is the only predicate that can generate AVL trees
% OTHER CALLS: partition_wf, member of empty set or singleton set, CLPFD operators: TO DO add

kernel_call_predicate(not_equal_object_wf(A,B,WF),kernel_objects,WF,Expr) :- !,
   get_wait_flag0(WF,WF0),not_eq_wf(A,B,WF,WF0,Expr).
kernel_call_predicate(partition_wf(A,B,PWF),kernel_objects,WF,Expr) :- !,
   (nonvar(B),B=[B1|T],T==[] % just one set
      -> kernel_call_predicate(equal_object_optimized(A,B1),kernel_objects,WF,Expr)
       ; all_but_one_set_known([A|B],Known),
         kernel_call(kernel_objects:partition_wf(A,B,PWF),Known,WF,Expr)).
kernel_call_predicate(less_than(A,B),kernel_objects,_WF,_Expr) :- !,  kernel_objects:less_than(A,B).
kernel_call_predicate(less_than_equal(A,B),kernel_objects,_WF,_Expr) :- !,  kernel_objects:less_than_equal(A,B).
kernel_call_predicate(greater_than(A,B),kernel_objects,_WF,_Expr) :- !,  kernel_objects:less_than(B,A).
kernel_call_predicate(greater_than_equal(A,B),kernel_objects,_WF,_Expr) :- !,  kernel_objects:less_than_equal(B,A).
kernel_call_predicate(check_element_of_wf(A,B,PWF),kernel_objects,WF,Expr) :- !,
   get_wait_flag0(WF,WF0), check_element_of_wf0(A,B,PWF,WF0,Expr).
kernel_call_predicate(Call,Module,WF,_Expr) :- get_wait_flag0(WF,WF0),
    %hit_profiler:add_profile_hit(kernel_call_predicate(Call,WF0),2),
    call_wf0(Module:Call,WF0).
%kernel_call_predicate(Call,HasToBeGround,WF,Expr) :- kernel_call(Call,HasToBeGround,WF,Expr).

:- if(debug:use_timer).
:- block call_wf0(?,-).
call_wf0(X,_) :- debug:timer_call(X). %debug:watch(10,X).
:- else.
:- block call_wf0(?,-).
call_wf0(Call,_) :-   call(Call).
:- endif.

% unfolded version of above (for efficiency):
kernel_call_predicate_equal_object_optimized(A,B,WF) :-  %print('EQUAL'),nl,
     get_wait_flag0(WF,WF0),equal_obj(A,B,WF0,WF).
kernel_call_predicate_check_element_of_wf(A,B,WF,Span) :-
   get_wait_flag0(WF,WF0), check_element_of_wf0(A,B,WF,WF0,Span).
kernel_call_predicate_not_equal_object(A,B,WF,Expr) :-
   get_wait_flag0(WF,WF0),not_eq_wf(A,B,WF,WF0,Expr).
kernel_call_predicate_not_element_of_wf(A,B,WF) :-
     get_wait_flag0(WF,WF0),not_element_of_wf0(A,B,WF,WF0).

:- block not_element_of_wf0(?,?,?,-).
not_element_of_wf0(ElemValue,SetValue,WF,_) :- kernel_objects:not_element_of_wf(ElemValue,SetValue,WF).

kernel_call_predicate3(KernelCall,Arg1,Arg2,Arg3,Module,WF,_Expression) :-
    get_wait_flag0(WF,WF0),
    %print(kernel_call3(KernelCall,Arg1,Arg2,Arg3)),nl,
    call_wf0_3(Module:KernelCall,Arg1,Arg2,Arg3,WF,WF0).
:- block call_wf0_3(?, ?,?,?,?, -).
call_wf0_3(Call,A1,A2,A3,WF,_) :-   call(Call,A1,A2,A3,WF).

kernel_call_predicate2(KernelCall,Arg1,Arg2,Module,WF,_Expression) :-
    get_wait_flag0(WF,WF0),
    %print(kernel_call3(KernelCall,Arg1,Arg2,Arg3)),nl,
    call_wf0_2(Module:KernelCall,Arg1,Arg2,WF,WF0).
:- block call_wf0_2(?, ?,?,?, -).
call_wf0_2(Call,A1,A2,WF,_) :-   call(Call,A1,A2,WF).

:- use_module(custom_explicit_sets,[singleton_set/2]).
:- block check_element_of_wf0(?,-,?,-,?).
check_element_of_wf0(_A,B,_WF,_,_) :- B==[],!,fail. % nothing can be element of empty set;  CAN PREVENT WD-ERRORS from being detected !
check_element_of_wf0(A,B,WF,_WF0,_) :-
    singleton_set(B,B1),!, /* also does var check */
    equal_object_optimized_wf(A,B1,check_element_of_wf0,WF). %was equal_obj(A,B1,WF0) but as B1 is ground value will always call equal_object_optimized directly
check_element_of_wf0(A,B,WF,WF0,_) :- nonvar(WF0),!,
    kernel_objects:check_element_of_wf(A,B,WF).
% below B must be nonvar
check_element_of_wf0(A,global_set(GS),WF,_,Span) :- nonvar(A),
    !,  % be sure that we check something like int(-1) : NATURAL
    %kernel_objects:check_element_of_wf(A,global_set(GS),WF).
    kernel_objects:element_of_global_set_wf(A,GS,WF,Span). % Span for enumeration warnings
check_element_of_wf0(A,avl_set(AVL),WF,_,_) :- simple_ground_value(A),
    !, % we can already check this efficiently in phase 0
    custom_explicit_sets:element_of_avl_set_wf(AVL,A,WF). % we could even call avl_fetch directly
    % kernel_objects:check_element_of_wf(A,avl_set(AVL),WF).
check_element_of_wf0(A,B,WF,WF0,Span) :-
   check_element_of_wf0_aux(A,B,WF,WF0,Span).
:- block check_element_of_wf0_aux(?,?,?,-,?).
check_element_of_wf0_aux(A,global_set(GS),WF,_WF0,Span) :- !,
    kernel_objects:element_of_global_set_wf(A,GS,WF,Span). % Span for enumeration warnings
check_element_of_wf0_aux(A,B,WF,_WF0,_Span) :- kernel_objects:check_element_of_wf(A,B,WF).

simple_ground_value(A) :- nonvar(A), simple_ground_value_aux(A).
simple_ground_value_aux(int(X)) :- number(X).
simple_ground_value_aux(fd(X,_)) :- number(X).
simple_ground_value_aux(string(S)) :- ground(S).
simple_ground_value_aux((A,B)) :- simple_ground_value(A), simple_ground_value(B).
simple_ground_value_aux(rec(F)) :- simple_ground_fields(F).
simple_ground_value_aux([]).

simple_ground_fields(Var) :- var(Var),!,fail. % TO DO: we could activate check as soon as it becomes ground
simple_ground_fields([]).
simple_ground_fields([field(N,V)|T]) :- nonvar(N),simple_ground_value(V),simple_ground_fields(T).

:- assert_must_succeed( kernel_mappings:not_eq_wf(int(1),int(2),_,_,_) ).
:- assert_must_succeed( kernel_mappings:not_eq_wf([int(1)],[int(2)],_,0,_) ).
:- block not_eq_wf(-,?,?,-,?), not_eq_wf(?,-,?,-,?).
not_eq_wf(A,B,WF,_,_) :- efficient_value(A),efficient_value(B),!, %tools_printing:print_term_summary(not_eq_wf(A,B)),
    kernel_objects:not_equal_object_wf(A,B,WF).
not_eq_wf(A,B,WF,WF0,Expr) :- not_eq_wf0(A,B,WF,WF0,Expr).
:- block not_eq_wf0(?,?,?,-,?).
not_eq_wf0(A,B,WF,_,_Expr) :-
   %((var(A),var(B)) -> instantiate_argument(Expr,A) ; true),
   kernel_objects:not_equal_object_wf(A,B,WF).

/* not required anymore ?!
% instantiate argument according to type of expression to ensure not_equal_object (or other calls) are triggered and do not wait setting up dif constraints
instantiate_argument(not_equal(AE,_),A) :-
   bsyntaxtree:get_texpr_type(AE,Type),!, print(instantiate(Type,A)),nl,trace,
   instantiate_arg(Type,A).
instantiate_argument(_,_).

instantiate_arg(string,R) :- !, R=string(_).
instantiate_arg(integer,R) :- !, R=int(_).
% missing: more cases global(G),.... ?
instantiate_arg(_,_).
*/

% check whether disequality can be efficiently checked (already in WF0 phase)
efficient_value(X) :- var(X),!,fail.
efficient_value(int(_)).
efficient_value(avl_set(_)).
efficient_value([]).
efficient_value(fd(_,_)).
efficient_value(string(_)).
efficient_value((A,B)) :- nonvar(A),nonvar(B),efficient_value(A), efficient_value(B).
efficient_value(pred_true). efficient_value(pred_false).




%:- meta_predicate kernel_call_or(0,*,*,*).
kernel_call(Call,HasToBeGround,WF,Expr) :-
     get_wait_flag0(WF,WF0),
     %hit_profiler:add_profile_hit(kernel_call(Call,WF),2),
     (nonvar(WF0)
      -> ~~kernel_det_call(Call,WF,WF0,Expr)
       ; value_variables(HasToBeGround,G1),
         when((nonvar(WF0) ; ground(G1)) ,
            (  %(var(WF0) -> print('enter  :'),tools_printing:print_term_summary(Call) ; true),
          %  print(kernel_det_call(Call,WF,WF0,Expr)),nl,
            ~~kernel_det_call(Call,WF,WF0,Expr)
           % ,print(finished_kernel_det_call(Call,WF,WF0)),nl
             %,(var(WF0) -> print('result :'),tools_printing:print_term_summary(Call) ; true)
            ) )
     ).

:- use_module(kernel_tools,[ground_value_check/2]).
% optimized custom versions of kernel_call for performacne:
:- use_module(bsets_clp,[apply_to/6]).
kernel_call_apply_to(FValue,ArgValue,Res,FunctionType,Span,WF) :-
   %hit_profiler:add_profile_hit(kernel_call_apply_to(FValue,ArgValue,Res,WF)),
   get_wait_flag0(WF,WF0),
   (var(WF0) %(var(WF0), \+ ground(ArgValue), \+ ground(FValue))
     -> ground_value_check((FValue,ArgValue),WF1), % often FValue is an avl_set, avoid pending co-routines later by checking it first for groundness
        call_apply_to(FValue,ArgValue,Res,FunctionType,Span,WF,WF0,WF1)
      ; apply_to(FValue,ArgValue,Res,FunctionType,Span,WF)
    ).
% TO DO: maybe transmit Type information for using with element/3

:- block call_apply_to(?,?,?,?,?,?,-,-).
call_apply_to(FValue,ArgValue,Res,FunctionType,Span,WF,_,_Gr) :-
   %discard_ground_value_check(Gr),
   bsets_clp:apply_to(FValue,ArgValue,Res,FunctionType,Span,WF).


/* ------------------------------- */

:- use_module(kernel_objects,[equal_object_optimized_wf/4]).
:- block equal_obj(-,-,-,?).

equal_obj(A,B,WF0,WF) :- nonvar(WF0),!,
    equal_object_optimized_wf(A,B,equal_obj1,WF).
equal_obj(A,B,WF0,WF) :- %hit_profiler:add_profile_hit(eq(A,B)),
    %print(eq_obj(A,B)),nl,
    non_simple_value_variables(A,VA),
    (VA==[] -> equal_object_optimized_wf(A,B,equal_obj2,WF)
     ; non_simple_value_variables(B,VB),
       (VB==[] -> equal_object_optimized_wf(A,B,equal_obj3,WF)
        ;  when((nonvar(WF0);ground(VA);ground(VB)), kernel_objects:equal_object_optimized_wf(A,B,equal_obj4,WF))
       )
    ).

% a version of value_variables that also returns [] in some cases where there is a variable but equal_object can be called
non_simple_value_variables(V,Vars) :- var(V),!, Vars=[V].
non_simple_value_variables(pred_false,Vars) :- !, Vars=[].
non_simple_value_variables(pred_true,Vars) :- !, Vars=[].
non_simple_value_variables(int(_),Vars) :- !, Vars=[]. % may contain var; but we can unify nonetheless
non_simple_value_variables(string(_),Vars) :- !, Vars=[]. % ditto
non_simple_value_variables(fd(_,_),Vars) :- !, Vars=[]. % ditto
non_simple_value_variables(A,Vars) :- !, value_variables(A,Vars).


:- if(debug:global_debug_flag).
kernel_call_or(Call,HasToBeGround1,HasToBeGround2,WF,Expr) :-
     get_wait_flag0(WF,WF0),
    % print(kernel_call_start(Call)),nl,
    % det_call(Call)
    value_variables(HasToBeGround1,G1),
    (G1==[] -> G2=[] ; value_variables(HasToBeGround2,G2)),
     when((nonvar(WF0) ; ground(G1) ; ground(G2) ) ,
        ( %(var(WF0) -> (print('enter o:'),tools_printing:print_term_summary(Call)) ; true),
        (var(WF0) -> ~~must_succ(~~kernel_det_call(Call,WF,WF0,Expr),Expr)
                   ; ~~kernel_det_call(Call,WF,WF0,Expr))
         %,(var(WF0) -> (print('resulto:'),tools_printing:print_term_summary(Call)) ; true)
        ) )
     .
:- else.
kernel_call_or(Call,HasToBeGround1,HasToBeGround2,WF,_Expr) :-
     get_wait_flag0(WF,WF0),
     value_variables(HasToBeGround1,G1),
     (G1==[] -> G2=[] ; value_variables(HasToBeGround2,G2)),
     %hit_profiler:add_profile_hit(kernel_call_or(Call,WF0),2),
     when( (nonvar(WF0) ; ground(G1) ; ground(G2) ) , call(Call)).
:- endif.


% special case for calling binary integer functions: here we can use more efficient block declarations
call_determined_integer_function(KF,V1,V2,Val,WF0,WF,Span) :-
    determined(V1,V2,Val,Determined,WF0),
    call_determined_integer_function2(KF,V1,V2,Val,WF0,Determined,WF,Span).
:- block call_determined_integer_function2(?,?,?,?,-,-,?,?).
%call_determined_integer_function2(KF,V1,V2,V,WF0,Det,_,_) :-
%    print(kernel_det_call(KF,V1,V2,V,WF0,Det)),nl,fail.
call_determined_integer_function2(division,V1,V2,Val,_,_,WF,Span) :-
    kernel_objects:division(V1,V2,Val,Span,WF).
call_determined_integer_function2(floored_division,V1,V2,Val,_,_,WF,Span) :-
    kernel_objects:floored_division(V1,V2,Val,Span,WF).
call_determined_integer_function2(int_minus,V1,V2,Val,_,_,_WF,_Span) :-
    kernel_objects:int_minus(V1,V2,Val).
call_determined_integer_function2(int_plus,V1,V2,Val,_,_,_WF,_Span) :-
%    kernel_call_determined(kernel_objects:int_plus(V1,V2,Val),V1,V2,Val,WF,integer).
    kernel_objects:int_plus(V1,V2,Val).
call_determined_integer_function2(int_power,V1,V2,Val,_,_,WF,Span) :-
    kernel_objects:int_power(V1,V2,Val,Span,WF).
call_determined_integer_function2(modulo,V1,V2,Val,_,_,WF,Span) :-
    kernel_objects:modulo(V1,V2,Val,Span,WF).
call_determined_integer_function2(times,V1,V2,Val,_,_,_WF,_Span) :-
    kernel_objects:times(V1,V2,Val).

% sets last argument to determined without instantiating its first three args
% it is set when two integer values are known
:- block determined(-,-,?,?,-), determined(-,?,-,?,-), determined(?,-,-,?,-).
%:- block determined(-,-,?,?,?), determined(-,?,-,?,?), determined(?,-,-,?,?).
%determined(Y,V,X,Det,WF0) :- print(determined(Y,V,X,Det,WF0)),nl,fail.
determined(_X,_Y,_Z,Det,WF0) :- nonvar(WF0),!,Det=determined.
% ((nonvar(_X),nonvar(_Y), \+ ground(_Y), var(_V)) -> trace ; true).
determined(X,Y,V,Det,_) :- var(X),!, Y=int(YY),V=int(VV), %(YY=0 -> X=V ; true),
    determined2(X,YY,VV,Det).
determined(Y,X,V,Det,_) :- var(X),!, Y=int(YY),V=int(VV), %(YY=0 -> X=V ; true),
    determined2(X,YY,VV,Det).
determined(Y,V,X,Det,_) :-  Y=int(YY),V=int(VV), %(YY=0 -> X=V ; VV=0 -> X=Y),
    determined2(X,YY,VV,Det).
:- block determined2(-,-,?,?), determined2(-,?,-,?), determined2(?,-,-,?).
%determined2(Y,V,X,Det) :- print(determined2(Y,V,X,Det)),nl,fail.
determined2(X,_,_,Det) :- var(X),!,Det=determined.
determined2(int(XX),YY,VV,Det) :- determined3(XX,YY,VV,Det).
:- block determined3(-,-,?,?), determined3(-,?,-,?), determined3(?,-,-,?).
%determined3(Y,V,X,Det) :- print(determined3(Y,V,X,Det)),nl,fail.
determined3(_,_,_,determined).


:- use_module(translate,[print_unwrapped_bexpr_with_limit/1]).
:- public kernel_det_call/4.
%kernel_det_call(X,_,_,_) :- !, call(X). %% comment in to avoid checks below
:- if(debug:use_timer).
kernel_det_call(X,_,_,_) :- debug:timer_call(X). %debug:watch(10,X).
:- else.
kernel_det_call(X,WF,WF0,_Expr) :- var(WF0),!,
   % print('+enter  :'),tools_printing:print_term_summary(X),print('+  for  :'), translate:print_unwrapped_bexpr_with_limit(_Expr)
   call(X), %debug:watch(10,X),
   % print('+result :'),tools_printing:print_term_summary(X),
   (self_check:run_time_type_check_mode(off) ->
       true
   ; ( (ground(X) ; (term_variables(X,XV), term_variables(WF,WV),XV==WV)) ->
         true
     ; add_message(kernel_det_call,'### Kernel Call does not ground variables: ',X)
     % note: can also occur when operator is undefined for args (front_of_sequence([],...)
     %       in this case it is not an error
     )
   ).
kernel_det_call(X,_,_,_) :- call(X).
:- endif.


:- public must_succ/2.
:- meta_predicate must_succ(0,*).
must_succ(X,_Expr) :- self_check:run_time_type_check_mode(off),!, call(X).
must_succ(X,Expr) :-
  if(call(X),true, (nl,print('FAILED: '),nl,print_unwrapped_bexpr_with_limit(Expr),print(X),nl,nl,fail)).


:- use_module(error_manager).

invalid_call(X) :- \+ check_valid_kernel_call(X,_).
check_valid_kernel_call(X,unary) :- unary_function(_,_,X).
check_valid_kernel_call(X,binary) :- binary_function(_,_,X).
check_valid_kernel_call(X,_) :- add_error(check_valid_kernel_call,'Illegal Kernel Call: ',X),fail.
:- assert_must_fail((kernel_mappings:expects_waitflag_and_span(X),
                     kernel_mappings:invalid_call(X))).

expects_waitflag_and_span(first_sequence).
expects_waitflag_and_span(last_sequence).
expects_waitflag_and_span(tail_sequence).
expects_waitflag_and_span(front_sequence).
expects_waitflag_and_span(minimum_of_set).
expects_waitflag_and_span(maximum_of_set).
expects_waitflag_and_span(intersection_generalized_wf).

expects_waitflag_and_span(division).
expects_waitflag_and_span(floored_division).
expects_waitflag_and_span(modulo).
expects_waitflag_and_span(int_power).
expects_waitflag_and_span(singleton_set_element).

:- assert_must_fail((kernel_mappings:expects_waitflag(X),
                     kernel_mappings:invalid_call(X))).
:- assert_must_fail((kernel_mappings:expects_waitflag(X),
                     kernel_mappings:expects_waitflag_and_span(X))).
/*
comment in if we want to ensure that all kernel predicates take at least the WF
:- assert_must_fail((kernel_mappings:unary_function(_,_,X),
                    \+ (kernel_mappings:expects_waitflag(X) ;
                         kernel_mappings:expects_waitflag_and_span(X)),
                         print(unary_no_waitflag(X)),nl )).
:- assert_must_fail((kernel_mappings:binary_function(_,_,X),
                    \+ (kernel_mappings:expects_waitflag(X) ;
                         kernel_mappings:expects_waitflag_and_span(X)),
                         print(binary_no_waitflag(X)),nl )).
*/

expects_waitflag(append_sequence).
expects_waitflag(bag_items).
expects_waitflag(compaction).
expects_waitflag(concat_sequence).
expects_waitflag(concatentation_of_sequences).
expects_waitflag(difference_set_wf).
expects_waitflag(direct_product_wf).
expects_waitflag(parallel_product_wf).
expects_waitflag(domain_restriction_wf).
expects_waitflag(domain_subtraction_wf).
expects_waitflag(domain_wf).
expects_waitflag(finite_cardinality_as_int).
expects_waitflag(identity_relation_over_wf).
expects_waitflag(image_wf).
expects_waitflag(intersection).
expects_waitflag(invert_relation_wf).
expects_waitflag(relational_trans_closure_wf).
expects_waitflag(override_relation).
expects_waitflag(prefix_sequence_wf).
expects_waitflag(prepend_sequence).
expects_waitflag(range_restriction_wf).
expects_waitflag(range_subtraction_wf).
expects_waitflag(range_wf).
expects_waitflag(rel_composition_wf).
%expects_waitflag(rel_iterate_wf).
expects_waitflag(rev_sequence).
expects_waitflag(size_of_sequence).
expects_waitflag(square).
expects_waitflag(suffix_sequence).
expects_waitflag(unary_minus_wf).
expects_waitflag(union_generalized_wf).
expects_waitflag(union_wf).

expects_waitflag_and_type(rel_iterate_wf).

:- use_module(bsyntaxtree,[get_texpr_set_type/2]).
% only used when not in CLPFD mode:
special_binary_predicate(greater(b(card(X),integer,_),Y), kernel_objects:cardinality_greater,X,TypeX,Y,integer) :-
   get_texpr_set_type(X,TypeX).
special_binary_predicate(less(Y,b(card(X),integer,_)),    kernel_objects:cardinality_greater,X,TypeX,Y,integer) :-
   get_texpr_set_type(X,TypeX).
special_binary_predicate(greater_equal(b(card(X),integer,_),Y), kernel_objects:cardinality_greater_equal,X,TypeX,Y,integer) :-
   get_texpr_set_type(X,TypeX).
special_binary_predicate(less_equal(Y,b(card(X),integer,_)),    kernel_objects:cardinality_greater_equal,X,TypeX,Y,integer) :-
   get_texpr_set_type(X,TypeX).



/* EXPRESSIONS */


:- assert_must_fail((kernel_mappings:reversible_unary_function(X),
                     kernel_mappings:invalid_call(X))).

reversible_unary_function(invert_relation_wf).
reversible_unary_function(rev_sequence).
reversible_unary_function(unary_minus_wf).

unary_function(reverse,bsets_clp,invert_relation_wf).
unary_function(closure,bsets_clp,relational_trans_closure_wf). % this is closure1
%unary_function(reflexive_closure,bsets_clp,relational_reflexive_closure). % now done in ast_cleanup
unary_function(domain,bsets_clp,domain_wf).
unary_function(range,bsets_clp,range_wf).

unary_function(general_union,kernel_objects,union_generalized_wf).
unary_function(general_intersection,kernel_objects,intersection_generalized_wf).

unary_function(first,bsets_clp,first_sequence).
unary_function(size,bsets_clp,size_of_sequence).
unary_function(front,bsets_clp,front_sequence).
unary_function(tail,bsets_clp,tail_sequence).
unary_function(rev,bsets_clp,rev_sequence).
unary_function(last,bsets_clp,last_sequence).

unary_function(card,kernel_objects,finite_cardinality_as_int).
unary_function(pow_subset,kernel_objects,power_set). % not really necessary anymore
unary_function(fin_subset,kernel_objects,power_set).  % not really necessary anymore
unary_function(pow1_subset,kernel_objects,non_empty_power_set).  % not really necessary anymore
unary_function(fin1_subset,kernel_objects,non_empty_power_set). % not really necessary anymore
unary_function(identity,bsets_clp,identity_relation_over_wf).
%unary_function(perm,bsets_clp,enumerate_permutation_sequence).
% unary_function(iseq,bsets_clp,enumerate_injective_sequence).
%unary_function(iseq1,bsets_clp,enumerate_non_empty_injective_sequence).
%unary_function(seq,bsets_clp,enumerate_sequence).
%unary_function(seq1,bsets_clp,enumerate_non_empty_sequence).
%unary_function(struct,kernel_records,enumerate_records).

unary_function(general_concat,bsets_clp,concatentation_of_sequences).
unary_function(unary_minus,kernel_objects,unary_minus_wf).
unary_function(square_multiplication,kernel_objects,square).  % artificial construct, detected when X*X is evaluated by interpreter; but square is directly called
unary_function(min,kernel_objects,minimum_of_set).
unary_function(max,kernel_objects,maximum_of_set).

unary_function(first_of_pair,kernel_objects,first_of_pair).
unary_function(second_of_pair,kernel_objects,second_of_pair).
unary_function(mu,kernel_objects,singleton_set_element). % Z MU operator

unary_function(compaction,kernel_z,compaction).
unary_function(bag_items,kernel_z,bag_items).


:- assert_must_fail((kernel_mappings:determined_binary_integer_function(X),
                     kernel_mappings:invalid_call(X))).
/* binary functions where 1 argument is determined from the other two */
determined_binary_integer_function(division). % isn't really determined X / 10 = 2 --> many possible values
determined_binary_integer_function(int_minus).
determined_binary_integer_function(int_plus).
determined_binary_integer_function(int_power).
determined_binary_integer_function(modulo).   % isn't really determined; X mod 10 = 0 --> many possible values
determined_binary_integer_function(times).

%:- assert_must_fail((kernel_mappings:determined_binary_function(X),
%                     kernel_mappings:invalid_call(X))).
%determined_binary_function(nat_range).
%determined_binary_function(concat_sequence).
%determined_binary_function(prepend_sequence).
%determined_binary_function(cartesian_product).

% set up any special propagation rules which should trigger before Args1 and 2 are fully known:
% we should probably move these rules down to kernel_objects, bsets_clp ... and validate them
% maybe we can even store them in B format, prove them and read them in
special_propagate_binary_function(union_wf,A,B,Res,WF) :- !,
    propagate_empty_set(Res,A,WF),  % A\/B ={}  => A={}
    propagate_empty_set(Res,B,WF).  % A\/B ={}  => B={}
%special_propagate_binary_function(override_relation,A,B,Res,WF) :- !, % causes small slowdown for ticket 184
%    propagate_empty_set(Res,A,WF), propagate_empty_set(Res,B).
special_propagate_binary_function(intersection,A,B,Res,WF) :- !,
    propagate_empty_set(A,Res,WF),  % A={}  =>  A /\ B = {}
    propagate_empty_set(B,Res,WF).  % B={}  =>  A /\ B = {}
special_propagate_binary_function(difference_set_wf,A,B,Res,WF) :- !,
    propagate_empty_set(A,Res,WF),  %  A={}  =>  A \ B = {}
    propagate_empty_set_eq(B,A,Res,WF).  % B={}  =>  A \ B = A
special_propagate_binary_function(image_wf,A,B,Res,WF) :- !,
    propagate_empty_set(B,Res,WF),  % B={}  =>  A[B] = {}
    propagate_id_eq(A,B,Res,WF).    % (A={}  =>  A[B] = {})  &  (A=id =>  A[B] = B)
special_propagate_binary_function(domain_restriction_wf,S,_R,Res,WF) :- !,
     propagate_empty_set(S,Res,WF). % if S ={}  =>  S <| R = {}
special_propagate_binary_function(domain_subtraction_wf,S,R,Res,WF) :- !,
     propagate_empty_set_eq(S,R,Res,WF). % if S ={}  =>  S <<| R = R
special_propagate_binary_function(range_restriction_wf,_R,S,Res,WF) :- !,
     propagate_empty_set(S,Res,WF). % if S ={}  =>  R |> S = {}
special_propagate_binary_function(range_subtraction_wf,R,S,Res,WF) :- !,
     propagate_empty_set_eq(S,R,Res,WF). % if S ={}  =>   R |>> S = R
special_propagate_binary_function(_,_,_,_,_WF). % no propagators available
:- block propagate_empty_set(-,?,?).
propagate_empty_set([],A,WF) :- !,
  % (A==[] -> true ; print(prop_empty(A)),nl), %%
  kernel_objects:empty_set_wf(A,WF).
propagate_empty_set(_,_,_).
:- block propagate_empty_set_eq(-,?,?,?).
propagate_empty_set_eq([],A,B,WF) :- !,kernel_objects:equal_object_wf(A,B,propagate_empty_set_eq,WF).
propagate_empty_set_eq(_,_,_,_).
:- block propagate_id_eq(-,?,?,?).
propagate_id_eq([],_A,B,WF) :- !,kernel_objects:empty_set_wf(B,WF).
propagate_id_eq(closure(Par,Types,Body),A,B,WF) :- custom_explicit_sets:is_full_id_closure(Par,Types,Body),
   % print(prop_id_eq(A,B)),nl,
   !,kernel_objects:equal_object_wf(A,B,propagate_id_eq,WF).
propagate_id_eq(_,_,_,_).

binary_function(concat,bsets_clp,concat_sequence). % ^
binary_function(insert_front,bsets_clp,prepend_sequence).
binary_function(insert_tail,bsets_clp,append_sequence).
binary_function(restrict_front,bsets_clp,prefix_sequence_wf). % s /|\ n
binary_function(restrict_tail,bsets_clp,suffix_sequence). % WF
binary_function(union,kernel_objects,union_wf).
binary_function(intersection,kernel_objects,intersection).
binary_function(set_subtraction,kernel_objects,difference_set_wf).
binary_function(cartesian_product,kernel_objects,cartesian_product). % always converted into closure
binary_function(add,kernel_objects,int_plus).
binary_function(multiplication,kernel_objects,times).
binary_function(minus,kernel_objects,int_minus).
binary_function(modulo,kernel_objects,modulo).
binary_function(power_of,kernel_objects,int_power).
binary_function(div,kernel_objects,division).
binary_function(floored_div,kernel_objects,floored_division).
%binary_function(interval,kernel_objects,nat_range).  % always converted into closure, unless empty or singleton
binary_function(image,bsets_clp,image_wf).
%binary_function(function,apply_to).
binary_function(domain_restriction,bsets_clp,domain_restriction_wf).
binary_function(domain_subtraction,bsets_clp,domain_subtraction_wf).
binary_function(range_subtraction,bsets_clp,range_subtraction_wf).
binary_function(range_restriction,bsets_clp,range_restriction_wf).
binary_function(overwrite,bsets_clp,override_relation).
binary_function(composition,bsets_clp,rel_composition_wf).
binary_function(iteration,bsets_clp,rel_iterate_wf).
binary_function(direct_product,bsets_clp,direct_product_wf).
binary_function(parallel_product,bsets_clp,parallel_product_wf).
% functions below always converted into closure
%binary_function(total_function,bsets_clp,enumerate_total_function).
%binary_function(partial_function,bsets_clp,enumerate_partial_function).
%binary_function(total_bijection,bsets_clp,enumerate_total_bijection).
%binary_function(relations,bsets_clp,enumerate_relation).
%binary_function(partial_injection,bsets_clp,enumerate_partial_injection).
%binary_function(partial_surjection,bsets_clp,enumerate_partial_surjection).
%binary_function(total_surjection,bsets_clp,enumerate_total_surjection).
%binary_function(total_relation,bsets_clp,enumerate_total_relation).
%binary_function(total_injection,bsets_clp,enumerate_total_injection).
%binary_function(partial_bijection,bsets_clp,enumerate_partial_bijection).
%binary_function(surjection_relation,bsets_clp,enumerate_surjection_relation).
%binary_function(total_surjection_relation,bsets_clp,enumerate_total_surjection_relation).

% check if the first argument determines the value of a binary_function:
binary_arg1_determines_value(multiplication,A1,int(0)) :- zero(A1).
binary_arg1_determines_value(intersection,A1,[]) :- A1==[].
binary_arg1_determines_value(set_subtraction,A1,[]) :- A1==[].
binary_arg1_determines_value(domain_restriction,A1,[]) :- A1==[]. % first arg is set
binary_arg1_determines_value(range_subtraction,A1,[]) :- A1==[]. % first arg is rel
binary_arg1_determines_value(range_restriction,A1,[]) :- A1==[]. % first arg is rel
binary_arg1_determines_value(cartesian_product,A1,[]) :- A1==[].
binary_arg1_determines_value(composition,A1,[]) :- A1==[].
binary_arg1_determines_value(direct_product,A1,[]) :- A1==[].
binary_arg1_determines_value(parallel_product,A1,[]) :- A1==[].


zero(X) :- nonvar(X), X=int(X1),X1==0.

/* add PartialBijection .... */

/* BOOLEAN EXPRESSIONS */


:- assert_must_succeed(kernel_mappings:symbolic_closure_binary_operator(partial_function)).
:- assert_must_succeed(kernel_mappings:symbolic_closure_binary_operator(total_function)).

symbolic_closure_binary_operator(Op) :-
   binary_in_boolean_type(Op,_),
   binary_not_in_boolean_type(Op,_),
   \+ not_symbolic_binary(Op).
/* note every symbolic closure operator must be dealt with by a binary_in and
   a binary_not_in clause below, otherwise we have a problem */

%not_symbolic_binary(set_subtraction). % initially test 426 failed or runtime went from 6 to 12 secs ; now with do_not_keep_symbolic rules below it is fine and runtime is 3 secs
%not_symbolic_binary(intersection). % seems to be ok to allow symbolic treatment
not_symbolic_binary(range_subtraction). % important for test 1001
not_symbolic_binary(range_restriction).
not_symbolic_binary(domain_subtraction). % important for test 1001
not_symbolic_binary(domain_restriction).
 /* interval symbolic: because we often have expressions of the form a..b where b is > MAXINT */

% return true if a symbolic closure for a binary operator would be definitely empty:
is_definitely_empty(interval,int(A),int(B)) :- number(A),number(B), A>B.
is_definitely_empty(cartesian_product,A,B) :- (B==[]->true ; A==[]). % A==[] actually already checked in binary_arg1_determines_value
is_definitely_empty(parallel_product,A,B) :- (A==[]->true ; B==[]).
is_definitely_empty(total_relation,A,B) :- B==[], definitely_not_empty(A).
is_definitely_empty(total_function,A,B) :- B==[], definitely_not_empty(A). % {} --> {}  = {{}}
is_definitely_empty(total_injection,A,B) :- B==[], definitely_not_empty(A).
is_definitely_empty(total_surjection,A,B) :- B==[], definitely_not_empty(A).
is_definitely_empty(total_bijection,A,B) :- B==[], definitely_not_empty(A).
is_definitely_empty(intersection,A,B) :- (A==[] ; B==[]).
is_definitely_empty(set_subtraction,A,B) :-
     (A==[]  ; nonvar(B), custom_explicit_sets:is_definitely_maximal_set(B)).

:- use_module(kernel_freetypes,[is_non_empty_freetype/1]).
definitely_not_empty(X) :- nonvar(X), definitely_not_empty_aux(X).
definitely_not_empty_aux([_|_]).
definitely_not_empty_aux(avl_set(_)).
definitely_not_empty_aux(global_set(_)).
definitely_not_empty_aux(freetype(ID)) :- is_non_empty_freetype(ID).

% check if it is worthwhile keeping a binary operator symbolic, if so: return value
do_not_keep_symbolic(cartesian_product,Set1,Set2,Res,_WF) :- %print(cart(Set1,Set2)),nl,
    (singleton_set(Set1,_) -> small_known_set(Set2)
     ; singleton_set(Set2,_) -> small_known_set(Set1)
    ),
    kernel_objects:cartesian_product(Set1,Set2,Res).
do_not_keep_symbolic(interval,int(A),int(B),Res,_WF) :-
  ((number(A),number(B),A=B) % or B-A < some limit
     -> custom_explicit_sets:convert_to_avl([int(A)],Res)).
do_not_keep_symbolic(parallel_product,Rel1,Rel2,Res,WF) :-
   not_closure(Rel1), not_closure(Rel2),
   bsets_clp:parallel_product_wf(Rel1,Rel2,Res,WF).
do_not_keep_symbolic(intersection,Set1,Set2,_Res,_WF) :- nonvar(Set1), nonvar(Set2),
   % only keep symbolic if at least one non-interval closure
   nonvar_non_interval_closure(Set1,Set2),
   (nonvar_non_interval_closure(Set1) ->
      \+ small_known_set(Set2) % intersection with singleton set can always be computed
    ; nonvar_non_interval_closure(Set2) -> \+ small_known_set(Set1)),
   !,
   %% print(symbolic_intersection),nl, translate:print_bvalue(Set1),nl, translate:print_bvalue(Set2),nl,nl, %%
   fail.
do_not_keep_symbolic(intersection,Set1,Set2,Res,WF) :-  % otherwise: compute intersection normally
   Call=kernel_objects:intersection(Set1,Set2,Res,WF),
   must_succ_kernel_call(Call,[Set1,Set2],WF,Call).
do_not_keep_symbolic(set_subtraction,Set1,Set2,Res,_WF) :- Set2 == [],!,
    Res = Set1.
do_not_keep_symbolic(set_subtraction,Set1,Set2,_Res,_WF) :- nonvar(Set1),
  % only keep symbolic if at least one non-interval closure
   nonvar_non_interval_closure(Set1,Set2),
   % check if we cannot construct a complement closure:
   \+ can_construct_complement_closure(Set1,Set2),
   \+ small_known_set(Set1),
   !,
   %% nl,print(symbolic_set_subtr(Set1,Set2)),nl,nl, %%
   fail.
do_not_keep_symbolic(set_subtraction,Set1,Set2,Res,WF) :- % otherwise: compute difference normally
   Call=kernel_objects:difference_set_wf(Set1,Set2,Res,WF),  % print(diff_set(Set1,Set2,Res)),nl,
   must_succ_kernel_call(Call,[Set1,Set2],WF,Call).

% should succeed if we do not want to keep a unary operator symbolic
do_not_keep_symbolic_unary(_,ArgValue) :- ArgValue==[],!.
do_not_keep_symbolic_unary(closure,Set1) :- % this is closure1
   very_small_known_set(Set1).


can_construct_complement_closure(Set1,Set2) :-
   nonvar(Set1), nonvar(Set2), Set2 = avl_set(_),
   custom_explicit_sets:is_very_large_maximal_global_set(Set1,_).

nonvar_non_interval_closure(A,B) :- (nonvar_non_interval_closure(A) -> true ; nonvar_non_interval_closure(B)).
nonvar_non_interval_closure(A) :- nonvar(A), A=closure(_,_,_),
   \+ custom_explicit_sets:is_interval_closure_or_integerset(A,_,_),
   \+ custom_explicit_sets:is_not_member_value_closure(A,_Type,_MS2). % there isn't full support for this for intersection yet ! TODO fix


:- use_module(custom_explicit_sets,[quick_custom_explicit_set_approximate_size/2]).
small_known_set(AVL) :- nonvar(AVL), AVL=avl_set(A),
   quick_custom_explicit_set_approximate_size(avl_set(A),Size),
   %print(size(Size)),nl,
   Size < 150.
very_small_known_set(AVL) :- nonvar(AVL), AVL=avl_set(A),
   quick_custom_explicit_set_approximate_size(avl_set(A),Size),
   %print(size(Size)),nl,
   Size < 50.

not_closure(X) :- nonvar(X), \+ functor(X,closure,3). %X \= closure(_,_,_).

% explicit_set([]). explicit_set([_|_]). explicit_set(avl_set(_)).

:- assert_must_fail((kernel_mappings:binary_in_boolean_type(Op,_),
                    \+(kernel_mappings:binary_not_in_boolean_type(Op,_)))).
:- assert_must_fail((kernel_mappings:binary_not_in_boolean_type(Op,_),
                     \+(kernel_mappings:binary_in_boolean_type(Op,_)))).
:- assert_must_fail((kernel_mappings:binary_in_boolean_type(Op,_), \+(kernel_mappings:binary_function(Op,_,_)),
                     kernel_mappings:not_symbolic_binary(Op) )).

binary_in_boolean_type(partial_function,bsets_clp:partial_function_wf).
binary_in_boolean_type(total_function,bsets_clp:total_function_wf).
binary_in_boolean_type(cartesian_product,kernel_objects:is_cartesian_pair_wf).
binary_in_boolean_type(partial_surjection,bsets_clp:partial_surjection_wf).
binary_in_boolean_type(partial_injection,bsets_clp:partial_injection_wf).
binary_in_boolean_type(total_injection,bsets_clp:total_injection_wf). % >->
binary_in_boolean_type(partial_bijection,bsets_clp:partial_bijection_wf).
binary_in_boolean_type(total_surjection,bsets_clp:total_surjection_wf). % -->>
binary_in_boolean_type(total_bijection,bsets_clp:total_bijection_wf).  % >->>
binary_in_boolean_type(relations,bsets_clp:relation_over_wf).
binary_in_boolean_type(total_relation,bsets_clp:total_relation_wf).
binary_in_boolean_type(surjection_relation,bsets_clp:surjection_relation_wf).
binary_in_boolean_type(total_surjection_relation,bsets_clp:total_surjection_relation_wf).
binary_in_boolean_type(interval,kernel_objects:in_nat_range_wf).
binary_in_boolean_type(parallel_product,bsets_clp:in_parallel_product_wf).
binary_in_boolean_type(set_subtraction,kernel_objects:in_difference_set_wf).
binary_in_boolean_type(intersection,kernel_objects:in_intersection_set_wf).
%binary_in_boolean_type(union,bsets_clp:in_union_set_wf). % makes some CBC checks fail; TO DO: investigate
binary_in_boolean_type(range_subtraction,bsets_clp:in_range_subtraction_wf).
binary_in_boolean_type(range_restriction,bsets_clp:in_range_restriction_wf).
binary_in_boolean_type(domain_subtraction,bsets_clp:in_domain_subtraction_wf).
binary_in_boolean_type(domain_restriction,bsets_clp:in_domain_restriction_wf).
%binary_function(overwrite,bsets_clp,in_override_relation). % TO DO: add this, for a more principled symbolic treatment (see test 1491)

% check whether a binary operator is definitely true; irrespective of arguments
binary_in_definitely_true(BOP,ValueOfEl) :- nonvar(ValueOfEl),
    binary_definitely_true_based_on_element_value(BOP,ValueOfEl).

binary_definitely_true_based_on_element_value(partial_function,[]). % {} : A +-> B
binary_definitely_true_based_on_element_value(partial_injection,[]). % {} : A >+> B
binary_definitely_true_based_on_element_value(relations,[]). % {} : A <-> B

% all make use of WF now:
binary_not_in_boolean_type(interval,kernel_objects:not_in_nat_range_wf).
binary_not_in_boolean_type(partial_function,bsets_clp:not_partial_function).
binary_not_in_boolean_type(partial_surjection,bsets_clp:not_partial_surjection_wf).
binary_not_in_boolean_type(total_surjection,bsets_clp:not_total_surjection_wf).
binary_not_in_boolean_type(partial_injection,bsets_clp:not_partial_injection).
binary_not_in_boolean_type(total_injection,bsets_clp:not_total_injection).
binary_not_in_boolean_type(partial_bijection,bsets_clp:not_partial_bijection).
binary_not_in_boolean_type(cartesian_product,kernel_objects:not_is_cartesian_pair).
binary_not_in_boolean_type(relations,bsets_clp:not_relation_over).
binary_not_in_boolean_type(total_function,bsets_clp:not_total_function).
binary_not_in_boolean_type(total_bijection,bsets_clp:not_total_bijection).
binary_not_in_boolean_type(total_relation,bsets_clp:not_total_relation_wf).
binary_not_in_boolean_type(surjection_relation,bsets_clp:not_surjection_relation_wf).
binary_not_in_boolean_type(total_surjection_relation,bsets_clp:not_total_surjection_relation_wf).
binary_not_in_boolean_type(parallel_product,bsets_clp:not_in_parallel_product_wf).
binary_not_in_boolean_type(set_subtraction,kernel_objects:not_in_difference_set_wf).
binary_not_in_boolean_type(intersection,kernel_objects:not_in_intersection_set_wf).
%binary_not_in_boolean_type(union,bsets_clp:not_in_union_set_wf).
binary_not_in_boolean_type(range_subtraction,bsets_clp:not_in_range_subtraction_wf).
binary_not_in_boolean_type(range_restriction,bsets_clp:not_in_range_restriction_wf).
binary_not_in_boolean_type(domain_subtraction,bsets_clp:not_in_domain_subtraction_wf).
binary_not_in_boolean_type(domain_restriction,bsets_clp:not_in_domain_restriction_wf).



symbolic_closure_unary_operator(Op) :-
   (unary_in_boolean_type(Op,_)),
    unary_not_in_boolean_type(Op,_),
   \+ not_symbolic_unary(Op).

not_symbolic_unary(domain).
%not_symbolic_unary(identity). % for Siemens Models

:- assert_must_fail((kernel_mappings:unary_in_boolean_type(Op,_),
 \+(kernel_mappings:unary_not_in_boolean_type(Op,_)))).
/* note every symbolic closure operator must be dealt with by a unary_in and
   a unary_not_in clause below, otherwise we have a problem */
:- assert_must_fail((kernel_mappings:unary_in_boolean_type(UnOp,_),
 \+ kernel_mappings:symbolic_closure_unary_operator(UnOp),
 \+ kernel_mappings:unary_function(UnOp,_,_) )).


/* a waitflags version is available  for these taking one extra argument (the waitflags): */
unary_in_boolean_type(pow_subset,kernel_objects:check_subset_of_wf).
unary_in_boolean_type(fin_subset,kernel_objects:check_finite_subset_of_wf).
unary_in_boolean_type(pow1_subset,kernel_objects:check_non_empty_subset_of_wf).
unary_in_boolean_type(fin1_subset,kernel_objects:check_finite_non_empty_subset_of_wf).
unary_in_boolean_type(struct,kernel_records:is_a_record_wf).
unary_in_boolean_type(seq,bsets_clp:is_sequence_wf). /* last arg: basic type */
unary_in_boolean_type(iseq,bsets_clp:injective_sequence_wf).
unary_in_boolean_type(perm,bsets_clp:permutation_sequence_wf).
unary_in_boolean_type(domain, bsets_clp:in_domain_wf).
unary_in_boolean_type(seq1,bsets_clp:finite_non_empty_sequence).
unary_in_boolean_type(iseq1,bsets_clp:injective_non_empty_sequence).
unary_in_boolean_type(identity,bsets_clp:in_identity).
unary_in_boolean_type(closure,bsets_clp:in_closure1_wf). %element_of_closure1_wf). % this is closure1
%unary_in_boolean_type(reflexive_closure,bsets_clp:element_of_reflexive_closure_wf). % this is closure
% we could add : inverse?, general_concat, .... ? range: in_range_wf


% check whether a unary operator is definitely true; irrespective of arguments
unary_in_definitely_true(BOP,ValueOfEl) :- nonvar(ValueOfEl),
    unary_definitely_true_based_on_element_value(BOP,ValueOfEl).

unary_definitely_true_based_on_element_value(pow_subset,[]).
unary_definitely_true_based_on_element_value(fin_subset,[]).
unary_definitely_true_based_on_element_value(seq,[]).
unary_definitely_true_based_on_element_value(iseq,[]).

unary_not_in_boolean_type(domain, bsets_clp:not_in_domain_wf).
unary_not_in_boolean_type(iseq,bsets_clp:not_injective_sequence).
unary_not_in_boolean_type(struct,kernel_records:not_is_a_record_wf).
unary_not_in_boolean_type(pow_subset,kernel_objects:not_subset_of_wf).
unary_not_in_boolean_type(pow1_subset,kernel_objects:not_non_empty_subset_of_wf).
unary_not_in_boolean_type(fin_subset,kernel_objects:not_finite_subset_of_wf).
unary_not_in_boolean_type(fin1_subset,kernel_objects:not_non_empty_finite_subset_of_wf).
unary_not_in_boolean_type(seq,bsets_clp:not_is_sequence_wf).
unary_not_in_boolean_type(seq1,bsets_clp:not_is_non_empty_sequence_wf).
unary_not_in_boolean_type(iseq1,bsets_clp:not_non_empty_injective_sequence).
unary_not_in_boolean_type(perm,bsets_clp:not_permutation_sequence).
unary_not_in_boolean_type(identity,bsets_clp:not_in_identity).
unary_not_in_boolean_type(closure,bsets_clp:not_in_closure1_wf). %not_element_of_closure1_wf). % this is closure1
%unary_not_in_boolean_type(reflexive_closure,bsets_clp:not_element_of_reflexive_closure_wf).


:- assert_must_fail(( kernel_mappings:cst_in_boolean_type(X,_),
                      \+ kernel_mappings:cst_not_in_boolean_type(X,_) )).
:- assert_must_fail(( kernel_mappings:cst_not_in_boolean_type(X,_),
                      \+ kernel_mappings:cst_in_boolean_type(X,_) )).

cst_in_boolean_type(string_set,kernel_objects:is_string).
cst_in_boolean_type(integer_set(IntSet),Call) :- integerset_in_boolean_type(IntSet,Call).


:- assert_must_fail(( kernel_mappings:integerset_in_boolean_type(X,_),
                      \+ kernel_mappings:integerset_not_in_boolean_type(X,_) )).
:- assert_must_fail(( kernel_mappings:integerset_not_in_boolean_type(X,_),
                      \+ kernel_mappings:integerset_in_boolean_type(X,_) )).

integerset_in_boolean_type('INTEGER',kernel_objects:is_integer).
integerset_in_boolean_type('NATURAL',kernel_objects:is_natural).
integerset_in_boolean_type('NATURAL1',kernel_objects:is_natural1).
integerset_in_boolean_type('INT',kernel_objects:is_implementable_int).
integerset_in_boolean_type('NAT',kernel_objects:is_implementable_nat).
integerset_in_boolean_type('NAT1',kernel_objects:is_implementable_nat1).

cst_not_in_boolean_type(string_set,kernel_objects:is_not_string).
cst_not_in_boolean_type(integer_set(IntSet),Call) :- integerset_not_in_boolean_type(IntSet,Call).

integerset_not_in_boolean_type('INTEGER',kernel_objects:is_not_integer).
integerset_not_in_boolean_type('NATURAL',kernel_objects:is_not_natural).
integerset_not_in_boolean_type('NATURAL1',kernel_objects:is_not_natural1).
integerset_not_in_boolean_type('INT',kernel_objects:is_not_implementable_int).
integerset_not_in_boolean_type('NAT',kernel_objects:is_not_implementable_nat).
integerset_not_in_boolean_type('NAT1',kernel_objects:is_not_implementable_nat1).

% binary_boolean_operator(B-AST-NAME, Module:Predicate, WF_ENABLED)
%binary_boolean_operator(member,kernel_objects:element_of). /* not really used: interpreter deals with In explicitly */
%binary_boolean_operator(not_member,kernel_objects:not_element_of_wf,yes). /* not really used: interpreter deals with In explicitly */
binary_boolean_operator(not_equal,kernel_objects:not_equal_object_wf,yes).
binary_boolean_operator(equal,kernel_objects:equal_object_optimized,no).
binary_boolean_operator(less,kernel_objects:less_than,no).
binary_boolean_operator(less_equal,kernel_objects:less_than_equal,no).
binary_boolean_operator(greater,kernel_objects:greater_than,no).
binary_boolean_operator(greater_equal,kernel_objects:greater_than_equal,no).
binary_boolean_operator(subset,kernel_objects:check_subset_of_wf,yes).
binary_boolean_operator(subset_strict,kernel_objects:strict_subset_of_wf,yes).
binary_boolean_operator(not_subset_strict,kernel_objects:not_strict_subset_of_wf,yes).
binary_boolean_operator(not_subset,kernel_objects:not_subset_of_wf,yes).

:- assert_must_fail(( kernel_mappings:binary_boolean_operator(X,_,_),
                      \+ kernel_mappings:negate_binary_boolean_operator(X,_) )).
:- assert_must_fail(( kernel_mappings:binary_boolean_operator(X,_,_),
                      kernel_mappings:negate_binary_boolean_operator(X,Y), Y=X )).
:- assert_must_fail(( kernel_mappings:negate_binary_boolean_operator(X,X) )).

negate_binary_boolean_operator(X,Y) :-
  ( negate_binary_boolean_operator2(X,Y) -> true
          ; negate_binary_boolean_operator2(Y,X)).
negate_binary_boolean_operator2(member,not_member).
negate_binary_boolean_operator2(subset,not_subset).
negate_binary_boolean_operator2(subset_strict,not_subset_strict).

negate_binary_boolean_operator2(equal,not_equal).
negate_binary_boolean_operator2(less,greater_equal).
negate_binary_boolean_operator2(less_equal,greater).




% input must be list of sets -> second argument set to true if all but one set is known
:- block all_but_one_set_known(-,?).
all_but_one_set_known([],true).
all_but_one_set_known([H|T],KNOWN) :-
  (T==[] -> KNOWN=true
          ; all_but_one(T,H,KNOWN)).
:- block all_but_one(-,?,?).
all_but_one([],_H,true).
all_but_one([H2|T],H1,KNOWN) :- known_value(H2,K2), known_value(H1,K1), all_but_one_wait(K1,K2,T,KNOWN).

:- block all_but_one_wait(-,-,?,?), all_but_one_wait(?,?,-,?).
all_but_one_wait(K1,K2,[],true) :- K1=true,K2=true.
all_but_one_wait(K1,K2,[H3|T],KNOWN) :- known_value(H3,K3),
   (var(K2) -> all_but_one_wait(K2,K3,T,KNOWN) ; all_but_one_wait(K1,K3,T,KNOWN)).

:- block known_value(-,-).
known_value(_,T) :- T==true,!.
known_value(int(X),K) :- !, known_atomic(X,K).
known_value(pred_true /* bool_true */,K) :- !, K=true.
known_value(pred_false /* bool_false */,K) :- !, K=true.
known_value(string(X),K) :- !, known_atomic(X,K).
known_value([],K) :- !, K=true.
known_value(avl_set(_A),K) :- !, K=true.
known_value(global_set(X),K) :- !,known_atomic(X,K).
known_value(freetype(X),K) :- !,known_atomic(X,K).
known_value(A,K) :- when(ground(A),K=true).

:- block known_atomic(-,?).
known_atomic(_,true).


% some code to automatically generate documentation for kernel predicates:
kernel_predicate(M,P,expression,1,BAST,BOP,WF) :-
   unary_function(BAST,M,P), get_waitflag_info(P,WF),
   (translate:unary_prefix(BAST,BOP,_) -> true
    ; translate:unary_postfix(BAST,BOP,_) -> true
    ; translate:function_like(BAST,BOP) -> true
    ; BOP = '??').
kernel_predicate(M,P,expression,2,BAST,BOP,WF) :-
   binary_function(BAST,M,P), get_waitflag_info(P,WF),
   (translate:binary_infix(BAST,BOP,_,_) -> true
    ; translate:function_like(BAST,BOP) -> true ; BOP = '??').
kernel_predicate(M,P,predicate,2,BAST,BOP,WF) :-
   binary_boolean_operator(BAST,M:P,WF),
   (translate:binary_infix(BAST,BOP,_,_) -> true
    ; translate:function_like(BAST,BOP) -> true ; BOP = '??').

get_waitflag_info(P,WF) :-
    (expects_waitflag_and_span(P) -> WF = wf_span ;
    expects_waitflag_and_span(P) -> WF = yes ;
    WF = no).
:- public generate_documentation/0.
generate_documentation :-
    kernel_predicate(M,P,_T,ARITY,BAST,BOP,_WF),
    (  symbolic_closure_unary_operator(BOP) -> Sym=' (symbolic_unary)'
     ; symbolic_closure_binary_operator(BOP) -> Sym=' (symbolic_binary)' % these not printed as binary_function does not include them
     ; Sym= ''),
    format(' ~w (~w/~w) --> ~w:~w ~w~n',[BOP,BAST,ARITY,M,P,Sym]),
    fail.
generate_documentation.

:- assert_must_succeed(kernel_mappings:generate_documentation).

