
% (c) 2015-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_ordering,[ordered_value/2, leq_ordered_value/3]).

:- use_module(self_check).

:- use_module(debug).
:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides term ordering utilities.').

:- use_module(b_global_sets,[b_get_fd_type_bounds/3]).
:- use_module(kernel_objects,[less_than_direct/2, less_than_equal_direct/2]).

% check that values are strictly ordered:
:- block ordered_value(-,?).
ordered_value(pred_false,Y) :- !, Y=pred_true.
ordered_value(pred_true,_Y) :- !, fail.
ordered_value(fd(X,Type),FDY) :- nonvar(Type),
   b_get_fd_type_bounds(Type,LowBnd,UpBnd),!,
   (LowBnd+1 =:= UpBnd -> X=LowBnd, FDY=fd(UpBnd,Type) ;
    less_than_direct(X,UpBnd),
    FDY = fd(Y,Type), less_than_direct(X,Y)).
ordered_value(X,Y) :-  % nl,nl,print(order2(X,Y)),nl,nl,
   ordered_value2(X,Y).

%:- use_module(library(avl),[avl_to_list/2]).
:- block ordered_value2(?,-).
ordered_value2(int(X),int(Y)) :- !, less_than_direct(X,Y).
ordered_value2(fd(X,T),fd(Y,T)) :- !, less_than_direct(X,Y).
ordered_value2(string(X),string(Y)) :- !, prolog_order_less(X,Y).
ordered_value2([],avl_set(_)) :- !.
ordered_value2([],[_|_]) :- !.
ordered_value2([],[]) :- !,fail.
ordered_value2(avl_set(_),[]) :- !,fail.
ordered_value2([_|_],[]) :- !,fail.
%ordered_value2(avl_set(A),avl_set(B)) :- !,
%    avl_to_list(A,AL),avl_to_list(B,BL), AL @<BL. % we need to normalize ! we could check avl_size ?
ordered_value2((X1,X2),(Y1,Y2)) :- !, %print(or(X1,X2,Y1,Y2)),nl,
   leq_ordered_value(X1,Y1,EqPred), opt_ordered_value(EqPred,X2,Y2).
ordered_value2(rec(XFields),rec(YFields)) :- !, ordered_fields(XFields,YFields).
%ordered_value2(A,B) :- print(uncovered_ordered_value(A,B)),nl,fail.
ordered_value2(_,_). % TO DO : treat more: sets,...
% TO DO: use something like lex_chain([[X,X],[Y,Z]],[op(#<)]) for pairs/records of FD values

% check ordered_value only if first arg = pred_true
:- block opt_ordered_value(-,?,?).
opt_ordered_value(pred_true,X2,Y2) :- !, ordered_value(X2,Y2).
opt_ordered_value(_,_,_).

:- block prolog_order_less(-,?), prolog_order_less(?,-).
prolog_order_less(Atom1,Atom2) :- Atom1 @< Atom2.
:- block prolog_order_less_equal(-,?), prolog_order_less_equal(?,-).
prolog_order_less_equal(Atom1,Atom2) :- Atom1 @=< Atom2.

:- use_module(bool_pred,[negate/2]).
% check that values are ordered (<=) and return pred_true if equal, pred_false if strictly <, and unknown if it cannot be determined:
:- block leq_ordered_value(-,?,?).
leq_ordered_value(pred_true,Y,PredRes) :- !, (Y,PredRes)=(pred_true,pred_true).
leq_ordered_value(pred_false,Y,PredRes) :- !, negate(Y,PredRes).
leq_ordered_value(X,Y,PredRes) :- leq_ordered_value2(X,Y,PredRes).

:- block leq_ordered_value2(?,-,?).
leq_ordered_value2(int(X),int(Y),PredRes) :- !,
    less_than_equal_direct(X,Y), atomic_eq_check(X,Y,PredRes).
leq_ordered_value2(fd(X,T),fd(Y,T),PredRes) :- !,
    less_than_equal_direct(X,Y), atomic_eq_check(X,Y,PredRes).
leq_ordered_value2(string(X),string(Y),PredRes) :- !,
    prolog_order_less_equal(X,Y), atomic_eq_check(X,Y,PredRes).
leq_ordered_value2((X1,X2),(Y1,Y2),PredRes) :- !, %print(or(X1,X2,Y1,Y2)),nl,
    leq_ordered_value(X1,Y1,EqPred), opt_leq_ordered_value(EqPred,X2,Y2,PredRes).
leq_ordered_value2(rec(FX),rec(FY),PredRes) :- !, leq_ordered_fields(FX,FY,PredRes).
%leq_ordered_value2(A,B,R) :- nl,print_term_summary(leq_ordered_value2(A,B,R)),nl,fail.
leq_ordered_value2(_,_,pred_unknown).  % TO DO : treat more

atomic_eq_check(X,Y,EqRes) :- when(?=(X,Y),(X=Y -> EqRes=pred_true ; EqRes=pred_false)).

% check ordered_value only if first arg = pred_true
:- block opt_leq_ordered_value(-,?,?,?).
opt_leq_ordered_value(pred_true,X2,Y2,PredRes) :- !, % if component 1 equal: check component 2 leq
    leq_ordered_value(X2,Y2,PredRes).
opt_leq_ordered_value(Res,_,_,Res). % component 1 less or unknown: return result unmodified

%  ordering for records:


:- block ordered_fields(-,-).
ordered_fields([],[]).
ordered_fields([FX|TX],[FY|TY]) :- %print(chk1(FX,TX,FY,TY)),nl,
   ordered_fields_aux(FX,TX,FY,TY).

:- use_module(kernel_records,[check_field_name_compatibility/2]).
:- block ordered_fields_aux(-,?,-,?).
ordered_fields_aux(field(Name1,FX),TX,field(Name2,FY),TY) :-
   check_field_name_compatibility(Name1,Name2),
   (TX==[] -> ordered_value(FX,FY),TY=[] % no other field remaining
    ; leq_ordered_value(FX,FY,EqPred),
      opt_ordered_fields(EqPred,TX,TY)).

% check ordered_value only if first arg = pred_true
:- block opt_ordered_fields(-,?,?).
opt_ordered_fields(pred_true,X2,Y2) :- !, ordered_fields(X2,Y2).
opt_ordered_fields(_,_,_).


:- block leq_ordered_fields(-,-,?).
leq_ordered_fields([],[],pred_true). % are equal
leq_ordered_fields([FX|TX],[FY|TY],PredRes) :- %print(chk1(FX,TX,FY,TY)),nl,
   leq_ordered_fields_aux(FX,TX,FY,TY,PredRes).

:- block leq_ordered_fields_aux(-,?,-,?,?).
leq_ordered_fields_aux(field(Name1,FX),TX,field(Name2,FY),TY,PredRes) :-
   check_field_name_compatibility(Name1,Name2),
   (TX==[] -> leq_ordered_value(FX,FY,PredRes),TY=[] % no other field remaining
    ; leq_ordered_value(FX,FY,EqPred),
      opt_leq_ordered_fields(EqPred,TX,TY,PredRes)).

% check leq_ordered_fields only if first arg = pred_true
:- block opt_leq_ordered_fields(-,?,?,?).
opt_leq_ordered_fields(pred_true,X2,Y2,EqPred) :- !, leq_ordered_fields(X2,Y2,EqPred).
opt_leq_ordered_fields(P,_,_,P).
