:- module(kodkod_annotator, [annotate_kodkod_constraints/2]).

:- use_module(probsrc(bsyntaxtree), [find_typed_identifier_uses/3, get_texpr_type/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,kodkod).
:- module_info(description,'Annotate specific constraints within a typed B or Event-B AST with external predicate calls to be solved by Kodkod.').

%% join_using_couples(+Ids, -Couples).
%
% Join a non-empty list of identifiers to a single (possibly nested) couple.
join_using_couples([Id|T], Couples) :-
  join_using_couples(T, Id, Couples).

join_using_couples([], Couples, Couples).
join_using_couples([Id|T], Acc, Couples) :-
  get_texpr_type(Id, Type1),
  get_texpr_type(Acc, Type2),
  NewAcc = b(couple(Id, Acc), couple(Type1, Type2), []),
  join_using_couples(T, NewAcc, Couples).

%% annotate_kodkod_constraints(+Ast, -AnnotatedAst).
%
% Annotate specific constraints to be solved by Kodkod.
% Currently preferred by Kodkod: implication, equal, not_equal, card
% Exclude: Operations on integers and higher-order quantification.
annotate_kodkod_constraints(Ast, AnnotatedAst) :-
  annotate_kodkod_constraints(Ast, [], NewVarsAcc, TAnnotatedAst),
  NewVarsAcc = [CoupledVars],
  EmptySet = b(set_extension([]),set(any),[]),
  KodkodSolve = b(external_pred_call('KODKOD_SOLVE',[b(integer(1),integer,[]), CoupledVars, EmptySet]),pred,[]),
  AnnotatedAst = b(conjunct(TAnnotatedAst, KodkodSolve),pred,[]).
annotate_kodkod_constraints(Ast, Ast).

% TODO: find more suitable operators for Kodkod annotations

%% annotate_kodkod_constraints(+Ast, +VarsAcc, -NewVarsAcc, -AnnotatedAst).
%
% True if NewVarsAcc is a singleton list containing a couple of local identifiers of annotated constraints,
% and AnnotatedAst the equivalent ast with Kodkod annotations.
annotate_kodkod_constraints(b(forall(Ids, Lhs, Rhs),pred,Info), VarsAcc, NewVarsAcc, AnnotatedAst) :-
  !,
  annotate_kodkod_constraints(Lhs, VarsAcc, VarsAcc1, ALhs),
  annotate_kodkod_constraints(Rhs, VarsAcc1, NewVarsAcc, ARhs),
  AnnotatedAst = b(forall(Ids, ALhs, ARhs),pred,Info).
annotate_kodkod_constraints(b(exists(Ids, Pred),pred,Info), VarsAcc, NewVarsAcc, AnnotatedAst) :-
  !,
  annotate_kodkod_constraints(Pred, VarsAcc, NewVarsAcc, APred),
  AnnotatedAst = b(exists(Ids, APred),pred,Info).
annotate_kodkod_constraints(b(implication(Lhs,Rhs),Type,Info), VarsAcc, NewVarsAcc, AnnotatedAst) :-
  \+contains_quant_or_int_ast([Lhs]),
  \+contains_quant_or_int_ast([Rhs]),
  annotate_kodkod_constraint(b(implication(Lhs,Rhs),Type,Info), VarsAcc, NewVarsAcc, AnnotatedAst).
annotate_kodkod_constraints(b(Node,Type,Info), VarsAcc, NewVarsAcc, AnnotatedAst) :-
  (   Node = equal(Lhs,Rhs)
  ;   Node = not_equal(Lhs,Rhs)),
  \+contains_quant_or_int_ast([Lhs]),
  \+contains_quant_or_int_ast([Rhs]),
  !,
  annotate_kodkod_constraint(b(Node,Type,Info), VarsAcc, NewVarsAcc, AnnotatedAst).
annotate_kodkod_constraints(Ast, VarsAcc, VarsAcc, Ast) :-
  Ast = b(card(b(identifier(_),_,_)),_,_),
  !.
annotate_kodkod_constraints(b(card(Set),Type,Info), VarsAcc, NewVarsAcc, AnnotatedAst) :-
  annotate_kodkod_constraint(b(card(Set),Type,Info), VarsAcc, NewVarsAcc, AnnotatedAst),
  !.
annotate_kodkod_constraints(b(Node,Type,Info), VarsAcc, NewVarsAcc, AnnotatedAst) :-
  Node =.. [Functor, Lhs, Rhs],
  !,
  annotate_kodkod_constraints(Lhs, VarsAcc, VarsAcc1, ALhs),
  annotate_kodkod_constraints(Rhs, VarsAcc1, NewVarsAcc, ARhs),
  NNode =.. [Functor, ALhs, ARhs],
  AnnotatedAst = b(NNode,Type,Info).
annotate_kodkod_constraints(Ast, VarsAcc, VarsAcc, Ast).

annotate_kodkod_constraint(Ast, VarsAcc, [CoupledVars], AnnotatedAst) :-
  AstAsBool = b(convert_bool(Ast),boolean,[]),
  find_typed_identifier_uses(Ast, [], LocalIds),
  join_using_couples(LocalIds, CoupledLocalIds),
  join_using_couples([CoupledLocalIds|VarsAcc], CoupledVars),
  EmptySet = b(set_extension([]),set(any),[]),
  AnnotatedAst = b(external_pred_call('KODKOD',[b(integer(1),integer,[]), CoupledLocalIds, EmptySet, AstAsBool]),pred,[]).

%% contains_quant_or_int_ast(+ListOfAst).
%
% True if list contains an ast of type integer or an higher order quantification.
contains_quant_or_int_ast([b(Node,Type,Info)|_]) :-
  get_texpr_type(b(Node,Type,Info), Type),
  Type == integer,
  !.
contains_quant_or_int_ast([Ast|_]) :-
  Ast =.. [Functor,Ids|_],
  (   Functor == forall
  ;   Functor == exists),
  length(Ids, LIds),
  LIds > 1,
  !.
contains_quant_or_int_ast([Ast|T]) :-
  Ast =.. [_|Args],
  (   contains_quant_or_int_ast(Args),
      !
  ;   contains_quant_or_int_ast(T)).
