% (c) 2016-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(latex_processor, [  process_latex_file/2
                    ]).


:- use_module(module_information).
:- module_info(group,visualization).
:- module_info(description,'This module processes Latex Files for ProB commands.').


:- use_module(debug).
:- use_module(translate).
:- use_module(error_manager).
:- use_module(preferences).
:- use_module(library(lists)).
:- use_module(library(codesio),[with_output_to_codes/4]).

process_latex_file(FileIn,FileOut) :-
    reset_latex_processor_errors,
    temporary_set_preference(repl_cache_parsing,true,CHNG),
    get_preference(latex_encoding,EncodingPref),
    (EncodingPref=auto
      -> open(FileIn,read,StreamIn),
         stream_property(StreamIn,encoding(Encoding))
      ;  open(FileIn,read,StreamIn,[encoding(EncodingPref)]),
         Encoding = EncodingPref
    ),
    %stream_property(StreamIn,encoding_signature(BOM)),
    debug_println(19,encoding(Encoding,FileIn)),
    open(FileOut,write,StreamOut,[encoding(Encoding)]), % use same encoding
    update_latex_output_file(FileOut),
    call_cleanup(process_lines_loop(1,StreamIn,StreamOut),
      (close_stream(StreamIn),
       close_stream(StreamOut),
       reset_temporary_preference(repl_cache_parsing,CHNG),
       latex_processor_errors(Errors),
       (Errors > 0 -> format(user_error,'~w error(s) occurred while processing ~w~n',[Errors,FileIn]),
                       add_error(latex_processor,'Latex Processor Errors: ',Errors)
        ; true)
       )),
     !.
process_latex_file(FileIn,_FileOut) :- add_error(process_latex_file,'Processing failed: ',FileIn).

process_lines_loop(Nr,Stream,StreamOut) :-
   on_exception(E,read_line(Stream,Line),
     (add_latex_processor_error('Exception reading line: ',Nr:E),Line=end_of_file)),
   (Line = end_of_file -> true
    ; %format("Read line ~w : ~s~n",[Nr,Line]),
      (process_line(StreamOut,Nr,Line,[]) -> true
        ; add_latex_processor_error('Processing of line failed: ',Nr),
          format(StreamOut,'ERROR\\marginpar{ProB Error on line ~w} ',[Nr])
      ),
      nl(StreamOut),
      N1 is Nr+1,
      !,
      process_lines_loop(N1,Stream,StreamOut)).

:- use_module(eval_strings,[eval_codes/6, eval_expression_codes/6, set_eval_dot_file/1, unset_eval_dot_file/0]).
:- use_module(probsrc(system_call),[system_call/5]).
:- use_module(bsyntaxtree,[is_set_type/2]).
:- use_module(translate,[translate_bexpression/2]).

eval_codes_success(LineNr,Codes,OuterQuantifier,StringResult,EnumWarning,LocalState) :-
   eval_codes(Codes,OuterQuantifier,StringResult,EnumWarning,LocalState,TypeInfo),
   debug:debug_println(5,eval_codes_output(LineNr,StringResult,TypeInfo)),
   (error_code(TypeInfo) ->
       add_latex_processor_error('ProB Evaluation returned error code: ',TypeInfo),
       fail
    ; string_result_unknown(StringResult) ->
       add_latex_processor_error('ProB Evaluation result unknown on line: ',LineNr),
       fail
     ; true),
   !.
eval_codes_success(LineNr,_Codes,_,StringResult,EnumWarning,LocalState) :-
   add_warning(process_latex_file,'ProB Evaluation failed on line: ',LineNr),
   StringResult = 'FAIL', LocalState = [], EnumWarning=false.

eval_expression_codes_success(Codes,StringResult,EnumWarning,LocalState,Typed) :-
   eval_expression_codes(Codes,StringResult,EnumWarning,LocalState,Typed,TypeInfo),
   debug:debug_println(4,output_expr(TypeInfo)),
   \+ error_code(TypeInfo).

string_result_unknown('UNKNOWN').
string_result_unknown('NOT-WELL-DEFINED').

error_code(error).
error_code(syntax_error).

string_result_false_or_unknown('UNKNOWN').
string_result_false_or_unknown('NOT-WELL-DEFINED').
string_result_false_or_unknown('FALSE').
string_result_false_or_unknown('FAIL').

write_string_result(Stream,Args,StringResult) :-
   write_optional_leadsto(Stream,Args),
   write_stream(Stream,StringResult).

write_optional_leadsto(Stream,Args) :-
  (member("print",Args)
    -> (member("ascii",Args) -> write(Stream,' == ') ; write(Stream,' \\leadsto '))
    ; true).

:- use_module(bvisual2,[set_bvisual2_translation_mode/1]). % in case external functions FORMULA_INFOS,... called

% TO DO: maybe process other comments (\begin{comment}...\end{comment}); but then we need multi-line processing
% TO DO: \probinclude{FILE} include file with processing; allows to reuse code multiple times; e.g, when drawing diagrams a la BMS in Latex
% TO DO: \probprolog{Stream}{PrologCodeUsingStream} ?? very dangerous
process_line(Stream,_) --> "%",!, ignore_rest(R), {format_stream(Stream,"%~s",[R])}.
process_line(Stream,LineNr) --> prob_command(LineNr,CMD,Codes,Args),!,
     {filter_preferences(Args,TempPrefs,RealArgs,LineNr)},
     process_command_with_catch(CMD,Codes,RealArgs,Stream,LineNr),
     !,
     {maplist(reset_preference,TempPrefs)}, % reset to old values
     process_line(Stream,LineNr).
process_line(Stream,LineNr) --> [Code],!, {put_code(Stream,Code)}, process_line(Stream,LineNr).
process_line(_Stream,_) --> []. % do not generate newline, processline also used for single lines in \probdot

% detect set-pref commands:
filter_preferences([],[],[],_).
filter_preferences(["set-pref",P,Val|T],[pref(Pref,CHNG)|PT],Args,LineNr) :-
    atom_codes(PA,P),
    (preference_description(PA,_Desc) -> Pref=PA
     ; eclipse_preference(PA,Pref),preference_description(Pref,_Desc)
     ; tools:ajoin(['Unknown preference for set-pref argument on line ',LineNr,': '],Msg),
         add_latex_processor_error(Msg,PA),
         fail),
    atom_codes(NewValueAtom,Val),
    temporary_set_preference(Pref,NewValueAtom,CHNG),!,
     print(set_pref(Pref,NewValueAtom)),nl,
    filter_preferences(T,PT,Args,LineNr).
filter_preferences([H|T],P,[H|AT],LineNr) :- filter_preferences(T,P,AT,LineNr).

reset_preference(pref(PA,CHNG)) :- reset_temporary_preference(PA,CHNG).

process_command_with_catch(CMD,Codes,Args,Stream,LineNr,In,Out) :-
     on_exception(E,
        process_command(CMD,Codes,Args,Stream,LineNr,In,Out),
        (tools:ajoin(['Exception for ProB command ',CMD,' on line ',LineNr,': '],Msg),
         add_latex_processor_error(Msg,E),
         format(Stream,' ProB Exception on line ~w ',[LineNr]),
         In=Out)),
         !.
process_command_with_catch(CMD,_Codes,_Args,Stream,LineNr,In,Out) :-
         tools:ajoin(['Failed to execute ProB command on line ',LineNr,': '],Msg),
         add_latex_processor_error(Msg,CMD),
         format(Stream,' ProB Failure on line ~w ',[LineNr]),
         In=Out.

:- dynamic latex_processor_errors/1.
latex_processor_errors(0).
reset_latex_processor_errors :- retractall(latex_processor_errors(_)), assert(latex_processor_errors(0)).
add_latex_processor_error(Msg,E) :-
   add_error(latex_processor,Msg,E),
   retract(latex_processor_errors(N)), N1 is N+1, % we need to count as eval_strings/REPL will reset error manager !
   assert(latex_processor_errors(N1)).

:- use_module(system_call, [system_call/5]).
:- use_module(table_tools,[print_value_as_table/4]).

% process commands:
process_command(system,Codes,Args,Stream,LineNr) -->
   !,
    {atom_codes(Cmd,Codes),
     maplist(codes_to_arg,Args,Options),
     format_stream(Stream,'>>> ~w ~w~n',[Cmd,Options]),
     system_call(Cmd,Options,Text,ErrorTextAsCodeList,JExit),
     format_stream(Stream,'~s~n',[Text]),
     (ErrorTextAsCodeList=[] -> true ;
      add_latex_processor_error('Error executing command on line: ',LineNr)),
     (JExit=exit(0) -> true ;
      add_latex_processor_error('Error return code executing command on line: ',LineNr:JExit))}.
process_command(let,Codes,[Expr|Args],Stream,LineNr) --> % \problet calls \probrepl with print option and prints let id = ...
    !,
    { atom_codes(ID,Codes),
     %(member("ascii",Args) -> set_bvisual2_translation_mode(ascii) ; set_latex_mode,set_bvisual2_translation_mode(latex)),
      translate:translate_bexpression(b(identifier(ID),any,[]),IDS),
     format_stream(Stream,'let\\ ~w = ',[IDS]),
     append(["let ",Codes," = ",Expr],NewCodes)},
    process_command(repl,NewCodes,["print"|Args],Stream,LineNr).
process_command(repl,Codes,Args,Stream,LineNr) -->  % \probrepl
    { dformat(user_output,"executing ProB REPL command on line ~w: ~s (~w)~n",[LineNr,Codes,Args]),
     setup_prefs_for_command(Args,CHNG),
     get_preference(expand_avl_upto,Max),
     set_preference(expand_avl_upto,-1), % means no limit
     eval_codes_success(LineNr,Codes,exists,StringResult,_EnumWarning,LocalState),
     (member("print",Args) ->
        eval_strings:last_expression(_,TypedExpr,_),
        print_expr(Stream,TypedExpr,Args)
        ; true),
     (member("silent",Args) -> true
       ; string_result_false_or_unknown(StringResult) -> write_string_result(Stream,Args,StringResult)
       ; member("solution",Args) -> write_solution(Stream,Args,LocalState)
       ; write_string_result(Stream,Args,StringResult)),
     reset_prefs_for_command(Args,CHNG),
     set_preference(expand_avl_upto,Max),
     ((member("check-true",Args), StringResult \== 'TRUE')
           -> tools:ajoin(['Result not TRUE on line ',LineNr,': '],Msg),
              add_latex_processor_error(Msg,StringResult)
        ; true),
     (member("store",Args),
        (LocalState = [] ->  add_latex_processor_error('No bindings to store on line: ',LineNr) ; true),
        member(bind(ID,IDValue),LocalState),
        (kernel_objects:infer_value_type(IDValue,IDType) -> true ; IDType=any),
        store_let_value(ID,IDType,IDValue),fail
      ; true),
     print_last_eval_time(Stream,Args)
     }.
process_command(requires,Codes,_Args,_Stream,LineNr) --> % \probrequires
    !,
    { version:version_str(CurVersion),
      split_chars(Codes,".",VCNrs),
      maplist(codes_to_number,VCNrs,VNrs),
      dformat(user_output,"checking ProB version on line ~w: ~w (against ~w)~n",[LineNr,VNrs,CurVersion]),
      version:compare_current_version(VNrs,CompareResult),
      (CompareResult = current_older
        -> add_latex_processor_error('Version of probcli is older than requested: ',CurVersion)
        ; CompareResult = current_newer ->  format(user_output,"~n*** ProB version newer than required on line ~w: ~s (against ~w)~n~n",[LineNr,Codes,CurVersion])
        ; true)}.

process_command(print,Codes,Args,Stream,LineNr) --> % \probprint
    !,
    { dformat(user_output,"pretty-printing ProB expression on line ~w: ~s (~w)~n",[LineNr,Codes,Args]),
     setup_prefs_for_command(Args,CHNG),
     (member("pred",Args) ->eval_strings:repl_parse_predicate(Codes,exists,TypedExpr,_TypeInfo)
        ; member("expr",Args) -> eval_strings:repl_parse_expression(Codes,TypedExpr,_Type,_Error)
        ; member("subst",Args) -> eval_strings:repl_parse_substitution(Codes,TypedExpr)
        ; eval_strings:repl_parse_expression(Codes,TypedExpr,_Type,_Error) -> true
        ; eval_strings:repl_parse_predicate(Codes,exists,TypedExpr,_TypeInfo)
     ),
     print_expr(Stream,TypedExpr,Args),
     reset_prefs_for_command(Args,CHNG)
     }.
process_command(expr,Codes,Args,Stream,LineNr) -->
    !,
    { dformat(user_output,"evaluating ProB expression on line ~w: ~s (~w)~n",[LineNr,Codes,Args]),
     setup_prefs_for_command(Args,CHNG),
     get_preference(expand_avl_upto,Max),
     set_preference(expand_avl_upto,-1), % means no limit
     eval_expression_codes_success(Codes,StringResult,_EnumWarning,_LocalState,Typed),
     translate_bexpression(Typed,PS),
     %tools:print_bt_message(trans(PS)),
     (member("value",Args) -> true % we only print value
      ; member("silent",Args) -> true % we only print value
      ; write_stream(Stream,PS), write_stream(Stream,' = ')),
     ((member("string",Args),eval_strings:get_last_result_value(_,string,string(Value)))
       -> write_stream(Stream,Value)  % remove string quotes
        ; write_stream(Stream,StringResult)),
     reset_prefs_for_command(Args,CHNG),
     print_last_eval_time(Stream,Args),
     set_preference(expand_avl_upto,Max)}.
process_command(table,Codes,Args,Stream,LineNr) -->
    % options for \probtable : no-tabular, no-hline, no-headings, no-row-numbers, max-table-size=NR
    !,
    { dformat(user_output,"evaluating ProB expression as table on line ~w: ~s (~w)~n",[LineNr,Codes,Args]),
     set_latex_mode,
     eval_codes_success(LineNr,Codes,exists,_StringResult,_EnumWarning,_LocalState),
     eval_strings:get_last_result_value(Expr,_,Value),
     maplist(codes_to_arg,Args,Options),
     print_value_as_table(Stream,Expr,Value,[latex|Options]),
     unset_latex_mode
     }.
process_command(dotcmd,Codes,Args,_Stream,LineNr) -->  % probdotcmd: call ProB Dot generation command such as state_space, signature_merge, dfa_merge, state_as_graph, invariant, properties, ...
    {atom_codes(Cmd,Codes), process_dot_command(Cmd,Args,LineNr)}.
process_command(dot,Codes,Args,_Stream,LineNr) -->
    { dformat(user_output,"evaluating ProB expression or predicate as dot on line ~w: ~s (~w)~n",[LineNr,Codes,Args]),
     atom_codes(Atom,Codes),
     (Args = [FileCodes|T]
      -> % process Filename in case \probexpr used there; tell used to re-direct any normal print,... away from FCStream
         with_output_to_codes((tell(user_output),process_line(FCStream,LineNr,FileCodes,[]),told),
                              FCStream,NewFileCodes,[]),
         atom_codes(File,NewFileCodes), %user:tcltk_show_expression_as_dot(Atom,File),
         expand_file_relativ_to_latex_file(File,EFile),
         set_eval_dot_file(EFile),
         call_cleanup(eval_codes_success(LineNr,Codes,exists,_StringResult,_EnumWarning,_LocalState),
           unset_eval_dot_file),
         % we could call graph_canon:print_cstate_graph(_LocalState,File)
         (T= [PDF_Codes|RestArgs] ->
            % also process PDF_Codes: in case \probexpr used
            with_output_to_codes((tell(user_output),process_line(FCStream2,LineNr,PDF_Codes,[]),told),
                                 FCStream2,NewPDF_Codes,[]),
            dformat(user_output,' ~s ==> ~s ~n',[PDF_Codes,NewPDF_Codes]),
            atom_codes(PDFFile,NewPDF_Codes),
            expand_file_relativ_to_latex_file(PDFFile,EPDFFile),
            gen_pdf(EFile,RestArgs,EPDFFile)
          ; true % user will have to generate PDF in Makefile
          )
      ; add_latex_processor_error('No file provided for probdot and expression: ',Atom)
     )
     }.
process_command(if,TestC,Args,Stream,LineNr) -->
    { dformat(user_output,"evaluating ProB if-construct on line ~w: ~s (~w)~n",[LineNr,TestC,Args]),
      Args = [ LatexTRUE, LatexFALSE],
     eval_codes_success(LineNr,TestC,exists,StringResult,_EnumWarning,_LocalState), % TO DO: refactor eval_codes to return various number of results in list
     dformat(user_output,'Value for probif: ~w~n',[StringResult]),
     (StringResult = 'TRUE' -> process_line(Stream,LineNr,LatexTRUE,[])
      ; StringResult = 'FALSE' -> process_line(Stream,LineNr,LatexFALSE,[])
      ; StringResult = 'UNKNOWN' -> add_latex_processor_error('UNKNOWN predicate or BOOL expression result for probif: ',LineNr)
      ; add_latex_processor_error('Not a predicate or BOOL expression for probif: ',LineNr)
      )
     }.
process_command(for,VARC,Args,Stream,LineNr) -->
    { dformat(user_output,"evaluating ProB for-construct on line ~w: ~s (~w)~n",[LineNr,VARC,Args]),
      Args = [ EXPRC, LATEX],
      atom_codes(ID,VARC), % the iteration variable
     eval_expression_codes_success(EXPRC,_StringResult,_EnumWarning,_LocalState,_Typed),
     eval_strings:get_last_result_value(_,Type,Value), %print(value(Value,Type)),nl,
     is_set_type(Type,IDType),
     custom_explicit_sets:expand_custom_set_to_list(Value,ExpValue,_,probfor),
     prob_foreach(ExpValue,ID,IDType,LATEX,Stream,LineNr)
     }.
% TO DO: add command to register latex pretty printing modifications
% e.g., \probppid{epsilon}{\epsilon}

:- use_module(tools,[get_filename_extension/2]).
:- use_module(meta_interface,[command_arguments/2, %command_optional_arguments/4,
                          command_category/2,
                          is_dot_command/1, is_dot_command_for_expr/1,
                          call_command/5]).
% this works for is_dot_command and is_dot_command_for_expr
process_dot_command(Cmd,Args,LineNr) :- print(process(Cmd,LineNr)),nl,
   command_category(Cmd,dot),
   command_arguments(Cmd,FormalArgs),
   dformat(user_output,"evaluating DOT command ~w using ~w~n",[Cmd,FormalArgs]),
   !,
   same_length(FormalArgs,ActualArgsC),
   (append(ActualArgsC,OptionalArgsC,Args)
    -> maplist(codes_to_atom,ActualArgsC,ActualArgs0),
       include(latex_dot_arg,OptionalArgsC,LatexArgs), % arguments for latex_processor here
       exclude(latex_dot_arg,OptionalArgsC,OptionalArgsC2),
       maplist(codes_to_arg,OptionalArgsC2,OptionalArgsA),
       expand_relative_files(FormalArgs,ActualArgs0,ActualArgs),
       ((get_dot_file(FormalArgs,ActualArgs,DOTFILE),
         append([PDFFILE],OptionalArgs,OptionalArgsA),
         get_filename_extension(PDFFILE,'pdf') )
         ->  debug:debug_println(19,will_generate_pdf(PDFFILE,OptionalArgs)) % first argument is PDFFile
          ;  OptionalArgs=OptionalArgsA
        ),
       debug:debug_println(19,call_command(Cmd,FormalArgs,ActualArgs,OptionalArgs)),
       (is_dot_command(Cmd) -> true
        ; is_dot_command_for_expr(Cmd) -> true
        ; add_latex_processor_error('Unknown dotcmd: ',Cmd)),
       setup_prefs_for_command(["ascii"|LatexArgs],CHNG), % DOT does not support Latex
       (call_command(dot,Cmd,FormalArgs,ActualArgs,OptionalArgs)
        -> reset_prefs_for_command(["ascii"|LatexArgs],CHNG),
           (var(PDFFILE) -> true ; expand_file_relativ_to_latex_file(PDFFILE,EPDFILE),
                                   gen_pdf(DOTFILE,LatexArgs,EPDFILE))
         ; reset_prefs_for_command(["ascii"],CHNG),
           add_latex_processor_error('Command failed: ',Cmd),fail)
       % TO DO: optionally generate PDF
    ;  tools:ajoin(['Not enough arguments for dot command on line ',LineNr,': '],Msg),
       add_latex_processor_error(Msg,Cmd),fail
   ).
process_dot_command(Cmd,_Args,LineNr) :-
   tools:ajoin(['Not a valid dot command on line ',LineNr,': '],Msg),
   add_latex_processor_error(Msg,Cmd),fail.

:- use_module(tools_commands,[get_dot_engine_options/2]).
gen_pdf(DOTFILE,Args,PDFFILE) :-
     (member("sfdp",Args) -> get_dot_engine_options(sfdp,Opts) ; get_dot_engine_options(default,Opts)),
     get_preference(path_to_dot,DOTCMD),
     dformat(user_output,'Calling dot: ~w to generate ~w~n',[DOTCMD,PDFFILE]),
     append(Opts,['-Tpdf', '-o', PDFFILE, DOTFILE],FullOpts),
     system_call(DOTCMD, FullOpts,_TextOut,_ErrText,_JExit).
% try and find the dotfile in the arguments
get_dot_file([dotfile|_],[H|_],Res) :- !, Res=H.
get_dot_file([_|FT],[_|T],Res) :- get_dot_file(FT,T,Res).

expand_relative_files([],[],[]).
expand_relative_files([dotfile|T],[A|AT], [EA|EAT]) :-  !,
     expand_file_relativ_to_latex_file(A,EA), expand_relative_files(T,AT,EAT).
expand_relative_files([_|FT],[AA|AT],[AA|EAT]) :- expand_relative_files(FT,AT,EAT).

% list of arguments processed by latex
latex_dot_arg(X) :- member(X,["nopt","prop-logic","sfdp","dot"]).  % dot ist the default

% used for maplist
codes_to_atom(C,A) :- atom_codes(A,C).
codes_to_number(C,A) :- number_codes(A,C).

% printing expression or substitution:
print_expr(Stream,TypedExpr,Args) :-
    (member("strip-exists",Args)
        -> strip_exists(TypedExpr,TE2) % sometimes we do not want to see the outer existential quantifiers
         ; TE2=TypedExpr),
     translate:translate_subst_or_bexpr(TE2,PS),
     %dformat(user_output,"pretty-printing :: ~w ~n",[PS]),
     format_stream(Stream,'~w ',[PS]).

:- use_module(tools,[split_chars/3, safe_number_codes/2,safe_atom_codes/2]).
% used for table argument processing
codes_to_arg(Codes,argument_value(L,R)) :- split_chars(Codes,"=",[Left,Right]), !, % detect Option=Value
   atom_codes(L,Left),
   (safe_number_codes(R,Right) -> true ; atom_codes(R,Right)).
codes_to_arg(C,A) :- atom_codes(A,C).

prob_command(LineNr,CMD,Codes,Args) --> "\\prob",cmd(CMD),"{", get_command(LineNr,0,Codes,Args),
   {check_for_invalid_args(CMD,Args,LineNr)}.

check_for_invalid_args(CMD,Args,LineNr) :- invalid_arg(CMD,Args,IllegalArg),
    tools:ajoin(['Illegal argument for ProB Latex command on line ',LineNr,': '],Msg),
    safe_atom_codes(Arg,IllegalArg),
    add_latex_processor_error(Msg,CMD:Arg),fail.
check_for_invalid_args(_,_,_).


cmd(dot) --> "dot".
cmd(dotcmd) --> "dotcmd".
cmd(expr) --> "expr".
cmd(for) --> "for".
cmd(if) --> "if".
cmd(let) --> "let".
cmd(print) --> "print".
cmd(repl) --> "repl".
cmd(table) --> "table".
cmd(requires) --> "requires".
cmd(system) --> "system".

argmember(Cmd,["set-pref",_,_|T]) :- !, argmember(Cmd,T).
argmember(Cmd,[Cmd|_]).
argmember(Cmd,[_|T]) :- argmember(Cmd,T).

invalid_arg(let,[_|Args],IllegalArg) :- argmember(IllegalArg,Args),
      nonmember(IllegalArg,["ascii","silent","solution","store","time","nopt","prop-logic"]).
invalid_arg(repl,Args,IllegalArg) :- argmember(IllegalArg,Args),
      nonmember(IllegalArg,["ascii","check-true","print","silent","solution","store","strip-exists",
                            "time","nopt","prop-logic","check"]).
invalid_arg(print,Args,IllegalArg) :- argmember(IllegalArg,Args),
      nonmember(IllegalArg,["ascii","pred","expr","nopt","prop-logic","strip-exists","subst"]). % expr is default
invalid_arg(expr,Args,IllegalArg) :- argmember(IllegalArg,Args),
      nonmember(IllegalArg,["ascii","silent","string","time","value","nopt","prop-logic"]).
%invalid_arg(table,Args,IllegalArg) :- member(IllegalArg,Args),nonmember(IllegalArg,["no-headings","no-hline",...]).
invalid_arg(if,[_,_,IllegalArg|_],IllegalArg). % too many args
invalid_arg(for,[_,_,IllegalArg|_],IllegalArg).

get_command(LineNr,Nr,[123|T],Arg) --> "{", !, {N1 is Nr+1}, get_command(LineNr,N1,T,Arg).
get_command(LineNr,0,[],Arg) --> "}",!, get_arguments(LineNr,Arg).
get_command(LineNr,Nr,[125|T],Arg) --> "}", !, {N1 is Nr-1}, get_command(LineNr,N1,T,Arg).
get_command(LineNr,Nr,[H|T],Arg) --> [H],!, get_command(LineNr,Nr,T,Arg).
get_command(LineNr,_,[],[]) --> {add_error(get_command,'prob command not closed: ',LineNr)}.

get_arguments(LineNr,[Arg|T]) --> "{", get_arg(Arg,0,LineNr),!, get_arguments(LineNr,T).
get_arguments(_,[]) --> [].

get_arg([],0,_) --> "}".
get_arg([123|T],Level,LineNr) --> "{", !, {L1 is Level+1}, get_arg(T,L1,LineNr).
get_arg([125|T],Level,LineNr) --> "}", {Level>0}, !, {L1 is Level-1}, get_arg(T,L1,LineNr).
get_arg([H|T],Level,LineNr) --> [H],!,get_arg(T,Level,LineNr).
get_arg([],_,LineNr) --> {add_error(get_arg,'prob argument not closed: ',LineNr)}.


ignore_rest([H|T]) --> [H], !, ignore_rest(T).
ignore_rest([]) --> [].


setup_prefs_for_command(Args,(CHNG1,CHNG2)) :-
     (member("ascii",Args)
       -> (latex_mode -> unset_latex_mode ; true),
          set_bvisual2_translation_mode(ascii) % in bvisual
        ; set_latex_mode,
          set_bvisual2_translation_mode(latex)),
     (member("nopt",Args) -> temporary_set_preference(optimize_ast,false,CHNG1) ; true),
     (member("prop-logic",Args) -> temporary_set_preference(pp_propositional_logic_mode,true,CHNG2) ; true).
reset_prefs_for_command(Args,(CHNG1,CHNG2)) :-
     (member("ascii",Args) -> true ; unset_latex_mode), % after each command we unset latex mode
     (member("nopt",Args) -> reset_temporary_preference(optimize_ast,CHNG1) ; true),
     (member("prop-logic",Args) -> reset_temporary_preference(pp_propositional_logic_mode,CHNG2) ; true).

% ------------------------------
:- use_module(debug,[debug_mode/1]).

dformat(Stream,F,A) :- (debug_mode(on) -> format_stream(Stream,F,A) ; true).

format_stream(Stream,F,A) :- on_exception(E, format(Stream,F,A),
         (add_latex_processor_error('Exception while writing to stream with format: ',E),
          write(Stream,' *** EXCEPTION *** '))).

write_stream(Stream,T) :- on_exception(E, write(Stream,T),
         (add_latex_processor_error('Exception while writing to stream: ',E),
          write(Stream,' *** EXCEPTION *** '))).

close_stream(Stream) :- on_exception(E, close(Stream),
         (add_latex_processor_error('Exception while closing stream: ',E))).

print_last_eval_time(Stream,Args) :- member("time",Args),
     eval_strings:last_eval_time(_Time,WTime),!,
     (member("ascii",Args) -> format_stream(Stream,'\\ (~w ms)',[WTime]) % includes parsing time
         ; format_stream(Stream,' \\mathit{\\ (~w ms)}',[WTime])).
print_last_eval_time(Stream,Args) :- member("time",Args),!, write(Stream,' (?? ms) ').
print_last_eval_time(_,_).

:- use_module(translate,[translate_bvalue/2]).
write_solution(Stream,Args,LocalState) :-
   write_optional_leadsto(Stream,Args),
   (member("ascii",Args) -> FormatStr = ' ~w = ~w '; FormatStr = ' \\mathit{~w} = ~w '),
   write_sol_aux(LocalState, Stream, FormatStr).
write_sol_aux([bind(ID,IDV)|T],Stream,FormatStr) :-
   translate_bvalue(IDV,VS),
   format_stream(Stream,FormatStr,[ID,VS]),
   (T=[] -> true ; write(Stream,' \\wedge'), write_sol_aux(T,Stream,FormatStr)).
write_sol_aux([],Stream,_) :- write(Stream,'TRUE').

% ------------------------------


prob_foreach([],_ID,_IDT,_Ltx,_Stream,_).
prob_foreach([Value|TVal], ID, IDType,LATEX, Stream,LineNr) :-
    %print(process_foreach(ID,Value)),nl,
    store_let_value(ID,IDType,Value),
    %print(process(LATEX)),nl,
    % now process line with ID = Value
    process_line(Stream,LineNr,LATEX,[]), %nl(Stream),
    %print(done),nl,
    prob_foreach(TVal,ID, IDType,LATEX,Stream,LineNr).



store_let_value(ID,IDType,Value) :-
    retractall(eval_strings:stored_let_value(ID,_,_)),  % TO DO: provide proper predicate in eval_strings for this
    assert(eval_strings:stored_let_value(ID,IDType,Value)).

strip_exists(b(exists(_,Body),pred,_),Res) :- !, strip_exists(Body,Res).
strip_exists(B,B).

% ----------------------

:- use_module(tools_strings,[string_concatenate/3]).
:- use_module(tools,[get_parent_directory/2, is_absolute_path/1]).
:- dynamic latex_output_file/2.
update_latex_output_file(FileOut) :-
   retractall(latex_output_file(_,_)),
   get_parent_directory(FileOut,Dir),
   assert(latex_output_file(FileOut,Dir)).

expand_file_relativ_to_latex_file(File,ExpandedFile) :- \+ is_absolute_path(File),
   latex_output_file(_,Dir),
   !,
   string_concatenate(Dir,File,ExpandedFile).
expand_file_relativ_to_latex_file(F,F).
