% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(logger, [set_log_file/1, get_log_file/1, set_logging_mode/1, reset_logger/0,
                              writeln_log/1,
                              writeln_log_time/1,
                              write_xml_element_to_log/2,
                              write_prolog_term_as_xml_to_log/1,
                              write_bstate_to_log/1, write_bstate_to_log/2,
                              start_xml_group_in_log/1, start_xml_group_in_log/3,
                              stop_xml_group_in_log/1, stop_xml_group_in_log_no_statistics/1,
                              close_all_xml_groups_in_log/0,
                              logging_is_enabled/0,
                              read_xml_log_file/2]).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'This module is responsible for (xml and prolog) logging.').


:- use_module(self_check).
:- use_module(error_manager).
:- use_module(library(xml)).
:- use_module(library(lists)).

:- dynamic logfile/1.
set_log_file(F) :- retractall(logfile(_)), assert(logfile(F)).
get_log_file(F) :- logfile(F).

:- dynamic logging_mode/1.
logging_mode(prolog).
% valid modes: prolog and xml
set_logging_mode(Mode) :- retractall(logging_mode(_)), assert(logging_mode(Mode)),
   (Mode=xml -> format_log_header(reset,'<?xml version="1.0" encoding="UTF-8"?>~n',[]) ; true).

   % TO DO: use:
    %get_preference(xml_encoding,EncodingPref),

logging_is_enabled :- logfile(_),!.

prolog_log_file(F) :- logfile(F), logging_mode(prolog).

reset_logger :- retractall(open_xml_group(_,_)),
   retractall(logfile(_)), set_logging_mode(prolog).

% TO DO: try and move most writeln_log calls to write_xml_element_to_log format
writeln_log(Term) :-
    (prolog_log_file(F)
      ->  open(F,append,S,[encoding('UTF-8')]),
          write_term(S,Term,[quoted(true)]), write(S,'.'),nl(S),
          close(S)
     ; true
    ).

open_logfile(Stream) :- logfile(F), open(F,append,Stream,[encoding('UTF-8')]).

format_log(FormatString,Args) :- %format(FormatString,Args),nl,
    (logfile(F)
      ->  open(F,append,S,[encoding('UTF-8')]),
          format(S,FormatString,Args),
          close(S)
     ; true
    ).

writeln_log_time(Term) :-
    (prolog_log_file(_) ->
       statistics(runtime,[Time,_]),
       statistics(walltime,[WTime,_]),
       statistics(memory_used,M), MB is M  / 1000000, % used instead of deprecated 1048576
       Term=..[H|Args],
       append(Args,[Time,WTime,mb(MB)],NArgs),
       NT =.. [H|NArgs],
       writeln_log(NT)
    ;  true).

:- use_module(library(file_systems),[file_exists/1]).
format_log_header(reset,FormatString,Args) :- !, % reset means we want to start with a fresh log file
    (logfile(F)
      ->  open(F,write,S,[encoding('UTF-8')]),
          format(S,FormatString,Args),
          close(S)
     ; true
    ).
format_log_header(_,FormatString,Args) :-
% like format_log, but only writes if the file does not exist yet
    (logfile(F), \+ file_exists(F)
      ->  open(F,append,S,[encoding('UTF-8')]),
          format(S,FormatString,Args),
          close(S)
     ; true
    ).

:- assert_must_succeed((logger:xml_encode_text("b<>c",R),R=="b&lt;&gt;c")).
xml_encode_text(Codes,Res) :- XML = xml([],[pcdata(Codes)]), xml_parse(Encoded,XML),!,Res=Encoded.
xml_encode_text(Codes,Encoded) :- format(user_error,'Could not encode for XML: ~s~n',[Codes]),
   Encoded=Codes.
% we could also use xml:pcdata_generation(Codes, Encoded, []).

:- assert_must_succeed((logger:xml_encode_element(check_goal,[true/1],R),R=="<check_goal true=\"1\" />")).
xml_encode_element(Tag,Attributes,Encoded2) :-
   maplist(prepare_attribute,Attributes,XMLAttr),
   XML = xml([],[element(Tag,XMLAttr,[])]),
   xml_parse(Encoded,XML),
   peel_off_leading_newline(Encoded,Encoded2).

peel_off_leading_newline([10|T],R) :- !, peel_off_leading_newline(T,R).
peel_off_leading_newline([13|T],R) :- !, peel_off_leading_newline(T,R).
peel_off_leading_newline(R,R).

% write a tag with attributes to the log file
write_xml_element_to_log(_,_) :- \+ logfile(_), !.
write_xml_element_to_log(Tag,Attributes) :- logging_mode(xml),!,
   xml_encode_element(Tag,Attributes,Encoded),
   indent_log(WS),
   format_log("~s~s~n",[WS,Encoded]).
write_xml_element_to_log(Tag,Attributes) :-
   Term =.. [Tag,Attributes],
   format_log("~w.~n",[Term]).

% write a Prolog Term either as Prolog Term in Prolog mode or in nested XML form
%write_term_to_log(Term) :- logging_mode(xml),!,
%   write_prolog_term_as_xml_to_log(Term).
%write_term_to_log(Term) :- writeln_log(Term).

write_prolog_term_as_xml_to_log(A) :- number(A),!,
   indent_log(WS),
   format_log("~s<number>~w</number>~n",[WS,A]).
write_prolog_term_as_xml_to_log(A) :- var(A),!,
   indent_log(WS),
   format_log("~s<variable>~w</variable>~n",[WS,A]).
write_prolog_term_as_xml_to_log(A) :- atomic(A),!, convert_to_codes(A,Codes),
   xml_encode_text(Codes,Encoded),
   indent_log(WS),
   (is_a_file_path(Encoded) -> format_log("~s<path>~s</path>~n",[WS,Encoded])
    ; format_log("~s<atom>~s</atom>~n",[WS,Encoded])).
write_prolog_term_as_xml_to_log(A/B) :- !,
   start_xml_group_in_log(bind),
   write_prolog_term_as_xml_to_log(A),
   write_prolog_term_as_xml_to_log(B),
   stop_xml_group_in_log_no_statistics(bind).
write_prolog_term_as_xml_to_log([H|T]) :- !, % Note: we assume we have a proper list !
   start_xml_group_in_log(list),
   maplist(write_prolog_term_as_xml_to_log,[H|T]),
   stop_xml_group_in_log_no_statistics(list).
write_prolog_term_as_xml_to_log(T) :- T =.. [Functor|Args],
   %TO DO, something like: escape / xml_encode_text(Functor,EFunc),
   encode_functor(Functor,XML_Functor),
   start_xml_group_in_log(XML_Functor),
   maplist(write_prolog_term_as_xml_to_log,Args),
   stop_xml_group_in_log_no_statistics(XML_Functor).

encode_functor('-',R) :- !, R='prolog-'.
encode_functor(X,X).

write_bstate_to_log(State) :- write_bstate_to_log(State,'').

% in response to logxml_write_vars
write_bstate_to_log(State,Prefix) :- logging_mode(xml),!,
   start_xml_group_in_log(state),
   atom_codes(Prefix,PrefixCodes),
   (State=root -> start_xml_group_in_log(root), stop_xml_group_in_log_no_statistics(root)
     ; maplist(write_b_binding_as_xml_to_log(PrefixCodes),State) -> true
     ; add_internal_error('Could not write state to xml logfile: ',write_bstate_to_log(State))),
   stop_xml_group_in_log_no_statistics(state).
write_bstate_to_log(_,_Prefix).

write_b_binding_as_xml_to_log(Prefix,bind(VarName,Value)) :- % TO DO: encode VarName
   atom_codes(VarName,Codes),
   append(Prefix,_,Codes), % check that variable name starts with prefix
   !,
   xml_encode_text(Codes,ECodes),
   atom_codes(EVarName,ECodes),
   start_xml_group_in_log(variable,name,EVarName),
   xml_write_b_value_to_log(Value),
   stop_xml_group_in_log_no_statistics(variable).
write_b_binding_as_xml_to_log(_,_).

xml_write_b_value_to_log(Value) :-
   open_logfile(Stream),
   indent_log(WS),format(Stream,'~s ',[WS]),
   xml_write_b_value(Value,Stream),
   format(Stream,'~n',[]),
   close(Stream).

:- use_module(probsrc(custom_explicit_sets),[expand_custom_set_to_list/2]).
:- use_module(probsrc(translate),[translate_bvalue_to_codes/2]).
xml_write_b_value_map(Stream,O) :- xml_write_b_value(O,Stream).
xml_write_b_value(Var,Stream) :- var(Var),!,
    add_internal_error('Illegal variable value:',xml_write_b_value(Var,Stream)),
    format(Stream,'<value>~w</value>',[Var]).
xml_write_b_value((Fst,Snd),Stream) :- !,
    write(Stream,'<pair><fst>'),
    xml_write_b_value(Fst,Stream),
    write(Stream,'</fst><snd>'),
    xml_write_b_value(Snd,Stream),
    write(Stream,'</snd></pair> ').
xml_write_b_value([],Stream) :- !,write(Stream,'<empty_set></empty_set> ').
xml_write_b_value(CS,Stream) :- custom_set_to_expand(CS),!,
    expand_custom_set_to_list(CS,Elements),
    write(Stream,'<set>'),
    maplist(xml_write_b_value_map(Stream),Elements),
    write(Stream,'</set> ').
xml_write_b_value([H|T],Stream) :- !,
    write(Stream,'<set>'),
    maplist(xml_write_b_value_map(Stream),[H|T]),
    write(Stream,'</set> ').
xml_write_b_value(rec(Fields),Stream) :- !,
    write(Stream,'<record>'),
    maplist(xml_write_b_field_value(Stream),Fields),
    write(Stream,'</record> ').
xml_write_b_value(string(S),Stream) :- !,
    atom_codes(S,Codes),
    xml_encode_text(Codes,Encoded),
    format(Stream,'<string>~s</string>',[Encoded]).
xml_write_b_value(int(N),Stream) :- !,
    format(Stream,'<integer>~w</integer>',[N]).
xml_write_b_value(pred_true,Stream) :- !,
    format(Stream,'<bool>TRUE</bool>',[]).
xml_write_b_value(pred_false,Stream) :- !,
    format(Stream,'<bool>FALSE</bool>',[]).
xml_write_b_value(fd(Nr,Type),Stream) :- !,
    translate_bvalue_to_codes(fd(Nr,Type),SValue),
    format(Stream,"<enum type=\"~w\" nr=\"~w\">~s</enum>",[Type,Nr,SValue]).
xml_write_b_value(Value,Stream) :-
    is_custom_explicit_set(Value,xml_write),
    is_interval_closure(Value,Low,Up),
    !,
    write(Stream,'<interval_set><from>'),
    xml_write_b_value(int(Low),Stream),
    write(Stream,'</from><to>'),
    xml_write_b_value(int(Up),Stream),
    write(Stream,'</to></interval_set>').
xml_write_b_value(Value,Stream) :- % other value, freetype, freeval, closure, ...
    is_custom_explicit_set(Value,xml_write),
    !,
    translate_bvalue_to_codes(Value,SValue),
    xml_encode_text(SValue,Encoded),
    format(Stream,'<symbolic_set>~s</symbolic_set>',[Encoded]).
xml_write_b_value(Value,Stream) :- % other value freeval, ...
    translate_bvalue_to_codes(Value,SValue),
    xml_encode_text(SValue,Encoded),
    format(Stream,'<value>~s</value>',[Encoded]).
% TO DO: check if there are uncovered values, e.g.,  freeval(ID,Case,Value)

:- use_module(custom_explicit_sets,[is_interval_closure/3,
   is_custom_explicit_set/2, dont_expand_this_explicit_set/2]).
custom_set_to_expand(avl_set(_)).
custom_set_to_expand(CS) :- nonvar(CS),
   is_custom_explicit_set(CS,xml_write),
   \+ dont_expand_this_explicit_set(CS,1000).

xml_write_b_field_value(Stream,field(Name,Val)) :-
   atom_codes(Name,Codes), xml_encode_text(Codes,Encoded),
   format(Stream,'<field name=\"~s\">',[Encoded]),
   xml_write_b_value(Val,Stream), write(Stream,'</field>').

% ---------------------------

:- use_module(tools,[host_platform/1]).
is_a_file_path(Codes) :- member(47,Codes).
is_a_file_path(Codes) :- host_platform(windows), member(92,Codes). % windows

prepare_attribute('='(Tag,Atom),'='(Tag,Codes)) :- convert_to_codes(Atom,Codes).
prepare_attribute('/'(Tag,Atom),'='(Tag,Codes)) :- convert_to_codes(Atom,Codes).

:- use_module(library(codesio),[write_to_codes/2]).
convert_to_codes(V,Codes) :- var(V),!,Codes="_".
convert_to_codes([H|T],Codes) :- number(H),!, Codes=[H|T].
convert_to_codes(N,Codes) :- number(N),!, number_codes(N,Codes).
convert_to_codes(A,Codes) :- atom(A),!,atom_codes(A,Codes).
convert_to_codes(A,Codes) :- write_to_codes(A,Codes).

:- dynamic open_xml_group/2, nesting_level/1.
nesting_level(0).
update_nesting_level(X) :- retract(nesting_level(Y)),
  New is Y+X, assert(nesting_level(New)).

space(32).
indent_log(WS) :- nesting_level(Lvl), length(WS,Lvl),
    maplist(space,WS).

check_and_generate_group_stats(Group,Stats) :- open_xml_group(A,_),!,
   (A=Group
    -> retract(open_xml_group(Group,WTimeStart)),
       (Stats=no_statistics -> true
        ; statistics(walltime,[WTimeEnd,_]),
          Delta is WTimeEnd - WTimeStart,
          statistics(memory_used,M),
          write_xml_element_to_log(statistics,[walltime/Delta,walltime_since_start/WTimeEnd,memory_used/M])
        ),
       update_nesting_level(-1)
    ; add_internal_error('XML closing tag mismatch: ', Group/A),
      stop_xml_group_in_log(A,Stats), % close offending group and try again
      check_and_generate_group_stats(Group,Stats)
    ).
check_and_generate_group_stats(Group,_) :-
    add_internal_error('XML closing tag error, no tag open: ', Group).

start_xml_group_in_log(Group) :- logging_mode(xml),!,
   statistics(walltime,[WTime,_]),
   indent_log(WS),
   asserta(open_xml_group(Group,WTime)),
   update_nesting_level(1),
   format_log("~s<~w>~n",[WS,Group]).
start_xml_group_in_log(_).

% we currently only support a single attribute and value
start_xml_group_in_log(Group,Attr,Value) :- logging_mode(xml),!,
   statistics(walltime,[WTime,_]),
   indent_log(WS),
   asserta(open_xml_group(Group,WTime)),
   update_nesting_level(1),
   convert_to_codes(Value,ValueC),
   xml_encode_text(ValueC,EValueC),
   format_log("~s<~w ~w=\"~s\">~n",[WS,Group,Attr,EValueC]).
start_xml_group_in_log(_,_,_).

stop_xml_group_in_log(Group,Stats) :- logging_mode(xml),!,
    check_and_generate_group_stats(Group,Stats),
    indent_log(WS),
    format_log("~s</~w>~n",[WS,Group]).
stop_xml_group_in_log(_,_).


stop_xml_group_in_log(Group) :- stop_xml_group_in_log(Group,statistics).
stop_xml_group_in_log_no_statistics(Group) :- stop_xml_group_in_log(Group,no_statistics).

% call if you need to prematurely exit probcli
close_all_xml_groups_in_log :- open_xml_group(Group,_), !,
   stop_xml_group_in_log(Group),
   close_all_xml_groups_in_log.
close_all_xml_groups_in_log.



:- use_module(library(xml),[xml_parse/3]).
:- use_module(tools, [read_string_from_file/3]).
:- use_module(debug, [formatsilent/2]).

read_xml_log_file(File,[errors/NrErrors,warnings/NrWarnings,expected_errors/NrExpErrors]) :-
    Encoding=auto,  % (must be "auto", "UTF-8", "UTF-16", "ISO-8859-1",...)
    statistics(walltime,_),
    absolute_file_name(File,AFile),
    read_string_from_file(AFile,Encoding,Codes),
    (xml_parse(Codes,xml(_Atts,Content),[format(true)]) -> true
       ; add_error(read_xml,'Converting file contents to XML failed: ',AFile),fail),
    statistics(walltime,[_,W2]),
    formatsilent('% Walltime ~w ms to parse and convert XML in ~w~n',[W2,AFile]),
    Content = [element('probcli-run',[],InnerContent)|_],
    check_log(InnerContent,0),
    member(element('probcli-errors',[errors=EC,warnings=WC|TErrs],[]),InnerContent),
    number_codes(NrErrors,EC),
    number_codes(NrWarnings,WC),
    (member(expected_errors=ExpE,TErrs),number_codes(NrExpErrors,ExpE) -> true ; NrExpErrors=0).

%extract_xml_log_file([element('probcli-run',[],Cont],Entries) :-
%extract_entry(element(Name,Attrs,Content)

extract_attribute(Attr=Codes,Attr=Atom) :- atom_codes(Atom,Codes).

% TO DO: extract interesting information, e.g., for test_runner
check_log([],_) :- !.
check_log([H|T],Level) :- !, L1 is Level+1, check_log(H,L1),
   check_log(T,Level).
check_log(element(Name,Attrs,Cont),Level) :- maplist(extract_attribute,Attrs,EAttrs),
   !,
   indentws(Level), format('<~w  ~w>~n',[Name,EAttrs]),
   L1 is Level + 1,
   check_log(Cont,L1).
check_log(pcdata(L),Level) :- !, indentws(Level), format('~s~n',[L]).
check_log(X,Level) :- indentws(Level),print(X),nl.

indentws(0) :- !.
indentws(X) :- X>0, print('  '), X1 is X-1, indentws(X1).

 /*
 <?xml version="1.0" encoding="ASCII"?>

  | ?- xml_parse("<PT ID=\"2\" stID=\"3\"/>",R).
R = xml([],[element('PT',['ID'=[50],stID=[51]],[])]) ?

<PointsTelegram elementID="W90" stationID="FR" interlockingID="FR" interlockingElementID="W90"/>


*/

