% (c) 2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(state_space_explorer,[explore_state_space/5, compute_transitions_opt/3,
                                get_b_optimisation_options/2]).

/* ProB Libraries */
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(tools),[print_message/1,print_mb/1]).
:- use_module(probsrc(state_space),[clear_context_state/0,visited_expression/2,state_error/3,set_context_state/1]).
:- use_module(probsrc(state_space_exploration_modes),[get_open_node_to_check/2,depth_breadth_first_mode/1]).
:- use_module(probsrc(specfile),[animation_mode/1]).
:- use_module(probsrc(preferences),[get_preference/2]).

%:- use_module(probporsrc(static_analysis),[compute_dependendency_relation_of_all_events_in_the_model/3]).
%:- use_module(probpgesrc(pge_algo), [compute_transitions_pge/2, compute_operation_relations_for_pge/1]).
%:- use_module(probporsrc(ample_sets),[check_cycle_proviso/0]).
:- use_module(probsrc(model_checker),[compute_transitions_opt/3, perform_static_analyses/3]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides predicates for state space exploration without performing any checks.').

explore_state_space(Nr,LimitNr,NodesAnalysed,LimitTime,Res) :-
  animation_mode(MODE),
  get_b_optimisation_options(MODE,Optimisations),
  % 0 for not including the invariant into the respective analyses
  perform_static_analyses(MODE,0,Optimisations),
  depth_breadth_first_mode(DFMODE),
  statistics(walltime,[ReferenceTime,_]),
  model_checker: get_model_check_stats(States,_Transitions,_ProcessedTotal,Percentage),
  RefStatus=refstatus(ReferenceTime,Percentage,States,ReferenceTime),
  explore_one_state_stats(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Optimisations,DFMODE,Res),
  clear_context_state.

explore_one_state_stats(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Optimisations,DFMODE,Res) :-
  (Nr < LimitNr ; LimitNr=:= -1),!, % -1 for infinite
  statistics(walltime,[NewTime,_]),
  print_stats_and_update_refstatus(Nr,RefStatus,NewTime,RefStatus1),
  ((number(LimitTime), NewTime>LimitTime, Nr>0) ->
    Res = [timeout,Nr],
    NodesAnalysed=Nr
  ; otherwise ->
    explore_one_state(Nr,LimitNr,RefStatus1,NodesAnalysed,LimitTime,Optimisations,DFMODE,Res)
  ).
explore_one_state_stats(N,_LimitNr,_RefStatus,N,_,_,_,no).

explore_one_state(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Optimisations,DFMODE,Res) :-
  get_open_node_to_check(DFMODE,ID),!,
  set_context_state(ID),
  visited_expression(ID,CurState),!,
  compute_transitions_opt(Optimisations,ID,CurState),
  %user:do_one_gui_event,
  (state_exploration_found_error(ID,Error)
   -> model_checker: print_found_error(Error,ID),
      clear_context_state,
      NodesAnalysed=Nr
    ; N1 is Nr + 1,
      explore_one_state_stats(N1,LimitNr,RefStatus,NodesAnalysed,LimitTime,Optimisations,DFMODE,Res)
  ).
explore_one_state(N,_LimitNr,_RefStatus,N,_,_,_,all) :-
  % in case get_open_node_to_check/2 at the previous clause fails we do not have any other states to explore
  print_message('All open nodes visited').

:- use_module(probsrc(debug),[formatsilent/2]).
print_stats_and_update_refstatus(Nr,RefStatus,NewTime,RefStatus1) :-
  ( (Nr mod 1000 =:= 0, Nr\=0)
    ->  RefStatus = refstatus(RefTime,_RefPerc,PrevStates,StartTime),
        Delta is NewTime-RefTime,
        (Delta>20000  % more than 20 seconds since last message
         -> model_checker: get_model_check_stats(States,Transitions,ProcessedTotal,Perc),
            formatsilent('~n* ~w states explored (~1f% of total ~w), ~w transitions,',[ProcessedTotal,Perc,States,Transitions]),
            statistics(memory_used,M),
            print_mb(M),
            DeltaStates is (States-PrevStates),
            NodesPerSec is DeltaStates*1000 / (NewTime-RefTime),
            formatsilent(', current ~2F states/sec~n',[NodesPerSec]),
            RefStatus1 = refstatus(NewTime,Perc,States,StartTime)
        ; RefStatus1 = RefStatus)
     ;   RefStatus1 = RefStatus).

state_exploration_found_error(ID,Error) :-
   state_error(ID,_,Err),
   Error = state_error(Err).
state_exploration_found_error(_ID,Error) :-
   real_error_occurred,
   Error = general_error_occurred.

get_b_optimisation_options(MODE,Optimisations) :-
  (MODE == b),!,
  Optimisations = opt(PORPrefs,PGEPref),
  PORPrefs = por(POR,UseEnableGraph,Depth,PGEPref),
  get_preference(por,POR),
  get_preference(enable_graph,UseEnableGraph),
  get_preference(enable_graph_depth,Depth),
  get_preference(pge,PGEPref).
get_b_optimisation_options(_,Optimisations) :-
  Optimisations = opt(off,off).
