% (c) 2014-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_simplifier, [simplify_b_predicate/2, tcltk_simplify_b_predicate/2, tcltk_simplify_b_predicate_error/2]).

/* Modules and Infos for the code coverage analysis */
:- use_module(probsrc(module_information)).
:- module_info(group,model_checker).
:- module_info(description,'This module provides predicates for simplifying B predicates').

% Classical B prolog modules
:- use_module(probsrc(bsyntaxtree), [get_texpr_expr/2, get_texpr_type/2, create_texpr/4,
                                     same_texpr/2, different_texpr_values/2,
                                     is_truth/1, is_falsity/1, is_negation_of/2,
                                     is_a_conjunct/3, is_a_disjunct/3, is_an_implication/3,
                                     conjunct_predicates/2, disjunct_predicates/2,
                                     %% extract_info/3,
                                     create_negation/2,
                                     create_implication/3, create_equivalence/3,
                                     create_forall/3, create_exists/3]).
:- use_module(probsrc(bmachine),[b_parse_machine_predicate/2]).
:- use_module(probsrc(kernel_objects),[element_of_global_set/2]).

%% :- use_module(probsrc(parsercall), [parse_predicate/2]).

:- use_module(probsrc(preferences),[temporary_set_preference/3, reset_temporary_preference/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
%% :- use_module(probsrc(debug), [debug_print/2, debug_println/2, debug_mode/1]).
:- use_module(probsrc(error_manager),[add_internal_error/2]).

% Importing unit tests predicates
:- use_module(probsrc(self_check)).

% SICSTUS libraries
:- use_module(library(avl)).

%% atom_codes(String,Codes),
%% debug_println(20,parsing_as_predicate(Codes)),
%% eval_strings: repl_typing_scope(TypingScope),
%% b_parse_machine_predicate_from_codes_open(exists,Codes, % will also mark outer variables so that they are not removed
%%                                                [],TypingScope,Typed),
%% debug_println(20,eval_predicate(Typed)),

tcltk_simplify_b_predicate_error(String,StringResult) :-
  if(tcltk_simplify_b_predicate(String,StringResult),true,
     (StringResult = 'Error or User-Interrupt')).

% exported for the Tcl/Tk interface
tcltk_simplify_b_predicate(String,StringResult) :-
  % parse also boolean expressions as predicates
  temporary_set_preference(bool_expression_as_predicate,true,Chg),
  %% atom_codes(String,Codes),
  %% parse_predicate(Codes,Tree),
  call_cleanup((b_parse_machine_predicate(String,Tree),
               simplify_b_predicate(Tree,Res),
               translate_bexpression(Res,StringResult)),
               reset_temporary_preference(bool_expression_as_predicate,Chg)).
  %extract_type_information(Typed,TypeInfo).

simplify_b_predicate(Pred,Res) :-
  get_texpr_expr(Pred,Expr),
  get_texpr_type(Pred,Type),
  simplify_b_predicate2(Expr,Type,Res).

simplify_b_predicate2(truth,Type,Res) :- !,
  create_texpr(truth,Type,[],Res).
simplify_b_predicate2(falsity,Type,Res) :- !,
  create_texpr(falsity,Type,[],Res).
simplify_b_predicate2(negation(BExpr),_Type,Res) :- !,
  simplify_negation(BExpr,Res).
simplify_b_predicate2(conjunct(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  ( is_false_constant(SimplifiedLHS) -> % FALSE & f == FALSE
    create_texpr(falsity,Type,[],Res)
  ; is_true_constant(SimplifiedLHS) -> % TRUE & f == f
    simplify_b_predicate(RHS,Res)
  ; otherwise ->
    simplify_b_predicate(RHS,SimplifiedRHS),
    ( is_false_constant(SimplifiedRHS) -> % f & FALSE == FALSE
      create_texpr(falsity,Type,[],Res)
    ; is_true_constant(SimplifiedRHS) -> % f & TRUE == f
      Res = SimplifiedLHS
    ; same_texpr(SimplifiedLHS,SimplifiedRHS) -> % f & f == f (idempotence)
      Res = SimplifiedLHS
    ; is_absorption_rule_conjunct(SimplifiedLHS,SimplifiedRHS,Res) -> % f & (f or g) == f (absorption)
      true
    ; test_transitivity(SimplifiedLHS,SimplifiedRHS,Res) -> % (f => g) & (g => h) == (f => h)
      true
    ; otherwise ->
      conjunct_predicates([SimplifiedLHS,SimplifiedRHS],Res)
    )
  ).
simplify_b_predicate2(disjunct(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  ( is_true_constant(SimplifiedLHS) -> % TRUE or f == TRUE
    create_texpr(truth,Type,[],Res)
  ; is_false_constant(SimplifiedLHS) -> % FALSE or f == f
    simplify_b_predicate(RHS,Res)
  ; otherwise ->
    simplify_b_predicate(RHS,SimplifiedRHS),
    ( is_true_constant(SimplifiedRHS) -> % f or TRUE == TRUE
      create_texpr(truth,Type,[],Res)
    ; is_false_constant(SimplifiedRHS) -> % f or FALSE == f
      Res = SimplifiedLHS
    ; same_texpr(SimplifiedLHS,SimplifiedRHS) -> % f or f == f (idempotence)
      Res = SimplifiedLHS
    ; is_absorption_rule_disjunct(LHS,RHS,Res) -> % f or (f & g) == f (absorption)
      true
    ; otherwise ->
      disjunct_predicates([SimplifiedLHS,SimplifiedRHS],Res)
    )
  ).
simplify_b_predicate2(implication(LHS,RHS),_Type,Res) :- !,
  simplify_b_predicate(RHS,SimplifiedRHS),
  simplify_implication(LHS,SimplifiedRHS,Res).
simplify_b_predicate2(equivalence(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  simplify_b_predicate(RHS,SimplifiedRHS),
  ( same_texpr(SimplifiedLHS,SimplifiedRHS) -> % f <=> f == TRUE
    create_texpr(truth,Type,[],Res)
  ; is_negation_of(SimplifiedLHS,SimplifiedRHS) -> % f <=> not(f) == FALSE
    create_texpr(falsity,Type,[],Res)
  ; otherwise ->
    create_equivalence(SimplifiedLHS,SimplifiedRHS,Res)
  ).
simplify_b_predicate2(equal(LHS,RHS),Type,Res) :- !,
  print(equal(LHS,RHS)),nl,
  simplify_b_predicate(LHS,SimplifiedLHS),
  simplify_b_predicate(RHS,SimplifiedRHS),
  (same_texpr(SimplifiedLHS,SimplifiedRHS) ->
    create_texpr(truth,Type,[],Res)
  ; different_texpr_values(SimplifiedLHS,SimplifiedRHS) ->
    create_texpr(falsity,Type,[],Res)
  ; otherwise ->
    compare('==',SimplifiedLHS,SimplifiedRHS,equal,Type,Res)
    %% extract_info(SimplifiedLHS,SimplifiedRHS,Info),
    %% create_texpr(equal(SimplifiedLHS,SimplifiedRHS),Type,Info,Res)
  ).
simplify_b_predicate2(not_equal(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  simplify_b_predicate(RHS,SimplifiedRHS),
  (same_texpr(SimplifiedLHS,SimplifiedRHS) ->
    create_texpr(falsity,Type,[],Res)
  ; different_texpr_values(SimplifiedLHS,SimplifiedRHS) ->
    create_texpr(truth,Type,[],Res)
  ; otherwise ->
    compare('=\\=',SimplifiedLHS,SimplifiedRHS,not_equal,Type,Res)
    %% extract_info(SimplifiedLHS,SimplifiedRHS,Info),
    %% create_texpr(not_equal(SimplifiedLHS,SimplifiedRHS),Type,Info,Res)
  ).
simplify_b_predicate2(greater(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  simplify_b_predicate(RHS,SimplifiedRHS),
  compare('>',SimplifiedLHS,SimplifiedRHS,greater,Type,Res).
simplify_b_predicate2(greater_equal(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  simplify_b_predicate(RHS,SimplifiedRHS),
  compare('>=',SimplifiedLHS,SimplifiedRHS,greater_equal,Type,Res).
simplify_b_predicate2(less(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  simplify_b_predicate(RHS,SimplifiedRHS),
  compare('<',SimplifiedLHS,SimplifiedRHS,less,Type,Res).
simplify_b_predicate2(less_equal(LHS,RHS),Type,Res) :- !,
  simplify_b_predicate(LHS,SimplifiedLHS),
  simplify_b_predicate(RHS,SimplifiedRHS),
  compare('=<',SimplifiedLHS,SimplifiedRHS,less_equal,Type,Res).
/* Add more sophisticated implementations for simplifying predicates with quatifiers:
  E.g.: !(x).(x:S => x<21) == max(S) < 21
        !(x).(x:S => x/=3) == 3/:S */
% Quantified predcates
% !(x).(P => Q)
simplify_b_predicate2(forall(Ids,LHS,RHS),Type,Res) :- !,
%%  trace,
  simplify_b_predicate(RHS,SimplifiedRHS),
  (is_true_constant(SimplifiedRHS) ->
    create_texpr(truth,Type,[],Res)
  ; otherwise ->
    simplify_b_predicate(LHS,SimplifiedLHS),
    create_implication(SimplifiedLHS,SimplifiedRHS,ImplicationRes),
    create_forall(Ids,ImplicationRes,Res)
  ).
% #(x).(P)
simplify_b_predicate2(exists(Ids,Pred),Type,Res) :- !,
  simplify_b_predicate(Pred,SimplifiedPred),
  (is_true_constant(SimplifiedPred) ->
    create_texpr(truth,Type,[],Res)
  ; is_false_constant(SimplifiedPred) ->
    create_texpr(falsity,Type,[],Res)
  ; otherwise ->
    create_exists(Ids,SimplifiedPred,Res)
  ).
% Simplifying membership relations
simplify_b_predicate2(member(Element,Set),Type,Res) :- !,
  simplify_membership_predicate(Element,Set,Type,Res).
simplify_b_predicate2(not_member(Element,Set),Type,Res) :- !,
  simplify_non_membership_predicate(Element,Set,Type,Res).
% TODO: do we need to safe additional information about the predicates
simplify_b_predicate2(BExpr,Type,Res) :-
  create_texpr(BExpr,Type,[],Res).
  %% (debug_mode(on) ->
  %%   print('Construct not covered yet (b_simplifier):'),nl,
  %%   print(BExpr),nl,
  %%   translate:print_bexpr(Res),nl
  %% ; true).

simplify_negation(negation(BExpr),Res) :- !, % double negation "not not f == f"
  simplify_b_predicate(BExpr,Res).
simplify_negation(BExpr,Res) :-
  simplify_b_predicate(BExpr,SimplifiedBExpr),
  % Simplifications like "not(FALSE) == TRUE" and "not(TRUE) == FALSE" will be applied by create_negation/2
  create_negation(SimplifiedBExpr,Res).

simplify_implication(_LHS,SimplifiedRHS,Res) :-
  is_true_constant(SimplifiedRHS),!, % f => TRUE == TRUE
  create_texpr(truth,pred,[],Res).
simplify_implication(LHS,SimplifiedRHS,Res) :-
  is_false_constant(SimplifiedRHS),!, % f => FALSE == neg(f)
  simplify_b_predicate(LHS,SimplifiedLHS),
  create_negation(SimplifiedLHS,Res).
simplify_implication(LHS,SimplifiedRHS,Res) :-
  simplify_b_predicate(LHS,SimplifiedLHS),
  (is_false_constant(SimplifiedLHS) -> % FALSE => f == TRUE
    create_texpr(truth,pred,[],Res)
  ; is_true_constant(SimplifiedLHS) -> % TRUE => f == f
    Res = SimplifiedRHS
  ; same_texpr(SimplifiedLHS,SimplifiedRHS) -> % f => f == TRUE
    create_texpr(truth,pred,[],Res)
  ; otherwise ->
    create_implication(SimplifiedLHS,SimplifiedRHS,Res)
  ).


%%%%%%%%% This is too trivial, we should consider the absorption and transition rules w.r.t. the origin predicate %%%%%%%%%
% P or (P & Q) == P
is_absorption_rule_disjunct(LHS,RHS,Res) :-
  is_a_conjunct(RHS,P,Q),
  (same_texpr(LHS,P); same_texpr(LHS,Q)),!,
  Res = LHS.
% (P & Q) or P == P
is_absorption_rule_disjunct(LHS,RHS,Res) :-
  is_a_conjunct(LHS,P,Q),
  (same_texpr(RHS,P); same_texpr(RHS,Q)),!,
  Res = RHS.

% P & (P or Q) == P
is_absorption_rule_conjunct(LHS,RHS,Res) :-
  is_a_disjunct(RHS,P,Q),
  (same_texpr(LHS,P); same_texpr(LHS,Q)),!,
  Res = LHS.
% (P or Q) & P == P
is_absorption_rule_conjunct(LHS,RHS,Res) :-
  is_a_disjunct(LHS,P,Q),
  (same_texpr(RHS,P); same_texpr(RHS,Q)),!,
  Res = RHS.

% (P => Q) & (Q => R) == (P => R)
test_transitivity(LHS,RHS,Res) :-
  is_an_implication(LHS,P,Q),
  is_an_implication(RHS,Q1,R),
  same_texpr(Q,Q1),!,
  create_implication(P,R,Res).
% (Q => R) & (P => Q) == (P => R)
test_transitivity(LHS,RHS,Res) :-
  is_an_implication(LHS,Q,R),
  is_an_implication(RHS,P,Q1),
  same_texpr(Q,Q1),!,
  create_implication(P,R,Res).


simplify_non_membership_predicate(Element,Set,Type,Res) :-
  b_integer_value(Element,IsIntValue,EX,_NewElement),
  IsIntValue,
  test_integer_membership(Set,EX,Type,Res1),!,
  (is_true_constant(Res1) ->
    create_texpr(falsity,Type,[],Res)
  ; is_false_constant(Res1) ->
    create_texpr(truth,Type,[],Res)
  ; otherwise ->
    add_internal_error('Internal error: Unexpected result from test_integer_membership/4 predicate in simplify_membership_predicate/4: ',Res1)
  ).
simplify_non_membership_predicate(Element,Set,Type,Res) :-
  b_integer_value(Element,_IsIntValue,_EX,NewElement),
  create_texpr(not_member(NewElement,Set),Type,[],Res).

simplify_membership_predicate(Element,Set,Type,Res) :-
  b_integer_value(Element,IsIntValue,EX,_NewElement),
  IsIntValue,
  test_integer_membership(Set,EX,Type,Res),!.
simplify_membership_predicate(Element,Set,Type,Res) :-
  b_integer_value(Element,_IsIntValue,_EX,NewElement),
  create_texpr(member(NewElement,Set),Type,[],Res).


test_integer_membership(b(empty_set,set(integer),_),_X,Type,Res) :- !,
  create_texpr(falsity,Type,[],Res).
test_integer_membership(b(value(avl_set(AVLSet)),set(integer),_),X,Type,Res) :- !,
  (avl_fetch(int(X),AVLSet) ->
    create_texpr(truth,Type,[],Res)
  ; otherwise ->
    create_texpr(falsity,Type,[],Res)
  ).
test_integer_membership(Set,X,Type,Res) :-
  is_interval_set(Set,Low,Up),!,
  ((Low=<X,X=<Up) ->
    create_texpr(truth,Type,[],Res)
  ; otherwise ->
    create_texpr(falsity,Type,[],Res)
  ).
test_integer_membership(Set,X,Type,Res) :-
  is_global_integer_set(Set,GlobalSet),!,
  (element_of_global_set(int(X),GlobalSet) ->
    create_texpr(truth,Type,[],Res)
  ; otherwise ->
    create_texpr(falsity,Type,[],Res)
  ).

is_interval_set(b(interval(b(Low,integer,_),b(Up,integer,_)),set(integer),_),Low,Up).

is_global_integer_set(Set,GlobalSet) :-
  get_texpr_expr(Set,SetExpr),
  SetExpr = integer_set(GlobalSet),
  memberchk(GlobalSet,['NATURAL','NAT','INTEGER','INT']).

is_true_constant(Expr) :-
  (is_truth(Expr); Expr=boolean_true).

is_false_constant(Expr) :-
  (is_falsity(Expr); Expr=boolean_false).

compare(Operator,LHS,RHS,Functor,Type,Res) :-
  b_integer_value(LHS,IsIntValueX,EX,NewLHS),
  b_integer_value(RHS,IsIntValueY,EY,NewRHS),
  ((IsIntValueX,IsIntValueY) ->
    (memberchk(Operator,['<','>','=<','>=','==','=\\=']) ->
      Call =.. [Operator,EX,EY],
      (Call -> create_texpr(truth,Type,[],Res); create_texpr(falsity,Type,[],Res))
    ; otherwise ->
      add_internal_error('Internal error: Unexpected arithmetic comparison Operator: ', Operator),
      BExpr =.. [Functor,NewLHS,NewRHS],
      create_texpr(BExpr,Type,[],Res)
    )
  ; otherwise ->
    BExpr =.. [Functor,NewLHS,NewRHS],
    create_texpr(BExpr,Type,[],Res)
  ).

b_integer_value(BExpr,IsIntValue,Res,NewBExpr) :-
  b_arithmetic_expression(BExpr,X),!,
  Res is X,IsIntValue=true,
  create_texpr(integer(Res),integer,[],NewBExpr).
b_integer_value(BExpr,IsIntValue,Res,NewBExpr) :-
  Res = BExpr,IsIntValue=false,
  NewBExpr=BExpr.

b_arithmetic_expression(BExpr,Res) :-
  get_texpr_expr(BExpr,Expr),
  get_texpr_type(BExpr,integer), % only an integer expression
  b_arithmetic_expression1(Expr,Res).

b_arithmetic_expression1(integer(X),X) :- !,
  number(X).
b_arithmetic_expression1(unary_minus(E),Res) :- !,
  b_arithmetic_expression(E,ERes),
  Res = '-'(ERes).
b_arithmetic_expression1(add(E1,E2),Res) :- !,
  b_arithmetic_expression(E1,E1Res),
  b_arithmetic_expression(E2,E2Res),
  Res = '+'(E1Res,E2Res).
b_arithmetic_expression1(minus(E1,E2),Res) :- !,
  b_arithmetic_expression(E1,E1Res),
  b_arithmetic_expression(E2,E2Res),
  Res = '-'(E1Res,E2Res).
b_arithmetic_expression1(multiplication(E1,E2),Res) :- !,
  b_arithmetic_expression(E1,E1Res),
  b_arithmetic_expression(E2,E2Res),
  Res = '*'(E1Res,E2Res).
% it is not the same div as in Prolog X/Y will be rounded down
%% b_arithmetic_expression1(div(E1,E2),Res) :- !,
%%   b_arithmetic_expression(E1,E1Res),
%%   b_arithmetic_expression(E2,E2Res),
%%   Res = '//'(E1Res,E2Res).
