% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(predicate_analysis,[predicate_analysis/5
                             ,predicate_analysis_with_global_sets/5
                             ,info_conjunct/3
                             ,info_disjunct/3
                             ,test_predicate_analysis/0
                             ,max_type_cardinality/2
                             ]).

:- use_module(library(lists)).
:- use_module(library(avl)).
:- use_module(library(ordsets)).
:- use_module(library(terms)).

:- use_module(bsyntaxtree).
:- use_module(btypechecker).
:- use_module(preferences).
:- use_module(tools).
:- use_module(translate).
:- use_module(bmachine).
:- use_module(b_global_sets, [is_b_global_constant/3,b_global_set_cardinality/2]).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(debug,[debug_println/2]).

:- use_module(inf_arith).
:- use_module(interval_calc).

% for interpreting value(...) components
:- use_module(store, [empty_state/1]).
:- use_module(b_interpreter, [b_compute_expression_nowf/4]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,analysis).
:- module_info(description,'This is a module and plugin to extract information about expressions (e.g. integer bounds (aka intervals), cardinality) by static analysis. Used by Kodkod.').

:- dynamic debugging_enabled/0.
:- load_files(library(system), [when(compile_time), imports([environ/2])]).
% avl_apply
% similar to avl_fetch_pair: but checks whether we have a function
% and whether the function is defined for the key
:- if(environ(debug_kodkod,true)).
debugging_enabled. % commment in to have dot files written to debug-constraint-anim folder (interval analysis for kodkod)
:- endif.

% the following is called by kodkod:
predicate_analysis_with_global_sets(Tree,ROIdentifiers,Identifiers,NewIdentifiers,NewTree) :-
    findall(X,
            (bmachine_is_precompiled,
             b_get_machine_set(Set),
                b_global_set_cardinality(Set,Card),
                create_texpr(identifier(Set),set(global(Set)),[analysis([card:irange(Card,Card)])],X)),
            Sets),
    findall(X,
            (bmachine_is_precompiled,
             is_b_global_constant(Set,_Index,Elem),
                create_texpr(identifier(Elem),global(Set),[],X)),
            Elements),
    append([ROIdentifiers,Sets,Elements],ROIds2),
    predicate_analysis(Tree,ROIds2,Identifiers,NewIdentifiers,NewTree).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% static analysis of integers

:- volatile nodecounter/1, debug_node/2, debug_constraints/1.
:- dynamic nodecounter/1, debug_node/2, debug_constraints/1.

predicate_analysis(Tree,ROIdentifiers,Identifiers,NewIdentifiers,NewTree) :-
    reset_nodecounter,
    debug_println(15,starting_predicate_analysis),
    start_ms_timer(Timer1),
    create_node_environment(ROIdentifiers,Identifiers,NewIdentifiers,NodeEnv,Infos1),
    (debug:debug_mode(on) -> stop_ms_timer_with_msg(Timer1,create_node_environment) ; true),
    start_ms_timer(Timer2),
    extract_nodes(Tree,NodeEnv,_Node,_Mode,NewTree,(Constraints,Infos),Infos1),
    (debug:debug_mode(on) -> stop_ms_timer_with_msg(Timer2,extract_nodes) ; true),
    start_ms_timer(Timer3),
    create_constraint_tree(Constraints,Nodes,CTree),
    (debug:debug_mode(on) -> stop_ms_timer_with_msg(Timer3,create_constraint_tree) ; true),
    retractall(debug_constraints(_)),assert(debug_constraints(Constraints)),
    create_initial_information(Nodes,Empty),
    apply_start_constraints(Constraints,ChangedNodes,Empty,Initial),
    compute_maximum_iterations(Nodes,MaxIter),
    debug_println(15,predicate_analysis:do_analysis(MaxIter)),
    start_ms_timer(Timer4),
    do_analysis(ChangedNodes,2,MaxIter,CTree,Initial,Values),
    (debug:debug_mode(on) -> stop_ms_timer_with_msg(Timer4,do_analysis) ; true),
    insert_info(Infos,Values).

create_node_environment(ROIdentifiers,Identifiers,NewIdentifiers,NodeEnv,Infos) :-
    initial_node_env(InitNodeEnv),
    create_identifier_nodes2(ROIdentifiers,_,InitNodeEnv,RONodeEnv1,Infos1,([],[])),
    identifier_nodes_readonly(RONodeEnv1,RONodeEnv),
    create_identifier_nodes2(Identifiers,NewIdentifiers,RONodeEnv,NodeEnv,Infos2,Infos1),
    foldl(add_id_basic_constraint(NodeEnv),Identifiers,Infos,Infos2).
add_id_basic_constraint(NodeEnv,Identifier,In,Out) :-
    get_texpr_id(Identifier,Id),
    identifier_node_lookup(Id,NodeEnv,_NodeId,Nodes,_Mode),
    add_type_constraints(Identifier,Nodes,In,Out).

% just a heuristic to avoid a non-terminating run
compute_maximum_iterations(Nodes,MaxIter) :-
    length(Nodes,N), MaxIter is N*10.

reset_nodecounter :-
    retractall(nodecounter(_)),assert(nodecounter(1)),
    retractall(debug_node(_,_)),
    retractall(debug_constraints(_)).

% create a numbered node for each identifier
initial_node_env(NodeEnv) :- empty_avl(NodeEnv).

%create_identifier_nodes(Ids,NIds,Env,IIn,IOut) :-
%    empty_avl(Start), create_identifier_nodes2(Ids,NIds,Start,Env,IIn,IOut).

create_identifier_nodes2([],[],Env,Env,Info,Info).
create_identifier_nodes2([TId|Rest],[NTId|NRest],In,Out,IIn,IOut) :-
    create_identifier_nodes3(TId,NTId,In,Inter,IIn,IInter),
    create_identifier_nodes2(Rest,NRest,Inter,Out,IInter,IOut).
create_identifier_nodes3(TId,NTId,In,Out,IIn,IOut) :-
    get_texpr_id(TId,Id),
    get_texpr_type(TId,Type),
    new_node_scope(Type,NodeId,Nodes),
    avl_store(Id,In,nodes(NodeId,Nodes,rw),Out),
    texpr_add_node_infos(TId,NodeId,NTId,IIn,IOut).

% look up the node of an identifier
identifier_node_lookup(Id,Env,NodeId,Nodes,Mode) :-
    (avl_fetch(Id,Env,nodes(NodeId,Nodes,Mode)) -> true
     ; add_warning(kodkod,'Could not lookup identifier: ',Id),fail).

% set the complete environment to read-only,
% read-only nodes are not used for output of constraint nodes
identifier_nodes_readonly(OldEnv,NewEnv) :-
    avl_map(set_readonly,OldEnv,NewEnv).
set_readonly(nodes(Id,Nodes,_Mode),nodes(Id,Nodes,ro)).

compute_integer_set('INTEGER',-inf,inf).
compute_integer_set('NATURAL',0,inf).
compute_integer_set('NATURAL1',1,inf).
compute_integer_set('INT',Min,Max) :-
    get_preference(minint,Min),
    get_preference(maxint,Max).
compute_integer_set('NAT',0,Max) :-
    get_preference(maxint,Max).
compute_integer_set('NAT1',1,Max) :-
    get_preference(maxint,Max).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% computation on information

info_conjunct(irange(X,Y),irange(A,B),Result) :-
    Max infis max(X,A), Min infis min(Y,B),
    normalise_irange(irange(Max,Min),Result).
info_disjunct(irange(X,Y),irange(A,B),Result) :-
    ( is_bottom(irange(X,Y)) -> Result = irange(A,B)
    ; is_bottom(irange(A,B)) -> Result = irange(X,Y)
    ; otherwise ->
        Min infis min(X,A), Max infis max(Y,B),
        normalise_irange(irange(Min,Max),Result)).

normalise_irange(In,Result) :-
    ( is_bottom(In) -> Result=irange(0,-1)
    ; otherwise     -> Result=In).
is_bottom(irange(A,B)) :- infless(B,A).


has_more_information(irange(X,_Y),irange(A,_B)) :- infless(A,X),!.
has_more_information(irange(_X,Y),irange(_A,B)) :- infless(Y,B),!.
has_more_information(set(A),elem(B)) :- has_more_information(A,B).
has_more_information(left(A),left(B)) :- has_more_information(A,B).
has_more_information(right(A),right(B)) :- has_more_information(A,B).

% what kind of information may be stored in a node of a certain data type
get_type_scopes(Type,Scopes) :-
    findall(S,get_type_scope(Type,S),Scopes1),
    sort([possible_values|Scopes1],Scopes).
get_type_scope(integer,interval).
get_type_scope(set(T),elem(S)) :-
    get_type_scope(T,S).
get_type_scope(set(_),card).
get_type_scope(seq(S),T) :-
    get_type_scope(set(couple(integer,S)),T).
get_type_scope(couple(A,_),left(S)) :-
    get_type_scope(A,S).
get_type_scope(couple(_,B),right(S)) :-
    get_type_scope(B,S).

% get the top element of a scope
get_scope_top(interval,irange(-inf,inf)).
get_scope_top(possible_values,irange(0,inf)).
get_scope_top(elem(S),T) :- get_scope_top(S,T).
get_scope_top(left(S),T) :- get_scope_top(S,T).
get_scope_top(right(S),T) :- get_scope_top(S,T).
get_scope_top(card,irange(0,inf)).

get_scope_bottom(interval,irange(0,-1)).
get_scope_bottom(possible_values,T) :- get_scope_bottom(interval,T).
get_scope_bottom(elem(S),T) :- get_scope_bottom(S,T).
get_scope_bottom(left(S),T) :- get_scope_bottom(S,T).
get_scope_bottom(right(S),T) :- get_scope_bottom(S,T).
get_scope_bottom(card,T) :- get_scope_bottom(interval,T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% operations on possible infinite numbers
%irange_union(irange(Am,AM),irange(Bm,BM),irange(Cm,CM)) :-
%    Cm infis min(Am,Bm), CM infis max(AM,BM).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% new extract

new_node_scope(Type,Id,Nodes) :-
    new_node2(Id),get_type_scopes(Type,Scopes),
    findall(S:Id,member(S,Scopes),Nodes).

new_node2(Id):-
    retract(nodecounter(Id)),
    Id2 is Id+1, assert(nodecounter(Id2)).

extract_nodes(TExpr,NodeEnv,Nodes,Mode,NTExpr,In,Out) :-
    get_texpr_expr(TExpr,Expr),
    extract_nodes2(Expr,TExpr,NodeEnv,Nodes,Mode,NTExpr,In,Inter),
    add_type_constraints(TExpr,Nodes,Inter,Out),
    store_debug_info(Nodes,TExpr).
extract_nodes2(identifier(Id),TExpr,NodeEnv,Nodes,Mode,NTExpr,In,Out) :- !,
    identifier_node_lookup(Id,NodeEnv,NodeId,Nodes,Mode),!,
    texpr_add_node_infos(TExpr,NodeId,NTExpr,In,Out).
extract_nodes2(conjunct(A,B),TExpr,NodeEnv,[],rw,NTExpr,In,Out) :- !,
    get_texpr_info(TExpr,Infos),
    create_texpr(conjunct(NA,NB),pred,Infos,NTExpr),
    extract_nodes(A,NodeEnv,_NodesA,_ModeA,NA,In,Inter),
    extract_nodes(B,NodeEnv,_NodesB,_ModeB,NB,Inter,Out).
extract_nodes2(forall(Ids,Pre,Cond),TExpr,NodeEnv,[],rw,NTExpr,In,Out) :- !,
    get_texpr_info(TExpr,Info),
    create_texpr(forall(NIds,NPre,NCond),pred,Info,NTExpr),
    identifier_nodes_readonly(NodeEnv,SubEnv1),
    create_identifier_nodes2(Ids,NIds,SubEnv1,SubEnv,In,Inter1),
    extract_nodes(Pre,SubEnv,_NodesPre,_ModePre,NPre,Inter1,Inter2),
    identifier_nodes_readonly(SubEnv,ROEnv),
    extract_nodes(Cond,ROEnv,_NodesCond,_ModeCond,NCond,Inter2,Out).
extract_nodes2(exists(Ids,Cond),TExpr,NodeEnv,[],rw,NTExpr,In,Out) :- !,
    get_texpr_info(TExpr,Info),
    create_texpr(exists(NIds,NCond),pred,Info,NTExpr),
    create_identifier_nodes2(Ids,NIds,NodeEnv,SubEnv,In,Inter1),
    extract_nodes(Cond,SubEnv,_NodesCond,_ModeCond,NCond,Inter1,Out).
extract_nodes2(general_sum([TId],TPred,TExpr),TSigma,NodeEnv,Nodes,rw,NTSigma,In,Out) :- !,
    nodes_for_general_sum_or_mult(mult,TId,TPred,TExpr,TSigma,NodeEnv,Nodes,NTSigma,In,Out).
extract_nodes2(general_product([TId],TPred,TExpr),TSigma,NodeEnv,Nodes,rw,NTSigma,In,Out) :- !,
    nodes_for_general_sum_or_mult(power_rule,TId,TPred,TExpr,TSigma,NodeEnv,Nodes,NTSigma,In,Out).
extract_nodes2(comprehension_set(Ids,Cond),TExpr,NodeEnv,Nodes,rw,NTExpr,In,Out) :- !,
    get_texpr_info(TExpr,Info),
    get_texpr_type(TExpr,Type),
    identifier_nodes_readonly(NodeEnv,SubEnv1),
    create_identifier_nodes2(Ids,NIds,SubEnv1,SubEnv,In,Inter1),
    extract_nodes(Cond,SubEnv,_NodesCond,_ModeCond,NCond,Inter1,Inter2),
    new_node_scope(Type,NodeId,Nodes),
    compset_constraints(Nodes,Ids,SubEnv,Constraints),
    add_all(Constraints,Inter2,Inter3),
    add_node_infos(Info,NodeId,NInfo,Inter3,Out),
    create_texpr(comprehension_set(NIds,NCond),Type,NInfo,NTExpr).
extract_nodes2(set_extension(L),TExpr,NodeEnv,LocalNodes,rw,NTExpr,In,Out) :- !,
    get_texpr_info(TExpr,Info),
    get_texpr_type(TExpr,Type),
    new_node_scope(Type,NodeId,LocalNodes),
    extract_set_extension(L,NodeEnv,LocalNodes,NTSubs,In,Inter1),
    CardNode = card:_, memberchk(CardNode,LocalNodes),
    length(L,MaxCard),
    add_all([constraint([],[CardNode],constant(irange(1,MaxCard)))],
            Inter1,Inter2),
    add_node_infos(Info,NodeId,NewInfo,Inter2,Out),
    create_texpr(set_extension(NTSubs),Type,NewInfo,NTExpr).
extract_nodes2(value(Value),TExpr,NodeEnv,LocalNodes,RW,NTExpr,In,Out) :-
    get_texpr_info(TExpr,Infos), member(was_identifier(ID),Infos),
    \+ ground(Value),!,
    %print(treating_value_as_id(Value,ID)),nl,
    extract_nodes2(identifier(ID),TExpr,NodeEnv,LocalNodes,RW,NTExpr,In,Out).
extract_nodes2(value(Value),TExpr,_NodeEnv,LocalNodes,rw,NTExpr,In,Out) :- nonvar(Value),
    !,
    get_texpr_info(TExpr,Info),
    get_texpr_type(TExpr,Type),
    new_node_scope(Type,NodeId,LocalNodes),
    findall(C,extract_constraint_for_value(Type,TExpr,LocalNodes,C),Constraints),
    add_all(Constraints,In,Inter),
    add_node_infos(Info,NodeId,NewInfo,Inter,Out),
    create_texpr(value(Value),Type,NewInfo,NTExpr).
extract_nodes2(partition(S,Es),TExpr,NodeEnv,[],rw,NTExpr,In,Out) :- !,
    extract_nodes_l([S|Es],NodeEnv,[SNodes|EsNodeLists],_Modes,[NS|NEs],In,Inter1),
    % subset constraints
    findall(constraint(CI,CO,CC),
            (member(Scope:SId,SNodes),
                member(NodeList,EsNodeLists),member(Scope:Elem,NodeList),
                is_subset_extract(Scope:Elem,Scope:SId,CI,CO,CC)),
            SubsetConstraints),
    add_all(SubsetConstraints,Inter1,Inter2),
    % cardinality constraints
    memberchk(card:SNodeId,SNodes),
    findall(card:Elem2,(member(NodeList2,EsNodeLists),member(card:Elem2,NodeList2)),
            CardNodes),
    create_add_list_constraints(CardNodes,card:SNodeId,CardConstraints),
    add_all(CardConstraints,Inter2,Inter3),
    % union constraints
    findall(constraint(ElemNodes,[elem(Scope3):SId3],union),
            ( member(elem(Scope3):SId3,SNodes),
                findall(elem(Scope3):Elem3,
                        (member(NodeList3,EsNodeLists),member(elem(Scope3):Elem3,NodeList3)),
                        ElemNodes)),
            UnionConstraints),
    add_all(UnionConstraints,Inter3,Out),
    get_texpr_info(TExpr,Info),
    create_texpr(partition(NS,NEs),pred,Info,NTExpr).
extract_nodes2(sequence_extension(L),TExpr,NodeEnv,LocalNodes,rw,NTExpr,In,Out) :- !,
    get_texpr_info(TExpr,Info),
    get_texpr_type(TExpr,Type),
    new_node_scope(Type,NodeId,LocalNodes),
    extract_seq_extension(L,NodeEnv,LocalNodes,NTSubs,In,Inter1),
    CardNode = card:_, memberchk(CardNode,LocalNodes),
    DomNode = elem(left(interval)):_, memberchk(DomNode,LocalNodes),
    length(L,Card),
    add_all([constraint([],[CardNode],constant(irange(Card,Card))),
             constraint([],[DomNode], constant(irange(1,   Card)))],
            Inter1,Inter2),
    add_node_infos(Info,NodeId,NewInfo,Inter2,Out),
    create_texpr(sequence_extension(NTSubs),Type,NewInfo,NTExpr).
extract_nodes2(empty_set,TExpr,_NodeEnv,Nodes,rw,NTExpr,In,Out) :- !,
    get_texpr_type(TExpr,Type), new_node_scope(Type,NodeId,Nodes),
    texpr_add_node_infos(TExpr,NodeId,NTExpr,In,Inter),
    get_type_scopes(Type,Scopes),
    findall(constraint([],[elem(S):NodeId],constant(Bottom)),
            (member(elem(S),Scopes),get_scope_bottom(S,Bottom)),
            Bottoms),
    add_all([constraint([],[card:NodeId],constant(irange(0,0)))|Bottoms],Inter,Out).
extract_nodes2(Expr,TExpr,NodeEnv,Nodes,rw,NTExpr,In,Out) :-
    same_functor(Expr,Tmp),
    extract(Tmp,_,_,_,_),!,
    syntaxtransformation(Expr,Subs,[],NSubs,NExpr),
    copy_term((NSubs,NExpr),(NSubsRO,NExprRO)),
    copy_term((NSubs,NExpr),(NewSubs,NewExpr)),
    extract_nodes_l(Subs,NodeEnv,SubNodeLists,Modes,NewSubs,In,Inter),
    ignore_readonly_nodes(Modes,SubNodeLists,SubNodeListsRO),
    get_texpr_info(TExpr,Info),
    get_texpr_type(TExpr,Type),
    new_node_scope(Type,NodeId,Nodes),
    add_node_infos(Info,NodeId,NewInfos,Inter,Inter2),
    create_texpr(NewExpr,Type,NewInfos,NTExpr),
    findall(constraint(InputNodes,OutputNodes,Transition),
            (   extract(NExpr,NodeId,InputNodes,OrigOutputNodes,Transition),
                copy_term(extract(NExpr,NodeId,InputNodes,OrigOutputNodes,Transition),
                          extract(NExprRO,NodeId,_,OutputNodes,Transition)),
                choose_one(SubNodeLists,SubNodeListsRO,NSubs,NSubsRO)),
            Constraints),
    %error_manager:print_error_span(TExpr),nl,print(Constraints),nl,
    add_all(Constraints,Inter2,Out).
extract_nodes2(Expr,TExpr,NodeEnv,[],rw,NTExpr,In,Out) :-
    remove_bt(TExpr,_,NExpr,NTExpr),
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr),
    identifier_nodes_readonly(NodeEnv,SubEnv1),
    create_identifier_nodes2(Names,_NIds,SubEnv1,SubEnv,In,Inter1),!,
    extract_nodes_l(Subs,SubEnv,_SubNodeLists,_Modes,NSubs,Inter1,Out).
extract_nodes2(_,TExpr,_,_,_,_,_,_) :-
    throw(failed_analysis(TExpr)).

add_type_constraints(TExpr,Nodes,In,Out) :-
    get_texpr_type(TExpr,Type),
    findall(C, type_constraint(Type,Nodes,C), Constraints),
    add_all(Constraints,In,Out).

type_constraint(Type,Nodes,constraint([],[possible_values:N],constant(irange(1,Card)))) :-
    member(possible_values:N,Nodes),
    max_type_cardinality(Type,Card).
type_constraint(set(T),Nodes,constraint([],[card:N],constant(irange(0,Card)))) :-
    member(card:N,Nodes),
    max_type_cardinality(T,Card).
type_constraint(integer,Nodes,constraint([interval:N],[possible_values:N],interval_possibilities)) :-
    member(interval:N,Nodes),member(possible_values:N,Nodes).

max_type_cardinality(global(G),Card) :-
    % TODO: This should be switched off if we try to determine the cardinality of the deferred sets
    b_global_set_cardinality(G,Card).
max_type_cardinality(boolean,2).
max_type_cardinality(set(T),Card) :-
    max_type_cardinality(T,C1),
    % catch/3 to avoid overflow errors
    catch( Card is ceiling(2 ** C1),
           error(_,_),
           fail).
max_type_cardinality(couple(A,B),Card) :-
    max_type_cardinality(A,CardA),
    max_type_cardinality(B,CardB),
    Card is CardA * CardB.

add_existing_constraints(Info,NodeId,In,Out) :-
    memberchk(analysis(Analysis),Info),!,
    findall( constraint([],[Scope:NodeId],constant(Value)),
             member(Scope:Value,Analysis),
             Constraints),
    add_all(Constraints,In,Out).
add_existing_constraints(_Info,_NodeId,In,In).

extract_constraint_for_value(Type,Value,LocalNodes,constraint([],[NodeType:NodeId],Constant)) :-
    extract_constraint_for_value2(Type,Value,NodeType,Constant),
    memberchk(NodeType:NodeId,LocalNodes).
extract_constraint_for_value2(set(_Type),Value,card,Constant) :-
    compute_integer(card(Value),Card),
    Constant = constant(irange(Card,Card)).
extract_constraint_for_value2(set(integer),Value,elem(interval),constant(irange(Min,Max))) :-
    compute_integer(min(Value),Min),
    compute_integer(max(Value),Max).
extract_constraint_for_value2(set(couple(TA,_)),Value,elem(left(DN)),Constant) :-
    compute_expression(domain(Value),set(TA),Dom),
    create_texpr(value(Dom),set(TA),[],BDom),
    extract_constraint_for_value2(set(TA),BDom,elem(DN),Constant).
extract_constraint_for_value2(set(couple(_,TB)),Value,elem(right(RN)),Constant) :-
    compute_expression(range(Value),set(TB),Ran),
    create_texpr(value(Ran),set(TB),[],BRan),
    extract_constraint_for_value2(set(TB),BRan,elem(RN),Constant).

compute_integer(Expr,Result) :-
    compute_expression(Expr,integer,int(Result)).
compute_expression(Expr,Type,Result) :-
    empty_state(Empty),
    create_texpr(Expr,Type,[],BExpr),
    b_compute_expression_nowf(BExpr,Empty,Empty,Result).

compset_constraints(SetNodes,Ids,NodeEnv,Constraints) :-
    lookup_identifier_nodes(Ids,NodeEnv,NodeLists),
    findall( C,
             (SetNode=elem(Scope):_SetNodeId, member(SetNode,SetNodes),
                 compset_constraint(Scope,NodeLists,ElemNode),
                 member(C, [constraint([SetNode],[ElemNode],id),
                            constraint([ElemNode],[SetNode],id)])),
             Constraints).
compset_constraint(left(Scope),NodeLists,Node) :- !,
    last(Begin,_,NodeLists),
    compset_constraint(Scope,Begin,Node).
compset_constraint(right(Scope),NodeLists,Node) :- !,
    last(_,Last,NodeLists),
    compset_constraint(Scope,[Last],Node).
compset_constraint(Scope,[Nodes],Scope:NodeId) :-
    memberchk(Scope:NodeId,Nodes).

lookup_identifier_nodes([],_Env,[]).
lookup_identifier_nodes([TId|Irest],Env,[Nodes|Nrest]) :-
    get_texpr_id(TId,Id),
    identifier_node_lookup(Id,Env,_NodeId,Nodes,_Mode),
    lookup_identifier_nodes(Irest,Env,Nrest).

/*
extract_set_extension([L],NodeEnv,Nodes,[NTExpr],In,Out) :- !,
    extract_nodes(L,NodeEnv,Nodes,_Mode,NTExpr,In,Out).
extract_set_extension([Elem|Rest],NodeEnv,TmpNodes,[NTExpr|NTrest],In,Out) :-
    extract_nodes(Elem,NodeEnv,ElemNodes,_Mode,NTExpr,In,Inter1),
    extract_set_extension(Rest,NodeEnv,RestNodes,NTrest,Inter1,Inter2),
    %append([TmpNodes,ElemNodes,RestNodes],Nodes),
    get_texpr_type(Elem,Type),
    new_node_scope(Type,NodeId,TmpNodes),
    add_node_infos([],NodeId,_,Inter2,Inter3),
    findall(C,( EN = Scope:_,
                RN = Scope:_,
                ON = Scope:_,
                member(EN,ElemNodes),
                memberchk(RN,RestNodes),
                memberchk(ON,TmpNodes),
                member(C,[constraint([EN,RN],[ON],union),
                          constraint([ON],[EN],id),
                          constraint([ON],[RN],id)])),
            Constraints),
    add_all(Constraints,Inter3,Out).
*/
extract_set_extension(Elements,NodeEnv,SeqNodes,NElements,In,Out) :-
    extract_extension_nodes(Elements,NodeEnv,ElemNodes,NElements,In,Inter),
    findall(n(Scope,elem(Scope):N),member(elem(Scope):N,SeqNodes),Scopes),
    extract_ext_nodes2(Scopes,ElemNodes,Constraints),
    add_all(Constraints,Inter,Out).

extract_seq_extension(Elements,NodeEnv,SeqNodes,NElements,In,Out) :-
    extract_extension_nodes(Elements,NodeEnv,ElemNodes,NElements,In,Inter),
    findall(n(Scope,elem(right(Scope)):N),member(elem(right(Scope)):N,SeqNodes),Scopes),
    extract_ext_nodes2(Scopes,ElemNodes,Constraints),
    add_all(Constraints,Inter,Out).

extract_ext_nodes2([],_ElemNodes,[]).
extract_ext_nodes2([Scope|Srest],ElemNodes,Constraints) :-
    extract_ext_nodes3(Scope,ElemNodes,C1),
    append(C1,C2,Constraints),
    extract_ext_nodes2(Srest,ElemNodes,C2).
extract_ext_nodes3(n(Scope,SeqNode),ElemNodes,[Union|Constraints]) :-
    findall(Scope:E,
            ( member(NodeList,ElemNodes), memberchk(Scope:E,NodeList)),
            AllElemNodes),
    Union = constraint(AllElemNodes,[SeqNode],union),
    % for set extensions like { min(0..7) } this generates an empty list AllElemenNodes; TO DO: fix
    findall(constraint([SeqNode],[E],id), member(E,AllElemNodes), Constraints).

extract_extension_nodes([],_NodeEnv,[],[],X,X).
extract_extension_nodes([Elem|Rest],NodeEnv,[ElemNodes|RestNodes],[NElem|NRest],In,Out) :-
    extract_nodes(Elem,NodeEnv,ElemNodes,_Mode,NElem,In,Inter),
    extract_extension_nodes(Rest,NodeEnv,RestNodes,NRest,Inter,Out).


texpr_add_node_infos(TExpr,NodeId,NTExpr,In,Out) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    get_texpr_info(TExpr,Info),
    add_node_infos(Info,NodeId,NewInfo,In,Inter),
    create_texpr(Expr,Type,NewInfo,NTExpr),
    add_existing_constraints(Info,NodeId,Inter,Out).
add_node_infos(OldInfo,NodeId,NewInfos,
               (Constraints,[infonode(NodeId,NewInfos,OldInfo)|Out]),
               (Constraints,Out)).

add_all([],In,In).
add_all([E|Rest],([E|In],Infonodes),Out) :-
    add_all(Rest,(In,Infonodes),Out).
extract_nodes_l([],_,[],[],[],In,In).
extract_nodes_l([TExpr|ERest],NodeEnv,[Node|NRest],[Mode|MRest],[NTExpr|NTRest],In,Out) :-
    extract_nodes(TExpr,NodeEnv,Node,Mode,NTExpr,In,Inter),
    extract_nodes_l(ERest,NodeEnv,NRest,MRest,NTRest,Inter,Out).
choose_one([],[],[],[]).
choose_one([List1|L1Rest],[List2|L2Rest],[Elem1|E1Rest],[Elem2|E2Rest]) :-
    choose_one2(List1,List2,Elem1,Elem2),
    choose_one(L1Rest,L2Rest,E1Rest,E2Rest).
choose_one2(NodeList,NodeListRO,Elem,ElemRO) :-
    ( Elem == (*) -> true
    ; otherwise -> nth1(N,NodeList,Elem), nth1(N,NodeListRO,ElemRO)).
ignore_readonly_nodes([],[],[]).
ignore_readonly_nodes([Mode|MRest],[InNodes|InRest],[OutNodes|OutRest]) :-
    ignore_readonly_nodes2(Mode,InNodes,OutNodes),
    ignore_readonly_nodes(MRest,InRest,OutRest).
ignore_readonly_nodes2(ro,Nodes,Ignored) :- ignore_readonly_nodes3(Nodes,Ignored).
ignore_readonly_nodes2(rw,Nodes,Nodes).
ignore_readonly_nodes3([],[]).
ignore_readonly_nodes3([Scope:_Node|NRest],[Scope:ignored|IRest]) :-
    ignore_readonly_nodes3(NRest,IRest).

nodes_for_general_sum_or_mult(Rule,TId,TPred,TExpr,TOrig,NodeEnv,Nodes,TNew,In,Out) :-
    identifier_nodes_readonly(NodeEnv,SubEnv1),
    create_identifier_nodes2([TId],[NTId],SubEnv1,SubEnv,In,Inter1),
    extract_nodes_l([TId,TPred,TExpr],SubEnv,[IdNodes,_PredNodes,ExprNodes],_Modes,
                    [NTId,NTPred,NTExpr],Inter1,Inter2),
    get_texpr_info(TOrig,Info),
    memberchk(interval:ExprNodeId,ExprNodes),
    memberchk(possible_values:IdId,IdNodes),
    new_node_scope(integer,NodeId,Nodes),
    add_node_infos(Info,NodeId,NInfo,Inter2,Inter3),
    add_all([constraint([possible_values:IdId,interval:ExprNodeId],
                        [interval:NodeId],Rule)],
            Inter3,Out),
    get_texpr_expr(TOrig,OrigExpr),functor(OrigExpr,Functor,_),
    NExpr =.. [Functor,[NTId],NTPred,NTExpr],
    create_texpr(NExpr,integer,NInfo,TNew).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% extract constraints from syntax elements
% extract(NExpr,NodeId,InputNodes,OrigOutputNodes,Transition)

extract(Expr,R,X,Y,Op) :-
    extract_symmetric(Expr,R,X1,Y1,Op1),
    symmetric_constraint(Op1,Op2),
    ( Op=Op1,X=X1,Y=Y1
    ; Op=Op2,X=Y1,Y=X1).

extract(integer(I),R,[],[interval:R],constant(irange(I,I))).
extract(value(IV),R,[],[interval:R],constant(irange(I,I))) :- nonvar(IV), IV = int(I). % TO DO: other vals
extract(interval(interval:A,interval:B),R,[interval:A,interval:B],[card:R],interval_card).
extract(member(S:A,elem(S):B),_,[elem(S):B],[S:A],id).
extract(subset(elem(S):A,elem(S):B),_,[elem(S):B],[elem(S):A],id).
extract(subset_strict(elem(S):A,elem(S):B),_,[elem(S):B],[elem(S):A],id).
extract(add(interval:A,interval:B),R,I,O,C) :- addition_constraint(interval:A,interval:B,interval:R,I,O,C).
extract(minus(interval:A,interval:B),R,I,O,C) :- addition_constraint(interval:B,interval:R,interval:A,I,O,C).
extract(minus(interval:A,interval:B),R,[interval:R,interval:B],[interval:A],add).
extract(minus(interval:A,interval:B),R,[interval:A,interval:R],[interval:B],sub).
extract(multiplication(interval:A,interval:B),R,[interval:A,interval:B],[interval:R],mult).
extract(div(interval:A,interval:B),R,[interval:A,interval:B],[interval:R],divrange).
extract(floored_div(interval:A,interval:B),R,[interval:A,interval:B],[interval:R],divrange). % TO DO: provide proper treatment of floored_div
extract(modulo(*,interval:B),R,[interval:B],[interval:R],lt).
extract(modulo(*,*),R,[],[interval:R],constant(irange(0,inf))).
extract(power_of(interval:Base,interval:Exp),R,[interval:Base,interval:Exp],[interval:R],power).
extract(integer_set(Set),R,[],[elem(interval):R],constant(irange(Min,Max))) :-
    compute_integer_set(Set,Min,Max).
extract(bool_set,R,[],[card:R],constant(irange(2,2))).
extract(pow1_subset(*),R,[],[elem(card):R],constant(irange(1,inf))).
extract(fin_subset(A),R,AL,RL,Op) :- extract(pow_subset(A),R,AL,RL,Op).
extract(fin1_subset(A),R,AL,RL,Op) :- extract(pow1_subset(A),R,AL,RL,Op).
extract(partial_function(A,B),R,AL,RL,Op) :- extract(relations(A,B),R,AL,RL,Op).
extract(total_function(A,B),R,AL,RL,Op) :- extract(relations(A,B),R,AL,RL,Op).
extract(partial_injection(A,B),R,AL,RL,Op) :- extract(partial_function(A,B),R,AL,RL,Op).
extract(total_injection(A,B),R,AL,RL,Op) :- extract(total_function(A,B),R,AL,RL,Op).
extract(partial_surjection(A,B),R,AL,RL,Op) :- extract(partial_function(A,B),R,AL,RL,Op).
extract(total_surjection(A,B),R,AL,RL,Op) :- extract(total_function(A,B),R,AL,RL,Op).
extract(total_bijection(A,B),R,AL,RL,Op) :- extract(total_function(A,B),R,AL,RL,Op).
extract(partial_bijection(A,B),R,AL,RL,Op) :- extract(partial_function(A,B),R,AL,RL,Op).
extract(cartesian_product(card:A,card:B),R,[card:A,card:B],[card:R],mult).
extract(cartesian_product(card:A,card:B),R,[card:R,card:A],[card:B],cart_div).
extract(cartesian_product(card:A,card:B),R,[card:R,card:B],[card:A],cart_div).
extract(set_subtraction(elem(S):A,*),R,[elem(S):A],[elem(S):R],id).
extract(image(elem(right(S)):A,*),R,[elem(right(S)):A],[elem(S):R],id).
extract(intersection(S:A,*),R,I,O,C) :- is_subset_extract(S:R,S:A,I,O,C).
extract(intersection(*,S:B),R,I,O,C) :- is_subset_extract(S:R,S:B,I,O,C).
extract(union(elem(S):A,elem(S):B),R,[elem(S):A,elem(S):B],[elem(S):R],union).
extract(union(S:A,*),R,I,O,C) :- is_subset_extract(S:A,S:R,I,O,C).
extract(union(*,S:B),R,I,O,C) :- is_subset_extract(S:B,S:R,I,O,C).
extract(domain_restriction(S:A,*),R,I,O,C) :- is_subset_extract(left(S):R,S:A,I,O,C). % ALWAYS FAILS !!!!! SOMETHING IS WRONG HERE; TODO: fix
extract(domain_restriction(*,S:B),R,I,O,C) :- is_subset_extract(S:R,S:B,I,O,C).
extract(domain_subtraction(*,S:B),R,I,O,C) :- is_subset_extract(S:R,S:B,I,O,C).
extract(range_restriction(*,S:B),R,I,O,C) :- is_subset_extract(right(S):R,S:B,I,O,C). % ALWAYS FAILS !!!!!
extract(range_restriction(S:A,*),R,I,O,C) :- is_subset_extract(S:R,S:A,I,O,C).
extract(range_subtraction(S:A,*),R,I,O,C) :- is_subset_extract(S:R,S:A,I,O,C).
extract(closure(elem(S):A),R,[elem(S):A],[elem(S):R],id).
extract(composition(elem(left(S)):A,*),R,I,O,C) :- is_subset_extract(elem(left(S)):R,elem(left(S)):A,I,O,C).
extract(composition(*,elem(right(S)):B),R,I,O,C) :- is_subset_extract(elem(right(S)):R,elem(right(S)):B,I,O,C).
extract(function(elem(right(S)):A,*),R,[elem(right(S)):A],[S:R],id).
extract(concat(card:A,card:B),R,I,O,C) :- addition_constraint(card:A,card:B,card:R,I,O,C).
extract(concat(Right:A,Right:B),R,[Right:A,Right:B],[Right:R],union) :- Right=elem(right(_)).
extract(concat(S:A,*),R,I,O,C) :- S=elem(right(_)),is_subset_extract(S:A,S:R,I,O,C).
extract(concat(*,S:B),R,I,O,C) :- S=elem(right(_)),is_subset_extract(S:B,S:R,I,O,C).
extract(concat(LI:A,LI:B),R,[LI:A,LI:B],[LI:R],concat_index) :- LI=elem(left(interval)). % TODO: More rules like addition
extract(seq(*),R,[],[elem(elem(left(interval))):R],constant(irange(1,inf))).
extract(seq1(A),R,I,O,C) :- extract(seq(A),R,I,O,C).
extract(seq1(*),R,[],[elem(card):R],constant(irange(1,inf))).
extract(size(A),R,I,O,C) :- extract(card(A),R,I,O,C).

addition_constraint(A,B,R,[A,B],[R],add).
addition_constraint(A,B,R,[R,B],[A],sub).
addition_constraint(A,B,R,[R,A],[B],sub).

is_subset_extract(A,B,I,O,C) :-extract(subset(A,B),_,I,O,C).

extract_symmetric(equal(S:A,S:B),_,[S:A],[S:B],id).
extract_symmetric(subset(card:A,card:B),_,[card:A],[card:B],gte).
extract_symmetric(subset_strict(card:A,card:B),_,[card:A],[card:B],gt).
extract_symmetric(less(interval:A,interval:B),_,[interval:B],[interval:A],lt).
extract_symmetric(less_equal(interval:A,interval:B),_,[interval:B],[interval:A],lte).
extract_symmetric(greater(interval:A,interval:B),_,[interval:B],[interval:A],gt).
extract_symmetric(greater_equal(interval:A,interval:B),_,[interval:B],[interval:A],gte).
extract_symmetric(unary_minus(interval:X),R,[interval:X],[interval:R],negate).
extract_symmetric(card(card:X),R,[card:X],[interval:R],id).
extract_symmetric(cartesian_product(elem(S):A,*),R,[elem(S):A],[elem(left(S)):R],id).
extract_symmetric(cartesian_product(*,elem(S):B),R,[elem(S):B],[elem(right(S)):R],id).
extract_symmetric(couple(S:A,*),R,[S:A],[left(S):R],id).
extract_symmetric(couple(*,S:B),R,[S:B],[right(S):R],id).
extract_symmetric(pow_subset(elem(S):A),R,[elem(S):A],[elem(elem(S)):R],id).
extract_symmetric(pow_subset(card:A),R,[card:A],[elem(card):R],lte).
extract_symmetric(pow1_subset(S),R,[A],[B],C) :- extract_symmetric(pow_subset(S),R,[A],[B],C).
extract_symmetric(seq(A),R,[elem(S):A],[elem(elem(right(S))):R],id).
extract_symmetric(relations(elem(S):A,*),R,[elem(S):A],[elem(elem(left(S))):R],id).
extract_symmetric(relations(*,elem(S):B),R,[elem(S):B],[elem(elem(right(S))):R],id).
extract_symmetric(interval(interval:A,interval:_),R,[interval:A],[elem(interval):R],gte).
extract_symmetric(interval(interval:_,interval:B),R,[interval:B],[elem(interval):R],lte).
extract_symmetric(partial_function(card:A,card:_),R,[card:A],[elem(card):R],lte).
extract_symmetric(total_function(card:A,card:_),R,[card:A],[elem(card):R],id).
extract_symmetric(partial_injection(card:_,card:B),R,[card:B],[elem(card):R],lte).
extract_symmetric(total_injection(card:_,card:B),R,[card:B],[elem(card):R],lte).
extract_symmetric(total_surjection(card:_,card:B),R,[card:B],[elem(card):R],gte).
extract_symmetric(total_bijection(card:A,card:_),R,[card:A],[elem(card):R],id).
extract_symmetric(total_bijection(card:_,card:B),R,[card:B],[elem(card):R],id).
extract_symmetric(domain(elem(left(S)):A),R,[elem(left(S)):A],[elem(S):R],id).
extract_symmetric(range(elem(right(S)):A),R,[elem(right(S)):A],[elem(S):R],id).
extract_symmetric(reverse(elem(left(S)):A),R,[elem(left(S)):A],[elem(right(S)):R],id).
extract_symmetric(reverse(elem(right(S)):A),R,[elem(right(S)):A],[elem(left(S)):R],id).
extract_symmetric(reverse(card:A),R,[card:A],[card:R],id).
extract_symmetric(set_subtraction(card:A,*),R,[card:A],[card:R],gte).
extract_symmetric(closure(elem(left(S)):A),R,[elem(left(S)):A],[elem(left(S)):R],id).
extract_symmetric(closure(elem(right(S)):A),R,[elem(right(S)):A],[elem(right(S)):R],id).
extract_symmetric(closure(card:A),R,[card:A],[card:R],gte).
extract_symmetric(image(card:A,*),R,[card:R],[card:A],gte).
extract_symmetric(first_of_pair(left(S):A),R,[left(S):A],[S:R],id).
extract_symmetric(second_of_pair(right(S):A),R,[right(S):A],[S:R],id).

symmetric_constraint(A,B) :- symmetric_constraint2(A,B),!.
symmetric_constraint(A,B) :- symmetric_constraint2(B,A).
symmetric_constraint2(id,id).
symmetric_constraint2(lt,gt).
symmetric_constraint2(lte,gte).
symmetric_constraint2(negate,negate).


% this seems to take up most of the time:
create_constraint_tree(Constraints,Nodes,Tree) :-
    empty_avl(EmptyTree),
    sort(Constraints,SConstraints), % for SAT_Generator sorting reduces runtime from 22 sec to 18 sec for n=8; TO DO: investigate
    (debug:debug_mode(on) -> length(Constraints,Len), format('Processing ~w constraints~n',[Len]) ; true),
    create_constraint_tree1(SConstraints,[],UnsortedNodes,EmptyTree,Tree),
    (debug:debug_mode(on) -> length(UnsortedNodes,Len2), format('Sorting ~w nodes~n',[Len2]) ; true),
    sort(UnsortedNodes,Nodes),
    (debug:debug_mode(on) -> length(Nodes,Len3), format('Done sorting ~w nodes~n',[Len3]) ; true).
    %,set_prolog_flag(profiling,off),print_profile.

create_constraint_tree1([],Nodes,Nodes,Tree,Tree).
create_constraint_tree1([C|CRest],InNodes,OutNodes,InTree,OutTree) :-
    create_constraint_tree2(C,InNodes,InterNodes,InTree,InterTree),
    create_constraint_tree1(CRest,InterNodes,OutNodes,InterTree,OutTree).

create_constraint_tree2(Constraint,InNodes,OutNodes,InTree,OutTree) :-
    Constraint = constraint(Input,Output,_Rule),
    ( all_nodes_ignored(Output) ->
        InTree = OutTree, InNodes=OutNodes
    ; %otherwise ->
        add_all_to_constraint_tree(Input,Constraint,InTree,OutTree),
        %ord_union(Input,InNodes,Nodes),ord_union(Output,Nodes,OutNodes) % this can be very expensive !
        append(Input,InNodes,Nodes), append(Output,Nodes,OutNodes) % just append the nodes and sort them later
        ).

add_all_to_constraint_tree([],_,Tree,Tree).
add_all_to_constraint_tree([Node|NRest],Constraint,InTree,OutTree) :-
    ( avl_change(Node,InTree,Old,InterTree,[Constraint|Old]) ->
        true
    ; %otherwise ->
        avl_store(Node,InTree,[Constraint],InterTree)),
    add_all_to_constraint_tree(NRest,Constraint,InterTree,OutTree).

all_nodes_ignored([]).
all_nodes_ignored([_:ignored|Rest]) :-
    all_nodes_ignored(Rest).


create_initial_information(Nodes,InitEnv) :-
    empty_avl(Empty),
    foldl(create_initial_information2,Nodes,Empty,InitEnv).
create_initial_information2(Scope:Node,In,Out) :-
    get_scope_top(Scope,Top),
    avl_store(Scope:Node,In,Top,Out).

apply_start_constraints(Constraints,ChangedNodes,In,Out) :-
    find_start_constraints(Constraints,Start),
    writetodot(1,[],Start,In),
    compute_constraints(Start,ChangedNodes1,[],In,Out),
    list_to_ord_set(ChangedNodes1,ChangedNodes).
find_start_constraints(Constraints,Start) :-
    findall(constraint([],Output,Rule),
            member(constraint([],Output,Rule),Constraints),
            Start).

do_analysis([],Counter,_MaxIter,_CTree,Values,Values) :-
    !,writetodot(Counter,[],[],Values).
do_analysis([ChangedNode|CNRest],Counter,MaxIter,CTree,In,Out) :-
    Counter < MaxIter,!,
    ( avl_fetch(ChangedNode,CTree,Constraints) -> true
    ; otherwise -> Constraints = []),
    writetodot(Counter,[ChangedNode|CNRest],Constraints,In),
    compute_constraints(Constraints,ChangedNodes1,[],In,Inter),
    list_to_ord_set(ChangedNodes1,ChangedNodes),
    ord_union(CNRest,ChangedNodes,NewChangedNodes),
    Counter2 is Counter+1,
    do_analysis(NewChangedNodes,Counter2,MaxIter,CTree,Inter,Out).
do_analysis(_ChangedNodes,Counter,MaxIter,_CTree,Values,Values) :-
    Counter >= MaxIter.
compute_constraints([],Changed,Changed,Values,Values).
compute_constraints([Constraint|CRest],Cin,Cout,Vin,Vout) :-
    compute_constraint(Constraint,Cin,Cinter,Vin,Vinter),
    compute_constraints(CRest,Cinter,Cout,Vinter,Vout).
compute_constraint(constraint(Input,Output,Rule),Cin,Cout,Vin,Vout) :-
    get_input_values(Input,Vin,InValues),
    same_length(Output,OutValues),
    call_rule(Rule,InValues,OutValues),
    apply_output(Output,OutValues,Cin,Cout,Vin,Vout).
get_input_values([],_,[]).
get_input_values([Node|NRest],Values,[Value|VRest]) :-
    avl_fetch(Node,Values,Value),
    get_input_values(NRest,Values,VRest).
call_rule(Rule,InValues,OutValues) :-
    append_arguments(Rule,InValues,CallRule),
    if(rule(CallRule,OutValues), true,
       (add_warning(predicate_analysis,'No rule for: ',CallRule),fail)).
apply_output([],[],Changed,Changed,Values,Values).
apply_output([Node|NRest],[Value|VRest],Cin,Cout,Vin,Vout) :-
    ( Node = _:ignored ->
        Vin = Vinter, Cin=Cinter
    ; otherwise ->
        avl_change(Node,Vin,OldValue,Vinter,NewValue),
        info_conjunct(OldValue,Value,NewValue),
        ( has_more_information(NewValue,OldValue) ->
            Cin = [Node|Cinter]
        ; otherwise ->
            Cin = Cinter)),
    apply_output(NRest,VRest,Cinter,Cout,Vinter,Vout).

append_arguments(Orig,Args,Result) :-
    Orig =.. FA,
    append_arguments2(FA,Args,AllArgs),
    Result =.. AllArgs.
append_arguments2([F|FA],Args,AllArgs) :-
    multi_rule(F),!,
    append([F|FA],[Args],AllArgs).
append_arguments2(FA,Args,AllArgs) :-
    append(FA,Args,AllArgs).


create_add_list_constraints([],_Result,Constraints) :- !,Constraints=[].
create_add_list_constraints([Node],Result,Constraints) :- !,
    Constraints = [constraint([Node],[Result],id),constraint([Result],[Node],id)].
create_add_list_constraints(Nodes,Result,Constraints) :- !,
    create_add_list_constraints2(Nodes,Result,Constraints,[]).
create_add_list_constraints2([A,B],Result) --> !,
    findall(constraint(I,O,C),addition_constraint(A,B,Result,I,O,C)).
create_add_list_constraints2([A|Nodes],Result) -->
    {new_node_scope(integer,NodeId,_Nodes)},
    findall(constraint(I,O,C),addition_constraint(A,interval:NodeId,Result,I,O,C)),
    create_add_list_constraints2(Nodes,interval:NodeId).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% compute a single constraint

rule(id(Info),[Info]).
rule(constant(Info),[Info]).
rule(add(A,B),[C]) :- interval_addition(A,B,C).
rule(sub(A,B),[C]) :- interval_subtraction(A,B,C).
rule(mult(A,B),[C]) :- interval_multiplication(A,B,C).
rule(power(Base,Exp),[Result]) :-
    ( power_rule(Base,Exp,Result) -> true
    ; otherwise                   -> Result = irange(-inf,inf)).
rule(cart_div(irange(Cm,CM),irange(Am,AM)),[irange(Bm,BM)]) :-
    ( CM = inf ->
        BM = inf
    ; Am = 0 ->
        BM = inf
    ; otherwise ->
        BM infis (CM+Am-1) // Am),
    ( AM = 0 -> Bm = inf
    ; AM = inf -> Bm = inf
    ; otherwise ->
        Bm infis Cm // AM).
rule(negate(A),[C]) :- interval_negation(A,C).
rule(lt(irange(_,AM)),[irange(-inf,CM)]) :- CM infis AM-1.
rule(lte(irange(_,AM)),[irange(-inf,AM)]).
rule(gt(irange(Am,_)),[irange(Cm,inf)]) :- Cm infis Am+1.
rule(gte(irange(Am,_)),[irange(Am,inf)]).
rule(interval_card(irange(Am,AM),irange(Bm,BM)),[irange(Cm,CM)]) :-
    Cm infis max(Bm-AM+1,0), CM infis max(BM-Am+1,0).
rule(interval_possibilities(irange(Am,AM)),[irange(0,M)]) :-
    M infis max(AM-Am+1,0).
rule(union([A|Rest]),[C]) :- union2(Rest,A,C).
rule(union([]),[irange(-inf,inf)]) :- print(empty_kodkod_union_constraint),nl. % TO DO: check if ok !!
    % happens e.g., for set extensions like { min(0..7) } ; should not occur
rule(divrange(irange(Am,AM),irange(Bm,_BM)),[C]) :-
    % rough approximation to division
    ( number(Bm),Bm>=0 ->
        C = irange(Am,AM)
    ; number(Am),number(AM) ->
        M is max(abs(Am),abs(AM)),
        C = irange(-M,M)
    ; otherwise ->
        C = irange(-inf,inf)).
rule(concat_index(irange(_,AM),irange(_,BM)),[irange(1,CM)]) :-
    ( number(AM),number(BM) -> CM is AM+BM
    ; otherwise             -> CM = inf).


union2([],Acc,R) :- !,Acc=R.
union2([H|T],Acc,R) :- info_disjunct(H,Acc,I),union2(T,I,R).

power_rule(irange(Basem,BaseM),irange(Expm,ExpM),irange(Rm,RM)) :-
    number(Basem),Basem > 0,
    number(Expm),Expm >= 0,
    Rm is floor( Basem ** Expm),
    ( (BaseM=inf; ExpM=inf) -> RM=inf
    ; otherwise             -> RM is floor(BaseM ** ExpM)).

multi_rule(union).


insert_info(Infonodes,Values) :-
    create_info_lookup_table(Values,ValueTable),
    insert_info2(Infonodes,ValueTable).
insert_info2([],_).
insert_info2([infonode(Node,New,Old)|Rest],Values) :-
    find_all_node_information(Node,Values,Info),
    add_info(Info,Old,New),
    insert_info2(Rest,Values).
find_all_node_information(Node,Values,Info) :-
    ( avl_member(Node,Values,RawInfos) -> filter_tops(RawInfos,Info)
    ; otherwise -> Info = []).
filter_tops([],[]).
filter_tops([Scope:RI|RRest],Result) :-
    get_scope_top(Scope,Top),
    ( has_more_information(RI,Top) -> Result = [Scope:RI|RestResult]
    ; otherwise -> Result = RestResult),
    filter_tops(RRest,RestResult).
add_info([],Old,Old) :- !.
add_info(Info,Old,[analysis(Info)|Old]).

create_info_lookup_table(Values,Nodes) :-
    avl_to_list(Values,ValueList),
    empty_avl(Empty),
    create_info_lookup_table2(ValueList,Empty,Nodes).
create_info_lookup_table2([],Nodes,Nodes).
create_info_lookup_table2([(Scope:Node)-I|Rest],In,Out) :-
    ( avl_fetch(Node,In,Old) -> true
    ; otherwise -> Old = []),
    avl_store(Node,In,[Scope:I|Old],Inter),
    create_info_lookup_table2(Rest,Inter,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dot output

writetodot(Counter,ChangedNodes,ActiveConstraints,Values) :-
    debugging_enabled,!,
    debug_constraints(Constraints),
    writetodot(Counter,Constraints,ChangedNodes,ActiveConstraints,Values).
writetodot(_Counter,_ChangedNodes,_ActiveConstraints,_Values).

writetodot(Counter,Constraints,ChangedNodes,ActiveConstraints,Values) :-
    printnulls(3,Counter,CStr),
    ajoin(['debug-constraint-anim/constraints-',CStr,'.dot'],Filename),
    open(Filename,write,S),
    write(S,'digraph constraints {\n'),
    get_nodes(Constraints,INodes,CNodes),
    write_inodes(INodes,ChangedNodes,Values,S),
    write_cnodes(CNodes,INodes,ActiveConstraints,S),
    write(S,'}\n'),
    close(S).
get_nodes(Constraints,INodes,RNodes) :-
    findall(CN,
            (member(constraint(I,O,_),Constraints),
                (member(CN,I);member(CN,O))),
            Nodes1),
    sort(Nodes1,Nodes2),
    number_nodes(Nodes2,1,INodes),
    number_nodes(Constraints,1,RNodes).
number_nodes([],_,[]).
number_nodes([N|Rest],Nr,[node(Nr,N)|RRest]) :-
    Nr2 is Nr+1,
    number_nodes(Rest,Nr2,RRest).
write_inodes([],_,_,_).
write_inodes([node(Nr,N)|Rest],ChangedNodes,Values,S) :-
    write(S,'  i'),write(S,Nr),write(S,'[shape=box,'),
    ( ChangedNodes = [N|_] ->
        !,write(S,'fillcolor=green,style=filled,')
    ; member(N,ChangedNodes) ->
        !,write(S,'fillcolor=yellow,style=filled,')
    ; otherwise -> true),
    write(S,'label=\"'),
    write(S,N),write(S,'\\n'),
    ( debug_node(N,TExpr) ->
        translate_bexpression(TExpr,Text)
    ; otherwise ->
        Text = ('?')),
    write(S,Text),write(S,'\\n'),
    ( avl_fetch(N,Values,Value) ->
        write(S,Value)
    ; otherwise -> write(S,'?')),
    write(S,'\"];\n'),
    write_inodes(Rest,ChangedNodes,Values,S).
write_cnodes([],_,_,_).
write_cnodes([node(Nr,constraint(In,Out,R))|Rest],INodes,ActiveConstraints,S) :-
    write(S,'  c'),write(S,Nr),write(S,'[shape=diamond,'),
    ( member(constraint(In,Out,R),ActiveConstraints) ->
        !,write(S,'fillcolor=lightblue,style=filled,')
    ; otherwise -> true),
    write(S,'label=\"'),
    write(S,R),write(S,'\"];\n'),
    write_inputs(In,INodes,Nr,S),
    write_outputs(Out,INodes,Nr,S),
    write_cnodes(Rest,INodes,ActiveConstraints,S).
write_inputs([],_,_,_).
write_inputs([I|Rest],INodes,RNr,S) :-
    member(node(INr,I),INodes),!,
    write(S,'  i'),write(S,INr),write(S,' -> '),
    write(S,'  c'),write(S,RNr),write(S,';\n'),
    write_inputs(Rest,INodes,RNr,S).
write_outputs([],_,_,_).
write_outputs([O|Rest],INodes,RNr,S) :-
    member(node(ONr,O),INodes),!,
    write(S,'  c'),write(S,RNr),write(S,' -> '),
    write(S,'  i'),write(S,ONr),write(S,';\n'),
    write_outputs(Rest,INodes,RNr,S).

printnulls(N,Number,Str) :-
    number_codes(Number,NCodes),
    length(NCodes,L),
    R is N-L,
    fillnulls(R,NCodes,Codes),
    atom_codes(Str,Codes).
fillnulls(R,Codes,Codes) :- R =< 0,!.
fillnulls(R,NCodes,[48|Codes]) :-
    R2 is R-1,
    fillnulls(R2,NCodes,Codes).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic debug_predicate_analysis/0.
% assert this fact to enable dot visualization using writetodot below

store_debug_info(I,T) :- debug_predicate_analysis,!,store_debug_info2(I,T).
store_debug_info(_,_).

store_debug_info2([],_).
store_debug_info2([Node|Rest],TExpr) :-
    assert(debug_node(Node,TExpr)),
    store_debug_info2(Rest,TExpr).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% test case specification for predicate analysis

:- use_module(junit_tests).

test_predicate_analysis :-
    b_machine_has_constants,
    b_get_machine_constants(Constants),
    b_get_properties_from_machine(Properties),
    b_get_static_assertions_from_machine(Assertions),
    Assertions = [_|_],!,
    test_predicate_analysis2(Constants,Properties,Assertions,Junits),
    print_junit(Junits,predicate_analysis).
test_predicate_analysis :-
    add_error(predicate_analysis,
              'Machine must have constants and assertions to be a valid predicate analysis test case.'),
    fail.
test_predicate_analysis2(Constants,Properties,Assertions,Junits) :-
    predicate_analysis_with_global_sets(Properties,[],Constants,TaggedConstants,_TaggedProperties),!,
    test_predicate_analysis3(Assertions,TaggedConstants,Junits).
test_predicate_analysis2(_,_,_,[Junit]) :-
    in_junit_mode,!,
    pa_junit(setup,error('Predicate analysis failed'),Junit).
test_predicate_analysis2(_,_,_,_) :-
    add_error(predicate_analysis, 'Predicate analysis of properties failed'),
    fail.
test_predicate_analysis3([],_Constants,[]).
test_predicate_analysis3([Assertion|Arest],Constants,[Junit|Jrest]) :-
    test_predicate_analysis_testcase(Assertion,Constants,Junit),
    test_predicate_analysis3(Arest,Constants,Jrest).

test_predicate_analysis_testcase(TAssertion,Constants,Junit) :-
    translate_bexpression(TAssertion,Name),
    ( catch( ( test_predicate_analysis_testcase1(TAssertion,Constants),
                 pa_junit(Name,pass,Junit)),
             junit(Msg),
             handle_error(Name,Msg,Junit)) -> true
    ; otherwise ->
        pa_junit(Name,error('Test failed.'),Junit)).
handle_error(Name,Msg,Junit) :-
    in_junit_mode,!,
    pa_junit(Name,error(Msg),Junit).
handle_error(Name,Msg,_Junit) :-
    add_error(predicate_analysis, Name, Msg), fail.

test_predicate_analysis_testcase1(TAssertion,Constants) :-
    extract_predicate_analysis_testcases(TAssertion,[],Testcases),
    maplist(run_testcase(Constants),Testcases).
run_testcase(Constants,pa_testcase(Id,Scope,ExpectedValue)) :-
    format('Predicate analysis testcase: check if "~w" of "~w" is ~w~N', [Scope,Id,ExpectedValue]),
    expected_value_is_top(ExpectedValue,Scope,IsTop),
    extract_analysis(Id,Constants,Scope,FoundValue,IsTop),
    ( FoundValue == ExpectedValue -> true
    ; otherwise ->
        ajoin(['unexpected information for constant ',Id,': expected ', ExpectedValue,
               ', but was ', FoundValue],Msg),
        throw(junit(Msg))).
expected_value_is_top(ExpectedValue,Scope,true) :-
    ground(ExpectedValue),get_scope_top(Scope,ExpectedValue),!.
expected_value_is_top(_ExpectedValue,_Scope,false).

extract_predicate_analysis_testcases(TExpr,QRefs,Testcases) :-
    get_texpr_expr(TExpr,Expr),
    extract_predicate_analysis_testcases2(Expr,QRefs,Testcases).
extract_predicate_analysis_testcases2(forall(Ids,TPre,Cond),QRefs,Testcases) :-
    get_texpr_expr(TPre,Pre),introduce_qrefs(Pre,Ids,QRefs,SubQRefs),
    extract_predicate_analysis_testcases(Cond,SubQRefs,Testcases).
extract_predicate_analysis_testcases2(equal(Ref,Spec),QRefs,[pa_testcase(Constant,Scope,irange(Value,Value))]) :- !,
    extract_test_reference(Ref,QRefs,Constant,Scope),
    extract_integer(Spec,Value).
extract_predicate_analysis_testcases2(less_equal(Ref,Spec),QRefs,[pa_testcase(Constant,Scope,irange(-inf,Value))]) :- !,
    extract_test_reference(Ref,QRefs,Constant,Scope),
    extract_integer(Spec,Value).
extract_predicate_analysis_testcases2(greater_equal(Ref,Spec),QRefs,[pa_testcase(Constant,Scope,irange(Value,inf))]) :- !,
    extract_test_reference(Ref,QRefs,Constant,Scope),
    extract_integer(Spec,Value).
extract_predicate_analysis_testcases2(member(Ref,TSpec),QRefs,[pa_testcase(Constant,Scope,Info)]) :-
    get_texpr_expr(TSpec,Spec),
    extract_test_reference(Ref,QRefs,Constant,Scope),
    integer_interval_membership(Spec,Scope,Info),!.
extract_predicate_analysis_testcases2(not_member(Ref,TSpec),QRefs,[pa_testcase(Constant,Scope,Top)]) :-
    get_texpr_expr(TSpec,empty_set),
    extract_test_reference(Ref,QRefs,Constant,Scope),
    get_scope_top(Scope,Top).
extract_predicate_analysis_testcases2(conjunct(A,B),QRefs,Testcases) :- !,
    extract_predicate_analysis_testcases(A,QRefs,TestcasesA),
    extract_predicate_analysis_testcases(B,QRefs,TestcasesB),
    append(TestcasesA,TestcasesB,Testcases).
extract_predicate_analysis_testcases2(_Expr,_QRefs,_Testcases) :-
    throw(junit('Unsupported test case pattern')).

integer_interval_membership(empty_set,Scope,Bottom) :- get_scope_bottom(Scope,Bottom).
integer_interval_membership(interval(Min,Max),_Scope,irange(VMin,VMax)) :-
    extract_integer(Min,VMin),extract_integer(Max,VMax).
integer_interval_membership(integer_set(I),_Scope,irange(Min,Max)) :-
    compute_integer_set(I,Min,Max).

introduce_qrefs(member(E2,Ref),Ids,QRefs,NewQRefs) :-
    lookup_ref(Ref,QRefs,Path,Constant),append(Path,[elem],Path2),
    couple_pathes(E2,Path2,Constant,Ids,NewQRefs,QRefs).
introduce_qrefs(equal(E2,Ref),Ids,QRefs,NewQRefs) :-
    lookup_ref(Ref,QRefs,Path,Constant),
    couple_pathes(E2,Path,Constant,Ids,NewQRefs,QRefs).

couple_pathes(TExpr,Path,Constant,TIdentifiers) -->
    {get_texpr_ids(TIdentifiers,Identifiers)},
    couple_pathes2(TExpr,Path,Constant,Identifiers).
couple_pathes2(TExpr,Path,Constant,Identifiers) -->
    {get_texpr_expr(TExpr,Expr)},couple_pathes3(Expr,TExpr,Path,Constant,Identifiers).
couple_pathes3(identifier(Id),_,Path,Constant,Identifiers) -->
    {memberchk(Id,Identifiers),!},[Id-ref(Path,Constant)].
couple_pathes3(couple(A,B),_,Path,Constant,Identifiers) --> !,
    {append(Path,[left],PathA),append(Path,[right],PathB)},
    couple_pathes2(A,PathA,Constant,Identifiers),
    couple_pathes2(B,PathB,Constant,Identifiers).
couple_pathes3(_,TExpr,_Path,_Constant,_Identifiers,_,_) :-
    translate_bexpression(TExpr,String),
    atom_concat('Expected quantified identifier or couple in test specification, but was: ',
                String,Msg),
    throw(junit(Msg)).


lookup_ref(Expr,QRefs,Path,Constant) :-
    get_texpr_id(Expr,Id),
    memberchk(Id-ref(Path1,Constant1),QRefs),!,
    Path1=Path,Constant1=Constant.
lookup_ref(Expr,_QRefs,[],Id) :-
    get_texpr_id(Expr,Id).

extract_analysis(Id,Constants,Scope,FoundValue,IsTop) :-
    get_texpr_id(Constant,Id),
    ( memberchk(Constant,Constants) ->
        get_texpr_info(Constant,Info),
        ( memberchk(analysis(Analysis),Info) -> true
        ; otherwise                          -> Analysis=[]),
        ( memberchk(Scope:FoundValue,Analysis) -> true
        ; IsTop == true -> get_scope_top(Scope,FoundValue) /* no information expected */
        ; otherwise ->
            ajoin(['No analysis information found for constant ',Id,', scope ',Scope],Msg),
            print(extract_analysis(Id,Constants,Scope,FoundValue,IsTop)),nl,
            throw(junit(Msg)))
    ; otherwise ->
        ajoin(['Constant ',Id,' not found'],Msg),
        throw(junit(Msg))).

extract_test_reference(TRef,QRefs,Id,Scope) :-
    get_texpr_expr(TRef,Ref),
    extract_test_reference2(Ref,QRefs,BaseScope,Id,ScopeList),
    construct_scope(ScopeList,BaseScope,Scope),!.
extract_test_reference(TRef,_QRefs,_Id,_Scope) :-
    translate_bexpression(TRef,String),
    atom_concat('Unsupported reference on left side: ', String,Msg),
    throw(junit(Msg)).
extract_test_reference2(card(TRef),QRefs,card,Id,Scope) :-
    !,get_texpr_expr(TRef,Ref),Ref=identifier(_),
    extract_test_reference2(Ref,QRefs,_,Id,Scope).
extract_test_reference2(identifier(Id),QRefs,interval,Constant,Scope) :-
    memberchk(Id-ref(Scope,Constant),QRefs),!.
extract_test_reference2(identifier(Id),_QRefs,interval,Id,[]) :- !.


:- assert_must_succeed(( construct_scope([elem,right],interval,S),
                         S==elem(right(interval)) )).
construct_scope([],Scope,Scope).
construct_scope([F|Rest],Base,Scope) :-
    functor(Scope,F,1),arg(1,Scope,Subscope),
    construct_scope(Rest,Base,Subscope).

extract_integer(Spec,Value) :-
    ( get_texpr_expr(Spec,integer(Value)) -> true
    ; get_texpr_expr(Spec,unary_minus(Spec2)) ->
        extract_integer(Spec2,Value2),Value is -Value2
    ; otherwise ->
        throw(junit('Integer value expected in right side of assertion'))).

in_junit_mode :-
    junit_mode(_),!.

pa_junit(Name,Verdict,Junit) :-
    b_machine_name(MachineName),
    create_junit_result(Name,predicate_analysis,MachineName,0,Verdict,Junit).
