/* a file to start prob_cli with coverage/profiling info enabled (for SICStus 4.2 and later) */

:- prolog_flag(compiling,_,debugcode).
:- prolog_flag(source_info,_,on).
:- prolog_flag(profiling,_,on).

:- include(coverage_term_expander).

:- ensure_loaded(prob_cli).

:- use_module(coverage_tools).
:- use_module(test_runner).
:- use_module(tools_strings,[ajoin/2]).
:- use_module(version).
:- use_module(extension('pltables/pltables')).

:- use_module(library(samsort), [merge/3]).
:- use_module(library(system)).
:- use_module(library(codesio)).
:- use_module(library(lists)).
:- use_module(library(file_systems)).
:- use_module(library(sets),[subtract/3]).

% use module_info/3 because we are not in a module here
:- module_info(prob_cov,group,coverage_analysis).
:- module_info(prob_cov,description,'ProB start file with coverage tests activated. Includes predicates to execute certain tests. Used by Jenkins.').

generate_makefile :-
    generate_makefile('Makefile_Coverage_Groups', write_coverage_target).

write_coverage_target(G) :- format('~w:\n\tsicstus -l src/prob_cov.pl --goal "coverage_group_data(\'~w\'), halt ; halt."',[G,G]),nl,nl.

% compute coverage for a testcase category (see testcases.pl for categories)
coverage_group_data(Category) :-
    % ids_in_category(Category,IDs),
    % load_coverage_files(IDs,CoverageData),
    run_tests_by_category(Category,all),
    format('Obtaining coverage data for category ~w~n',[Category]),
    coverage_data(CoverageData),
    write_coverage_data(Category,CoverageData).

% write collected coverage information to a Datafile
write_coverage_data(FileName,Data) :-
    ajoin(['./temp_coverage_data/', FileName],Path),
    format('Writing coverage data to ~w~n',[Path]),
    open(Path,write,Stream),
    Out=cov(Data),
    write_canonical(Stream,Out),
    write(Stream,'.'),
    close(Stream).

read_coverage_data(Path,Data) :-
    open(Path,read,S),
    read(S,cov(Data)),
    close(S).

relevant_categories(units,[unit]).
relevant_categories(laws,[laws]).
relevant_categories(units-and-laws,[unit,laws]).
relevant_categories(regression,Groups) :-
    all_categories(All),
    subtract(All,[unit,laws],Groups).
relevant_categories(all,All) :-
    all_categories(All).
relevant_categories(data_validation,[data_validation]).
relevant_categories(X,[X]) :- atom(X).
relevant_categories(X,X) :- is_list(X).

header(units, 'Unit Tests').
header(laws, 'Law Checking').
header(units-and-laws,'Unit and Law Tests').
header(regression,'Regression Tests').
header(all,'All Tests').
header(data_validation,'Data Validation Tests').
%header(last,'Last Test Case').
header(X,Joined) :- atom(X), !, ajoin(['Testcases with Group ', X],Joined).
header(X,Joined) :- is_list(X), maplist(atom,X), !, ajoin(['Testcases with Groups ', X],Joined).

atom_header(Keyword, Header) :-
    header(Keyword,Tests),
    datime(datime(Year,Month,Day,Hour,Min,_Sec)),
    revision(CurrentCommit),
    format_to_codes('Coverage Report (~w, generated ~w.~w.~w ~w:~w, commit ~w)', [Tests,Day,Month,Year,Hour,Min,CurrentCommit], Codes),
    atom_codes(Header, Codes).

:- use_module(tools,[ensure_atom/2]).
load_coverage_files(IDs,CoverageData) :-
    maplist(ensure_atom,IDs,AtomicIDs),
    file_members_of_directory('./temp_coverage_data/',CoverageFiles),
    write('Reading Coverage Files'), nl,
    read_coverage_files(CoverageFiles,AtomicIDs,CoverageData).


read_coverage_files(Files,RelevantIDs,CoverageData) :-
    read_coverage_files(Files,RelevantIDs,[],CoverageData).

read_coverage_files([],_IDs,DataSoFar,DataSoFar).
read_coverage_files([Name-Path|Files],IDs,DataSoFar,CovData) :-
    member(Name,IDs), !,
    format('.... Reading ~w~n',[Name]),
    read_coverage_data(Path,Content),
    % coverage data should already be sorted, only merge-step
    format('     .... Merging Coverage Lists~n',[]),
    merge(DataSoFar,Content,AppendedData),
    format('     .... Merging Coverage Data~n',[]),
    merge_coverage_data(AppendedData,NewData),
    length(DataSoFar,L1), length(NewData,L2), L is L2-L1,
    format('     .... ~w tags in old, ~w tags in new, ~w difference~n',[L1,L2,L]),
    garbage_collect,
    read_coverage_files(Files,IDs,NewData,CovData).
read_coverage_files([_|Files],IDs,DataSoFar,CovData) :-
    read_coverage_files(Files,IDs,DataSoFar,CovData).

merge_coverage_data([],[]).
merge_coverage_data([C],[C]).
merge_coverage_data([H1,H2|T],Out) :-
    merge_coverage_counter(H1,H2,HMerged), !,
    merge_coverage_data([HMerged|T],Out).
merge_coverage_data([H1,H2|T],[H1|Out]) :-
    merge_coverage_data([H2|T],Out).

merge_coverage_counter(counter(File,PredSpec,ClauseNo,LineNr)-Tag1,
                       counter(File,PredSpec,ClauseNo,LineNr)-Tag2,
                       counter(File,PredSpec,ClauseNo,LineNr)-TagOut) :-
    merge_coverage_tags(Tag1,Tag2,TagOut).

merge_coverage_tags(Tag1,Tag2,TagOut) :-
    Tag1 =.. [Functor1,Count1], Tag2 =.. [Functor2,Count2], !,
    Count3 is Count1 + Count2,
    (Functor1 = Functor2
     -> TagOut =.. [Functor1,Count3]
     ;  TagOut =.. [nondet,Count3]).

generate_report(Now,descriptions) :- !,
    ajoin(['./coverage_reports/html/descriptions/descriptions-',Now,'.html'],FilenameHTML),
    ajoin(['./coverage_reports/tex/descriptions/descriptions-',Now,'.tex'],FilenameTEX),
    gen_pltables_descriptions_table(descriptionstable, [headline('Module Descriptions')],
                                    [kernel, interpreter]),
    tables_to_html([descriptionstable], FilenameHTML),
    tables_to_latex([descriptionstable], FilenameTEX).

generate_report(Now,Keyword) :-
    format('Generating coverage report for keyword: ~w~n',[Keyword]),
    relevant_categories(Keyword,Categories),
    load_coverage_files(Categories,CoverageInfo),
    % store coverage data of certain keywords, i.e.
    % to generate units-and-laws report by reading the units and the laws reports
    % and merging them
    write_coverage_data(Keyword,CoverageInfo),
    get_file_coverage_info(CoverageInfo,FinalCoverageInfo),
    atom_header(Keyword,Header),
    ajoin(['./coverage_reports/html/',Keyword,'/',Keyword,'-',Now,'.html'],FilenameHTML),
    ajoin(['./coverage_reports/tex/',Keyword,'/',Keyword,'-',Now,'.html'],FilenameTEX),
    ajoin(['./coverage_reports/xml/',Keyword,'/',Keyword,'-',Now,'.xml'],FilenameXML),
    ajoin(['./coverage_reports/color/',Keyword,'/'],ColorizedPath),
    ajoin(['../../color/',Keyword,'/'],ColorizedURL),
    ajoin(['./coverage_reports/emma/',Keyword,'.xml'],FilenameEmma),
    export(FinalCoverageInfo,Header,FilenameHTML,FilenameTEX,FilenameXML,ColorizedPath,ColorizedURL,FilenameEmma).

export(Info,Header,FilenameHTML,FilenameTEX,FilenameXML,ColorizedPath,ColorizedURL,EmmaPath) :-
    write('colorising source files line by line'), nl, full_flush,
    colorize_source_files(Info, ColorizedPath),
    write('colorising source files line by line done'), nl, full_flush,
    write('generation coverage report tables'), nl, full_flush,
    gen_pltables_coverage_rows(Info, [headline(Header)],[kernel,interpreter],ColorizedURL),
    write('generation coverage report tables done'), nl, full_flush,
    write('exporting tables to html, tex and xml'), nl, full_flush,
    export_tables(FilenameHTML,FilenameTEX,FilenameXML),
    write('exporting tables to html, tex and xml done'), nl, full_flush,
    write('exporting coverage data to emma'), nl, full_flush,
    export_emma(EmmaPath),
    write('exporting coverage data to emma done'), nl, full_flush.

% from old prob_cov.pl - now ported to new table module
pci :- print_coverage_info.
pci(X) :- print_coverage_info(X).
hci :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_html_stream(user_output).
hci_file :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_html('out.html').
lci :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_latex_stream(user_output).
lci_file :-
    get_file_coverage_info(Info),
    gen_pltables_coverage_rows(Info, [], [kernel, interpreter]),
    export_tables_latex('out.tex').
