% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(z_tools,
          [decorate_vars/3,
           find_unchanged/3,
           vars_of_decl/2,
           z_conjunction/2,
           z_conjunctiont/2,
           z_disjunction/2,
           z_counter/2,
           z_translate_identifier/2,
           z_translate_identifiers/2,
           z_translate_typed_identifiers/4,
           z_translate_typed_identifier/4,
           z_translate_identifier_list/2,
           append_not_duplicate/3,
           create_z_let/6]).

:- use_module(library(lists)).

:- use_module(probsrc(tools)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(error_manager),[add_warning/3]).

:- use_module(probsrc(module_information)).
:- module_info(group,proz).
:- module_info(description,'This module provides several helper predicates for ProZ.').


%*******************************************************************************
% translates an Z identifier of the form "name(Name,Appendix)" to a B identifier
z_translate_identifier(name(ZName,Deco), BName) :-
    atom_concat(ZName, Deco, BName).
z_translate_identifiers([],[]).
z_translate_identifiers([ZName|ZRest],[BName|BRest]) :-
    z_translate_identifier(ZName,BName),
    z_translate_identifiers(ZRest,BRest).

z_translate_identifier_list([],[]).
z_translate_identifier_list([ZName|ZRest],['Identifier'(BName)|BRest]) :-
    z_translate_identifier(ZName,BName),
    z_translate_identifier_list(ZRest,BRest).


z_translate_typed_identifier(name(ZName,Deco),Type,Infos,TExpr) :-
    ajoin([ZName,Deco],Id),
    create_texpr(identifier(Id),Type,[zname(ZName,Deco)|Infos],TExpr).

z_translate_typed_identifiers([],[],_,[]).
z_translate_typed_identifiers([ZName|Zrest],[Type|Trest],Infos,[TExpr|Brest]) :-
    z_translate_typed_identifier(ZName,Type,Infos,TExpr),
    z_translate_typed_identifiers(Zrest,Trest,Infos,Brest).

%*******************************************************************************
% Returns a term with a unique number appended
:- dynamic z_counter_db/1.
z_counter_db(1).

z_counter(Name,NewName) :-
    retract(z_counter_db(C)),
    C1 is C + 1,
    assert(z_counter_db(C1)),
    atom_codes(Name,NC), number_codes(C,CC),
    append(NC,CC,NewC),
    atom_codes(NewName,NewC).


%*******************************************************************************
% returns the list of variables that are declarated
vars_of_decl(Decls,Vars) :-
    findall(V,
            (member(decl(Vars,_Type),Decls),member(V,Vars)),
            Vars).


%*******************************************************************************
% appends a list to another while ignoring duplicate entries
append_not_duplicate(A,B,Result) :-
    remove_all(B,A,D),
    append(A,D,Result).


%*******************************************************************************
% decorates variables
decorate_vars(In,Deco,Out) :- (Deco = '' -> In=Out ; decorate_vars2(In,Deco,Out)).
decorate_vars2([],_,[]).
decorate_vars2([name(Name,D1)|Rest],Deco,[name(Name,D2)|DRest]) :-
    atom_concat(D1,Deco,D2),
    decorate_vars2(Rest,Deco,DRest).


%********************************************************************************
% find unchanged variables

find_unchanged(ti(body(Decls,Where),Type),UnchangedVars,ti(body(Decls,RestWhere),Type)) :-
    find_unchanged2(Where,UnchangedVars,RestWhere).
find_unchanged2([],[],[]) :- !.
find_unchanged2([ti(equal(ti(ref(name(Name,DecoA),[]),_),
                          ti(ref(name(Name,DecoB),[]),_)),bool)|ExprRest],
                [name(Name,'')|UnchangedRest],Rest) :-
    ( (DecoA = '', DecoB = '\'')    % TODO: Use library(lists)
    ; (DecoA = '\'', DecoB = '')),!,
    find_unchanged2(ExprRest,UnchangedRest,Rest).
find_unchanged2([Pred|Rest],Unchanged,[Pred|FRest]) :- find_unchanged2(Rest,Unchanged,FRest).


%*******************************************************************************
% build the conjunction/disjunction of a list of predicates
% the result is a single predicate

z_conjunction(List,Result) :- lcombine(List,and,truth,Result).
z_disjunction(List,Result) :- lcombine(List,or,falsity,Result).

lcombine(List,Op,Neutral,Result) :-
    lcombine2(List,Op,Neutral,Neutral,Result).
lcombine2([],_,_,Result,Result).
lcombine2([A|Rest],Op,Neutral,Acc,Result) :-
    lcombine3(Acc,A,Op,Neutral,Inter),
    lcombine2(Rest,Op,Neutral,Inter,Result).
lcombine3(Neutral,B,_,Neutral,B) :- !.
lcombine3(A,Neutral,_,Neutral,A) :- !.
lcombine3(A,B,Op,_,Result) :- Result =.. [Op,A,B].


z_conjunctiont(List,Result) :-
    z_conjunctiont2(List,ti(truth,bool),Result).
z_conjunctiont2([],Result,Result).
z_conjunctiont2([A|Rest],Acc,Result) :-
    z_conjunctiont3(Acc,A,Inter),
    z_conjunctiont2(Rest,Inter,Result).
z_conjunctiont3(ti(truth,bool),B,B) :- !.
z_conjunctiont3(A,ti(truth,bool),A) :- !.
z_conjunctiont3(A,B,ti(and(A,B),bool)).


% ****************************************************



% create Z-style let; see page 59 of Spivey, identifiers are not in scope for Value expressions
create_z_let(ExPr,Ids,Values,Inner1,Infos,Result) :-
    remove_unused_identifiers(Ids,Values,Inner1,RIds0,RValues0),
    remove_unnecessary_lets(RIds0,RValues0,RIds1,RValues),
    ( RIds1=[] -> Result = Inner1
    ;
     find_identifier_uses_l(RValues,[],RVars),
     gen_fresh_typed_params(RIds1,Inner1,RVars,RIds,Inner),
     (ExPr=expr ->
         get_texpr_type(Inner,Type),
        create_texpr(let_expression(RIds,RValues,Inner),Type,Infos,Result)
     ; ExPr=pred ->
        create_texpr(let_predicate(RIds,RValues,Inner),pred,Infos,Result)
     ),
% extra check: but gen_fresh_typed_params should ensure everything ok
       get_texpr_ids(RIds,S1), sort(S1,SortedIDs),
       (ord_intersection(SortedIDs,RVars,Clash), Clash \= []
         -> add_warning(create_let,'Variable clash in generated let:',Clash)
            % can cause problems if we use let_predicate/let_expression which allows introducing multiple, dependent ids
         ;  true)
    ).

:- use_module(library(lists)).
:- use_module(library(ordsets)).

%% rename parameters and direct definition body to use fresh ids
%% reason is that otherwise we repeatedly introduce the same variable and obtain nested lets with variable clashes
%% i.e., we may get lets of the form LET a ... LET a,b BE a=(a) & b=(a) ... where (a) is the outer value of a
%% Note: however, performance of test 1657 suffers if we systematically rename (hence we only rename VarsToAvoid)
gen_fresh_typed_params(TParams,TBody,VarsToAvoid,NewTParams,NewTBody) :-
   maplist(alpha_typed_id_rename(VarsToAvoid),TParams,NewTParams,RenameList),
   rename_bt(TBody,RenameList,NewTBody).

:- use_module(probsrc(gensym),[gensym/2]).
alpha_typed_id_rename(VarsToAvoid,b(identifier(OldID),T,I),b(identifier(FreshID),T,I),rename(OldID,FreshID)) :-
   (ord_member(OldID,VarsToAvoid) -> gensym(OldID,FreshID) ; FreshID=OldID).

remove_unused_identifiers(TLetIds1,TLetValues1,TDef,TLetIds,TLetValues) :-
    find_identifier_uses(TDef,[],UsedIds),
    maplist(combine_with_minus,TLetIds1,TLetValues1,IdValuePairs),
    include(id_is_used(UsedIds),IdValuePairs,NewIdValuePairs),
    maplist(combine_with_minus,TLetIds,TLetValues,NewIdValuePairs).

% remove let a=a, ...
remove_unnecessary_lets([],[],[],[]).
remove_unnecessary_lets([TID|T1],[TRHS|T2],IDS,RHSS) :-
   (unnecessary_let(TID,TRHS)
    -> remove_unnecessary_lets(T1,T2,IDS,RHSS)
     ; IDS=[TID|TIDS], RHSS = [TRHS|TRHSS], remove_unnecessary_lets(T1,T2,TIDS,TRHSS)
    ).
unnecessary_let(b(identifier(ID),T,_),b(identifier(ID),T,_)).
combine_with_minus(A,B,A-B).
id_is_used(UsedIds,TId-_Value) :-
    get_texpr_id(TId,Id),memberchk(Id,UsedIds).
