% (c) 2014-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(runtime_profiler,[reset_runtime_profiler/0,
                            profile_failure_driven_loop/1, profile_failure_driven_loop/2,
                            profile_single_call/2, profile_single_call/3,
                            register_profiler_runtime/4, % manually register runtimes
                            tcltk_get_profile_info/1,
                            runtime_profile_available/0,
                            print_runtime_profile/0,

                            runtime_observe_available/0,
                            observe_runtime/3
                            ]).

% TO DO: optionally load more precise microsecond timer

:- load_files(library(system), [when(compile_time), imports([environ/2])]).


:- if(\+ environ(prob_observe_runtime, true)).
runtime_observe_available :- fail.
observe_runtime(_StartMs,_Msg,_Span).
:- else.
:- use_module(error_manager,[add_message/4]).
runtime_observe_available.
observe_limit(100). % TO DO: set limit by SICStus environment variable or preference
observe_runtime(Start,Msg,Span) :-
    statistics(runtime,[End,_]), TotRuntime is End-Start,
    observe_limit(Limit),
    (TotRuntime>Limit -> add_message(delay,Msg,TotRuntime,Span) ; true).
:- endif.

:- if(\+ environ(prob_profile, true)).

reset_runtime_profiler.
:- meta_predicate profile_single_call(-,-,0).
profile_single_call(_Label,_ID,C) :- call(C).
:- meta_predicate profile_single_call(-,0).
profile_single_call(_Label,C) :- call(C).
register_profiler_runtime(_,_,_,_).
profile_failure_driven_loop(_).
profile_failure_driven_loop(_,_).
print_runtime_profile :-
    print('No profiling information available'),nl,
    print('Recompile ProB with -Dprob_profile=true'),nl.
tcltk_get_profile_info(list([I1,I2])) :-
    I1='No profiling information available',
    I2='Recompile ProB with -Dprob_profile=true'.
runtime_profile_available :- fail.

:- else.
runtime_profile_available.

:- volatile total_runtime/6.
:- dynamic total_runtime/6.

reset_runtime_profiler :- % print('RESET '),nl, print_runtime_profile,
  retractall(total_runtime(_,_,_,_,_,_)).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_runtime_profiler,
                           'Reset runtime profiler.').

:- meta_predicate profile_single_call(-,-,0).
profile_single_call(Label,ID,C) :-
   start_profile(Label,StartInfo),
   if(call(C),stop_profile(Label,ID,StartInfo),(stop_profile(Label,ID,StartInfo),fail)).

:- meta_predicate profile_single_call(-,0).
profile_single_call(Label,C) :- profile_single_call(Label,unknown,C).


% register runtime for a failure driven loop:
profile_failure_driven_loop(ActionName) :- profile_failure_driven_loop(ActionName,unknown).
profile_failure_driven_loop(ActionName,StateID) :-
   start_profile(ActionName,Info),
   %format('Start Action ----> ~w  (~w)~n',[ActionName,Info]),
   register_time(ActionName,StateID,Info).
register_time(_,_,_).
register_time(AN,StateID,Info) :-
   stop_profile(AN,StateID,Info),
   fail.

start_profile(_Label,[RT1,WT1]) :-
   statistics(runtime,[RT1,_Delta]),
   statistics(walltime,[WT1,_]).
stop_profile(Label,StateID,[RT1,WT1]) :-
   statistics(runtime,[RT2,_]),
   statistics(walltime,[WT2,_]),
   DRT is RT2-RT1, DWT is WT2-WT1,
   %format('Runtime for ~w : ~w ms (~w wall)~n',[AN,DRT,DWT]),
   register_profiler_runtime(Label,StateID,DRT,DWT).

% register_profiler_runtime(Label,StateID,DeltaRunTime,DeltaWallTime)
register_profiler_runtime(_,_,0,0) :- !. % comment out if you want to count all calls
register_profiler_runtime(PP,ID,RTime,WTime) :-
   (var(PP) -> PP2='unknown' ; PP2 = PP),
   (retract(total_runtime(PP2,OldR,OldW,Calls,Max,MaxID))
    -> true ; OldR=0,OldW=0,Max=0,MaxID=unknown,Calls=0),
   ((RTime<0 ; WTime <0) -> print(negative_runtime(PP,ID,RTime,WTime)),nl ; true),
   NewR is OldR+RTime, NewW is OldW+WTime, NewCalls is Calls+1,
   (WTime > Max -> NewMax = WTime, NewMaxID=ID ; NewMax = Max, NewMaxID=MaxID),
   assert(total_runtime(PP2,NewR,NewW,NewCalls,NewMax,NewMaxID)).


:- use_module(library(lists)).
print_runtime_profile :-
   format('-- ProB Runtime Profiler --~n',[]),
   findall(profile(RT,WT,Calls,MaxWT,PP,MaxID),total_runtime(PP,RT,WT,Calls,MaxWT,MaxID),List),
   sort(List,SL),
   reverse(SL,RSL),
   maplist(print_runtime_profile_info,RSL),
   findall(WT2,total_runtime(_PP1,_RT1,WT2,_Calls,_MaxWT1,_MaxID1),WList),
   sumlist(WList,TotalWT),
   findall(Calls2,total_runtime(_PP2,_RT2,WT2,Calls2,_MaxWT2,_MaxID2),WList2),
   sumlist(WList2,TotalCalls),
   format(' Total Walltime: ~w ms for #calls ~w~n',[TotalWT,TotalCalls]).

print_runtime_profile_info(profile(RT,WT,NrCalls,MaxWT,PP,unknown)) :-
   format(' ~w : ~w ms (~w ms walltime & ~w ms max. walltime; #calls ~w)~n',[PP,RT,WT,MaxWT,NrCalls]).
print_runtime_profile_info(profile(RT,WT,NrCalls,MaxWT,PP,MaxID)) :-
   format(' ~w : ~w ms (~w ms walltime & ~w ms max. walltime for ~w; #calls ~w)~n',[PP,RT,WT,MaxWT,MaxID,NrCalls]).

tcltk_get_profile_info(list([list(['Operation Name','Total Runtime (ms)','Total Walltime (ms)','# Calls','Max. Walltime (ms)','Max. Witness ID'])|L])) :-
   findall(profile(RT,WT,Calls,MaxWT,AN,MaxWitnessID),total_runtime(AN,RT,WT,Calls,MaxWT,MaxWitnessID),List),
   sort(List,SL),
   reverse(SL,RSL),
   print(RSL),nl,
   maplist(get_profile_list,RSL,L).
get_profile_list(profile(RT,WT,Calls,MaxWT,AN,MaxID),list([AN,RT,WT,Calls,MaxWT,MaxID])).


:- endif.
