% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(seq_rewriter, [rewrite_seq_to_set/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smt_solvers).
:- module_info(description,'This module implements transformations of sequences to sets as needed by the SMT-LIB translation.').

:- use_module(library(lists),[maplist/3,
                              append/2,
                              is_list/1,
                              include/3]).

:- use_module(probsrc(bsyntaxtree),[remove_bt/4,
                                    create_texpr/4,
                                    safe_create_texpr/3,
                                    get_texpr_expr/2,
                                    get_texpr_type/2,
                                    conjunct_predicates/2,
                                    disjunct_predicates/2,
                                    create_forall/3,
                                    create_exists/3,
                                    create_implication/3,
                                    syntaxtransformation/5,
                                    unique_typed_id/3,
                                    find_typed_identifier_uses/2]).
:- use_module(probsrc(error_manager), [add_internal_error/2,
                                       add_error_fail/3]).
:- use_module(probsrc(typing_tools), [normalize_type/2]).

:- use_module(smt_common_predicates).

rewrite_seq_to_set(Expr,Out) :-
    %translate:translate_bexpression(Expr,PPExpr),
    %write(before(PPExpr)),nl,
    %write(before(Expr)),nl,
    rewrite_seq_to_set(Expr,CExpr,[],AC,IDs),
    % translate:translate_bexpression(CExpr,PPCExpr),
    % write(after(PPCExpr)),nl,
    combine_pred_and_ac(CExpr,AC,IDs,Out). % Do not call regular ast_cleanup afterwards: It undoes some of the changes done here
rewrite_seq_to_set(Expr,CExprOut,Path,ACOut,IDsOut) :-
    rewrites(pre,Expr,[],TPExpr,Path,ACPre,IDsPre),
    remove_bt(TPExpr,PExpr,LExpr,TLExpr),
    syntaxtransformation(PExpr,Subs,_,NewSubs,LExpr),
    functor(PExpr,F,N),
    % recursively clean up sub-expressions
    maplist(rewrite_seq_to_set,ACPre,ACPreClean),
    rewrite_seq_to_set_l(Subs,NewSubs,F/N,1,Path,ACL,IDsL),
    maplist(rewrite_seq_to_set,ACL,ACLClean),
    rewrites(post,TLExpr,[],CExpr,Path,ACPost,IDsPost),
    maplist(rewrite_seq_to_set,ACPost,ACPostClean),
    append([ACPreClean,ACLClean,ACPostClean],AC),
    append([IDsPre,IDsL,IDsPost],IDs),
    (get_texpr_type(CExpr,pred)
     -> combine_pred_and_ac(CExpr,AC,IDs,CExprOut), IDsOut = [], ACOut = []
     ;  CExprOut = CExpr, ACOut = AC, IDsOut = IDs).

combine_pred_and_ac(Pred,AC,IDs,Out) :-
    conjunct_predicates(AC,ACPred),
    conjunct_predicates([Pred,ACPred],OutInnerPred),
    create_exists(IDs,OutInnerPred,Out).

rewrite_seq_to_set_l([],[],_Functor,_Nr,_Path,[],[]).
rewrite_seq_to_set_l([Expr|Rest],[CExpr|CRest],Functor,ArgNr,Path,AC,IDs) :-
    rewrite_seq_to_set(Expr,CExpr,[arg(Functor,ArgNr)|Path],ACC,IDsC),
    A1 is ArgNr+1,
    rewrite_seq_to_set_l(Rest,CRest,Functor,A1,Path,ACL,IDsL),
    append(ACC,ACL,AC),
    append(IDsC,IDsL,IDs).

rewrites(Phase,Expr,AppliedRules,Result,Path,ACS,IDs) :-
    assure_single_rules(AppliedRules,Rule),
    ( rewrite_phase(Phase,Expr,NExpr,Mode/Rule,Path,ACPhase,IDsPhase) -> % try to apply a rule (matching the current phase)
    ( Mode==single -> AppRules = [Rule|AppliedRules] % if the rule is marked as "single", we add to the list of already applied rules
    ; Mode==multi  -> AppRules = AppliedRules        % if "multi", we do not add it to the list, the rule might be applied more than once
    ; otherwise    -> add_error_fail(ast_cleanup_for_smt,'Unexpected rule mode ',Mode)),
    rewrites(Phase,NExpr,AppRules,Result,Path,ACRec,IDsRec), % continue recursively with the new expression
    append(ACPhase,ACRec,ACS), append(IDsPhase,IDsRec,IDs)
    ; otherwise ->              % if no rule matches anymore,
    Result = Expr,          %   we leave the expression unmodified
    ACS = [], IDs = [],
    Rule = none).           %   and unblock the co-routine (see assure_single_rules/2)

assure_single_rules([],_Rule) :- !.
assure_single_rules(AppliedRules,Rule) :-
    assure_single_rules2(AppliedRules,Rule).
:- block assure_single_rules2(?,-).
assure_single_rules2(_AppliedRules,none) :- !.
assure_single_rules2(AppliedRules,Rule) :-
    \+ member(Rule, AppliedRules).

rewrite_phase(Phase,OTExpr,NTExpr,Mode/Rule,Path,AC,IDs) :-
    create_texpr(OExpr,OType,OInfo,OTExpr),
    create_texpr(NExpr,NType,NInfo,NTExpr),
    rewrite_phase2(Phase,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path,AC,IDs).
rewrite_phase2(pre,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,_Path,AC,IDs) :-
    rewrite_pre(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule,AC,IDs),
    (functor(Mode_Rule,'/',2) -> Mode_Rule = Mode/Rule
    ; add_internal_error('Illegal cleanup_pre rule, missing mode: ',Mode_Rule),fail).
rewrite_phase2(post,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path,AC,IDs) :-
    rewrite_post_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule,Path,AC,IDs),
    (functor(Mode_Rule,'/',2)
    -> Mode_Rule = Mode/Rule %,format(' rewrite_post Rule: ~w/~w~n',[Mode,Rule])
    ; add_internal_error('Illegal rewrite_post rule, missing mode: ',Mode_Rule),fail).

rewrite_pre(empty_sequence,seq(T),I,empty_set,set(couple(integer,T)),I,multi/empty_sequence_to_set,[],[]).
rewrite_pre(sequence_extension(SeqExtValues),seq(Type),I,set_extension(SetExtValues),set(couple(integer,Type)),I,multi/seq_ext_to_set_ext,[],[]) :-
    seq_ext_to_set_ext(SeqExtValues,1,Type,SetExtValues).

% change type of values to set instead of seq
%rewrite_pre(value(A),seq(integer),I,value(A),set(couple(integer,integer)),I,multi/replace_type_identifier,[],[]).
%rewrite_pre(value(A),seq(string),I,value(A),set(couple(string,string)),I,multi/replace_type_identifier,[],[]).

% seq(S)
% {f | f : NATURAL1 +-> S & #n.(n : NATURAL & dom(f) = 1..n)}
% warum nicht:
% {f | #n.(n : NATURAL & f : 1..n --> S)}
% scheint kein speedup zu liefern....
rewrite_pre(seq(S),set(seq(Type)),I,comprehension_set([F],Pred),set(set(couple(integer,Type))),I,multi/rewrite_seq,[],[]) :-
    unique_typed_id("_smt_tmp",integer,N),
    unique_typed_id("_smt_tmp",set(couple(integer,Type)),F),

    % f : NATURAL1 +-> S
    safe_create_texpr(partial_function(b(integer_set('NATURAL1'),set(integer),[]),S),set(set(couple(integer,Type))),PartFunc),
    safe_create_texpr(member(F,PartFunc),pred,FMember),

    safe_create_texpr(member(N,b(integer_set('NATURAL'),set(integer),[])),pred,NMember),
    safe_create_texpr(interval(b(integer(1),integer,[]),N),set(integer),Interval),
    safe_create_texpr(domain(F),set(integer),Dom),
    safe_create_texpr(equal(Dom,Interval),pred,DomIsEqualToInterval),

    conjunct_predicates([NMember,DomIsEqualToInterval],ExistsInner),
    create_exists([N],ExistsInner,Exists),
    conjunct_predicates([Exists,FMember],Pred).

% seq1(S)
% {f | f : NATURAL1 +-> S & #n.(n : NATURAL1 & dom(f) = 1..n)}
% warum nicht:
% {f | #n.(n : NATURAL1 & f : 1..n +-> S)}
% scheint kein speedup zu liefern....
rewrite_pre(seq1(S),set(seq(Type)),I,comprehension_set([F],Pred),set(set(couple(integer,Type))),I,multi/rewrite_seq,[],[]) :-
    unique_typed_id("_smt_tmp",integer,N),
    unique_typed_id("_smt_tmp",set(couple(integer,Type)),F),

    % f : NATURAL1 +-> S
    safe_create_texpr(partial_function(b(integer_set('NATURAL1'),set(integer),[]),S),set(set(couple(integer,Type))),PartFunc),
    safe_create_texpr(member(F,PartFunc),pred,FMember),

    safe_create_texpr(member(N,b(integer_set('NATURAL1'),set(integer),[])),pred,NMember),
    safe_create_texpr(interval(b(integer(1),integer,[]),N),set(integer),Interval),
    safe_create_texpr(domain(F),set(integer),Dom),
    safe_create_texpr(equal(Dom,Interval),pred,DomIsEqualToInterval),

    conjunct_predicates([NMember,DomIsEqualToInterval],ExistsInner),
    create_exists([N],ExistsInner,Exists),
    conjunct_predicates([Exists,FMember],Pred).

% iseq(S)
% {f | f : NATURAL1 +-> S & #n.(n : NATURAL & dom(f) = 1..n)} /\ (NATURAL1 >+> S)
% warum nicht:
% {f | #n.(n : NATURAL & f : 1..n +-> S )} /\ (NATURAL1 >+> S)
% scheint kein speedup zu liefern....
rewrite_pre(iseq(S),set(seq(Type)),I,intersection(CompSet,PartInj),set(set(couple(integer,Type))),I,multi/rewrite_seq,[],[]) :-
    unique_typed_id("_smt_tmp",integer,N),
    unique_typed_id("_smt_tmp",set(couple(integer,Type)),F),

    % f : NATURAL1 +-> S
    safe_create_texpr(partial_function(b(integer_set('NATURAL1'),set(integer),[]),S),set(set(couple(integer,Type))),PartFunc),
    safe_create_texpr(member(F,PartFunc),pred,FMember),

    safe_create_texpr(member(N,b(integer_set('NATURAL'),set(integer),[])),pred,NMember),
    safe_create_texpr(interval(b(integer(1),integer,[]),N),set(integer),Interval),
    safe_create_texpr(domain(F),set(integer),Dom),
    safe_create_texpr(equal(Dom,Interval),pred,DomIsEqualToInterval),

    conjunct_predicates([NMember,DomIsEqualToInterval],ExistsInner),
    create_exists([N],ExistsInner,Exists),
    conjunct_predicates([Exists,FMember],Pred),
    safe_create_texpr(comprehension_set([F],Pred),set(set(couple(integer,Type))),CompSet),

    safe_create_texpr(partial_injection(b(integer_set('NATURAL1'),set(integer),[]),S),set(set(couple(integer,Type))),PartInj).

% perm(S)
% {f | f : NATURAL1 +-> S & #n.(n : NATURAL & dom(f) = 1..n)} /\ (NATURAL1 >+> S) /\ (NATURAL1 +->> S)
% warum nicht:
% {f | #n.(n : NATURAL & f : 1..n +-> S )} /\ (NATURAL1 >+> S) /\ (NATURAL1 +->> S)
% scheint kein speedup zu liefern....
rewrite_pre(perm(S),set(seq(Type)),I,intersection(CompSet,Intersec),set(set(couple(integer,Type))),I,multi/rewrite_seq,[],[]) :-
    unique_typed_id("_smt_tmp",integer,N),
    unique_typed_id("_smt_tmp",set(couple(integer,Type)),F),

    % f : NATURAL1 +-> S
    safe_create_texpr(partial_function(b(integer_set('NATURAL1'),set(integer),[]),S),set(set(couple(integer,Type))),PartFunc),
    safe_create_texpr(member(F,PartFunc),pred,FMember),

    safe_create_texpr(member(N,b(integer_set('NATURAL'),set(integer),[])),pred,NMember),
    safe_create_texpr(interval(b(integer(1),integer,[]),N),set(integer),Interval),
    safe_create_texpr(domain(F),set(integer),Dom),
    safe_create_texpr(equal(Dom,Interval),pred,DomIsEqualToInterval),

    conjunct_predicates([NMember,DomIsEqualToInterval],ExistsInner),
    create_exists([N],ExistsInner,Exists),
    conjunct_predicates([Exists,FMember],Pred),
    safe_create_texpr(comprehension_set([F],Pred),set(set(couple(integer,Type))),CompSet),

    safe_create_texpr(partial_injection(b(integer_set('NATURAL1'),set(integer),[]),S),set(set(couple(integer,Type))),PartInj),
    safe_create_texpr(partial_surjection(b(integer_set('NATURAL1'),set(integer),[]),S),set(set(couple(integer,Type))),TotalSurj),
    safe_create_texpr(intersection(PartInj,TotalSurj),pred,Intersec).

% replace size by card -> card performance really bad, timeout with around 20 elements
%rewrite_pre(size(A),integer,I,card(A),integer,I,multi/replace_size_card,[],[]).
% better take maximum of indices
%rewrite_pre(size(b(empty_sequence,_,_)),integer,I,b(integer(0),integer,[]),integer,I,multi/rewrite_size,[],[]).
rewrite_pre(size(S),integer,I,UnwrappedId,integer,I,multi/rewrite_size,[AC1,AC2],[UId]) :-
    unique_typed_id("_smt_tmp",integer,UId),
    get_texpr_expr(UId,UnwrappedId),
    % two axioms:
    % !(i,v).((i,v) : Set => uid >= x)
    % #(i,v).((i,v) : Set & uid = x)

    unique_typed_id("_smt_tmp",integer,Index),

    (get_texpr_type(S,seq(SeqType)), !
     ; get_texpr_type(S,set(couple(integer,SeqType)))),

    unique_typed_id("_smt_tmp",SeqType,Value),
    safe_create_texpr(couple(Index,Value),couple(integer,SeqType),QuantifiedVar),

    safe_create_texpr(greater_equal(UId,Index),pred,GreaterEqual),
    safe_create_texpr(equal(UId,Index),pred,Equal),
    safe_create_texpr(member(QuantifiedVar,S),pred,QuantifiedInSet),

    conjunct_predicates([QuantifiedInSet,Equal],ExistsInner),
    create_implication(QuantifiedInSet,GreaterEqual,ForallInner),

    create_exists([Index,Value],ExistsInner,AC1),
    create_forall([Index,Value],ForallInner,AC2).
% gets first element of non-empty seq
% card performance, solver answers unknown with >=13 elements (card timeout?)
%rewrite_pre(first(A),Type,I,UnwrappedId,Type,I,multi/rewrite_first,[AC],[UId]) :-
%    unique_typed_id("_smt_tmp",integer,UId),
%    get_texpr_expr(UId,UnwrappedId),

%    safe_create_texpr(couple(b(integer(1),integer,[]),UId),couple(integer,Type),TheCouple),
%    safe_create_texpr(member(TheCouple,A),pred,AC).
rewrite_pre(first(A),Type,I,function(A,b(integer(1),integer,[])),Type,I,multi/rewrite_first,[],[]).
    % this should be handled by the function(...) itself
    % safe_create_texpr(size(A),integer,Card).
    % create_texpr(greater_equal(Card,b(integer(1),integer,[])),pred,[],AC).
% gets last element of non-empty seq
% card performance, solver answers unknown with >=13 elements (card timeout?)
rewrite_pre(last(A),integer,I,function(A,Card),integer,I,multi/rewrite_last,[],[]) :-
    safe_create_texpr(size(A),integer,Card).
    % this should be handled by the function(...) itself
    % create_texpr(greater_equal(Card,b(integer(1),integer,[])),pred,[],AC).
% reverses sequence
rewrite_pre(rev(S),seq(Type),I,comprehension_set([Index,Value],QuantifiedInSet),set(couple(integer,Type)),I,multi/rewrite_rev,[AC1],[TmpCard]) :-
    % {(i,v) | (card(S)-i+1,b) : x}
    safe_create_texpr(size(S),integer,Card),
    unique_typed_id("_smt_tmp",Type,Index),
    unique_typed_id("_smt_tmp",Type,Value),

    % size is stored outside of the set comprehension using a temp. identifier
    % to increase performance
    unique_typed_id("_smt_tmp",integer,TmpCard),
    safe_create_texpr(equal(TmpCard,Card),pred,AC1),

    safe_create_texpr(minus(TmpCard,Index),integer,TmpIndex),
    safe_create_texpr(add(TmpIndex,b(integer(1),integer,[])),integer,NIndex),
    safe_create_texpr(couple(NIndex,Value),couple(integer,Type),TheCouple),

    safe_create_texpr(member(TheCouple,S),pred,QuantifiedInSet).
% concat returns S2 concatenated to S1 as a Sequence
rewrite_pre(concat(S1,S2),seq(Type),I,comprehension_set([Index,Value],Disjunct),set(couple(integer,Type)),I,multi/rewrite_concat,[AC1],[TmpCard]) :-
    % {(i,v) | (i,v) : S1 or (i-card(S1),v) : S2}
    unique_typed_id("_smt_tmp",integer,Index),
    unique_typed_id("_smt_tmp",Type,Value),

    safe_create_texpr(size(S1),integer,Card1),
    % size is stored outside of the set comprehension using a temp. identifier
    % to increase performance
    unique_typed_id("_smt_tmp",integer,TmpCard),
    safe_create_texpr(equal(TmpCard,Card1),pred,AC1),

    safe_create_texpr(couple(Index,Value),couple(integer,Type),QuantifiedVar),
    % (i,v) : S1
    safe_create_texpr(member(QuantifiedVar,S1),pred,QuantifiedInSet1),

    % (i-card(S1),v) : S2
    safe_create_texpr(minus(Index,TmpCard),integer,NIndex),
    safe_create_texpr(couple(NIndex,Value),couple(integer,Type),QuantifiedVar2),
    safe_create_texpr(member(QuantifiedVar2,S2),pred,QuantifiedInSet2),

    disjunct_predicates([QuantifiedInSet1,QuantifiedInSet2],Disjunct).
% restrict_front returns the first N elements of the sequence S as a Sequence
rewrite_pre(restrict_front(S,N),seq(Type),I,comprehension_set([Index,Value],Conjunct),set(couple(integer,Type)),I,multi/rewrite_restrict_front,[],[]) :-
    % {(i,v) | (i,v) : S & i <= N}
    unique_typed_id("_smt_tmp",integer,Index),
    unique_typed_id("_smt_tmp",Type,Value),
    safe_create_texpr(couple(Index,Value),couple(integer,Type),QuantifiedVar),
    % (i,v) : S
    safe_create_texpr(member(QuantifiedVar,S),pred,QuantifiedInSet),
    % i <= N
    safe_create_texpr(less_equal(Index,N),pred,LessEqual),

    conjunct_predicates([QuantifiedInSet,LessEqual],Conjunct).
% restrict_tail returns the sequence S without the first N elements
% special case to increase performance
rewrite_pre(restrict_tail(S,N),seq(Type),I,SExpr,set(couple(integer,Type)),I,multi/rewrite_restrict_tail_special,[],[]) :-
    get_texpr_expr(N,integer(0)), !,
    get_texpr_expr(S,SExpr).
rewrite_pre(restrict_tail(S,N),seq(Type),I,comprehension_set([Index,Value],Conjunct),set(couple(integer,Type)),I,multi/rewrite_restrict_tail,[],[]) :-
    % {(i,v) | (i+N,v) : S & i > 0}
    unique_typed_id("_smt_tmp",integer,Index),
    unique_typed_id("_smt_tmp",Type,Value),
    %create_texpr(couple(Index,Value),couple(integer,Type),[],AC),

    % NIndex = i+N
    create_texpr(add(N,Index),integer,[],NIndex),
    safe_create_texpr(couple(NIndex,Value),couple(integer,Type),TheCouple),
    safe_create_texpr(member(TheCouple,S),pred,QuantifiedInSet),

    % i > 0
    safe_create_texpr(integer(0),integer,Null),
    safe_create_texpr(greater(Index,Null),pred,GreaterEqual),

    conjunct_predicates([QuantifiedInSet,GreaterEqual],Conjunct).
% prepends element E to Sequence S
rewrite_pre(insert_front(E,S),seq(Type),I,union(AC,CompSet),set(couple(integer,Type)),I,multi/rewrite_insert_front,[],[]) :-
    % {(i,v) | (i-1,v) : S or (i = 1 & v = E)}
    % easier to code: {(i,v) | (i-1,v) : S} \/ {(1,E)}
    % {(1,E)}
    safe_create_texpr(set_extension([b(couple(b(integer(1),integer,[]),E),couple(integer,Type),[])]),set(couple(integer,Type)),AC),

    % {(i,v) | (i-1,v) : S}
    unique_typed_id("_smt_tmp",integer,Index),
    unique_typed_id("_smt_tmp",Type,Value),

    safe_create_texpr(minus(Index,b(integer(1),integer,[])),integer,NIndex),
    safe_create_texpr(member(TheCouple,S),pred,QuantifiedInSet),
    safe_create_texpr(couple(NIndex,Value),couple(integer,Type),TheCouple),

    safe_create_texpr(comprehension_set([Index,Value],QuantifiedInSet),set(couple(integer,Type)),CompSet).
% append element S <- E: S \/ {(size(S)+1,E)}
rewrite_pre(insert_tail(S,E),seq(Type),I,union(S,AC),set(couple(integer,Type)),I,multi/rewrite_insert_tail,[],[]) :-
    % NIndex = size(S)+1
    safe_create_texpr(size(S),integer,Card),
    safe_create_texpr(add(Card,b(integer(1),integer,[])),integer,NIndex),
    % AC = {(NIndex,E)}
    safe_create_texpr(set_extension([b(couple(NIndex,E),couple(integer,Type),[])]),set(couple(integer,Type)),AC).
% get front: S - {(size(S),function(S,size(S)))}
%rewrite_pre(front(A),seq(Type),I,restrict_front(A,N),seq(Type),I,multi/rewrite_front,[],[]) :-
%    safe_create_texpr(size(A),integer,Card),
%    safe_create_texpr(minus(Card,b(integer(1),integer,[])),integer,N).
rewrite_pre(front(A),seq(Type),I,set_subtraction(A,AC),set(couple(integer,Type)),I,multi/rewrite_front,[],[]) :-
    safe_create_texpr(size(A),integer,Card),
    safe_create_texpr(function(A,Card),Type,Last),
    % AC = {(size(A),function(A,size(A)))}
    safe_create_texpr(set_extension([b(couple(Card,Last),couple(integer,Type),[])]),set(couple(integer,Type)),AC).
rewrite_pre(tail(S),seq(Type),I,comprehension_set([Index,Value],Conjunct),set(couple(integer,Type)),I,multi/rewrite_tail,[],[]) :-
    % {(i,v) | (i+1,v) : S & i > 0}
    unique_typed_id("_smt_tmp",integer,Index),
    unique_typed_id("_smt_tmp",Type,Value),

    % (i+1,v) : S
    safe_create_texpr(add(Index,b(integer(1),integer,[])),integer,NIndex),
    safe_create_texpr(couple(NIndex,Value),couple(integer,Type),TheCouple),
    safe_create_texpr(member(TheCouple,S),pred,QuantifiedInSet),

    % i > 0
    safe_create_texpr(greater(Index,b(integer(0),integer,[])),pred,Greater),

    conjunct_predicates([QuantifiedInSet,Greater],Conjunct).
% Change seq -> set(couple(...))
rewrite_pre(A,TypeIn,I,A,TypeOut,I,multi/replace_seq_type_by_set_type,[],[]) :-
    normalize_type(TypeIn,TypeOut), TypeIn \= TypeOut.

% iseq and permutation introduce member of intersection. split again to make cleanup easier later on
rewrite_post_with_path(member(X,Intersection),pred,I,conjunct(Member1,Member2),pred,I,multi/split_member_of_intersection,_,[],[]) :-
    get_texpr_expr(Intersection,intersection(SetA,SetB)),
    safe_create_texpr(member(X,SetA),pred,Member1),
    safe_create_texpr(member(X,SetB),pred,Member2).

seq_ext_to_set_ext([],_,_,[]).
seq_ext_to_set_ext([H|T],Counter,Type,[b(couple(CounterAsBExpr,H),couple(integer,Type),[])|List]) :-
    NCounter is Counter + 1,
    safe_create_texpr(integer(Counter),integer,CounterAsBExpr),
    seq_ext_to_set_ext(T,NCounter,Type,List).
