% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(solver_interface, [solve_predicate/3, solve_predicate/4, solve_predicate/5,
                             set_up_typed_localstate_for_pred/4,
                             type_check_in_machine_context/2,
                             predicate_uses_unfixed_deferred_set/1, predicate_uses_unfixed_deferred_set/2,
                             unfixed_typed_id_in_list/3,
                             call_with_smt_mode_enabled/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module provides an interface to ProB\'s constraint solving kernel.').

:- use_module(library(lists), [include/3,keys_and_values/3]).

:- use_module(bmachine, [b_get_machine_variables/1, b_get_machine_constants/1, full_b_machine/1]).
:- use_module(btypechecker,[env_empty/1,opentype/5, add_identifiers_to_environment/3, env_store_operator/4, env_store_definition/3]).
:- use_module(b_global_sets, [is_b_global_constant/3,
                              b_global_set/1, unfixed_deferred_set/1,
                              contains_unfixed_deferred_set/1]).
:- use_module(b_interpreter, [set_up_typed_localstate/6]).
:- use_module(tools_timeout, [time_out_with_factor_call/4]).
:- use_module(bsyntaxtree, [find_typed_identifier_uses/3, find_typed_identifier_uses/2,
                 map_over_typed_bexpr/3, get_texpr_id/2, get_texpr_type/2, get_texpr_types/2]).
:- use_module(error_manager).
:- use_module(tools,[assert_once/1,foldl/5,foldl/4]).
:- use_module(translate, [translate_bvalue/2, translate_bvalue_with_limit/3]).
:- use_module(bmachine_structure,[get_section/3]).
:- use_module(debug,[debug_println/2]).
:- use_module(preferences).
:- use_module(eventhandling,[announce_event/1]).
:- use_module(smt_solvers_interface(smt_solvers_interface)).
:- use_module(store, [normalise_value_for_var/4]).

:- meta_predicate call_with_smt_mode_enabled(0).
%:- meta_predicate call_with_chr(0).
%call_with_chr(Call) :- call_with_preference(Call,use_chr_solver,true).


% :- meta_predicate call_with_smt_mode_enabled(0). % directive above for Spider
call_with_smt_mode_enabled(Call) :-
    temporary_set_preference(use_smt_mode,true,SMT),
    temporary_set_preference(use_clpfd_solver,true,CLPFD), % we also enable CLPFD
    % TO DO: also enable CHR
    call_cleanup( (Call,!),
                  (reset_temporary_preference(use_smt_mode,SMT),
                   reset_temporary_preference(use_clpfd_solver,CLPFD)
                  )).

% solve predicate finds *one* solution for predicate in State
solve_predicate(Pred,State,Res) :- solve_predicate(Pred,State,1,Res).
solve_predicate(Pred,State,TimeoutFactor,Res) :-
   solve_predicate(Pred,State,TimeoutFactor,[use_smt_mode/true,use_clpfd_solver/true],Res).

% one could add a preference to try Z3 instead/or first:
%solve_predicate(Pred,State,TimeoutFactor,Options,RealRes) :-
%    % preference(solve_predicates_with_z3,true),
%    !, print(solve_with_z3),nl, translate:print_bexpr(Pred),nl,
%    ((TimeoutFactor=1,Options=[]) -> true
%    ; format('TO DO; ignoring Timeout Factor ~w and Options ~w for Z3~n',[TimeoutFactor,Options])),
%   smt_solvers_interface:smt_solve_predicate(z3,Pred,State,RealRes),
%   print(z3_result(RealRes)),nl.
solve_predicate(Pred,State,TimeoutFactor,Options,RealRes) :-
   solve_predicate_with_options(Options,[],Pred,State,TimeoutFactor,Res), !,
   if(solver_result(Res,Pred,State,Options,RealResBeforeFallback),true,
      (add_internal_error('Unknown solver_result: ',Res),RealResBeforeFallback=error)),
   fallback_to_smt_if_requested(RealResBeforeFallback,Pred,State,RealRes).

% no fallback if solution / contradiction
fallback_to_smt_if_requested(solution(S),_,_,Res) :- !, Res=solution(S).
fallback_to_smt_if_requested(contradiction_found,_,_,Res) :- !,Res=contradiction_found.
fallback_to_smt_if_requested(S,_,_,Res) :-
    \+ preferences:get_preference(smt_supported_interpreter,true), ! , Res = S.
fallback_to_smt_if_requested(_,Pred,State,Res) :-
    smt_solve_predicate(z3,Pred,State,Res), !.

solve_predicate_with_options([],RestOptions,Pred,State,TimeoutFactor,Res) :-
   solve_predicate_direct(Pred,State,TimeoutFactor,RestOptions,Res).
solve_predicate_with_options([Option|T],RestOptions,Pred,State,TimeoutFactor,Res) :-
    solve_with_option(Option,T,RestOptions,Pred,State,TimeoutFactor,Res).

solve_with_option('CHR',T,Rest,Pred,State,TimeoutFactor,Res) :- !,
     solve_with_option(use_chr_solver/true,T,Rest,Pred,State,TimeoutFactor,Res).
solve_with_option('SMT',T,Rest,Pred,State,TimeoutFactor,Res) :- !,
     solve_with_option(use_smt_mode/true,T,Rest,Pred,State,TimeoutFactor,Res).
solve_with_option('KODKOD',T,Rest,Pred,State,TimeoutFactor,Res) :- !,
     solve_with_option(try_kodkod_on_load/true,T,Rest,Pred,State,TimeoutFactor,Res).
solve_with_option('SMT_SUPPORTED_INTERPRETER',T,Rest,Pred,State,TimeoutFactor,Res) :- !,
      solve_with_option(smt_supported_interpreter/true,T,Rest,Pred,State,TimeoutFactor,Res).
solve_with_option('CLPFD',T,Rest,Pred,State,TimeoutFactor,Res) :- !,
     solve_with_option(use_clpfd_solver/true,T,Rest,Pred,State,TimeoutFactor,Res).
solve_with_option(Pref/Val,T,Rest,Pred,State,TimeoutFactor,Res) :- !,
   call_with_preference(solve_predicate_with_options(T,Rest,Pred,State,TimeoutFactor,Res),Pref,Val).
solve_with_option(Opt,T,Rest,Pred,State,TimeoutFactor,Res) :-
   valid_option(Opt),!,
   solve_predicate_with_options(T,[Opt|Rest],Pred,State,TimeoutFactor,Res).
solve_with_option(Opt,T,Rest,Pred,State,TimeoutFactor,Res) :-
   add_warning(solve_with_option,'Unrecognised option: ',Opt),
   solve_predicate_with_options(T,Rest,Pred,State,TimeoutFactor,Res).

valid_option(full_machine_state).
valid_option(solve_in_visited_state(S)) :- atomic(S).
valid_option(truncate(Limit)) :- number(Limit). % truncate pretty printing
valid_option(truncate). % 10000
valid_option(force_evaluation). % force evaluation of symbolic results

solve_predicate_direct(Pred,State,TimeoutFactor,Options,Res) :-
    %temporary_set_preference(randomise_enumeration_order,true),
    %temporary_set_preference(find_abort_values,true),  % I am not sure we should do this; we want to find counter-examples quickly, we are less worried with finding WD-problems ?!

    %temporary_set_preference(minint,-2147483648), % disabled to pass tests 1004,1015,..
    %temporary_set_preference(maxint,2147483647),
    (real_error_occurred -> reset_real_error_occurred, PrevErr=true ; PrevErr=false),

    smt_solvers_interface:init,
    announce_event(start_solving),
    solve_predicate_aux(Pred,State,TimeoutFactor,Options,InnerRes),
    announce_event(end_solving),

    (InnerRes = overflow -> Res = overflow ; real_error_occurred -> Res = error; Res = InnerRes),
    (PrevErr -> assert_real_error_occurred(previous_error) ; true). % re-establish real-error occured

%:- use_module(clpfd_interface,[clpfd_overflow_error_message/0]).
solve_predicate_aux(Pred,State,TimeoutFactor,Options,Res) :-
    enter_new_error_scope(ScopeID,solve_predicate), %print(enter_scope(ScopeID)),nl,
    call_cleanup(solve_predicate_aux1(Pred,State,TimeoutFactor,Options,Res),
                 exit_error_scope(ScopeID,_,solve_predicate_aux)).

solve_predicate_aux1(Pred,State,TimeoutFactor,Options,Res) :-
    solve_predicate_aux2(Pred,State,TimeoutFactor,Options,Res),
    !, % we only look for one solution
    clear_enumeration_warnings. % avoid propagating them higher up
solve_predicate_aux1(_Pred,State,_TimeoutFactor,_Options,Res) :-
    critical_enumeration_warning_occured_in_error_scope(A,B,C,D), % we did not find a solution: check for enum warnings
    !,
    clear_enumeration_warnings,
    State = no_state,
    Res = virtual_time_out(enumeration_warning(A,B,C,D,critical)).
solve_predicate_aux1(_Pred,State,_TimeoutFactor,_Options,Res) :-
    State = no_state,
    Res = no.


:- use_module(clpfd_interface,[catch_clpfd_overflow_call2/2]).
solve_predicate_aux2(Pred,State,TimeoutFactor,Options,Res) :-
    %predicate_level_optimizations(Pred,OptPred),  % is now done in b_interpreter_components; to avoid that CSE disrupts partitioning
    catch_clpfd_overflow_call2(
               solve_predicate_aux3(Pred,State,TimeoutFactor,Options,Res),
               ( %print(overflow),nl,trace,
                 %clpfd_interface:clpfd_overflow_warning_message,  % message is already printed by catch_clpfd_overflow_call2
                 % we could try and solve without CLPFD ?!!
                 Res=overflow,State=no_state)).

/*
filter_global_identifiers([],[]).
filter_global_identifiers([Identifier|IDs],Out) :-
    filter_global_identifiers(IDs,TempOut),
    (is_global(Identifier)
     -> Out = TempOut
     ;  Out = [Identifier|TempOut]
    ).

is_global(Identifier) :-
    get_texpr_id(Identifier,Name), get_texpr_type(Identifier,Type),
    Type = global(Set), is_b_global_constant(Set,_Nr,Name).
is_global(Identifier) :-
    get_texpr_id(Identifier,Name), get_texpr_type(Identifier,Type),
    Type = set(global(Name)), b_global_set(Name).
*/

:- use_module(b_interpreter_components).
:- use_module('kodkod/kodkod',[apply_kodkod/3]).
:- use_module(specfile,[state_corresponds_to_set_up_constants/2]).
:- use_module(state_space,[visited_expression/2]).
:- use_module(library(lists),[maplist/2]).
solve_predicate_aux3(Pred,State,TimeoutFactor,Options,Res) :-
    % find all identifiers used in the predicate and create a state
    % was done by the caller, leading to code duplication
    (member(full_machine_state,Options) -> get_full_machine_state_typed_identifiers(FilteredIdentifiers)
      ; find_typed_identifier_uses(Pred,FilteredIdentifiers)),
    !,
    % global constants should not appear in the state
    % otherwise they treated are as variables by the solver
    %  -> we now call find_typed_identifier_uses; we used to call filter_global_identifiers(Identifiers,FilteredIdentifiers),
    % state may be predefined / partially set up
    apply_kodkod(FilteredIdentifiers,Pred,Pred2),
    (member(solve_in_visited_state(StateID),Options)
      ->  set_up_typed_localstate_for_pred(FilteredIdentifiers,Pred2,TypedVals,State),
          (visited_expression(StateID,StateExpr),
           state_corresponds_to_set_up_constants(StateExpr,FullStore)
            -> maplist(initialise_localstate(FullStore),State)
            ; StateID=root -> true
            ; add_internal_error('Illegal State:',use_visited_state(StateID))
           )
     ; var(State)
       -> set_up_typed_localstate_for_pred(FilteredIdentifiers,Pred2,TypedVals,State)
          % TypedVals contains typedval(Val,Type,Var,TriggersEnumWarning)
          % TO DO: add static_symmetry detection b_global_sets:static_symmetry_reduction_for_global_sets(State),
     ;  add_warning(solve_predicate,'Non-variable state: ',State)
         % TypedVals is unbound, is this a problem !?? TO DO: investigate
    ),
    %tools_printing:print_term_summary(b_trace_test_components(Pred2,State,TypedVals)),nl,
    b_interpreter_components:reset_component_info(false),
    time_out_with_factor_call(
        b_interpreter_components:b_trace_test_components(Pred2,State,TypedVals),
        TimeoutFactor, [silent],
        TO='time_out'),
    (TO==time_out -> Res = TO ; Res = yes).

% set up a complete typed solver state for machine with all constants and variables, even if typing invariants removed and no reference anymore to variable in Invariant
% TypedVals need to be enumerated after calling solve_predicate, in case some variables are not constrained
get_full_machine_state_typed_identifiers(TIds) :-
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants), append(Constants,Variables,TIds).

:- use_module(static_ordering,[sort_ids_by_usage/4,reorder_state/3]).
set_up_typed_localstate_for_pred(CV,Pred,TypedVals,State) :-
   (preferences:preference(use_static_ordering,true)
     -> static_ordering:sort_ids_by_usage(CV,Pred,SortedVarlist,no_warnings) ,
        set_up_typed_localstate(SortedVarlist,_FreshVars,TypedVals,[],SortedState,positive),
        reorder_state(CV,SortedState,State) % SortedState can have issues with pack_state skeleton
     ;  set_up_typed_localstate(CV,_FreshVars,TypedVals,[],State,positive)).

% initialise local state with values found in bind list
initialise_localstate(FullStore,bind(Var,Val)) :-
   (member(bind(Var,VarVal),FullStore)
     -> Val=VarVal ; true).

% translate solver_result for use by Disprover/Prover or similar:
solver_result(time_out,_P,_State,_,time_out).
solver_result(error,_P,_State,_,error).
solver_result(no,P,_State,_,Res) :-
    % there are two cases in which we might have a false negative:
    % we enumerated a critical variable
    %\+ event_occurred_in_error_scope(enumeration_warning(_,_,_,_,critical)), % we now check this in solve_predicate_aux1
    % or there are unfixed / deferred global sets involved.
    (predicate_uses_unfixed_deferred_set(P,CType)
        % no unfixed global sets involved -> real contradiction found
        -> debug_println(10,solver_predicate_contains_unfixed_deferred_set(CType)),
           Res=no_solution_found(unfixed_deferred_sets)
        ;  Res=contradiction_found
    ).
solver_result(virtual_time_out(Reason),_P,_State,_,no_solution_found(Reason)) :-
    true. % this is now checked in solve_predicate_aux1: event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)).
solver_result(overflow,_P,_State,_,no_solution_found(clpfd_overflow)).
solver_result(yes,_P,State,Options,solution(Solution)) :-
    findall(binding(Id,EValue,PPValue),
            (member(bind(Id,Value),State),
                (solver_pp_bvalue(Value,Options,EValue,PPValue) -> true ; PPValue='**pretty-print failed**')),
            Solution).

% expand and pretty print solver result values:
solver_pp_bvalue(Value,Options,EValue,PPValue) :- member(force_evaluation,Options),!,
    EXPAND=force,
    normalise_value_for_var(solver_result,EXPAND,Value,EValue),
    solver_pp_bvalue2(EValue,Options,PPValue).
solver_pp_bvalue(Value,Options,Value,PPValue) :- solver_pp_bvalue2(Value,Options,PPValue).

solver_pp_bvalue2(Value,Options,PPValue) :- member(truncate(Limit),Options),!,
    translate_bvalue_with_limit(Value,Limit,PPValue). % should we also control expand_avl_upto ??
solver_pp_bvalue2(Value,Options,PPValue) :- member(truncate,Options),!,
    translate_bvalue_with_limit(Value,10000,PPValue).
solver_pp_bvalue2(Value,_,PPValue) :- translate_bvalue(Value,PPValue).

% check if a typed expression references an unfixed deferred set in some inner node
predicate_uses_unfixed_deferred_set(TExpr) :- predicate_uses_unfixed_deferred_set(TExpr,_).

predicate_uses_unfixed_deferred_set(TExpr,CType) :-
   unfixed_deferred_set(_), % check first if some unfixed deferred set exists
   map_over_typed_bexpr(solver_interface:uses_unfixed_ds_aux,TExpr,CType).
:- public uses_unfixed_ds_aux/2.
uses_unfixed_ds_aux(TExpr,Type) :- get_texpr_type(TExpr,Type),
   contains_unfixed_deferred_set(Type).

% find a typed id whose type references an unfixed deferred set
unfixed_typed_id_in_list(TID,CType,Constants) :-
         unfixed_deferred_set(_), % check first if some unfixed deferred set exists
         member(TID,Constants), get_texpr_type(TID,CType),
         contains_unfixed_deferred_set(CType).

:- use_module(library(lists),[maplist/3, include/3]).
type_check_in_machine_context(Predicates, TPredicates) :-
  % create an empty type environment
  % and store the known types of variables / constants
  % otherwise, we might be unable to type variables from
  % machines somewhere in the refinement hierarchy
  % (their typing predicate might not be in the hypothesis)
  % Definitions are also added to the environment since the typechecker
  % inlines definitions.
  env_empty(E),
  b_get_machine_variables(V), b_get_machine_constants(C),
  % add theory operators
  full_b_machine(Machine),
  get_section(operators,Machine,Operators),
  keys_and_values(Operators,Ids,Ops),
  foldl(env_store_operator,Ids,Ops,E,EWithTheories),
  get_section(definitions, Machine, Definitions),
  foldl(env_store_definition, Definitions, EWithTheories, EWithDefinitions),
  % constants and variables
  findall(b(identifier(Name),global(GlobalSet),[enumerated_set_element]),is_b_global_constant(GlobalSet,_Nr,Name),GC),
  findall(b(identifier(Set),set(global(Set)),[given_set]),b_global_set(Set),GS),
  add_identifiers_to_environment(V,EWithDefinitions,EV),
  add_identifiers_to_environment(C,EV,EVC),
  add_identifiers_to_environment(GC,EVC,EVCGC),
  add_identifiers_to_environment(GS,EVCGC,EOut),
  opentype(Predicates, EOut, Identifiers, DeferredSets, TPredicates),
  % DeferredSets should be empty? I.e. should this be allowed to create new types?
  % currently this would not work as these types have not been added/registered in b_global_sets ! kernel_objects:basic_type would fail
  (DeferredSets = [] -> true
   ; include(problematic_id(DeferredSets),Identifiers,TErrIds),
     maplist(get_texpr_id,TErrIds,ErrIds),
     add_warning(solver_interface,'The types of some variables could not be determined: ',ErrIds),
   fail).
  % a possible refactoring for the above - however, it does not type with open identifiers
  % reverted it for now
  % bmachine:b_type_expressions(Predicates,[constants],_Types,TPredicates,Errors),
  % add_all_perrors(Errors,[],type_expression_error),
  % no_real_perror_occurred(Errors).

problematic_id(DeferredSets,TID) :- get_texpr_type(TID,Type),
   uses_def_set(Type,DeferredSets).

uses_def_set(global(T),DeferredSets) :- member(T,DeferredSets).
uses_def_set(set(A),DeferredSets) :- uses_def_set(A,DeferredSets).
uses_def_set(seq(A),DeferredSets) :- uses_def_set(A,DeferredSets).
uses_def_set(couple(A,_),DeferredSets) :- uses_def_set(A,DeferredSets).
uses_def_set(couple(_,A),DeferredSets) :- uses_def_set(A,DeferredSets).
uses_def_set(record(F),DS) :- member(field(_,T),F), uses_def_set(T,DS).
% TO DO: add freetypes
