% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(state_permuter,[permute_state/2]).
/* File: state_permuter.pl */
/* Created: 8/6/06 by Michael Leuschel */

% use_module(state_permuter), user:current_expression(_CurID,X), permute_state(X,PX).

 :- use_module(module_information).
:- module_info(group,symmetry).
:- module_info(description,'Computes all permutations of a B state; used by flooding symmetry.').


:- use_module(error_manager).

:- use_module(b_global_sets,[b_global_set/1, is_unused_b_global_constant/3,
  b_global_deferred_set/1,b_global_deferred_set_with_card_gt1/1,
  all_elements_of_type/2,b_exists_global_set_with_potential_symmetry/0]).

:- use_module(library(lists)).

:- use_module(store,[normalise_store/2]).

:- use_module(self_check).
:- use_module(custom_explicit_sets,[expand_custom_set_to_list/2]).

:- assert_must_succeed(( I=[bind(x,int(1))], state_permuter:permute_state(I,R),
 R==I )).
:- assert_must_succeed(( I=[bind(x,(pred_true,string(a)))],state_permuter:permute_state(I,R),
 R==I )).
permute_state(X,NPX) :-
  ( b_exists_global_set_with_potential_symmetry
   -> permute_state2(X,NPX) ; NPX=X).

/* Note : it can happen that the same state is generated twice:

  X={Name1,Name2} +--> will permute into itself twice

 */

:- use_module(specfile,[expand_const_and_vars_to_full_store/2]).

permute_state2(State,PermutedState) :-
  (State = concrete_constants(BState)
    -> PermutedState = concrete_constants(NPX)
    ;  (State = csp_and_b(CSPState,BState)
        -> PermutedState = csp_and_b(CSPState,NPX)
        ;  (State=BState, PermutedState=NPX)
       )
  ),
  expand_const_and_vars_to_full_store(BState,FullBState),  /* TO DO: better treatment of ConstID in const_and_vars */
  permute_state3(FullBState,[],PX),normalise_store(PX,NPX).


permute_state3([],Perm,[]) :- % print(labeling(Perm)),nl,
  label_permutation(Perm).
permute_state3([bind(X,Val)|T],InPerm,[bind(X,PermVal)|PT]) :-
  (permute_value(Val,InPerm,OutPerm,PermVal)
    -> true
    ;  add_error(state_permuter,'permute_value failed: ',permute_value(Val,InPerm,OutPerm,PermVal)),
       PermVal = Val
  ),
  %print(perm(Val,PermVal,OutPerm)),nl,
  permute_state3(T,OutPerm,PT).



permute_value(pred_false /* bool_false */,P,P,pred_false /* bool_false */).
permute_value(pred_true /* bool_true */,P,P,pred_true /* bool_true */).
permute_value(fd(X,Set),InPerm,OutPerm,PX) :-
   ((b_global_deferred_set_with_card_gt1(Set)
      ; is_unused_b_global_constant(Set,X,_) /* TO DO: only makes sense if more than one unused global constant */
    )
     -> (member(perm(X,PX,Set),InPerm)
          -> OutPerm  = InPerm
          ;  OutPerm = [perm(X,PX,Set)|InPerm]
        )
     ;  OutPerm=InPerm, PX=fd(X,Set)
   ).
permute_value(global_set(X),P,P,global_set(X)).
permute_value(freetype(X),P,P,freetype(X)).
permute_value(avl_set(X),In,Out,PermutedList) :-
   custom_explicit_sets:expand_custom_set_to_list(avl_set(X),Value),
  % print(perm(X,Value)),nl,
   permute_value(Value,In,Out,PermutedList).
permute_value([],P,P,[]).
permute_value(int(X),P,P,int(X)).
permute_value(term(X),P,P,term(X)).
permute_value(string(X),P,P,string(X)).
permute_value(rec(X),In,Out,rec(PX)) :-
  permute_fields(X,In,Out,PX).
permute_value(freeval(I,C,X),In,Out,freeval(I,C,PX)) :-
  permute_value(X,In,Out,PX).
permute_value([H|T],In,Out,[PH|PT]) :-
  permute_value(H,In,In2,PH),
  permute_value(T,In2,Out,PT).
%  (PT = [PPH|_] ->  check_normalised(PH,PPH)  /* ensure that we generate normalised sets only;
%   not all elements of set are either permuted or none;
%   only works for simple sets !!*/
%    ; true).
permute_value((H,T),In,Out,(PH,PT)) :-
  permute_value(H,In,In2,PH),
  permute_value(T,In2,Out,PT).
permute_value(closure(X,Y,Z),P,P,closure(X,Y,Z)).
permute_value(closure_x(X,Y,Z,E),P,P,closure_x(X,Y,Z,E)).
permute_value(abort(Err),P,P,abort(Err)) :- print(deprecated_abort_in_permute_value(Err)),nl.
  /* to do: permute values inside closure; or disallow permutation generation for
    states containing closures !!! */

permute_fields([],P,P,[]).
permute_fields([field(Name,Val)|T],In,Out,[field(Name,PVal)|PT]) :-
  permute_value(Val,In,In2,PVal),
  permute_fields(T,In2,Out,PT).


label_permutation(Perm) :-
  findall(vals(Set,Vals), (b_global_set(Set),set_of_values(Set,Vals)), AV),
 %print(label_perm(Perm,AV)),nl,
  label_permutation(Perm,AV).


set_of_values(Set,Vals) :-  b_global_deferred_set(Set),all_elements_of_type(Set,Vals).
set_of_values(Set,Vals) :-
   findall(fd(X,Set),is_unused_b_global_constant(Set,X,_), Vals),
   Vals \= [].

label_permutation([],_).
label_permutation([perm(X,PX,Set)|T], AV) :- % print(p(X,Set)),nl,
   extract_value(AV,X,Set,PX,AV2), %print(val(X,PX,AV2)),nl,
   label_permutation(T,AV2).

extract_value([vals(Set,Vals)|T],_X,Set,PX,AV2) :- !,
   remove(Vals,PX,Vals2),
   AV2 = [vals(Set,Vals2)|T].
extract_value([H|T],X,Set,PX,[H|T2]) :-
   extract_value(T,X,Set,PX,T2).


remove([X|T],X,T).
remove([Y|T],X,[Y|DT]) :- remove(T,X,DT).


%check_normalised(PH,PPH) :- when((ground(PH),ground(PPH)), PH @< PPH).

/*

use_module(state_permuter), permute_state([ bind(x,[ fd(1,'Name'), fd(2,'Name')]) ], R).
R = [bind(x,[fd(2,'Name'),fd(1,'Name')])] ? ;
R = [bind(x,[fd(1,'Name'),fd(2,'Name')])] ? ;
no


Mottisfont, Dual 2.7 GHz figures:

scheduler, 6 PIDs
 w/o perm reduction: 15.51 secs, 10498 states
 with perm reduction: 2.03 secs, 10498 states

Volvo Cruise_finite1.mch  (no deferred sets)
  w/o: 54 secs, 1361 states
  with: same

RussianPostalPuzzle, 4 keys
  w/o: 15.73 secs, 2325 states
  with: 3.06 secs, 5064 states (4816 permute states)

USB.mch, 2 TRANSFERS
  w/o: 8.76 secs, 415 nodes
  with: 5.53 secs, 415 nodes, 201 permute states
*/
