% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(solver_handling,[clauses_on_level/4,
                           add_clauses_to_level/4,
                           in_solver_on_level/3,
                           solve/2, solve_negated/2,
                           initial_state/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'Internal data structures for solver / prover').

:- use_module(library(avl)).
:- use_module(library(ordsets)).
:- use_module(library(lists)).

:- use_module(extension('counter/counter'),[inc_counter/1]).

:- use_module(smt_solvers_interface(smt_solvers_interface), [smt_solve_predicate/4]).

:- use_module(probsrc(atelierb_provers_interface),[prove_predicate/3]).
:- use_module(probsrc(preferences), [get_preference/2,
                                     temporary_set_preference/3,
                                     reset_temporary_preference/2]).
:- use_module(probsrc(solver_interface), [solve_predicate/5]).
:- use_module(probsrc(translate), [print_bexpr/1]).
:- use_module(probsrc(bsyntaxtree), [disjunct_predicates/2,
                                     conjunction_to_list/2,
                                     conjunct_predicates/2]).

:- use_module(symbolic_model_checker(predicate_handling)).

initial_state(Pred) :-
    get_initial_state_predicate(Init),
    conjunct_predicates([Init,Pred],IandPred),
    solve(IandPred,Result),
    Result = solution(_).

add_clauses_to_level(Level,FramesIn,Clauses,FramesOut) :-
    maplist(list_to_ord_set,Clauses,OrdClauses),
    clauses_on_level(Level,FramesIn,F_To,ClausesTo),
    list_to_ord_set(OrdClauses,SetToAdd),
    ord_union(ClausesTo,SetToAdd,NewClauses),
    exclude(is_subsumed(NewClauses),NewClauses,FilteredNewClauses),
    avl_store(Level,FramesIn,frame(F_To,FilteredNewClauses),FramesOut).

is_subsumed(AllClauses,C) :-
    member(C2,AllClauses),
    C \= C2,
    % C is subsumed by C2 if C = C2 + additional disjuncts
    ord_subtract(C2,C,[]).

clauses_on_level(Level,Frames,F_K,Clauses) :-
    avl_fetch(Level,Frames,frame(F_K,Clauses)).

in_solver_on_level(Level,Frames,InSolver) :-
    clauses_on_level(Level,Frames,F_K,Clauses),
    maplist(disjunct_predicates,Clauses,Conjuncts),
    append(F_K,Conjuncts,InSolver).

% the version below retuns clauses as negated conjunctions
% for some reason this seems to aid the solver?
% in_solver_on_level(Level,Frames,InSolver) :-
%     clauses_on_level(Level,Frames,F_K,Clauses),
%     (Clauses = [] -> InSolver = F_K ;
%     maplist(maplist(bsyntaxtree:create_negation),Clauses,T),
%     maplist(bsyntaxtree:conjunct_predicates,T,T2),
%     maplist(bsyntaxtree:create_negation,T2,Conjuncts),
%     append(F_K,Conjuncts,InSolver)).

% if our constraint solver fails, we try the atelierb provers
try_further_solvers(no_solution_found(_)).
try_further_solvers(time_out).

throw_if_solvers_too_weak(X,Pred) :-
    X \= contradiction_found(_), X \= solution(_), !,
    format('Solvers too weak for predicate:~n',[]),
    print_bexpr(Pred),nl,
    %translate:nested_print_bexpr(Pred),nl,
    throw(solver_and_provers_too_weak).
throw_if_solvers_too_weak(_,_).

print_result(Solver,FullResult) :-
    functor(FullResult,SRForPrint,_),
    format('solve/2: result of ~w: ~w~n',[Solver,SRForPrint]).

solve_negated(P,R) :-
    create_negation_and_cleanup(P,NP),
    solve(NP,R).
solve(Predicate,Result) :-
    inc_counter(ic3solvercalls),
    % debug output for all solver calls
    % translate:translate_bexpression(Predicate,PPPredicate),
    % format('Checking: ~w~n',[PPPredicate]),
    % calls through the solver interface
    % uses smt support and smt fallback
    Options = [],
    get_timeout_factor(Options,TimeoutFactor),
    solve_prob(Predicate,Options,TimeoutFactor,SolverResult),
    print_result(prob,SolverResult),
    solve2(Predicate,SolverResult,Result).

% other options to try if prob alone did not find a suitable result
% corresponding timeout factors for portfolio / prob only
options_to_try(['KODKOD']).
options_to_try(['SMT_SUPPORTED_INTERPRETER']).

get_timeout_factor([],Factor) :-
    get_preference(symbolic_mc_try_other_solvers,false)
    -> Factor = 1 ; Factor = 0.25.
get_timeout_factor(['KODKOD'],0.25).
get_timeout_factor(['SMT_SUPPORTED_INTERPRETER'],0.35). % last 0,15 reserved for ml/pp

solve2(Predicate,SolverResult,Result) :-
    try_further_solvers(SolverResult),
    get_preference(symbolic_mc_try_other_solvers,true),
    options_to_try(Options), % backtracks into different options
    inc_counter(ic3solvercalls),
    get_timeout_factor(Options,TimeoutFactor),
    solve_prob(Predicate,Options,TimeoutFactor,NewResult),
    print_result(prob,NewResult),
    \+ (try_further_solvers(NewResult)), !,
    Result = NewResult.
solve2(Predicate,SolverResult,Result) :-
    solve3(Predicate,SolverResult,Result).

solve3(Predicate,SolverResult,Result) :-
    try_further_solvers(SolverResult),
    get_preference(symbolic_mc_try_other_solvers,true),
    smt_solve_predicate(z3,Predicate,_,contradiction_found), !,
    print_result('smt_solve_predicate',contradiction_found),
    Result = contradiction_found(Predicate).
solve3(Predicate,SolverResult,Result) :-
    solve4(Predicate,SolverResult,Result).

solve4(Predicate,SolverResult,Result) :-
    try_further_solvers(SolverResult),
    get_preference(symbolic_mc_try_other_solvers,true), !,
    solve_provers(Predicate,Result),
    print_result('pp/ml',Result),
    throw_if_solvers_too_weak(Result,Predicate).
solve4(Predicate,Result,Result) :-
    throw_if_solvers_too_weak(Result,Predicate).

solve_prob(Predicate,Options,TimeoutFactor,ResultOut) :-
    temporary_set_preference(disprover_mode,true,Changed),
    % temporary_set_preference(smt_supported_interpreter,true),
    solve_predicate(Predicate,_State,TimeoutFactor,Options,Result), !,
    reset_temporary_preference(disprover_mode,Changed),
    % reset_temporary_preference(smt_supported_interpreter),
    (Result = contradiction_found -> ResultOut = contradiction_found(Predicate) ; % TODO: extract core
    otherwise -> ResultOut = Result).

solve_provers(Predicate,Result) :-
    conjunction_to_list(Predicate,Hyps),
    % no hypotheses. might want to consider refactoring the predicate in order to find them
    length(Hyps,L),write(len_of_hyps(L)),nl,
    % (L=1 -> write(Predicate),nl ; true),
    prove_predicate(Hyps,b(falsity,pred,[]),TResult), !,
    (TResult = proved
    -> Result = contradiction_found(Predicate) % TODO: extract core, etc.
    ;  Result = no_solution_found(atbprover_unproved)).
solve_provers(_,no_solution_found). % in case prove_predicate fails
