:- module(predicate_data_generator, [generate_synthesis_data_from_predicate/3,
                                     generate_synthesis_data_from_predicate/5]).

:- use_module(library(random)).
:- use_module(library(sets), [intersect/2]).

:- use_module(probsrc(bmachine), [b_load_machine_from_file/1,
                                  b_get_machine_variables/1,
                                  b_get_all_used_identifiers/1,
                                  b_get_main_filename/1,
                                  bmachine_is_precompiled/0,
                                  b_machine_precompile/0]).
:- use_module(probsrc(parsercall), [parse_formula/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(solver_interface), [solve_predicate/5, type_check_in_machine_context/2]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2, get_texpr_info/2, find_identifier_uses/3]).
:- use_module(probsrc(preferences), [set_preference/2,
                                     get_preference/2]).

:- use_module(synthesis('deep_learning/ground_truth')).
:- use_module(synthesis('deep_learning/b_machine_identifier_normalization')).
:- use_module(synthesis(synthesis_util), [get_input_nodes_from_bindings/2,
                                          create_equality_nodes_from_example/2,
                                          b_get_typed_invariant_from_machine/1]).

min_amount_of_examples(3).
max_amount_of_examples(positive, 8).
max_amount_of_examples(negative, 8).

augment_records(5).
solver_timeout_ms(10000).

get_random_amount_of_examples((PAmountOfExamples,NAmountOfExamples)) :-
    min_amount_of_examples(MinAmountOfExamples),
    max_amount_of_examples(positive, PTempMaxAmountOfExamples),
    max_amount_of_examples(negative, NTempMaxAmountOfExamples),
    PMaxAmountOfExamples is PTempMaxAmountOfExamples+1,
    NMaxAmountOfExamples is NTempMaxAmountOfExamples+1,
    random(MinAmountOfExamples, PMaxAmountOfExamples, PAmountOfExamples),
    random(MinAmountOfExamples, NMaxAmountOfExamples, NAmountOfExamples).

random_list_of_numbers(0, Acc, Acc) :-
    !.
random_list_of_numbers(C, Acc, L) :-
    C1 is C-1,
    get_random_amount_of_examples(R),
    \+ member(R, Acc),
    !,
    random_list_of_numbers(C1, [R|Acc], L).
random_list_of_numbers(C, Acc, L) :-
    random_list_of_numbers(C, Acc, L).

exclude_solution([], ExclusionPred, ExclusionPred) :-
    !.
exclude_solution(State, ExclusionPred, NewExclusionPred) :-
    create_equality_nodes_from_example(State, EqualityNodes),
    conjunct_predicates(EqualityNodes, EqualityConj),
    truth_or_exclude(EqualityConj,ExclusionPred, NewExclusionPred).

truth_or_exclude(b(truth,pred,[]), ExclusionPred, ExclusionPred) :-
    !.
truth_or_exclude(EqualityConj, ExclusionPred, b(conjunct(b(negation(EqualityConj),pred,[]),ExclusionPred),pred,[])).

filter_machine_var_states([], _, Acc, Acc).
filter_machine_var_states([VarState|T], MachineVars, Acc, States) :-
    get_texpr_info(VarState, Info),
    member(synthesis(machinevar, VarName), Info),
    member(b(identifier(VarName), _, _), MachineVars),
    !,
    filter_machine_var_states(T, MachineVars, [VarState|Acc], States).
filter_machine_var_states([_|T], MachineVars, Acc, States) :-
    filter_machine_var_states(T, MachineVars, Acc, States).

map_translate_var_state(Env, [], [], Env).
map_translate_var_state(Env, [Ast|T], [(VarName,Type,PrettyAst)|NT], NewEnv) :-
    normalize_ids_in_b_ast(Env, Ast, NAst, Env1),
    NAst = b(_, Type, Info),
    member(synthesis(machinevar, VarName), Info),
    translate_bexpression(NAst, PrettyAst),
    map_translate_var_state(Env1, T, NT, NewEnv).

get_amount_of_states_for_predicate(Env, 0, _, _, _, Acc, Acc, Env) :-
    !.
get_amount_of_states_for_predicate(Env, AmountOfExamples, ExclusionPred, MachineVars, VPredicateAst, Acc, ListOfExamples, NewEnv) :-
    AmountOfExamples1 is AmountOfExamples-1,
    solve_predicate(b(conjunct(ExclusionPred,VPredicateAst),pred,[]), _, 1, [force_evaluation], Solution),
    Solution = solution(Bindings),
    get_input_nodes_from_bindings(Bindings, TempState),
    filter_machine_var_states(TempState, MachineVars, [], State),
    State \== [],
    exclude_solution(State, ExclusionPred, NewExclusionPred),
    map_translate_var_state(Env, State, PrettyState, Env1),
    !,
    get_amount_of_states_for_predicate(Env1, AmountOfExamples1, NewExclusionPred, MachineVars, VPredicateAst, [PrettyState|Acc], ListOfExamples, NewEnv).
% cancel if no solution found
get_amount_of_states_for_predicate(Env, _, _, _, _, Acc, Acc, Env).

get_augmented_states_for_predicate(Env, AR, MachineVars, UsedComponents, PredicateAst, AugmentedSetOfData, NewEnv) :-
    random_list_of_numbers(AR, [], RandomNrList),
    get_augmented_states_for_predicate_rand(Env, RandomNrList, MachineVars, UsedComponents, PredicateAst, AugmentedSetOfData, NewEnv).

get_augmented_states_for_predicate_rand(Env, [], _, _, _, [], Env).
get_augmented_states_for_predicate_rand(Env, [AmountOfExamples|T], MachineVars, UsedComponents, PredicateAst, [(PositiveStates,NegativeStates,UsedComponents)|NT], NewEnv) :-
    AmountOfExamples = (PAmountOfExamples,NAmountOfExamples),
    get_amount_of_states_for_predicate(Env, PAmountOfExamples, b(truth,pred,[]), MachineVars, PredicateAst, [], PositiveStates, Env1),
    get_amount_of_states_for_predicate(Env1, NAmountOfExamples, b(truth,pred,[]), MachineVars, b(negation(PredicateAst),pred,[]), [], NegativeStates, Env2),
    length(PositiveStates, LPos),
    length(NegativeStates, LNeg),
    Amount is LPos+LNeg,
    min_amount_of_examples(MinAmountOfExamples),
    Amount >= MinAmountOfExamples,
    !,
    get_augmented_states_for_predicate_rand(Env2, T, MachineVars, UsedComponents, PredicateAst, NT, NewEnv).
get_augmented_states_for_predicate_rand(Env, [_|T], MachineVars, UsedComponents, PredicateAst, NT, NewEnv) :-
    get_augmented_states_for_predicate_rand(Env, T, MachineVars, UsedComponents, PredicateAst, NT, NewEnv).

%% remove_foolish_sub_predicates(+AllUsedIds, +TPredicateAst, -PredicateAst).
%
% Remove predicates that do not refer to any identifier within a machine (AllUsedIds).
% For instance, 'x: NAT & x < 10 & NAT \/ NAT1 = NAT' will be reduced to 'x: NAT & x < 10' with x being a machine variable.
% TODO: remove typing predicates ???
remove_foolish_sub_predicates(AllUsedIds, Ast, NewAst) :-
    Ast = b(Node,_,_),
    Node =.. [Functor, Lhs, Rhs],
    (Node == conjunct ; Node == disjunct),
    !,
    remove_foolish_sub_predicates_binary(AllUsedIds, Ast, Functor, Lhs, Rhs, NewAst).
% fail if sub-predicate does not use any ids from AllUsedIds
remove_foolish_sub_predicates(AllUsedIds, Ast, Ast) :-
    find_identifier_uses(Ast, [], UsedIds),
    intersect(AllUsedIds, UsedIds),
    !.

remove_foolish_sub_predicates_binary(AllUsedIds, b(_,Type,Info), Functor, Lhs, Rhs, b(NNode,Type,Info)) :-
    remove_foolish_sub_predicates(AllUsedIds, Lhs, NLhs),
    remove_foolish_sub_predicates(AllUsedIds, Rhs, NRhs),!,
    NNode =.. [Functor, NLhs, NRhs].
remove_foolish_sub_predicates_binary(AllUsedIds, _, _, Lhs, Rhs, NRhs) :-
    \+ remove_foolish_sub_predicates(AllUsedIds, Lhs, _),
    remove_foolish_sub_predicates(AllUsedIds, Rhs, NRhs),!.
remove_foolish_sub_predicates_binary(AllUsedIds, _, _, Lhs, Rhs, NLhs) :-
    remove_foolish_sub_predicates(AllUsedIds, Lhs, NLhs),
    \+ remove_foolish_sub_predicates(AllUsedIds, Rhs, _),!.
% fail if both sides do not use any ids from AllUsedIds

%% generate_synthesis_data_from_predicate(+RawPredicate, -GeneratedData).
%
% Generate positive and negative examples for a pretty-printed predicate and
% extract the ground truth B components as used by the program synthesis tool.
% Data is a list of triples (PositiveExamples, NegativeExamples, GroundTruth) considering
% data augmentation. Each example is a set of triples (MachineVar, Type, Value).
% Note: Predicate assumes that the B or Event-B machine is loaded that RawPredicate originates from.
%       Otherwise, fails silently.
generate_synthesis_data_from_predicate(MachinePath, RawPredicate, AugmentedSetOfData) :-
    augment_records(AR),
    solver_timeout_ms(SolverTimeoutMs),
    generate_synthesis_data_from_predicate(MachinePath, AR, SolverTimeoutMs, RawPredicate, AugmentedSetOfData).

load_b_machine_if_unloaded(MachinePath) :-
    bmachine_is_precompiled,
    b_get_main_filename(LoadedMachinePath),
    atom_concat(_, MachinePath, LoadedMachinePath),
    !.
load_b_machine_if_unloaded(MachinePath) :-
    b_load_machine_from_file(MachinePath),
    b_machine_precompile.

%% generate_synthesis_data_from_predicate(+AugmentRecords, +SolverTimeoutMs, +RawPredicate, -GeneratedData).
%
% AugmentRecords is the amount of data augmentations.
generate_synthesis_data_from_predicate(MachinePath, AugmentRecords, SolverTimeoutMs, RawPredicate, AugmentedSetOfData) :-
    load_b_machine_if_unloaded(MachinePath),
    \+ current_machine_uses_records,
    set_desired_preferences(SolverTimeoutMs, OldKodkodPref, OldTimeOutPref, OldRandPref),
    atom_codes(RawPredicate, RawPredicateCodes),
    parse_formula(RawPredicateCodes, ParsedPredicate),
    type_check_in_machine_context([ParsedPredicate], TParsedPredicates),
    TParsedPredicates = [TPredicateAst],
    b_get_machine_variables(MachineVars),
    b_get_all_used_identifiers(AllUsedIds),
    remove_foolish_sub_predicates(AllUsedIds, TPredicateAst, PredicateAst),
    get_library_components_from_pred_or_expr(PredicateAst, UsedComponents),
    get_normalized_id_name_mapping(NormalizedSets, NormalizedIds, NOperationNames, NRecordFieldNames),
    Env = [[], NormalizedSets, NormalizedIds, NOperationNames, NRecordFieldNames],
    b_get_typed_invariant_from_machine(Invariant),
    PredWithInv = b(conjunct(Invariant, PredicateAst),pred,[]),
    get_augmented_states_for_predicate(Env, AugmentRecords, MachineVars, UsedComponents, PredWithInv, AugmentedSetOfData, _),
    reset_old_preferences(OldKodkodPref, OldTimeOutPref, OldRandPref).

%% reset_old_preferences(+OldKodkodPref, +OldTimeOutPref, +OldRandPref).
%
reset_old_preferences(OldKodkodPref, OldTimeOutPref, OldRandPref) :-
    set_preference(try_kodkod_on_load, OldKodkodPref),
    set_preference(time_out, OldTimeOutPref),
    set_preference(randomise_enumeration_order, OldRandPref).

%% set_desired_preferences(-OldKodkodPref, -OldTimeOutPref, -OldRandPref).
%
% Set desired preferences and return the old ones.
set_desired_preferences(Timeout, OldKodkodPref, OldTimeOutPref, OldRandPref) :-
    get_preference(try_kodkod_on_load, OldKodkodPref),
    set_preference(try_kodkod_on_load, false),
    get_preference(randomise_enumeration_order, OldRandPref),
    set_preference(randomise_enumeration_order, true),
    get_preference(time_out, OldTimeOutPref),
    set_preference(time_out, Timeout).
