% (c) 2014-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(tcltk_tree_inspector,[
                 tti_number_of_columns/2,
                 %tti_column_name/3,
                 tti_column_info/6,
                 tti_get_node_info/6,
   reset_tcltk_tree_inspector/0]).

:- use_module(error_manager).
:- use_module(sap).
:- use_module(library(lists)).
:- use_module(ast_inspector).

:- use_module(module_information,[module_info/2]).
:- module_info(group,tcltk).
:- module_info(description,'This module provides the interface to the Tcl/Tk tree_inspector.').

% The module provides the descriptions and access to various trees displayed by tree_inspector.tcl
% We have various TreeTypes : ast, cbc_tests, ....
% Each Tree has a certain number of columns for the nodes described by tti_column_name/3
% Each Tree node can be inspected using tti_get_node_info
% The Paths in the tree (TTI_Path) record the number of each child selected (starting at 0)

tti_number_of_columns(empty,Nr) :- !, Nr=0.
tti_number_of_columns(Type,Nr) :-
   findall(C,(tti_column_name(Type,C,_), number(C)),L),
   length(L,Nr).

% column name and width descriptions:
tti_column_name(TreeType,Column,Name) :-
   tti_column_info(TreeType,Column,Name,_MinWidth,_Width,_Anchor).

tti_column_info(ast,header,'Node',100,300,w).
tti_column_info(ast,0,'Type',20,80,center).
tti_column_info(ast,1,'Infos',20,90,w).
tti_column_info(ast,2,'Quantified Ids',20,80,center).

tti_column_info(cbc_tests,header,'Event',100,400,w).
tti_column_info(cbc_tests,0,'Depth',      20,70,center).
tti_column_info(cbc_tests,1,'StateID',    25,70,center).
tti_column_info(cbc_tests,2,'Nr.Paths',   25,70,center).
tti_column_info(cbc_tests,3,'Nr.Tests',   25,70,center).

tti_column_info(_,_,'??',20,40,center).

% possible Tags are defined in file tree_inspector.tcl: error, inac, ptrue, pfalse, subst
% tti_get_node_info(Type,Pindex,Text,Columns,Subs,Tags) :-
%tti_get_node_info(Type,Pindex,Text,Columns,Subs,Tags) :- print(get(Type,Pindex,Text,Columns,Subs)),nl,fail.
tti_get_node_info(empty,_,'-',list([]),0,list([])) :- !.
tti_get_node_info(cbc_tests,Pindex,Text,list(Columns),NrOfSubs,list(Tags)) :- !,
   cbc_test_node_info(Pindex,Text,Columns,NrOfSubs,Tags).
tti_get_node_info(ast,Pindex,Text,list([Type,Infos,QIDs]),NrOfSubs,list(Tags)) :-
   tcltk_get_node_info(Pindex,Text,Type,Infos,QIDs,NrOfSubs),
   (Type = write(pred) -> Tags = [ptrue] ; Type = write(subst) -> Tags = [pfalse]
    ; Type = write('') -> Tags = [gray5]
    ; Tags = [darkblue]).


cbc_test_node_info([],'',[0,0,0,0],1,[]) :- % top-level; not shown in view
   assert(tti_path_to_term(cbc_tests,[],[])).
cbc_test_node_info([0|TTI_Path],Event,[Depth,Last,NrOfPaths,NrOfTests],NrOfSubs,Tags) :-
   %print(find(TTI_Path)),nl,
   cbc_path_info(TTI_Path,Event,Depth,Last,NrOfPaths,NrOfTests,NrOfSubs,Tags),
   %print(found(TTI_Path,Event,Depth,Last,NrOfPaths,NrOfSubs)),nl,
   true.

cbc_path_info(TTI_Path,Event,Depth,Last,NrOfPaths,NrOfTests,NrOfSubs,Tags) :-
   translate_tti_path_to_term(cbc_tests,TTI_Path,CBC_Path),
   %print(cbc(CBC_Path)),nl,
   cbc_get_path(Depth,CBC_Path,Last),
   (CBC_Path = [] -> Event = 'INITIALISATION' ; last(CBC_Path,Event)),
   append(CBC_Path,_,Prefix),
   findall(Last,sap:cb_path(_,Prefix,_),AllPathsWithPrefix),
   length(AllPathsWithPrefix,NrOfPaths),
   findall(Last,sap:stored_test_case_op_trace(_,Prefix),AllTestsWithPrefix),
   length(AllTestsWithPrefix,NrOfTests),
   findall(Desc,cbc_get_direct_descendant(CBC_Path,Desc),AllDirectDescPaths),
   store_tti_paths(AllDirectDescPaths,0,TTI_Path,cbc_tests),
   length(AllDirectDescPaths,NrOfSubs),
   (sap:stored_test_case_op_trace(_TestNr,CBC_Path)
     -> Tags = [ptrue] % we have generated a test-case from this path
     ; NrOfTests>0 -> Tags = []
      ; Tags = [darkgray]).

cbc_get_path(Depth,CBC_Path,Last) :- sap:cb_path(Depth,CBC_Path,Last).
cbc_get_path(Depth,CBC_Path,timeout) :- sap:cb_timeout_path(Depth,CBC_Path).
cbc_get_path(Depth,CBC_Path,Last) :- sap:cb_path_testcase_only(Depth,CBC_Path,Last).


cbc_get_direct_descendant(CBC_Path,NewPath) :-
   append(CBC_Path,[_],NewPath),
   cbc_get_path(_,NewPath,_).

% --------------------

% utilities of translating a TTI Path of argument positions to a Prolog Term
% TO DO: use term_hash or something like that + retractall after reset
:- dynamic tti_path_to_term/3.

store_tti_paths([],_,_,_).
store_tti_paths([DescendantTerm|T],ChildNr,TTI_Path,TTI_Tree_Type) :-
   append(TTI_Path,[ChildNr],NewPath),
   assert(tti_path_to_term(TTI_Tree_Type,NewPath,DescendantTerm)),
   %print(stored(NewPath,DescendantTerm)),nl,
   ChildNr1 is ChildNr+1,
   store_tti_paths(T,ChildNr1,TTI_Path,TTI_Tree_Type).

translate_tti_path_to_term(TTI_Tree_Type,TTI_Path,PrologTerm) :-
   (tti_path_to_term(TTI_Tree_Type,TTI_Path,R) -> PrologTerm=R
     ; add_internal_error('Cannot translate: ',
        translate_tti_path_to_term(TTI_Tree_Type,TTI_Path,PrologTerm)),
        PrologTerm = []).

reset_tcltk_tree_inspector :- retractall(tti_path_to_term(_,_,_)).
