% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(tools_commands,
    [ edit_file/1,edit_file/2, diff_files_with_editor/2,
      show_errors_with_bb_results/0, show_errors_with_bb_results/1, show_source_locations_with_bb_results/1,
      gen_dot_output/4, gen_dot_output_env/6,
      get_dot_engine_options/2,
      show_ps_file/1,
      show_dot_file/1 ]).


:- use_module(preferences).
:- use_module(debug).
:- use_module(tools).
:- use_module(error_manager).
:- use_module(system_call,[system_call/4]).


edit_file(F) :- edit_file(F,none).
edit_file(FILE,LINE) :-
     get_preference(path_to_text_editor_launch,EDITOR),
     edit_file_with_editor(EDITOR,FILE,LINE).

edit_file_with_editor('',FILE,_LINE) :- !,
     add_error(edit_file_with_editor,'No text editor set (be sure to set the ProB EDITOR preference), cannot edit: ',FILE).
edit_file_with_editor(EDITOR,FILE,LINE) :- number(LINE),
     tools:get_tail_filename(EDITOR,EdName),
     editor_understands_plus_line(EdName),!, % bbedit/vim can handle +LINE arg
     format('Opening file "~w" at LINE ~w using EDITOR "~w"~n',[FILE,LINE,EDITOR]),
     tools:string_concatenate('+',LINE,LinePosArg),
     system_call(EDITOR,[LinePosArg,FILE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).
edit_file_with_editor(EDITOR,FILE,LINE) :- number(LINE),
     tools:get_tail_filename(EDITOR,atom),!, % atom can handle FILE:LINE
     format('Opening file "~w" at LINE ~w using EDITOR "~w"~n',[FILE,LINE,EDITOR]),
     ajoin([FILE,':',LINE],FILE_LINE),
     system_call(EDITOR,[FILE_LINE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).
edit_file_with_editor(EDITOR,FILE,_) :-
     format('Opening file "~w" using EDITOR "~w"~n',[FILE,EDITOR]),
     system_call(EDITOR,[FILE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).


editor_understands_plus_line(bbedit).
editor_understands_plus_line(edit). % Text Wrangler
editor_understands_plus_line(vim).



diff_files_with_editor(FILE1,FILE2) :-
     DIFFER = '/usr/local/bin/bbdiff', % TO DO: provide preference
     system_call:system_call(DIFFER,[FILE1,FILE2],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).


:- use_module(library(process)).
:- use_module(system_call,[system_call_keep_open/7]).
show_errors_with_bb_results :- show_errors_with_bb_results([]).
show_errors_with_bb_results(Options) :-
    stored_error(_,_,_,Span,Options),
    extract_file_number_and_name(Span,_,_Filename),
    extract_line_col(Span,_,_,_,_),
    !,
    Command = '/usr/local/bin/bbresults', % TO DO: put in preference
    BBOptions = ['-p', 'flake8'], Env = [],
    on_exception(error(existence_error(_,_),E),
         (system_call_keep_open(Command,BBOptions,Process,STDIn,_STDOut,_STDErr,Env),
          write_errors_for_bbresults(STDIn,Options),
          close(STDIn),
          process_release(Process)),
          add_error(show_errors_with_bb_results,'Cannot start error viewer: ',Command:E)).
show_errors_with_bb_results(_). % no errors to show

stored_error(Source,S,TContext,Span,Options) :-
    (member(current,Options) -> logged_error(Source,S,TContext,Span)
     ; backed_up_error(Source,S,_TContext,Span)).

% ---------------------
% write errors in bbresults style (flake8)
% filename:line:column: W/E1 message

write_errors_for_bbresults(Stream,Options) :-
   stored_error(Source,S,_TContext,Span,Options),
   extract_file_number_and_name(Span,_,Filename),
   extract_line_col(Span,SL,SC,_EL,_EC),
   StartCol is SC+1,
   (Source = warning(_) -> Type='W1' ; Type='E1'),
   format(Stream,'~w:~w:~w: ~w ~w~n',[Filename,SL,StartCol,Type,S]),
   %format(user_output,'~w:~w:~w: ~w ~w~n',[Filename,SL,SC,Type,S]),
   fail.
write_errors_for_bbresults(_,_).

:- use_module(library(lists)).
% show a list of source locations
show_source_locations_with_bb_results(List) :-
    Command = '/usr/local/bin/bbresults', % TO DO: put in preference
    BBOptions = ['-p', 'flake8'], Env = [],
    on_exception(error(existence_error(_,_),E),
         (system_call_keep_open(Command,BBOptions,Process,STDIn,_STDOut,_STDErr,Env),
          maplist(write_src_loc_msg_for_bbresults(STDIn),List),
          close(STDIn),
          process_release(Process)),
          add_error(show_errors_with_bb_results,'Cannot start error viewer: ',Command:E)).

write_src_loc_msg_for_bbresults(Stream,src_loc_msg(S,Filename,SL,StartCol,_EndLine,_EndCol)) :-
   format(Stream,'~w:~w:~w: ~w ~w~n',[Filename,SL,StartCol,'W1',S]).

% ------------------------------

% show pdf/ps file using Preview/open/gsview/...
show_ps_file(FILE) :-
     get_preference(path_to_ps_viewer,PSVIEWER),
     system_call:system_call(PSVIEWER,[FILE],ErrorTextAsCodeList,ExitCode),
     (ErrorTextAsCodeList=[] -> true
      ; format('PS/PDF Viewer (~w) Error Result: ~s~nExit code:~w.~n',[PSVIEWER,ErrorTextAsCodeList,ExitCode]),
        fail
      ).

% show dot file using dotty
show_dot_file(FILE) :-
     get_preference(path_to_dotty,DOTTY),
     system_call:system_call(DOTTY,[FILE],ErrorTextAsCodeList,ExitCode),
     (ErrorTextAsCodeList=[] -> true
      ; format('Dot Viewer(~w) Error Result: ~s~nExit code:~w.~n',[DOTTY,ErrorTextAsCodeList,ExitCode]),
        fail
      ).

% ------------------------------

% call dot or sfdp to generate a PDF/PNG/SVG file:
:- use_module(probsrc(system_call),[system_call_env/6]).
gen_dot_output(DOTFILE,DotType,OutputType,OutputFILE) :-
   gen_dot_output_env(DOTFILE,DotType,OutputType,OutputFILE,[],_JExit).
% you can use JExit=no_process_wait and ENV = [detached(false)]
gen_dot_output_env(DOTFILE,DotType,OutputType,OutputFILE,ENV,JExit) :-
     get_preference(path_to_dot,DOTCMD),
     valid_dot_output_parameter(OutputType,Para),!,
     absolute_file_name(OutputFILE,AOutFile),
     format(user_output,'Calling dot: ~w to generate ~w from ~w~n',[DOTCMD,AOutFile,DOTFILE]),
     get_dot_engine_options(DotType,DotOptions),
     append([Para|DotOptions],['-o', AOutFile, DOTFILE],DotArgs),
     system_call_env(DOTCMD, DotArgs,ENV,_TextOut,ErrorTextAsCodeList,JExit),
     (ErrorTextAsCodeList=[] -> true
      ; format('Dot Error Result: ~s~n.',[ErrorTextAsCodeList]),
        fail
      ).

get_dot_engine_options(circo,R) :- !, R=['-Kcirco'].
get_dot_engine_options(dot,R) :- !, R=['-Kdot'].
get_dot_engine_options(fdp,R) :- !, R=['-Kfdp'].
get_dot_engine_options(neato,R) :- !, R=['-Kneato'].
get_dot_engine_options(sfdp,R) :- !, R=['-Ksfdp'].
get_dot_engine_options(twopi,R) :- !, R=['-Ktwopi'].
get_dot_engine_options(default,R) :- get_preference(dot_default_engine,Eng), Eng \= default,!,
    get_dot_engine_options(Eng,R).
get_dot_engine_options(OTHER,[]) :- add_error(get_dot_engine_options,'Unknown DOT Engine:',OTHER).

valid_dot_output_parameter(pdf,'-Tpdf').
valid_dot_output_parameter(png,'-Tpng').
valid_dot_output_parameter(svg,'-Tsvg').
valid_dot_output_parameter(dot,'-Tdot'). % dot with layout
valid_dot_output_parameter(canon,'-Tcanon'). % dot without layout
valid_dot_output_parameter(xdot,'-Txdot'). % dot with layout
% many more a supported: eps, xdot, canon, ...
valid_dot_output_parameter(T,_) :- add_internal_error('Illegal dot output type: ',T),fail.



